/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <criterion/criterion.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "common.h"

/* LCOV_EXCL_START */

static void fill_out(char **pout, unsigned int *plen, int pipe)
{
    char buf[4096];
    while (1)
    {
        int count = read(pipe, buf, sizeof(buf));
        char *tmp = 0;
        if (count == -1)
        {
            if (EINTR == errno)
            {
                continue;
            }
            cr_expect_fail("Cannot read() pout");
            break;
        }
        else if (count == 0)
        {
            break;
        }

        tmp = (char *)realloc(*pout, *plen + count);
        if (0 == tmp)
        {
            cr_expect_fail("Cannot realloc() pout");
            break;
        }
        memcpy(tmp + *plen, buf, count);
        *plen += count;
        *pout = tmp;
    }
}

static void verify_out(unsigned int expected_len, char *expected,
    unsigned int actual_len, char *actual)
{
    if (expected_len != actual_len || (0 < expected_len && 0 == actual) ||
        0 != strncmp(expected, actual, expected_len))
    {
        cr_expect_fail("Outputs different\n"
            "<expected %u>\n%.*s\n</expected>\n<actual %u>\n%.*s\n</actual>",
        expected_len, expected_len, expected,
        actual_len, actual_len, 0 != actual ? actual : "");
    }
    if (0 != actual)
    {
        free(actual);
    }
}

void run_with_args(char **argv, unsigned int expected_exit, char *expected_stdout,
    unsigned int expected_stdout_len, char *expected_stderr, unsigned int expected_stderr_len)
{
    pid_t pid = 0;
    int stdout[2];
    int stderr[2];
    int status = 0;
    int actual_exit = -1;
    char *actual_stdout = 0;
    char *actual_stderr = 0;
    unsigned int actual_stdout_len = 0;
    unsigned int actual_stderr_len = 0;
    unsigned int i;

    cr_log_info("# Run with args:");
    for (i = 0; 0 != argv[i]; i++)
    {
        cr_log_info("    argv[%u] = %s", i, argv[i]);
    }
    cr_log_info("    expected_exit = %d", expected_exit);
    cr_log_info("    expected_stdout_len = %u", expected_stdout_len);
    cr_log_info("    <expected_stdout>\n%.*s\n    </expected_stdout>", expected_stdout_len,
        expected_stdout);
    cr_log_info("    expected_stderr_len = %u", expected_stderr_len);
    cr_log_info("    <expected_stderr>\n%.*s\n    </expected_stderr>", expected_stderr_len,
        expected_stderr);

    if (-1 == pipe(stdout))
    {
        cr_expect_fail("Cannot pipe() stdout");
        return;
    }
    if (-1 == pipe(stderr))
    {
        cr_expect_fail("Cannot pipe() stderr");
        return;
    }

    pid = fork();

    if (pid == -1)
    {
        cr_expect_fail("Cannot fork()");
        return;
    }

    if (pid == 0)
    {
        dup2(stdout[1], 1);
        dup2(stderr[1], 2);
        close(stdout[1]);
        close(stderr[1]);
        execv(argv[0], argv);
        cr_expect_fail("Cannot execv()");
        _exit(1);
    }

    close(stdout[1]);
    close(stderr[1]);

    fill_out(&actual_stdout, &actual_stdout_len, stdout[0]);
    verify_out(expected_stdout_len, expected_stdout,
        actual_stdout_len, actual_stdout);

    fill_out(&actual_stderr, &actual_stderr_len, stderr[0]);
    verify_out(expected_stderr_len, expected_stderr,
        actual_stderr_len, actual_stderr);

    close(stdout[0]);
    close(stderr[0]);
    waitpid(pid, &status, 0);
    actual_exit = WEXITSTATUS(status);
    cr_expect_eq(expected_exit, actual_exit, "Exit codes different, expected %d, actual %d",
        expected_exit, actual_exit);
}

void expect_file(char *path)
{
    FILE *afile = fopen(path, "r");
    if (0 == afile)
    {
          cr_expect_fail("Cannot open file %s: %s", path, strerror(errno));
    }
    else
    {
        fclose(afile);
    }
}

/* LCOV_EXCL_STOP */
