/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <unistd.h>

#include <criterion/criterion.h>

#include <subprocess.h>

struct subprocess
{
    pid_t pid;
    int in;
    int out;
    int err;
};

struct subprocess_partial_stream
{
    unsigned int at;
    unsigned int len;
    char *data;
};

static void set_nonblocking(int output)
{
    errno = 0;
    int opts = fcntl(output, F_GETFL);
    if (0 > opts)
    {
        cr_log_error("Cannot fcntl F_GETFL: %d %s", errno, strerror(errno));
        return;
    }

    errno = 0;
    opts |= O_NONBLOCK;
    if (0 > fcntl(output, F_SETFL, opts))
    {
        cr_log_error("Cannot fcntl F_SETFL: %d %s", errno, strerror(errno));
    }
}

static int is_read_selected(int stream, unsigned int timeout)
{
    struct timeval second = {timeout, 0};
    fd_set aset;
    int ret;

    errno = 0;
    FD_ZERO(&aset);
    FD_SET(stream, &aset);

    ret = select(FD_SETSIZE, &aset, 0, 0, &second);
    if (0 > ret)
    {
        cr_expect_fail("Cannot select(): %d %s", errno, strerror(errno));
        return 0;
    }

    return 1 == ret;
}

static void reset_env(char **env)
{
    if (0 != env)
    {
        for (unsigned i = 0; 0 != env[i]; i += 2)
        {
            setenv(env[i], env[i + 1], 1);
        }
    }
}

static void log_new(char **argv, char **env,
    int will_read_from_stdout, int will_read_from_stderr)
{
    unsigned int i;
    cr_log_info("# Start process args:");
    for (i = 0; 0 != argv[i]; i++)
    {
        cr_log_info("    argv[%u] = <%s>", i, argv[i]);
    }
    if (1 == will_read_from_stdout)
    {
        cr_log_info("    Will read from stdout");
    }
    if (1 == will_read_from_stderr)
    {
        cr_log_info("    Will read from stderr");
    }
    if (0 != env)
    {
        cr_log_info("    Env:");
        for (i = 0; 0 != env[i]; i += 2)
        {
            cr_log_info("      <%s> = <%s>", env[i], env[i + 1]);
        }
    }
}

static void prepare_streams_in_child(int will_read_from_stdout, int stdout[2],
    int will_read_from_stderr, int stderr[2])
{
    if (1 == will_read_from_stdout)
    {
        dup2(stdout[1], 1);
        close(stdout[1]);
    }
    if (1 == will_read_from_stderr)
    {
        dup2(stderr[1], 2);
        close(stderr[1]);
    }
}

int subprocess_new(char **argv, char **env, int will_read_from_stdout, int will_read_from_stderr,
    subprocess **pthis)
{
    subprocess *this = 0;
    pid_t pid = 0;
    int stdout[2] = {-1. -1};
    int stderr[2] = {-1, -1};

    if (0 == argv || 0 == pthis || 0 != *pthis)
    {
        return 0;
    }

    log_new(argv, env, will_read_from_stdout, will_read_from_stderr);

    if (1 == will_read_from_stdout && -1 == pipe(stdout))
    {
        cr_expect_fail("Cannot pipe() stdout");
        return 0;
    }
    if (1 == will_read_from_stderr && -1 == pipe(stderr))
    {
        cr_expect_fail("Cannot pipe() stderr");
        return 0;
    }

    pid = fork();
    if (pid == -1)
    {
        cr_expect_fail("Cannot fork()");
        return 0;
    }

    if (pid == 0)
    {
        prepare_streams_in_child(will_read_from_stdout, stdout,
            will_read_from_stderr, stderr);
        reset_env(env);

        execv(argv[0], argv);
        cr_expect_fail("Cannot execv()");
        _exit(1);
    }

    cr_log_info("Forked as %d", pid);

    if (1 == will_read_from_stdout)
    {
        close(stdout[1]);
        set_nonblocking(stdout[0]);
    }

    if (1 == will_read_from_stderr)
    {
        close(stderr[1]);
        set_nonblocking(stderr[0]);
    }

    this = (subprocess *)malloc(sizeof(subprocess));
    this->pid = pid;
    this->out = stdout[0];
    this->err = stderr[0];

    *pthis = this;
    return 1;
}

int subprocess_free(subprocess **pthis)
{
    subprocess *this;

    if (0 == pthis || 0 == *pthis)
    {
        return 0;
    }

    this = *pthis;
    free(this);
    *pthis = 0;
    return 1;
}

int subprocess_kill(subprocess *this, int a_signal)
{
    if (0 == this)
    {
        return 0;
    }

    cr_log_info("# Kill process with signal=%d", a_signal);
    kill(this->pid, a_signal);
    return 1;
}

int subprocess_wait(subprocess *this, unsigned int expected_exit)
{
    int status = 0;
    int actual_exit;

    if (0 == this)
    {
        return 0;
    }

    cr_log_info("# Wait for exit process");
    cr_log_info("    expected_exit = %d", expected_exit);

    waitpid(this->pid, &status, 0);
    actual_exit = WEXITSTATUS(status);

    cr_log_info("    actual_exit = %d", actual_exit);
    cr_expect_eq(expected_exit, actual_exit,
        "Exit code different, expected %d, actual %d",
        expected_exit, actual_exit);
    return 1;
}

static int append_string(char **pout, unsigned int *plen, char *s, unsigned int slen)
{
    char *out = 0;
    unsigned int outlen = (*plen) + slen;

    out = (char *)realloc(*pout, sizeof(char) * (outlen + 1));
    if (0 == out)
    {
        return 0;
    }

    *pout = out;
    memcpy(out + (*plen), s, slen);
    out[(*plen) + slen] = 0;
    *plen = outlen;
    return 1;
}

static void get_most_output(int stream, unsigned int timeout, unsigned int *plen, char **pout)
{
    char buf[4096];
    unsigned int len = 0;
    char *out = 0;

    while (1)
    {
        int ret;

        if (0 == is_read_selected(stream, timeout))
        {
            break;
        }
        timeout = 1;

        errno = 0;
        ret = read(stream, buf, sizeof(buf));
        if (-1 == ret)
        {
            if (EWOULDBLOCK != errno && EAGAIN != errno)
            {
                cr_expect_fail("Stream error: %d %s\n", errno, strerror(errno));
                break;
            }
        }
        else if (0 == ret || 0 == append_string(&out, &len, buf, ret))
        {
            break;
        }
    }

    *pout = out;
    *plen = len;
}

int subprocess_read_output(subprocess *sub, unsigned int timeout,
    subprocess_partial_stream **pthis)
{
    char *out = 0;
    unsigned int len = 0;
    subprocess_partial_stream *this;

    get_most_output(sub->out, timeout, &len, &out);
    cr_log_info("# Read %u bytes from output: <%.*s>", len, len, out);

    this = (subprocess_partial_stream *)malloc(sizeof(subprocess_partial_stream));
    this->len = len;
    this->data = out;
    this->at = 0;
    *pthis = this;
    return 1;
}

int subprocess_read_error(subprocess *sub, unsigned int timeout,
    subprocess_partial_stream **pthis)
{
    char *out = 0;
    unsigned int len = 0;
    subprocess_partial_stream *this;

    get_most_output(sub->err, timeout, &len, &out);
    cr_log_info("# Read %u bytes from error: <%.*s>", len, len, out);

    this = (subprocess_partial_stream *)malloc(sizeof(subprocess_partial_stream));
    this->len = len;
    this->data = out;
    this->at = 0;
    *pthis = this;
    return 1;
}

int subprocess_partial_stream_free(subprocess_partial_stream **pthis)
{
    subprocess_partial_stream *this = *pthis;
    if (this->at != this->len)
    {
        cr_expect_fail("Stream not empty\n"
            "    At %u/%u bytes\n"
            "    Buffer before (%u): <%.*s>\n"
            "    Buffer after (%u): <%.*s>",
            this->at, this->len,
            this->at, this->at, this->data,
            this->len - this->at, this->len - this->at, this->data + this->at);
    }

    free(this->data);
    free(this);
    *pthis = 0;
    return 1;
}

int subprocess_partial_stream_get_for(subprocess_partial_stream *this, unsigned int len,
    char **pout)
{
    if (this->at + len > this->len)
    {
        cr_expect_fail("Read less than requested\n"
            "    At %u/%u bytes\n"
            "    Requested %u bytes\n"
            "    Buffer before (%u): <%.*s>\n"
            "    Buffer after (%u): <%.*s>",
            this->at, this->len,
            this->at, this->at, this->data,
            this->len - this->at, this->len - this->at, this->data + this->at);

        return 0;
    }

    *pout = this->data + this->at;
    this->at += len;
    return 1;
}

int subprocess_partial_stream_get_until(subprocess_partial_stream *this, char c,
    unsigned int *plen, char **pout)
{
    unsigned int until = -1;
    for (unsigned int i = this->at; i < this->len; i++)
    {
        if (c == this->data[i])
        {
            until = i;
            break;
        }
    }

    if (-1 == until)
    {
        cr_expect_fail("Did not read character %c (%u)\n"
            "    At %u/%u bytes\n"
            "    Buffer before (%u): <%.*s>\n"
            "    Buffer after (%u): <%.*s>",
            c, c,
            this->at, this->len,
            this->at, this->at, this->data,
            this->len - this->at, this->len - this->at, this->data + this->at);

        return 0;
    }

    *pout = this->data + this->at;
    *plen = until - this->at;
    this->at = until + 1;
    return 1;
}

int subprocess_partial_stream_cmp(subprocess_partial_stream *this, unsigned int len, char *str)
{
    if (this->at + len > this->len)
    {
        cr_expect_fail("Read less than requested\n"
            "    At %u/%u bytes\n"
            "    Expected: <%.*s>\n"
            "    Buffer before (%u): <%.*s>\n"
            "    Buffer after (%u): <%.*s>",
            this->at, this->len,
            len, str,
            this->at, this->at, this->data,
            this->len - this->at, this->len - this->at, this->data + this->at);
        return 0;
    }
    if (0 != memcmp(str, this->data + this->at, len))
    {
        cr_expect_fail("Read different %u bytes than requested\n"
            "    At %u/%u bytes\n"
            "    Expected: <%.*s>\n"
            "    Buffer before (%u): <%.*s>\n"
            "    Buffer after (%u): <%.*s>",
            len,
            this->at, this->len,
            len, str,
            this->at, this->at, this->data,
            this->len - this->at, this->len - this->at, this->data + this->at);
        return 0;
    }

    this->at += len;
    return 1;
}

/*
int subprocess_partial_stream_skip_for(subprocess_partial_stream *this, unsigned int len)
{
    if (this->at + len > this->len)
    {
        cr_expect_fail("Read less than requested\n"
            "    At %u/%u bytes\n"
            "    Requested to skip %u bytes\n"
            "    Buffer before (%u): <%.*s>\n"
            "    Buffer after (%u): <%.*s>",
            this->at, this->len, len,
            this->at, this->at, this->data,
            this->len - this->at, this->len - this->at, this->data + this->at);
        return 0;
    }

    this->at += len;
    return 1;

}
*/
int subprocess_partial_stream_skip_until(subprocess_partial_stream *this, char c)
{
    unsigned int until = -1;
    for (unsigned int i = this->at; i < this->len; i++)
    {
        if (c == this->data[i])
        {
            until = i;
            break;
        }
    }

    if (-1 == until)
    {
        cr_expect_fail("Did not read character '%c' (%u)\n"
            "    At %u/%u bytes\n"
            "    Buffer before (%u): <%.*s>\n"
            "    Buffer after (%u): <%.*s>",
            c, this->at, this->len,
            this->at, this->at, this->data,
            this->len - this->at, this->len - this->at, this->data + this->at);
        return 0;
    }

    this->at = until + 1;
    return 1;
}
