/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

enum parser_state
{
    PS_TEXT,
    PS_ONE_MUSTACHE,
    PS_TAG,
    PS_ONE_UNMUSTACHE
};

typedef struct parser parser;
struct parser
{
    enum parser_state state;
    int failed;

    unsigned int tag_len;
    char *tag;

    FILE *to;

    int did_dumpline_pre;
    int dumpline_i;
};

static void print_dumpline_pre(parser *a_parser)
{
    fprintf(a_parser->to, "    if (0 == tpl_append(pout, plen,\n");
    fprintf(a_parser->to, "        \"");
}

static void print_dumpline_post(parser *a_parser, unsigned int len)
{
    fprintf(a_parser->to, "\", %u))\n", len);
    fprintf(a_parser->to, "    {\n");
    fprintf(a_parser->to, "        free(*pout);\n");
    fprintf(a_parser->to, "        *pout = 0;\n");
    fprintf(a_parser->to, "        return 0;\n");
    fprintf(a_parser->to, "    }\n");
}

static void handle_enter(parser *a_parser)
{
    char *tag = a_parser->tag + 1;
    fprintf(a_parser->to, "    if (1 == tpl_enter(pout, plen, enter_f, userdata, \"%s\"))\n",
        tag);
    fprintf(a_parser->to, "    {\n");

    fprintf(a_parser->to, "        while (1 == tpl_next(pout, plen, next_f, userdata))\n");
    fprintf(a_parser->to, "        {\n");
}

static void handle_leave(parser *a_parser)
{
    fprintf(a_parser->to, "        }\n");
    fprintf(a_parser->to, "        tpl_leave(pout, plen, leave_f, userdata);\n");
    fprintf(a_parser->to, "    }\n");
}

static void handle_print_tag(parser *a_parser)
{
    fprintf(a_parser->to, "    if (0 == tpl_tag(pout, plen, tag_f, userdata, \"%s\"))\n",
        a_parser->tag);
    fprintf(a_parser->to, "    {\n");
    fprintf(a_parser->to, "        free(*pout);\n");
    fprintf(a_parser->to, "        *pout = 0;\n");
    fprintf(a_parser->to, "        return 0;\n");
    fprintf(a_parser->to, "    }\n");
}

static void examine_tag(parser *a_parser)
{
    char *tag = a_parser->tag;
    unsigned int len = strlen(tag);

    if (2 <= len && '#' == *tag)
    {
        handle_enter(a_parser);
    }
    else if (2 <= len && '/' == *tag)
    {
        handle_leave(a_parser);
    }
    else
    {
        handle_print_tag(a_parser);
    }
}

static void dump_one_char(parser *a_parser, char c)
{
    if (0 == a_parser->did_dumpline_pre)
    {
        print_dumpline_pre(a_parser);
        a_parser->did_dumpline_pre = 1;
        a_parser->dumpline_i = 0;
    }

    fprintf(a_parser->to, "\\x%.2X", c);
    a_parser->dumpline_i++;

    if ((a_parser->dumpline_i) % 20 == 19)
    {
        fprintf(a_parser->to, "\"\n");
        fprintf(a_parser->to, "        \"");
    }
}

static void append_tag(parser *a_parser, char c)
{
    char *ntag = 0;

    ntag = (char *)realloc(a_parser->tag, sizeof(char) * (a_parser->tag_len + 2));

    if (0 == ntag)
    {
        free(a_parser->tag);
        a_parser->tag = 0;
        a_parser->failed = 1;
    }
    else
    {
        a_parser->tag = ntag;
        a_parser->tag[a_parser->tag_len] = c;
        a_parser->tag[a_parser->tag_len + 1] = 0;
        a_parser->tag_len++;
    }
}

static void parse_text(parser *a_parser, char c)
{
    if ('{' != c)
    {
        dump_one_char(a_parser, c);
    }
    else
    {
        a_parser->state = PS_ONE_MUSTACHE;
    }
}

static void end_text(parser *a_parser)
{
    if (0 != a_parser->did_dumpline_pre)
    {
        print_dumpline_post(a_parser, a_parser->dumpline_i);
    }
}

static void end_one_mustache(parser *a_parser)
{
    a_parser->failed = 1;
    fprintf(stderr, "File ended with {, and it should not\n");
}

static void end_tag(parser *a_parser)
{
    free(a_parser->tag);
    a_parser->failed = 1;
    fprintf(stderr, "File ended inside tag, and it should not\n");
}

static void end_one_unmustache(parser *a_parser)
{
    free(a_parser->tag);
    a_parser->failed = 1;
    fprintf(stderr, "File ended inside tag and }, and it should not\n");
}

static void parse_one_mustache(parser *a_parser, char c)
{
    if ('{' != c)
    {
        a_parser->state = PS_TEXT;
        dump_one_char(a_parser, '{');
        dump_one_char(a_parser, c);
    }
    else
    {
        if (0 != a_parser->did_dumpline_pre)
        {
            print_dumpline_post(a_parser, a_parser->dumpline_i);
            a_parser->did_dumpline_pre = 0;
        }
        a_parser->dumpline_i = 0;
        a_parser->state = PS_TAG;
    }
}

static void parse_tag(parser *a_parser, char c)
{
    if ('}' != c)
    {
        append_tag(a_parser, c);
    }
    else
    {
        a_parser->state = PS_ONE_UNMUSTACHE;
    }
}

static void parse_one_unmustache(parser *a_parser, char c)
{
    if ('}' != c)
    {
        a_parser->state = PS_TEXT;
        append_tag(a_parser, '}');
        append_tag(a_parser, c);
    }
    else
    {
        if (0 != a_parser->tag)
        {
            examine_tag(a_parser);

            free(a_parser->tag);
            a_parser->tag = 0;
            a_parser->tag_len = 0;
        }
        a_parser->state = PS_TEXT;
    }
}

static void parse_char(parser *a_parser, char c)
{
    switch (a_parser->state)
    {
        case PS_TEXT:
            parse_text(a_parser, c);
            break;
        case PS_ONE_MUSTACHE:
            parse_one_mustache(a_parser, c);
            break;
        case PS_TAG:
            parse_tag(a_parser, c);
            break;
        case PS_ONE_UNMUSTACHE:
            parse_one_unmustache(a_parser, c);
            break;
        default:
            break;
    }
}

static void end_parse(parser *a_parser)
{
    switch (a_parser->state)
    {
        case PS_TEXT:
            end_text(a_parser);
            break;
        case PS_ONE_MUSTACHE:
            end_one_mustache(a_parser);
            break;
        case PS_TAG:
            end_tag(a_parser);
            break;
        case PS_ONE_UNMUSTACHE:
            end_one_unmustache(a_parser);
            break;
        default:
            break;
    }
}

static void parse(FILE *from, FILE *to)
{
    parser parser;

    parser.state = PS_TEXT;
    parser.failed = 0;
    parser.tag = 0;
    parser.tag_len = 0;
    parser.to = to;
    parser.did_dumpline_pre = 0;
    parser.dumpline_i = 0;

    while (0 == feof(from))
    {
        char buf[4096];
        int did_read = fread(buf, sizeof(char), 4096, from);
        if (0 >= did_read)
        {
            break;
        }
        for (unsigned int i = 0; i < did_read; i++)
        {
            parse_char(&parser, buf[i]);
        }
    }
    end_parse(&parser);
    /*
    if (0 != parser->failed)
    {
    }
*/
}

int main(int argc, char **argv)
{
    FILE *ifile = 0;
    FILE *cfile = 0;
    FILE *hfile = 0;

    if (5 != argc)
    {
        fprintf(stderr, "Usage: %s [input] [variable] [hfile] [cfile]\n\n",
            argv[0]);
        return 1;
    }

    ifile = fopen(argv[1], "rb");
    if (0 == ifile)
    {
        fprintf(stderr, "Cannot open file for reading: %s\n",
            strerror(errno));
        return 1;
    }

    hfile = fopen(argv[3], "wb");
    if (0 == hfile)
    {
        fclose(ifile);
        fprintf(stderr, "Cannot open hfile for writing: %s\n",
            strerror(errno));
        return 1;
    }

    cfile = fopen(argv[4], "wb");
    if (0 == cfile)
    {
        fclose(hfile);
        fclose(ifile);
        fprintf(stderr, "Cannot open cfile for writing: %s\n",
            strerror(errno));
        return 1;
    }

    fprintf(hfile, "#include <stdlib.h>\n");
    fprintf(hfile, "#include <tpl.h>\n");
    fprintf(hfile, "int tpl_%s(tpl_tag_f, tpl_enter_f, tpl_next_f, tpl_leave_f, \n", argv[2]);
    fprintf(hfile, "    void *, unsigned int *, char **);\n\n");
    fclose(hfile);

    fprintf(cfile, "#include \"%s\"\n", argv[3]);
    fprintf(cfile, "\n");
    fprintf(cfile, "int tpl_%s(tpl_tag_f tag_f, tpl_enter_f enter_f,\n", argv[2]);
    fprintf(cfile, "    tpl_next_f next_f, tpl_leave_f leave_f, void *userdata,\n");
    fprintf(cfile, "    unsigned int *plen, char **pout)\n");
    fprintf(cfile, "{\n");

    parse(ifile, cfile);
    fclose(ifile);
    fprintf(cfile, "    return 1;\n");
    fprintf(cfile, "}\n\n");
    fclose(cfile);
    return 0;
}
