/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sqlite3.h>

#include "common.h"
#include "task.h"
#include "task_diff.h"
#include "query.h"
#include "storage.h"
#include "sqlite.h"

typedef struct tm_sqlite_query tm_sqlite_query;

struct tm_sqlite_query
{
    unsigned long *pid_equals;
    char *name_contains;
    char *description_contains;
    char *status_contains;
    time_t *pdate_created_after;
    time_t *pdate_created_before;
    time_t *pdate_updated_after;
    time_t *pdate_updated_before;

    tm_query **and;
    unsigned int and_len;
};

static void sqlite_query_free(void *);

static char *query_create_table_version = "CREATE TABLE version (" \
    "id INTEGER PRIMARY KEY ASC AUTOINCREMENT DEFAULT 1, " \
    "value INTEGER NOT NULL" \
    ")";
static char *query_insert_version = "INSERT INTO version (value) VALUES (2)";

static char *query_does_version_table_exist = "SELECT 1 FROM sqlite_master " \
    "WHERE type = 'table' AND name = 'version'";
static char *query_get_tables_count = "SELECT COUNT(*) FROM sqlite_master " \
    "WHERE type = 'table'";

static char *query_create_table_task = "CREATE TABLE task (" \
    "id INTEGER PRIMARY KEY ASC AUTOINCREMENT DEFAULT 1, " \
    "date_created INTEGER NOT NULL"\
    ")";
static char *query_create_table_task_diff = "CREATE TABLE task_diff (" \
    "id INTEGER PRIMARY KEY ASC AUTOINCREMENT DEFAULT 1, " \
    "task INTEGER NOT NULL, " \
    "name TEXT NOT NULL, " \
    "value TEXT, " \
    "date INTEGER NOT NULL"\
    ")";

static char *query_get_tasks_prefix = "SELECT id, date_created FROM task";
static char *query_get_tasks_where_id_equals = "id = ?";
static char *query_get_tasks_where_name_contains = "id in (" \
        "SELECT diff.task FROM (" \
            "SELECT MAX(id), task, value " \
            "FROM task_diff " \
            "WHERE name = 'name' " \
            "GROUP BY task, name) " \
        "AS diff " \
        "WHERE diff.value GLOB ?" \
    ")";
static char *query_get_tasks_where_description_contains = "id in (" \
        "SELECT diff.task FROM (" \
            "SELECT max(id), task, value " \
            "FROM task_diff " \
            "WHERE name = 'description' " \
            "GROUP BY task, name)" \
        "AS diff " \
        "WHERE diff.value GLOB ?" \
    ")";
static char *query_get_tasks_where_status_contains = "id in (" \
        "SELECT diff.task FROM (" \
            "SELECT max(id), task, value " \
            "FROM task_diff " \
            "WHERE name = 'status' " \
            "GROUP BY task, name)" \
        "AS diff " \
        "WHERE diff.value GLOB ?" \
    ")";
static char *query_get_tasks_where_date_created_after = "date_created >= ?";
static char *query_get_tasks_where_date_created_before = "date_created < ?";
static char *query_get_tasks_where_date_updated_after = "id in (" \
        "SELECT diff.task FROM (" \
            "SELECT max(id), task, date " \
            "FROM task_diff " \
            "GROUP BY task, name)" \
        "AS diff " \
        "WHERE diff.date >= ?" \
    ")";
static char *query_get_tasks_where_date_updated_before = "id in (" \
        "SELECT diff.task FROM (" \
            "SELECT max(id), task, date " \
            "FROM task_diff " \
            "GROUP BY task, name)" \
        "AS diff " \
        "WHERE diff.date <= ?" \
    ")";

static char *query_get_tasks_v1 = "SELECT name FROM tasks";
static char *query_drop_tasks_v1 = "DROP TABLE tasks";

static char *query_get_task = "SELECT date_created FROM task WHERE id = ?";
static char *query_get_task_top_diffs = "SELECT a.id, a.name, a.value, a.date " \
    "FROM (" \
        "SELECT MAX(id) as last, name FROM task_diff " \
        "WHERE task = ? " \
        "GROUP BY name" \
    ") AS b " \
    "INNER JOIN task_diff AS a " \
    "ON b.last = a.id";
static char *query_get_task_diffs = "SELECT id, name, value, date FROM task_diff WHERE task = ? "
    " ORDER BY id DESC";

static char *query_add_task = "INSERT INTO task (date_created) VALUES (?)";
static char *query_add_new_task_diff_name = "INSERT INTO task_diff (task, name, value, date) " \
    "VALUES ((SELECT MAX(id) FROM task), 'name', ?, ?)";
static char *query_add_new_task_diff_description = "INSERT INTO task_diff " \
    "(task, name, value, date) " \
    "VALUES ((SELECT MAX(id) FROM task), 'description', ?, ?)";
static char *query_add_new_task_diff_status = "INSERT INTO task_diff (task, name, value, date) " \
    "VALUES ((SELECT MAX(id) FROM task), 'status', ?, ?)";

static char *query_add_existing_task_diff_name = "INSERT INTO task_diff " \
    "(task, name, value, date) " \
    "VALUES (?, 'name', ?, ?)";
static char *query_add_existing_task_diff_description = "INSERT INTO task_diff " \
    "(task, name, value, date) " \
    "VALUES (?, 'description', ?, ?)";
static char *query_add_existing_task_diff_status = "INSERT INTO task_diff " \
    "(task, name, value, date) " \
    "VALUES (?, 'status', ?, ?)";

static char *query_remove_task = "DELETE FROM task WHERE id = ?";
static char *query_remove_task_diffs = "DELETE FROM task_diff WHERE task = ?";

static void sqlite_free(tm_storage *this)
{
    sqlite3_close((sqlite3 *)this->data);
}

static int begin_transaction(sqlite3 *db)
{
    return SQLITE_OK == sqlite3_exec(db, "BEGIN TRANSACTION", 0, 0, 0);
}

static int end_transaction(sqlite3 *db)
{
    return SQLITE_OK == sqlite3_exec(db, "END TRANSACTION", 0, 0, 0);
}

static int rollback_transaction(sqlite3 *db)
{
    return SQLITE_OK == sqlite3_exec(db, "ROLLBACK TRANSACTION", 0, 0, 0);
}

static void serialize_query(char **pqtext, unsigned int *pqlen, tm_query *query)
{
    tm_sqlite_query *data = (tm_sqlite_query *)query->data;
    if (0 != data->pid_equals)
    {
        tm_compact_sprints(pqtext, pqlen, query_get_tasks_where_id_equals);
    }
    else if (0 != data->name_contains)
    {
        tm_compact_sprints(pqtext, pqlen, query_get_tasks_where_name_contains);
    }
    else if (0 != data->description_contains)
    {
        tm_compact_sprints(pqtext, pqlen, query_get_tasks_where_description_contains);
    }
    else if (0 != data->status_contains)
    {
        tm_compact_sprints(pqtext, pqlen, query_get_tasks_where_status_contains);
    }
    else if (0 != data->pdate_created_after)
    {
        tm_compact_sprints(pqtext, pqlen, query_get_tasks_where_date_created_after);
    }
    else if (0 != data->pdate_created_before)
    {
        tm_compact_sprints(pqtext, pqlen, query_get_tasks_where_date_created_before);
    }
    else if (0 != data->pdate_updated_after)
    {
        tm_compact_sprints(pqtext, pqlen, query_get_tasks_where_date_updated_after);
    }
    else if (0 != data->pdate_updated_before)
    {
        tm_compact_sprints(pqtext, pqlen, query_get_tasks_where_date_updated_before);
    }
    else if (0 != data->and)
    {
        tm_compact_sprints(pqtext, pqlen, "(");
        for (unsigned int i = 0; i < data->and_len; i++)
        {
            if (i > 0)
            {
                tm_compact_sprints(pqtext, pqlen, " AND ");
            }
            serialize_query(pqtext, pqlen, data->and[i]);
        }
        tm_compact_sprints(pqtext, pqlen, ")");
    }
}

static void bind_args_to_stmt(sqlite3_stmt *stmt, unsigned int *pi, tm_query *query)
{
    tm_sqlite_query *data = (tm_sqlite_query *)query->data;
    if (0 != data->pid_equals)
    {
        sqlite3_bind_int64(stmt, ++(*pi), *(data->pid_equals));
    }
    else if (0 != data->name_contains)
    {
        char *tmp = 0;
        unsigned int tmplen = 0;
        tm_compact_sprints(&tmp, &tmplen, "*");
        tm_compact_sprints(&tmp, &tmplen, data->name_contains);
        tm_compact_sprints(&tmp, &tmplen, "*");
        sqlite3_bind_text(stmt, ++(*pi), tmp, tmplen, free);
    }
    else if (0 != data->description_contains)
    {
        char *tmp = 0;
        unsigned int tmplen = 0;
        tm_compact_sprints(&tmp, &tmplen, "*");
        tm_compact_sprints(&tmp, &tmplen, data->description_contains);
        tm_compact_sprints(&tmp, &tmplen, "*");
        sqlite3_bind_text(stmt, ++(*pi), tmp, tmplen, free);
    }
    else if (0 != data->status_contains)
    {
        char *tmp = 0;
        unsigned int tmplen = 0;
        tm_compact_sprints(&tmp, &tmplen, "*");
        tm_compact_sprints(&tmp, &tmplen, data->status_contains);
        tm_compact_sprints(&tmp, &tmplen, "*");
        sqlite3_bind_text(stmt, ++(*pi), tmp, tmplen, free);
    }
    else if (0 != data->pdate_created_after)
    {
        sqlite3_bind_int64(stmt, ++(*pi), *(data->pdate_created_after));
    }
    else if (0 != data->pdate_created_before)
    {
        sqlite3_bind_int64(stmt, ++(*pi), *(data->pdate_created_before));
    }
    else if (0 != data->pdate_updated_after)
    {
        sqlite3_bind_int64(stmt, ++(*pi), *(data->pdate_updated_after));
    }
    else if (0 != data->pdate_updated_before)
    {
        sqlite3_bind_int64(stmt, ++(*pi), *(data->pdate_updated_before));
    }
    else if (0 != data->and)
    {
        for (unsigned int i = 0; i < data->and_len; i++)
        {
            bind_args_to_stmt(stmt, pi, data->and[i]);
        }
    }
}

static void build_stmt(sqlite3 *db, tm_query *query, sqlite3_stmt **pstmt)
{
    sqlite3_stmt *stmt = 0;
    char *qtext = strdup(query_get_tasks_prefix);
    unsigned int qlen = strlen(qtext);
    unsigned int i = 0;

    if (0 != query)
    {
        tm_compact_sprints(&qtext, &qlen, " WHERE ");
        serialize_query(&qtext, &qlen, query);
    }

    sqlite3_prepare_v2(db, qtext, strlen(qtext), &stmt, 0);
    free(qtext);
    if (0 != query)
    {
        bind_args_to_stmt(stmt, &i, query);
    }
    *pstmt = stmt;
}

static void process_get_task_row(sqlite3_stmt *stmt, tm_task *task)
{
    char *name = (char *)sqlite3_column_text(stmt, 1);
    char *value = (char *)sqlite3_column_text(stmt, 2);
    time_t date = (time_t)sqlite3_column_int64(stmt, 3);
    time_t task_date_updated = 0;

    tm_task_get_date_updated(task, &task_date_updated);
    if (date > task_date_updated)
    {
        tm_task_set_date_updated(task, date);
    }

    if (0 == strcmp("name", name))
    {
        tm_task_set_name(task, value);
    }
    else if (0 == strcmp("description", name))
    {
        tm_task_set_description(task, value);
    }
    else if (0 == strcmp("status", name))
    {
        tm_task_set_status(task, value);
    }
}

static int get_task(sqlite3 *db, unsigned int id, time_t date_created, tm_task **ptask)
{
    sqlite3_stmt *stmt = 0;
    int rc;
    tm_task *task = 0;

    tm_task_new(&task);
    tm_task_set_id(task, id);
    tm_task_set_date_created(task, date_created);

    sqlite3_prepare_v2(db, query_get_task_top_diffs, strlen(query_get_task_top_diffs),
        &stmt, 0);
    sqlite3_bind_int64(stmt, 1, id);

    do
    {
        rc = sqlite3_step(stmt);
        if (rc == SQLITE_ROW)
        {
            process_get_task_row(stmt, task);
        }
        else if (rc != SQLITE_DONE)
        {
            fprintf(stderr, "Cannot query task %u, step returned %d: %s\n",
                id, rc, sqlite3_errmsg(db));
            tm_task_free(&task);
            sqlite3_finalize(stmt);
            return 0;
        }
    } while (SQLITE_DONE != rc);
    sqlite3_finalize(stmt);
    tm_task_invalidate(task);
    *ptask = task;
    return 1;
}

static int process_get_tasks_row(sqlite3 *db, sqlite3_stmt *stmt, tm_task ***parr,
    unsigned int *parr_len, unsigned int *parr_allocated)
{
    unsigned int task_id = sqlite3_column_int64(stmt, 0);
    time_t date_created = (time_t)sqlite3_column_int64(stmt, 1);
    tm_task *task = 0;

    if (0 == get_task(db, task_id, date_created, &task))
    {
        return 0;
    }

    if (*parr_allocated == *parr_len)
    {
        *parr_allocated = (*parr_allocated) * 3 + 2;
        tm_task **tmp = (tm_task **)realloc(*parr, sizeof(tm_task *) * (*parr_allocated));

        if (0 == tmp)
        {
            fprintf(stderr, "Cannot realloc\n");
            return 0;
        }
        *parr = tmp;
    }
    (*parr)[(*parr_len)++] = task;
    return 1;
}

static int add_new_task_diff_name(sqlite3 *db, tm_task *task)
{
    char *name = 0;
    time_t date_created;
    sqlite3_stmt *stmt = 0;
    int rc;

    tm_task_get_name(task, &name);
    if (0 == name)
    {
        return 0;
    }

    tm_task_get_date_created(task, &date_created);

    sqlite3_prepare_v2(db, query_add_new_task_diff_name, strlen(query_add_new_task_diff_name),
        &stmt, 0);
    sqlite3_bind_text(stmt, 1, strdup(name), strlen(name), free);
    sqlite3_bind_int64(stmt, 2, date_created);

    rc = sqlite3_step(stmt);
    sqlite3_finalize(stmt);
    return SQLITE_DONE == rc;
}

static int add_existing_task_diff_name(sqlite3 *db, unsigned long task, char *name, time_t date)
{
    sqlite3_stmt *stmt = 0;
    int rc;

    if (0 == name)
    {
        return 1;
    }

    sqlite3_prepare_v2(db, query_add_existing_task_diff_name,
        strlen(query_add_existing_task_diff_name), &stmt, 0);
    sqlite3_bind_int64(stmt, 1, task);
    sqlite3_bind_text(stmt, 2, strdup(name), strlen(name), free);
    sqlite3_bind_int64(stmt, 3, date);

    rc = sqlite3_step(stmt);
    sqlite3_finalize(stmt);
    return SQLITE_DONE == rc;
}

static int add_new_task_diff_description(sqlite3 *db, tm_task *task)
{
    char *description = 0;
    time_t date_created;
    sqlite3_stmt *stmt = 0;
    int rc;

    tm_task_get_description(task, &description);
    if (0 == description)
    {
        return 1;
    }

    tm_task_get_date_created(task, &date_created);

    sqlite3_prepare_v2(db, query_add_new_task_diff_description,
        strlen(query_add_new_task_diff_description), &stmt, 0);
    sqlite3_bind_text(stmt, 1, strdup(description), strlen(description), free);
    sqlite3_bind_int64(stmt, 2, date_created);

    rc = sqlite3_step(stmt);
    sqlite3_finalize(stmt);
    return SQLITE_DONE == rc;
}

static int add_existing_task_diff_description(sqlite3 *db, unsigned long task, char *description,
    time_t date)
{
    sqlite3_stmt *stmt = 0;
    int rc;

    sqlite3_prepare_v2(db, query_add_existing_task_diff_description,
        strlen(query_add_existing_task_diff_description), &stmt, 0);

    sqlite3_bind_int64(stmt, 1, task);
    if (0 != description && 0 != *description)
    {
        sqlite3_bind_text(stmt, 2, strdup(description), strlen(description), free);
    }
    else
    {
        sqlite3_bind_null(stmt, 2);
    }
    sqlite3_bind_int64(stmt, 3, date);

    rc = sqlite3_step(stmt);
    sqlite3_finalize(stmt);
    return SQLITE_DONE == rc;
}

static int add_new_task_diff_status(sqlite3 *db, tm_task *task)
{
    char *status = 0;
    time_t date_created;
    sqlite3_stmt *stmt = 0;
    int rc;

    tm_task_get_status(task, &status);
    if (0 == status)
    {
        return 1;
    }

    tm_task_get_date_created(task, &date_created);

    sqlite3_prepare_v2(db, query_add_new_task_diff_status,
        strlen(query_add_new_task_diff_status), &stmt, 0);
    sqlite3_bind_text(stmt, 1, strdup(status), strlen(status), free);
    sqlite3_bind_int64(stmt, 2, date_created);

    rc = sqlite3_step(stmt);
    sqlite3_finalize(stmt);
    return SQLITE_DONE == rc;
}

static int add_existing_task_diff_status(sqlite3 *db, unsigned long task, char *status,
    time_t date)
{
    sqlite3_stmt *stmt = 0;
    int rc;

    sqlite3_prepare_v2(db, query_add_existing_task_diff_status,
        strlen(query_add_existing_task_diff_status), &stmt, 0);

    sqlite3_bind_int64(stmt, 1, task);
    if (0 != status)
    {
        sqlite3_bind_text(stmt, 2, strdup(status), strlen(status), free);
    }
    else
    {
        sqlite3_bind_null(stmt, 2);
    }
    sqlite3_bind_int64(stmt, 3, date);

    rc = sqlite3_step(stmt);
    sqlite3_finalize(stmt);
    return SQLITE_DONE == rc;
}

static int add_new_task(sqlite3 *db, tm_task *task)
{
    sqlite3_stmt *stmt = 0;
    int rc;
    time_t date_created;

    tm_task_get_date_created(task, &date_created);

    sqlite3_prepare_v2(db, query_add_task, strlen(query_add_task),
        &stmt, 0);
    sqlite3_bind_int64(stmt, 1, date_created);

    rc = sqlite3_step(stmt);
    sqlite3_finalize(stmt);
    if (SQLITE_DONE == rc)
    {
        unsigned long rowid = sqlite3_last_insert_rowid(db);
        tm_task_set_id(task, rowid);
        return 1;
    }
    return 0;
}

static int remove_task(sqlite3 *db, unsigned long task)
{
    sqlite3_stmt *stmt = 0;
    int rc;

    sqlite3_prepare_v2(db, query_remove_task, strlen(query_remove_task),
        &stmt, 0);
    sqlite3_bind_int64(stmt, 1, task);
    rc = sqlite3_step(stmt);
    sqlite3_finalize(stmt);
    return SQLITE_DONE == rc;
}

static int remove_task_diffs(sqlite3 *db, unsigned long task)
{
    sqlite3_stmt *stmt = 0;
    int rc;

    sqlite3_prepare_v2(db, query_remove_task_diffs,
        strlen(query_remove_task_diffs), &stmt, 0);
    sqlite3_bind_int64(stmt, 1, task);
    rc = sqlite3_step(stmt);
    sqlite3_finalize(stmt);
    return SQLITE_DONE == rc;
}

static int sqlite_get_tasks(tm_storage *this, tm_query *query, tm_task ***parr,
    unsigned int *plen)
{
    sqlite3 *db = (sqlite3 *)this->data;
    sqlite3_stmt *stmt = 0;
    int rc;
    unsigned int arr_len = 0;
    unsigned int arr_allocated = 10;
    tm_task **arr = 0;

    /* @todo how to do this more efifcient - both get tasks and newest diffs? */
    arr = (tm_task **)malloc(sizeof(tm_task *) * arr_allocated);
    build_stmt(db, query, &stmt);

    do
    {
        rc = sqlite3_step(stmt);
        if (rc == SQLITE_ROW &&
            0 == process_get_tasks_row(db, stmt, &arr, &arr_len, &arr_allocated))
        {
            free(arr);
            sqlite3_finalize(stmt);
            return 0;
        }
        else if (rc != SQLITE_ROW && rc != SQLITE_DONE)
        {
            fprintf(stderr, "Cannot query, step returned %d: %s\n",
                rc, sqlite3_errmsg(db));
            free(arr);
            sqlite3_finalize(stmt);
            return 0;
        }
    } while (SQLITE_DONE != rc);
    sqlite3_finalize(stmt);
    *parr = arr;
    *plen = arr_len;
    return 1;
}

static int sqlite_get_task(tm_storage *this, unsigned long id, tm_task **ptask)
{
    sqlite3 *db = (sqlite3 *)this->data;
    sqlite3_stmt *stmt = 0;
    int rc;
    int ret = 0;

    sqlite3_prepare_v2(db, query_get_task, strlen(query_get_task),
        &stmt, 0);
    sqlite3_bind_int64(stmt, 1, id);

    rc = sqlite3_step(stmt);

    if (rc == SQLITE_ROW)
    {
        time_t date_created = (time_t )sqlite3_column_int64(stmt, 0);
        ret = get_task(db, id, date_created, ptask);
    }
    else if (rc == SQLITE_DONE)
    {
        ret = 1;
        *ptask = 0;
    }
    else
    {
        fprintf(stderr, "Cannot get task, step returned %d: %s\n",
            rc, sqlite3_errmsg(db));
    }
    sqlite3_finalize(stmt);
    return ret;
}

static int process_get_task_diffs_row(sqlite3 *db, sqlite3_stmt *stmt, unsigned int task_id,
    tm_task_diff ***parr, unsigned int *parr_len, unsigned int *parr_allocated)
{
    tm_task_diff *a_diff = 0;

    unsigned long id = sqlite3_column_int64(stmt, 0);
    char *name = (char *)sqlite3_column_text(stmt, 1);
    char *value = (char *)sqlite3_column_text(stmt, 2);
    time_t date = (time_t )sqlite3_column_int64(stmt, 3);

    tm_task_diff_new(&a_diff);
    tm_task_diff_set_id(a_diff, id);
    tm_task_diff_set_task(a_diff, task_id);
    tm_task_diff_set_name(a_diff, name);
    tm_task_diff_set_value(a_diff, value);
    tm_task_diff_set_date(a_diff, date);

    if (*parr_allocated == *parr_len)
    {
        *parr_allocated = (*parr_allocated) * 3 + 2;
        tm_task_diff **tmp = (tm_task_diff **)realloc(*parr,
            sizeof(tm_task_diff *) * (*parr_allocated));

        if (0 == tmp)
        {
            fprintf(stderr, "Cannot realloc\n");
            return 0;
        }
        *parr = tmp;
    }
    (*parr)[(*parr_len)++] = a_diff;
    return 1;
}

static int sqlite_get_task_diffs(tm_storage *this, tm_task *task, tm_task_diff ***parr,
    unsigned int *plen)
{
    sqlite3 *db = (sqlite3 *)this->data;
    sqlite3_stmt *stmt = 0;
    int rc;
    unsigned long task_id = 0;
    tm_task_diff **arr = 0;
    unsigned int arr_len = 0;
    unsigned int arr_allocated = 10;

    arr = (tm_task_diff **)malloc(sizeof(tm_task_diff *) * arr_allocated);

    tm_task_get_id(task, &task_id);
    sqlite3_prepare_v2(db, query_get_task_diffs, strlen(query_get_task_diffs),
        &stmt, 0);
    sqlite3_bind_int64(stmt, 1, task_id);

    do
    {
        rc = sqlite3_step(stmt);
        if (rc == SQLITE_ROW &&
            0 == process_get_task_diffs_row(db, stmt, task_id, &arr, &arr_len, &arr_allocated))
        {
            free(arr);
            sqlite3_finalize(stmt);
            return 0;
        }
        else if (rc != SQLITE_ROW && rc != SQLITE_DONE)
        {
            fprintf(stderr, "Cannot query diffs for task %lu, step returned %d: %s\n",
                task_id, rc, sqlite3_errmsg(db));
            free(arr);
            tm_task_free(&task);
            sqlite3_finalize(stmt);
            return 0;
        }
    } while (SQLITE_DONE != rc);
    sqlite3_finalize(stmt);
    *parr = arr;
    *plen = arr_len;
    return 1;
}

static int modify_task(sqlite3 *db, tm_task *task)
{
    unsigned long id = -1;
    time_t now = time(0);

    tm_task_get_id(task, &id);

    if (0 != task->did_name_change)
    {
        char *name = 0;
        tm_task_get_name(task, &name);
        if (0 == add_existing_task_diff_name(db, id, name, now))
        {
            return 0;
        }
    }
    if (0 != task->did_description_change)
    {
        char *description = 0;
        tm_task_get_description(task, &description);
        if (0 == add_existing_task_diff_description(db, id, description, now))
        {
            return 0;
        }
    }
    if (0 != task->did_status_change)
    {
        char *status = 0;
        tm_task_get_status(task, &status);
        if (0 == add_existing_task_diff_status(db, id, status, now))
        {
            return 0;
        }
    }
    return 1;
}

static int sqlite_modify_task(tm_storage *this, tm_task *task)
{
    sqlite3 *db = (sqlite3 *)this->data;

    if (0 == begin_transaction(db))
    {
        fprintf(stderr, "Cannot modify task: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    if (0 == modify_task(db, task))
    {
        rollback_transaction(db);
        fprintf(stderr, "Cannot modify task: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    if (0 == end_transaction(db))
    {
        rollback_transaction(db);
        fprintf(stderr, "Cannot modify task: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    tm_task_invalidate(task);
    return 1;
}

static int sqlite_remove_task(tm_storage *this, tm_task *task)
{
    sqlite3 *db = (sqlite3 *)this->data;
    unsigned long id = -1;

    if (0 == begin_transaction(db))
    {
        fprintf(stderr, "Cannot remove task: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    tm_task_get_id(task, &id);
    if (0 == remove_task(db, id) || 0 == remove_task_diffs(db, id))
    {
        rollback_transaction(db);
        fprintf(stderr, "Cannot remove task: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    if (0 == end_transaction(db))
    {
        rollback_transaction(db);
        fprintf(stderr, "Cannot remove tasks: %s\n", sqlite3_errmsg(db));
        return 0;
    }
    return 1;
}

static int add_task(sqlite3 *db, tm_task *task)
{
    if (0 == add_new_task(db, task))
    {
        fprintf(stderr, "Cannot add task: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    if (0 == add_new_task_diff_name(db, task))
    {
        fprintf(stderr, "Cannot add task name: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    if (0 == add_new_task_diff_description(db, task))
    {
        fprintf(stderr, "Cannot add task description: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    if (0 == add_new_task_diff_status(db, task))
    {
        fprintf(stderr, "Cannot add task status: %s\n", sqlite3_errmsg(db));
        return 0;
    }
    return 1;
}

static int sqlite_add_tasks(tm_storage *this, tm_task **tasks, unsigned int len)
{
    sqlite3 *db = (sqlite3 *)this->data;
    unsigned int i;

    if (0 == begin_transaction(db))
    {
        fprintf(stderr, "Cannot add tasks: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    for (i = 0; i < len; i++)
    {
        tm_task *task = tasks[i];
        if (0 == add_task(db, task))
        {
            rollback_transaction(db);
            return 0;
        }
    }

    if (0 == end_transaction(db))
    {
        rollback_transaction(db);
        fprintf(stderr, "Cannot add tasks: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    for (i = 0; i < len; i++)
    {
        tm_task_invalidate(tasks[i]);
    }
    return 1;
}

static int sqlite_add_task(tm_storage *this, tm_task *task)
{
    sqlite3 *db = (sqlite3 *)this->data;

    if (0 == begin_transaction(db))
    {
        fprintf(stderr, "Cannot add tasks: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    if (0 == add_task(db, task))
    {
        rollback_transaction(db);
        return 0;
    }

    if (0 == end_transaction(db))
    {
        rollback_transaction(db);
        fprintf(stderr, "Cannot add task: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    tm_task_invalidate(task);
    return 1;
}

static int sqlite_modify_tasks(tm_storage *this, tm_task **tasks, unsigned int len)
{
    sqlite3 *db = (sqlite3 *)this->data;
    unsigned int i = 0;

    if (0 == begin_transaction(db))
    {
        fprintf(stderr, "Cannot modify tasks: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    for (i = 0; i < len; i++)
    {
        tm_task *task = tasks[i];
        if (0 == modify_task(db, task))
        {
            rollback_transaction(db);
            fprintf(stderr, "Cannot modify task %u/%u: %s\n", i + 1, len, sqlite3_errmsg(db));
            return 0;
        }
    }

    if (0 == end_transaction(db))
    {
        rollback_transaction(db);
        fprintf(stderr, "Cannot modify tasks: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    for (i = 0; i < len; i++)
    {
        tm_task_invalidate(tasks[i]);
    }
    return 1;
}

static int sqlite_remove_tasks(tm_storage *this, tm_task **tasks, unsigned int len)
{
    sqlite3 *db = (sqlite3 *)this->data;
    unsigned int i;

    if (0 == begin_transaction(db))
    {
        fprintf(stderr, "Cannot remove tasks: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    for (i = 0; i < len; i++)
    {
        tm_task *task = tasks[i];
        unsigned long id = -1;

        tm_task_get_id(task, &id);
        if (0 == remove_task(db, id) || 0 == remove_task_diffs(db, id))
        {
            rollback_transaction(db);
            fprintf(stderr, "Cannot remove task %u/%u: %s\n", i, len, sqlite3_errmsg(db));
            return 0;
        }
    }

    if (0 == end_transaction(db))
    {
        rollback_transaction(db);
        fprintf(stderr, "Cannot remove tasks: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    return 1;
}

static void sqlite_query_free(void *dat)
{
    tm_sqlite_query *data = (tm_sqlite_query *)dat;
    if (0 != data->pid_equals)
    {
        free(data->pid_equals);
    }
    if (0 != data->name_contains)
    {
        free(data->name_contains);
    }
    if (0 != data->description_contains)
    {
        free(data->description_contains);
    }
    if (0 != data->status_contains)
    {
        free(data->status_contains);
    }
    if (0 != data->pdate_created_after)
    {
        free(data->pdate_created_after);
    }
    if (0 != data->pdate_created_before)
    {
        free(data->pdate_created_before);
    }
    if (0 != data->pdate_updated_after)
    {
        free(data->pdate_updated_after);
    }
    if (0 != data->pdate_updated_before)
    {
        free(data->pdate_updated_before);
    }
    if (0 != data->and)
    {
        for (unsigned int i = 0; i < data->and_len; i++)
        {
            tm_query_free(&(data->and[i]));
        }
        free(data->and);
    }
    free(data);
}

static int sqlite_query_id_equals(tm_storage *this, unsigned long id, tm_query **pquery)
{
    tm_sqlite_query *data = (tm_sqlite_query *)malloc(sizeof(tm_sqlite_query));
    memset(data, 0, sizeof(tm_sqlite_query));
    data->pid_equals = (unsigned long *)malloc(sizeof(unsigned long));
    *(data->pid_equals) = id;
    return tm_query_new(data, sqlite_query_free, pquery);
}

static int sqlite_query_name_contains(tm_storage *this, char *name, tm_query **pquery)
{
    tm_sqlite_query *data = (tm_sqlite_query *)malloc(sizeof(tm_sqlite_query));
    memset(data, 0, sizeof(tm_sqlite_query));
    data->name_contains = strdup(name);
    return tm_query_new(data, sqlite_query_free, pquery);
}

static int sqlite_query_description_contains(tm_storage *this, char *description, tm_query **pquery)
{
    tm_sqlite_query *data = (tm_sqlite_query *)malloc(sizeof(tm_sqlite_query));
    memset(data, 0, sizeof(tm_sqlite_query));
    data->description_contains = strdup(description);
    return tm_query_new(data, sqlite_query_free, pquery);
}

static int sqlite_query_status_contains(tm_storage *this, char *status, tm_query **pquery)
{
    tm_sqlite_query *data = (tm_sqlite_query *)malloc(sizeof(tm_sqlite_query));
    memset(data, 0, sizeof(tm_sqlite_query));
    data->status_contains = strdup(status);
    return tm_query_new(data, sqlite_query_free, pquery);
}

static int sqlite_query_date_created_after(tm_storage *this, time_t date, tm_query **pquery)
{
    tm_sqlite_query *data = (tm_sqlite_query *)malloc(sizeof(tm_sqlite_query));
    memset(data, 0, sizeof(tm_sqlite_query));
    data->pdate_created_after = (time_t *)malloc(sizeof(time_t));
    *(data->pdate_created_after) = date;
    return tm_query_new(data, sqlite_query_free, pquery);
}

static int sqlite_query_date_created_before(tm_storage *this, time_t date, tm_query **pquery)
{
    tm_sqlite_query *data = (tm_sqlite_query *)malloc(sizeof(tm_sqlite_query));
    memset(data, 0, sizeof(tm_sqlite_query));
    data->pdate_created_before = (time_t *)malloc(sizeof(time_t));
    *(data->pdate_created_before) = date;
    return tm_query_new(data, sqlite_query_free, pquery);
}

static int sqlite_query_date_updated_after(tm_storage *this, time_t date, tm_query **pquery)
{
    tm_sqlite_query *data = (tm_sqlite_query *)malloc(sizeof(tm_sqlite_query));
    memset(data, 0, sizeof(tm_sqlite_query));
    data->pdate_updated_after = (time_t *)malloc(sizeof(time_t));
    *(data->pdate_updated_after) = date;
    return tm_query_new(data, sqlite_query_free, pquery);
}

static int sqlite_query_date_updated_before(tm_storage *this, time_t date, tm_query **pquery)
{
    tm_sqlite_query *data = (tm_sqlite_query *)malloc(sizeof(tm_sqlite_query));
    memset(data, 0, sizeof(tm_sqlite_query));
    data->pdate_updated_before = (time_t *)malloc(sizeof(time_t));
    *(data->pdate_updated_before) = date;
    return tm_query_new(data, sqlite_query_free, pquery);
}

static int sqlite_query_and(tm_storage *this, unsigned int clen, tm_query **cs, tm_query **pquery)
{
    tm_sqlite_query *data = (tm_sqlite_query *)malloc(sizeof(tm_sqlite_query));
    memset(data, 0, sizeof(tm_sqlite_query));
    data->and = cs;
    data->and_len = clen;
    return tm_query_new(data, sqlite_query_free, pquery);
}

static int install_tables(sqlite3 *db)
{
    char *errmsg = 0;
    if (0 != sqlite3_exec(db, query_create_table_version, 0, 0, &errmsg) ||
        0 != sqlite3_exec(db, query_insert_version, 0, 0, &errmsg) ||
        0 != sqlite3_exec(db, query_create_table_task, 0, 0, &errmsg) ||
        0 != sqlite3_exec(db, query_create_table_task_diff, 0, 0, &errmsg))
    {
        fprintf(stderr, "Cannot initialize database: %s\n", errmsg);
        return 0;
    }
    return 1;
}

static int migrate_task_v1_to_v2(sqlite3 *db, sqlite3_stmt *stmt)
{
    char *name = (char *)sqlite3_column_text(stmt, 0);
    int ret;
    tm_task *task = 0;

    tm_task_new(&task);
    tm_task_set_name(task, name);
    ret = add_task(db, task);
    tm_task_free(&task);
    return ret;
}

static int migrate_tasks_v1_to_v2(sqlite3 *db)
{
    sqlite3_stmt *stmt;
    int rc;

    sqlite3_prepare_v2(db, query_get_tasks_v1, strlen(query_get_tasks_v1), &stmt, 0);
    do
    {
        rc = sqlite3_step(stmt);
        if (rc == SQLITE_ROW)
        {
            if (0 == migrate_task_v1_to_v2(db, stmt))
            {
                sqlite3_finalize(stmt);
                return 0;
            }
        }
        else if (rc != SQLITE_DONE)
        {
            fprintf(stderr, "Cannot query v1 task, step returned %d: %s\n",
                rc, sqlite3_errmsg(db));
            sqlite3_finalize(stmt);
            return 0;
        }
    } while (SQLITE_DONE != rc);

    sqlite3_finalize(stmt);
    return 1;
}

static int drop_tasks_v1(sqlite3 *db)
{
    sqlite3_stmt *stmt;
    int rc;

    sqlite3_prepare_v2(db, query_drop_tasks_v1, strlen(query_drop_tasks_v1), &stmt, 0);
    rc = sqlite3_step(stmt);
    if (rc != SQLITE_DONE)
    {
        fprintf(stderr, "Cannot drop v1 tasks, step returned %d: %s\n",
            rc, sqlite3_errmsg(db));
        return 0;
    }
    sqlite3_finalize(stmt);
    return 1;
}

static int upgrade_tables_v1_to_v2(char *filename, sqlite3 *db)
{
    char *bck_filename = 0;
    sqlite3 *bck = 0;
    sqlite3_backup *backup = 0;
    int rc;

    bck_filename = (char *)malloc(sizeof(char) * (strlen(filename) + 11));
    sprintf(bck_filename, "%s.v1-v2.bak", filename);

    rc = sqlite3_open(bck_filename, &bck);
    free(bck_filename);

    if (0 != rc)
    {
        fprintf(stderr, "Cannot open backup database: %s\n", sqlite3_errmsg(bck));
        return 0;
    }

    backup = sqlite3_backup_init(bck, "main", db, "main");
    if (0 == backup)
    {
        sqlite3_close(bck);
        fprintf(stderr, "Cannot init backup: %s\n", sqlite3_errmsg(bck));
        return 0;
    }

    rc = sqlite3_backup_step(backup, -1);
    sqlite3_backup_finish(backup);
    sqlite3_close(bck);

    if (SQLITE_DONE != rc)
    {
        fprintf(stderr, "Cannot step backup: %d %s\n", rc, sqlite3_errmsg(bck));
        return 0;
    }

    return 0 != install_tables(db) &&
        0 != migrate_tasks_v1_to_v2(db) &&
        0 != drop_tasks_v1(db);
}

static int try_install_or_upgrade_tables_to_v1(char *filename, sqlite3 *db)
{
    sqlite3_stmt *stmt = 0;
    int rc;
    int ret = 0;

    sqlite3_prepare_v2(db, query_get_tables_count, strlen(query_get_tables_count),
        &stmt, 0);
    rc = sqlite3_step(stmt);

    if (rc == SQLITE_ROW)
    {
        int count = sqlite3_column_int(stmt, 0);
        sqlite3_finalize(stmt);
        if (0 == count)
        {
            ret = install_tables(db);
        }
        else if (count == 2)
        {
            ret = upgrade_tables_v1_to_v2(filename, db);
        }
        else
        {
            fprintf(stderr, "malformed table, not even v1\n");
        }
    }
    else
    {
        sqlite3_finalize(stmt);
        fprintf(stderr, "Cannot check database version, step returned %d: %s\n",
            rc, sqlite3_errmsg(db));
    }

    return ret;
}

static int try_upgrade_tables(char *filename, sqlite3 *db)
{
    sqlite3_stmt *stmt = 0;
    int rc;
    int ret = 0;

    sqlite3_prepare_v2(db, query_does_version_table_exist, strlen(query_does_version_table_exist),
        &stmt, 0);

    rc = sqlite3_step(stmt);
    if (rc == SQLITE_ROW)
    {
        ret = 1;
    }
    else if (rc == SQLITE_DONE)
    {
        ret = try_install_or_upgrade_tables_to_v1(filename, db);
    }
    else
    {
        fprintf(stderr, "Cannot check database version, step returned %d: %s\n",
            rc, sqlite3_errmsg(db));
    }

    sqlite3_finalize(stmt);
    return ret;
}

int tm_storage_sqlite(char *filename, tm_storage **pthis)
{
    tm_storage *this = 0;
    sqlite3 *db = 0;

    if (0 == filename || 0 == pthis || 0 != *pthis)
    {
        return 0;
    }

    if (0 != sqlite3_open(filename, &db))
    {
        fprintf(stderr, "Cannot open database: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    if (0 == try_upgrade_tables(filename, db))
    {
        sqlite3_close(db);
        return 0;
    }

    this = (tm_storage *)malloc(sizeof(struct tm_storage));
    this->data = (void *)db;
    this->free_f = sqlite_free;

    this->get_tasks_f = sqlite_get_tasks;

    this->get_task_f = sqlite_get_task;
    this->add_task_f = sqlite_add_task;
    this->modify_task_f = sqlite_modify_task;
    this->remove_task_f = sqlite_remove_task;

    this->get_task_diffs_f = sqlite_get_task_diffs;

    this->add_tasks_f = sqlite_add_tasks;
    this->modify_tasks_f = sqlite_modify_tasks;
    this->remove_tasks_f = sqlite_remove_tasks;

    this->query_id_equals_f = sqlite_query_id_equals;
    this->query_name_contains_f = sqlite_query_name_contains;
    this->query_description_contains_f = sqlite_query_description_contains;
    this->query_status_contains_f = sqlite_query_status_contains;
    this->query_date_created_after_f = sqlite_query_date_created_after;
    this->query_date_created_before_f = sqlite_query_date_created_before;
    this->query_date_updated_after_f = sqlite_query_date_updated_after;
    this->query_date_updated_before_f = sqlite_query_date_updated_before;
    this->query_and_f = sqlite_query_and;
    *pthis = this;
    return 1;
}
