/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <criterion/criterion.h>

#include <tm.h>

static char filename[80];
static tm_lib *this;

static void setup(void)
{
    tm_storage *storage = 0;
    sprintf(filename, "tmp-%d.db", getpid());
    cr_expect_eq(1, tm_storage_sqlite(filename, &storage));
    cr_expect_eq(1, tm_new(storage, &this));
}

static void teardown(void)
{
    tm_free(&this);
    remove(filename);
}

TestSuite(tasks, .init = setup, .fini = teardown);

static void expect_date_within_last_1_minute(time_t date)
{
    time_t now = time(0);
    cr_log_info("expect_date_within_last_1_minute now=%ld date=%ld (now-date)=%ld",
        now, date, now - date);
    cr_expect_geq(now, date);
    cr_expect_geq(date, now - 60);
}

static void expect_task_diff(tm_task_diff **diffs, unsigned int i, unsigned int expected_id,
    unsigned int expected_task, char *expected_name, char *expected_value)
{
    char *name = 0;
    char *value = 0;
    unsigned long id = -1;

    cr_log_info("expect_task_diff i=%u expected_id=%u expected_task=%u "
        "expected_name=%s expected_value=%s", i, expected_id, expected_task,
        expected_name, expected_value);

    if (0 == diffs)
    {
        return;
    }

    cr_expect_neq(0, diffs[i]);
    if (0 == diffs[i])
    {
        return;
    }

    cr_expect_eq(1, tm_task_diff_get_id(diffs[i], &id));
    cr_expect_eq(expected_id, id);
    cr_expect_eq(1, tm_task_diff_get_task(diffs[i], &id));
    cr_expect_eq(expected_task, id);
    cr_expect_eq(1, tm_task_diff_get_name(diffs[i], &name));
    cr_expect_str_eq(expected_name, name);
    cr_expect_eq(1, tm_task_diff_get_value(diffs[i], &value));
    cr_expect_str_eq(expected_value, value);
}

static void expect_task_diff_date_after(tm_task_diff **diffs, unsigned int i, time_t after)
{
    time_t date = 0;

    cr_log_info("expect_task_diff_date_after i=%u after=%ld", i, after);
    if (0 == diffs)
    {
        return;
    }

    cr_expect_neq(0, diffs[i]);
    if (0 == diffs[i])
    {
        return;
    }

    cr_expect_eq(1, tm_task_diff_get_date(diffs[i], &date));
    cr_expect_geq(date, after);
}

static void expect_add_task(void)
{
    tm_task *task = 0;
    unsigned long id = -1;
    cr_expect_eq(1, tm_task_new(&task));
    cr_expect_neq(0, task);
    cr_expect_eq(1, tm_task_set_name(task, "Buy milk"));
    cr_expect_eq(1, tm_task_set_description(task, "Goat milk"));
    cr_expect_eq(1, tm_add_task(this, task));
    cr_expect_eq(1, tm_task_get_id(task, &id));
    cr_expect_neq(-1, id);
    cr_expect_eq(1, tm_task_free(&task));
}

static void expect_get_task(tm_task **ptask)
{
    tm_task *task = 0;
    unsigned long id = -1;
    char *name = 0;
    char *description = 0;
    char *status = 0;
    time_t date_created = 0;
    time_t date_updated = 0;

    cr_expect_eq(1, tm_get_task(this, 1, &task));
    cr_expect_neq(0, task);
    cr_expect_eq(1, tm_task_get_id(task, &id));
    cr_expect_eq(1, id);
    cr_expect_eq(1, tm_task_get_name(task, &name));
    cr_expect_str_eq("Buy milk", name);
    cr_expect_eq(1, tm_task_get_description(task, &description));
    cr_expect_str_eq("Goat milk", description);
    cr_expect_eq(1, tm_task_get_status(task, &status));
    cr_expect_str_eq("new", status);
    cr_expect_eq(1, tm_task_get_date_created(task, &date_created));
    expect_date_within_last_1_minute(date_created);
    cr_expect_eq(1, tm_task_get_date_updated(task, &date_updated));
    expect_date_within_last_1_minute(date_updated);

    *ptask = task;
}

Test(tasks, add_modify_remove_task)
{
    tm_task *task = 0;
    tm_task_diff **diffs = 0;
    time_t date_updated = 0;
    unsigned int len = -1;

    expect_add_task();
    expect_get_task(&task);

    cr_expect_eq(1, tm_task_get_date_updated(task, &date_updated));

    cr_expect_eq(1, tm_get_task_diffs(this, task, &diffs, &len));
    cr_expect_neq(0, diffs);
    cr_expect_eq(3, len);
    expect_task_diff(diffs, 2, 1, 1, "name", "Buy milk");
    expect_task_diff(diffs, 1, 2, 1, "description", "Goat milk");
    expect_task_diff(diffs, 0, 3, 1, "status", "new");
    cr_expect_eq(1, tm_task_diff_arr_free(&diffs, len));
    cr_expect_eq(0, diffs);

    cr_expect_eq(1, tm_task_set_name(task, "Buy beer"));
    cr_expect_eq(1, tm_modify_task(this, task));
    cr_expect_eq(1, tm_get_task_diffs(this, task, &diffs, &len));
    cr_expect_neq(0, diffs);
    cr_expect_eq(4, len);
    expect_task_diff(diffs, 0, 4, 1, "name", "Buy beer");
    expect_task_diff_date_after(diffs, 0, date_updated);
    cr_expect_eq(1, tm_task_diff_arr_free(&diffs, len));
    cr_expect_eq(0, diffs);

    cr_expect_eq(1, tm_task_set_description(task, "Lager"));
    cr_expect_eq(1, tm_modify_task(this, task));
    cr_expect_eq(1, tm_get_task_diffs(this, task, &diffs, &len));
    cr_expect_neq(0, diffs);
    cr_expect_eq(5, len);
    expect_task_diff(diffs, 0, 5, 1, "description", "Lager");
    expect_task_diff_date_after(diffs, 0, date_updated);
    cr_expect_eq(1, tm_task_diff_arr_free(&diffs, len));
    cr_expect_eq(0, diffs);

    cr_expect_eq(1, tm_task_set_status(task, "open"));
    cr_expect_eq(1, tm_modify_task(this, task));
    cr_expect_eq(1, tm_get_task_diffs(this, task, &diffs, &len));
    cr_expect_neq(0, diffs);
    cr_expect_eq(6, len);
    expect_task_diff(diffs, 0, 6, 1, "status", "open");
    expect_task_diff_date_after(diffs, 0, date_updated);
    cr_expect_eq(1, tm_task_diff_arr_free(&diffs, len));
    cr_expect_eq(0, diffs);

    cr_expect_eq(1, tm_task_set_name(task, "Buy gas"));
    cr_expect_eq(1, tm_task_set_description(task, "PB95"));
    cr_expect_eq(1, tm_task_set_status(task, "done"));
    cr_expect_eq(1, tm_modify_task(this, task));
    cr_expect_eq(1, tm_get_task_diffs(this, task, &diffs, &len));
    cr_expect_neq(0, diffs);
    cr_expect_eq(9, len);
    expect_task_diff(diffs, 2, 7, 1, "name", "Buy gas");
    expect_task_diff_date_after(diffs, 2, date_updated);
    expect_task_diff(diffs, 1, 8, 1, "description", "PB95");
    expect_task_diff_date_after(diffs, 1, date_updated);
    expect_task_diff(diffs, 0, 9, 1, "status", "done");
    expect_task_diff_date_after(diffs, 0, date_updated);
    cr_expect_eq(1, tm_task_diff_arr_free(&diffs, len));
    cr_expect_eq(0, diffs);

    cr_expect_eq(1, tm_remove_task(this, task));
    cr_expect_eq(1, tm_task_free(&task));
    cr_expect_eq(0, task);

    cr_expect_eq(1, tm_get_task(this, 0, &task));
    cr_expect_eq(0, task);
}

static void expect_add_tasks(void)
{
    tm_task **tasks;
    unsigned long id = -1;

    tasks = (tm_task **)malloc(sizeof(tm_task *) * 3);
    memset(tasks, 0, sizeof(tm_task *) * 3);
    cr_expect_eq(1, tm_task_new(&(tasks[0])));
    cr_expect_eq(1, tm_task_set_name(tasks[0], "Buy milk"));
    cr_expect_eq(1, tm_task_set_description(tasks[0], "Goat milk"));
    cr_expect_eq(1, tm_task_new(&(tasks[1])));
    cr_expect_eq(1, tm_task_set_name(tasks[1], "Buy beer"));
    cr_expect_eq(1, tm_task_set_description(tasks[1], "Lager"));
    cr_expect_eq(1, tm_task_new(&(tasks[2])));
    cr_expect_eq(1, tm_task_set_name(tasks[2], "Buy gas"));
    cr_expect_eq(1, tm_task_set_description(tasks[2], "PB95"));
    cr_expect_eq(1, tm_add_tasks(this, tasks, 3));

    cr_expect_eq(1, tm_task_get_id(tasks[0], &id));
    cr_expect_neq(-1, id);
    id = -1;
    cr_expect_eq(1, tm_task_get_id(tasks[1], &id));
    cr_expect_neq(-1, id);
    id = -1;
    cr_expect_eq(1, tm_task_get_id(tasks[2], &id));
    cr_expect_neq(-1, id);

    cr_expect_eq(1, tm_task_arr_free(&tasks, 3));
    cr_expect_eq(0, tasks);
}

static void expect_get_3_tasks(tm_task ***ptasks, unsigned int *plen)
{
    tm_task **tasks;
    unsigned long id = -1;
    char *name = 0;
    char *description = 0;
    unsigned int len = -1;

    cr_expect_eq(1, tm_get_tasks(this, 0, &tasks, &len));
    cr_expect_neq(0, tasks);
    if (0 == tasks)
    {
        *ptasks = 0;
        *plen = 0;
        return;
    }

    cr_expect_eq(3, len);
    cr_expect_neq(0, tasks[0]);
    cr_expect_neq(0, tasks[1]);
    cr_expect_neq(0, tasks[2]);

    cr_expect_eq(1, tm_task_get_id(tasks[0], &id));
    cr_expect_eq(1, id);
    cr_expect_eq(1, tm_task_get_name(tasks[0], &name));
    cr_expect_str_eq("Buy milk", name);
    cr_expect_eq(1, tm_task_get_description(tasks[0], &description));
    cr_expect_str_eq("Goat milk", description);
    cr_expect_eq(1, tm_task_get_id(tasks[1], &id));
    cr_expect_eq(2, id);
    cr_expect_eq(1, tm_task_get_name(tasks[1], &name));
    cr_expect_str_eq("Buy beer", name);
    cr_expect_eq(1, tm_task_get_description(tasks[1], &description));
    cr_expect_str_eq("Lager", description);
    cr_expect_eq(1, tm_task_get_id(tasks[2], &id));
    cr_expect_eq(3, id);
    cr_expect_eq(1, tm_task_get_name(tasks[2], &name));
    cr_expect_str_eq("Buy gas", name);
    cr_expect_eq(1, tm_task_get_description(tasks[2], &description));
    cr_expect_str_eq("PB95", description);

    *ptasks = tasks;
    *plen = len;
}

Test(tasks, add_modify_remove_task_arr)
{
    tm_task **tasks;
    tm_task_diff **diffs = 0;
    unsigned int len = -1;

    expect_add_tasks();
    expect_get_3_tasks(&tasks, &len);
    if (0 == tasks)
    {
        return;
    }

    cr_expect_eq(1, tm_get_task_diffs(this, tasks[0], &diffs, &len));
    cr_expect_neq(0, diffs);
    cr_expect_eq(3, len);
    expect_task_diff(diffs, 2, 1, 1, "name", "Buy milk");
    expect_task_diff(diffs, 1, 2, 1, "description", "Goat milk");
    expect_task_diff(diffs, 0, 3, 1, "status", "new");
    cr_expect_eq(1, tm_task_diff_arr_free(&diffs, len));
    cr_expect_eq(0, diffs);
    cr_expect_eq(1, tm_get_task_diffs(this, tasks[1], &diffs, &len));
    cr_expect_neq(0, diffs);
    cr_expect_eq(3, len);
    expect_task_diff(diffs, 2, 4, 2, "name", "Buy beer");
    expect_task_diff(diffs, 1, 5, 2, "description", "Lager");
    expect_task_diff(diffs, 0, 6, 2, "status", "new");
    cr_expect_eq(1, tm_task_diff_arr_free(&diffs, len));
    cr_expect_eq(0, diffs);
    cr_expect_eq(1, tm_get_task_diffs(this, tasks[2], &diffs, &len));
    cr_expect_neq(0, diffs);
    cr_expect_eq(3, len);
    expect_task_diff(diffs, 2, 7, 3, "name", "Buy gas");
    expect_task_diff(diffs, 1, 8, 3, "description", "PB95");
    expect_task_diff(diffs, 0, 9, 3, "status", "new");
    cr_expect_eq(1, tm_task_diff_arr_free(&diffs, len));
    cr_expect_eq(0, diffs);

    cr_expect_eq(1, tm_task_arr_set_name(tasks, 3, "Belay that"));
    cr_expect_eq(1, tm_task_arr_set_description(tasks, 3, "Rowered"));
    cr_expect_eq(1, tm_task_arr_set_status(tasks, 3, "open"));
    cr_expect_eq(1, tm_modify_tasks(this, tasks, 3));

    cr_expect_eq(1, tm_get_task_diffs(this, tasks[0], &diffs, &len));
    cr_expect_eq(6, len);
    expect_task_diff(diffs, 2, 10, 1, "name", "Belay that");
    expect_task_diff(diffs, 1, 11, 1, "description", "Rowered");
    expect_task_diff(diffs, 0, 12, 1, "status", "open");
    cr_expect_eq(1, tm_task_diff_arr_free(&diffs, len));
    cr_expect_eq(0, diffs);
    cr_expect_eq(1, tm_get_task_diffs(this, tasks[1], &diffs, &len));
    cr_expect_eq(6, len);
    expect_task_diff(diffs, 2, 13, 2, "name", "Belay that");
    expect_task_diff(diffs, 1, 14, 2, "description", "Rowered");
    expect_task_diff(diffs, 0, 15, 2, "status", "open");
    cr_expect_eq(1, tm_task_diff_arr_free(&diffs, len));
    cr_expect_eq(0, diffs);
    cr_expect_eq(1, tm_get_task_diffs(this, tasks[2], &diffs, &len));
    cr_expect_eq(6, len);
    expect_task_diff(diffs, 2, 16, 3, "name", "Belay that");
    expect_task_diff(diffs, 1, 17, 3, "description", "Rowered");
    expect_task_diff(diffs, 0, 18, 3, "status", "open");
    cr_expect_eq(1, tm_task_diff_arr_free(&diffs, len));
    cr_expect_eq(0, diffs);

    cr_expect_eq(1, tm_remove_tasks(this, tasks, 3));
    cr_expect_eq(1, tm_task_arr_free(&tasks, 3));
    cr_expect_eq(0, tasks);

    cr_expect_eq(1, tm_get_tasks(this, 0, &tasks, &len));
    cr_expect_neq(0, tasks);
    cr_expect_eq(0, len);
    cr_expect_eq(1, tm_task_arr_free(&tasks, len));
    cr_expect_eq(0, tasks);
}

Test(tasks, get_tasks)
{
    tm_task *task = 0;
    tm_task **tasks = 0;
    tm_query *query = 0;
    tm_query **subqueries = 0;
    time_t date_created = 0;
    time_t date_updated = 0;
    unsigned int len = -1;

    cr_expect_eq(1, tm_task_new(&task));
    cr_expect_eq(1, tm_task_set_name(task, "Buy milk"));
    cr_expect_eq(1, tm_task_set_description(task, "Goat milk"));
    cr_expect_eq(1, tm_add_task(this, task));
    cr_expect_eq(1, tm_task_free(&task));

    cr_expect_eq(1, tm_get_task(this, 1, &task));
    cr_expect_eq(1, tm_task_get_date_created(task, &date_created));
    cr_expect_eq(1, tm_task_get_date_updated(task, &date_updated));
    cr_expect_eq(1, tm_task_free(&task));

    subqueries = (tm_query **)malloc(sizeof(tm_query *) * 8);
    memset(subqueries, 0, sizeof(tm_query *) * 8);

    cr_expect_eq(1, tm_query_id_equals(this, 1, &subqueries[0]));
    cr_expect_neq(0, subqueries[0]);
    cr_expect_eq(1, tm_query_name_contains(this, "milk", &subqueries[1]));
    cr_expect_neq(0, subqueries[1]);
    cr_expect_eq(1, tm_query_description_contains(this, "Goat", &subqueries[2]));
    cr_expect_neq(0, subqueries[2]);
    cr_expect_eq(1, tm_query_status_contains(this, "new", &subqueries[3]));
    cr_expect_neq(0, subqueries[3]);
    cr_expect_eq(1, tm_query_date_created_after(this, date_created - 1, &subqueries[4]));
    cr_expect_neq(0, subqueries[4]);
    cr_expect_eq(1, tm_query_date_created_before(this, date_created + 1, &subqueries[5]));
    cr_expect_neq(0, subqueries[5]);
    cr_expect_eq(1, tm_query_date_updated_after(this, date_updated - 1, &subqueries[6]));
    cr_expect_neq(0, subqueries[6]);
    cr_expect_eq(1, tm_query_date_updated_before(this, date_updated + 1, &subqueries[7]));
    cr_expect_neq(0, subqueries[7]);

    cr_expect_eq(1, tm_query_and(this, 8, subqueries, &query));
    cr_expect_neq(0, query);
    cr_expect_eq(1, tm_get_tasks(this, query, &tasks, &len));
    cr_expect_neq(0, tasks);
    cr_expect_eq(1, tm_query_free(&query));
    cr_expect_eq(0, query);
    cr_expect_eq(1, len);
    cr_expect_eq(1, tm_task_arr_free(&tasks, len));
    cr_expect_eq(0, tasks);
}

Test(tasks, run_query)
{
    tm_task *task = 0;
    tm_task **tasks = 0;
    tm_query *query = 0;
    tm_query **subqueries = 0;
    unsigned int len = -1;

    cr_expect_eq(1, tm_task_new(&task));
    cr_expect_eq(1, tm_task_set_name(task, "Buy milk"));
    cr_expect_eq(1, tm_task_set_description(task, "Goat milk"));
    cr_expect_eq(1, tm_add_task(this, task));
    cr_expect_eq(1, tm_task_free(&task));

    subqueries = (tm_query **)malloc(sizeof(tm_query *) * 3);
    memset(subqueries, 0, sizeof(tm_query *) * 3);
    cr_expect_eq(1, tm_query_id_equals(this, 1, &subqueries[0]));
    cr_expect_neq(0, subqueries[0]);
    cr_expect_eq(1, tm_query_name_contains(this, "milk", &subqueries[1]));
    cr_expect_neq(0, subqueries[1]);
    cr_expect_eq(1, tm_query_description_contains(this, "Goat", &subqueries[2]));
    cr_expect_neq(0, subqueries[2]);
    cr_expect_eq(1, tm_query_and(this, 3, subqueries, &query));
    cr_expect_neq(0, query);
    cr_expect_eq(1, tm_run_query(this, query, &tasks, &len));
    cr_expect_neq(0, tasks);
    cr_expect_eq(1, tm_query_free(&query));
    cr_expect_eq(0, query);
    cr_expect_eq(1, len);
    cr_expect_eq(1, tm_task_arr_free(&tasks, len));
    cr_expect_eq(0, tasks);
}
