/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <sys/types.h>
#include <unistd.h>

#include <criterion/criterion.h>

#include <tm.h>
#include <sqlite3.h>

static char filename[80];

static void setup(void)
{
    sprintf(filename, "tmp-%d.db", getpid());
}

static void teardown(void)
{
    remove(filename);
}

static int file_exists(char *filename)
{
    FILE *fp;
    fp = fopen(filename, "r");
    if (0 == fp)
    {
        return 0;
    }
    fclose(fp);
    return 1;
}

TestSuite(tm, .init = setup, .fini = teardown);

Test(tm, happy_path)
{
    tm_lib *this = 0;
    tm_storage *storage = 0;

    cr_expect_eq(1, tm_storage_sqlite(filename, &storage));
    cr_assert_neq(0, storage);
    cr_expect_eq(1, tm_new(storage, &this));
    cr_assert_neq(0, this);
    cr_expect_eq(1, file_exists(filename));

    cr_expect_eq(1, tm_free(&this));
    cr_assert_eq(0, this);

    storage = 0;
    cr_expect_eq(1, tm_storage_sqlite(filename, &storage));
    cr_assert_neq(0, storage);
    cr_expect_eq(1, tm_new(storage, &this));
    cr_assert_neq(0, this);
    cr_expect_eq(1, file_exists(filename));

    cr_expect_eq(1, tm_free(&this));
    cr_assert_eq(0, this);
}

static void create_table_task_v1(sqlite3 *db)
{
    char *query = "CREATE TABLE IF NOT EXISTS tasks (" \
        "id INTEGER PRIMARY KEY ASC AUTOINCREMENT DEFAULT 1, " \
        "name TEXT NOT NULL"\
        ")";
    sqlite3_stmt *stmt = 0;

    sqlite3_prepare_v2(db, query, strlen(query), &stmt, 0);
    cr_expect_eq(SQLITE_DONE, sqlite3_step(stmt));
    sqlite3_finalize(stmt);
}

static void create_a_task_v1(sqlite3 *db)
{
    char *query = "INSERT INTO tasks (name) VALUES ('Buy beer')";
    sqlite3_stmt *stmt = 0;

    sqlite3_prepare_v2(db, query, strlen(query), &stmt, 0);
    cr_expect_eq(SQLITE_DONE, sqlite3_step(stmt));
    sqlite3_finalize(stmt);
}

Test(tm, upgrade_v1_v2)
{
    tm_lib *this = 0;
    tm_storage *storage = 0;
    sqlite3 *db = 0;
    char backup_filename[128];

    sprintf(backup_filename, "%s.v1-v2.bak", filename);

    cr_expect_eq(0, sqlite3_open(filename, &db));
    cr_expect_neq(0, db);

    create_table_task_v1(db);
    create_a_task_v1(db);

    sqlite3_close(db);

    cr_expect_eq(1, tm_storage_sqlite(filename, &storage));
    cr_assert_neq(0, storage);
    cr_expect_eq(1, tm_new(storage, &this));
    cr_assert_neq(0, this);

    cr_expect_eq(1, file_exists(filename));
    cr_expect_eq(1, file_exists(backup_filename));

    cr_expect_eq(1, tm_free(&this));
    cr_assert_eq(0, this);

    remove(backup_filename);
}
