/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <tm_dateparser.h>

typedef struct date_parser date_parser;
enum date_parser_state
{
    DPS_NULL,
    DPS_YEAR,
    DPS_MONTH,
    DPS_DAY,
    DPS_HOUR,
    DPS_MINUTE,
    DPS_SECOND,
    DPS_SPECIAL
};

enum date_parser_special
{
    DPT_NULL,
    DPT_FRIDAY,
    DPT_MONDAY,
    DPT_NOW,
    DPT_S,
    DPT_SATURDAY,
    DPT_SUNDAY,
    DPT_T,
    DPT_THURSDAY,
    DPT_TO,
    DPT_TODAY,
    DPT_TOMORROW,
    DPT_TUESDAY,
    DPT_WEDNESDAY,
    DPT_YESTERDAY
};

struct date_parser
{
    enum date_parser_state state;
    unsigned int chars_in_state;

    int year;
    unsigned int month;
    unsigned int day;
    unsigned int hour;
    unsigned int minute;
    unsigned int second;
    enum date_parser_special special;
};

static int make_date_year(date_parser *parser, int is_after, struct tm *atime)
{
    atime->tm_year = parser->year - 1900 + (1 == is_after ? 0 : 1);
    atime->tm_mon = 0;
    atime->tm_mday = 1;
    atime->tm_hour = 0;
    atime->tm_min = 0;
    atime->tm_sec = 0;
    return 1;
}

static int make_date_year_month(date_parser *parser, int is_after, struct tm *atime)
{
    if (0 == parser->chars_in_state || 0 == parser->month)
    {
        return 0;
    }

    atime->tm_year = parser->year - 1900;
    atime->tm_mon = parser->month - 1;
    atime->tm_mday = 1;
    atime->tm_hour = 0;
    atime->tm_min = 0;
    atime->tm_sec = 0;

    if (0 == is_after)
    {
        if (11 == atime->tm_mon)
        {
            atime->tm_year++;
            atime->tm_mon = 0;
        }
        else
        {
            atime->tm_mon++;
        }
    }

    return 1;
}

static int is_leap_year(int year)
{
    int is_400 = (year % 400) == 0;
    int is_100 = (year % 100) == 0;
    int is_4 = (year % 4) == 0;
    int ret = 1 == is_400 || (1 == is_4 && 0 == is_100);
    return ret;
}

static void subtract_day(struct tm *atime)
{
    if (1 == atime->tm_mday)
    {
        if (0 == atime->tm_mon)
        {
            atime->tm_year--;
            atime->tm_mon = 11;
            atime->tm_mday = 31;
        }
        if (2 == atime->tm_mon)
        {
            atime->tm_mon--;
            atime->tm_mday = 1 == is_leap_year(atime->tm_year + 1900) ? 29 : 28;
        }
        else if (3 == atime->tm_mon || 5 == atime->tm_mon || 8 == atime->tm_mon ||
            10 == atime->tm_mon)
        {
            atime->tm_mon--;
            atime->tm_mday = 31;
        }
        else if (1 == atime->tm_mday)
        {
            atime->tm_mon--;
            atime->tm_mday = 30;
        }
    }
    else
    {
        atime->tm_mday--;
    }
}

static void add_day(struct tm *atime)
{
    if (1 == atime->tm_mon &&
        (1 == is_leap_year(atime->tm_year + 1900) ? 29 : 28) == atime->tm_mday)
    {
        atime->tm_mon++;
        atime->tm_mday = 1;
    }
    else if (11 == atime->tm_mon && 31 == atime->tm_mday)
    {
        atime->tm_year++;
        atime->tm_mon = 0;
        atime->tm_mday = 1;
    }
    else if (30 == atime->tm_mday &&
        (3 == atime->tm_mon || 5 == atime->tm_mon || 8 == atime->tm_mon ||
            10 == atime->tm_mon))
    {
        atime->tm_mon++;
        atime->tm_mday = 1;
    }
    else if (31 == atime->tm_mday)
    {
        atime->tm_mon++;
        atime->tm_mday = 1;
    }
    else
    {
        atime->tm_mday++;
    }
}

static int make_date_year_month_day(date_parser *parser, int is_after, struct tm *atime)
{
    if (0 == parser->chars_in_state || 0 == parser->day)
    {
        return 0;
    }

    atime->tm_year = parser->year - 1900;
    atime->tm_mon = parser->month - 1;
    atime->tm_mday = parser->day;
    atime->tm_hour = 0;
    atime->tm_min = 0;
    atime->tm_sec = 0;

    if (0 == is_after)
    {
        add_day(atime);
    }

    return 1;
}

static void add_hour(struct tm *atime)
{
    if (23 == atime->tm_hour)
    {
        add_day(atime);
        atime->tm_hour = 0;
    }
    else
    {
        atime->tm_hour++;
    }
}

static int make_date_year_month_day_hour(date_parser *parser, int is_after, struct tm *atime)
{
    if (0 == parser->chars_in_state)
    {
        return 0;
    }
    atime->tm_year = parser->year - 1900;
    atime->tm_mon = parser->month - 1;
    atime->tm_mday = parser->day;
    atime->tm_hour = parser->hour;
    atime->tm_min = 0;
    atime->tm_sec = 0;

    if (0 == is_after)
    {
        add_hour(atime);
    }

    return 1;
}

static void add_minute(struct tm *atime)
{
    if (59 == atime->tm_min)
    {
        add_hour(atime);
        atime->tm_min = 0;
    }
    else
    {
        atime->tm_min++;
    }
}

static int make_date_year_month_day_hour_minute(date_parser *parser, int is_after, struct tm *atime)
{
    if (0 == parser->chars_in_state)
    {
        return 0;
    }

    atime->tm_year = parser->year - 1900;
    atime->tm_mon = parser->month - 1;
    atime->tm_mday = parser->day;
    atime->tm_hour = parser->hour;
    atime->tm_min = parser->minute;
    atime->tm_sec = 0;

    if (0 == is_after)
    {
        add_minute(atime);
    }

    return 1;
}

static void add_second(struct tm *atime)
{
    if (59 == atime->tm_sec)
    {
        add_minute(atime);
        atime->tm_sec = 0;
    }
    else
    {
        atime->tm_sec++;
    }
}

static int make_date_year_month_day_hour_minute_second(date_parser *parser, int is_after,
    struct tm *atime)
{
    if (0 == parser->chars_in_state)
    {
        return 0;
    }
    atime->tm_year = parser->year - 1900;
    atime->tm_mon = parser->month - 1;
    atime->tm_mday = parser->day;
    atime->tm_hour = parser->hour;
    atime->tm_min = parser->minute;
    atime->tm_sec = parser->second;

    if (0 == is_after)
    {
        add_second(atime);
    }

    return 1;
}

static int make_date_special_today(date_parser *parser, int is_after,
    struct tm *atime)
{
    time_t date = time(0);
    localtime_r(&date, atime);

    if (5 > parser->chars_in_state)
    {
        return 0;
    }
    atime->tm_hour = 0;
    atime->tm_min = 0;
    atime->tm_sec = 0;

    if (0 == is_after)
    {
        add_day(atime);
    }
    return 1;
}

static int make_date_special_tomorrow(date_parser *parser, int is_after,
    struct tm *atime)
{
    time_t date = time(0);
    localtime_r(&date, atime);

    if (8 > parser->chars_in_state)
    {
        return 0;
    }
    atime->tm_hour = 0;
    atime->tm_min = 0;
    atime->tm_sec = 0;

    add_day(atime);
    if (0 == is_after)
    {
        add_day(atime);
    }
    return 1;
}

static int make_date_special_yesterday(date_parser *parser, int is_after,
    struct tm *atime)
{
    time_t date = time(0);
    localtime_r(&date, atime);

    if (9 > parser->chars_in_state)
    {
        return 0;
    }
    atime->tm_hour = 0;
    atime->tm_min = 0;
    atime->tm_sec = 0;

    if (1 == is_after)
    {
        subtract_day(atime);
    }
    return 1;
}

static int make_date_special_wday(date_parser *parser, int is_after, unsigned int chars_in_state,
    unsigned int wday, struct tm *atime)
{
    if (chars_in_state > parser->chars_in_state)
    {
        return 0;
    }
    atime->tm_hour = 0;
    atime->tm_min = 0;
    atime->tm_sec = 0;

    mktime(atime);
    while (wday != atime->tm_wday)
    {
        add_day(atime);
        mktime(atime);
    }
    if (0 == is_after)
    {
        add_day(atime);
    }
    return 1;
}

static int make_date_special(date_parser *parser, int is_after,
    struct tm *atime)
{
    time_t date = time(0);
    localtime_r(&date, atime);

    switch (parser->special)
    {
        case DPT_TODAY:
            return make_date_special_today(parser, is_after, atime);
        case DPT_YESTERDAY:
            return make_date_special_yesterday(parser, is_after, atime);
        case DPT_TOMORROW:
            return make_date_special_tomorrow(parser, is_after, atime);
        case DPT_MONDAY:
            return make_date_special_wday(parser, is_after, 6, 1, atime);
        case DPT_TUESDAY:
            return make_date_special_wday(parser, is_after, 7, 2, atime);
        case DPT_WEDNESDAY:
            return make_date_special_wday(parser, is_after, 8, 3, atime);
        case DPT_THURSDAY:
            return make_date_special_wday(parser, is_after, 8, 4, atime);
        case DPT_FRIDAY:
            return make_date_special_wday(parser, is_after, 6, 5, atime);
        case DPT_SATURDAY:
            return make_date_special_wday(parser, is_after, 8, 6, atime);
        case DPT_SUNDAY:
            return make_date_special_wday(parser, is_after, 6, 0, atime);
        case DPT_NOW:
            if (3 != parser->chars_in_state)
            {
                return 0;
            }
            break;
        default:
            break;
    }
    return 1;
}

static int parse_char_null(date_parser *parser, char c)
{
    parser->state = DPS_SPECIAL;
    parser->chars_in_state++;

    switch (c)
    {
        case 'f':
            parser->special = DPT_FRIDAY;
            break;
        case 'm':
            parser->special = DPT_MONDAY;
            break;
        case 'n':
            parser->special = DPT_NOW;
            break;
        case 's':
            parser->special = DPT_S;
            break;
        case 't':
            parser->special = DPT_T;
            break;
        case 'w':
            parser->special = DPT_WEDNESDAY;
            break;
        case 'y':
            parser->special = DPT_YESTERDAY;
            break;
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            parser->year = (int)(c - '0');
            parser->state = DPS_YEAR;
            parser->chars_in_state++;
            break;
        default:
            return 0;
    }
    return 1;
}

static int parse_char_year(date_parser *parser, char c)
{
    if (0 != parser->chars_in_state && '-' == c)
    {
        parser->state = DPS_MONTH;
        parser->chars_in_state = 0;
        return 1;
    }

    if ('0' > c || '9' < c)
    {
        return 0;
    }

    parser->year = parser->year * 10 + (int)(c - '0');
    parser->chars_in_state++;
    return 1;
}

static int parse_char_month(date_parser *parser, char c)
{
    if (0 != parser->chars_in_state && '-' == c)
    {
        parser->state = DPS_DAY;
        parser->chars_in_state = 0;
        return 1;
    }

    if ('0' > c || '9' < c)
    {
        return 0;
    }
    if (2 == parser->chars_in_state)
    {
        return 0;
    }
    if (1 == parser->chars_in_state)
    {
        if (1 < parser->month)
        {
            return 0;
        }
        if (0 == parser->month && '0' == c)
        {
            return 0;
        }
        if (1 == parser->month && '2' < c)
        {
            return 0;
        }
    }
    parser->month = parser->month * 10 + (int)(c - '0');
    parser->chars_in_state++;
    return 1;
}

static int parse_char_day1(date_parser *parser, char c)
{
    if (0 == parser->day && '0' == c)
    {
        return 0;
    }
    if ((1 == parser->month || 3 == parser->month || 5 == parser->month ||
        7 == parser->month || 8 == parser->month || 10 == parser->month ||
        12 == parser->month) && (3 < parser->day || (3 == parser->day && '1' < c)))
    {
        return 0;
    }
    if ((4 == parser->month || 6 == parser->month || 9 == parser->month ||
        11 == parser->month) && (3 < parser->day || (3 == parser->day && '0' != c)))
    {
        return 0;
    }
    if (2 == parser->month && (2 < parser->day ||
        (2 == parser->day && 0 == is_leap_year(parser->year) && '9' == c)))
    {
        return 0;
    }
    return 1;
}

static int parse_char_day(date_parser *parser, char c)
{
    if (c == ' ' && 0 < parser->chars_in_state)
    {
        parser->state = DPS_HOUR;
        parser->chars_in_state = 0;
        return 1;
    }

    if ('0' > c || '9' < c)
    {
        return 0;
    }


    if (2 == parser->chars_in_state)
    {
        return 0;
    }
    if (1 == parser->chars_in_state && 0 == parse_char_day1(parser, c))
    {
        return 0;
    }
    parser->day = parser->day * 10 + (int)(c - '0');
    parser->chars_in_state++;
    return 1;
}

static int parse_char_hour(date_parser *parser, char c)
{
    if (c == ':' && 0 < parser->chars_in_state)
    {
        parser->state = DPS_MINUTE;
        parser->chars_in_state = 0;
        return 1;
    }

    if ('0' > c || '9' < c)
    {
        return 0;
    }

    if (2 == parser->chars_in_state)
    {
        return 0;
    }
    if (1 == parser->chars_in_state && (2 < parser->hour || (2 == parser->hour && '3' < c)))
    {
        return 0;
    }
    parser->hour = parser->hour * 10 + (int)(c - '0');
    parser->chars_in_state++;
    return 1;
}

static int parse_char_minute(date_parser *parser, char c)
{
    if (c == ':' && 0 < parser->chars_in_state)
    {
        parser->state = DPS_SECOND;
        parser->chars_in_state = 0;
        return 1;
    }

    if ('0' > c || '9' < c)
    {
        return 0;
    }

    if (2 == parser->chars_in_state)
    {
        return 0;
    }
    if (1 == parser->chars_in_state && 5 < parser->minute)
    {
        return 0;
    }
    parser->minute = parser->minute * 10 + (int)(c - '0');
    parser->chars_in_state++;
    return 1;
}

static int parse_char_second(date_parser *parser, char c)
{
    if ('0' > c || '9' < c)
    {
        return 0;
    }

    if (2 == parser->chars_in_state)
    {
        return 0;
    }
    if (1 == parser->chars_in_state && 5 < parser->second)
    {
        return 0;
    }
    parser->second = parser->second * 10 + (int)(c - '0');
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special_monday(date_parser *parser, char c)
{
    char *a = "onday";
    if (5 < parser->chars_in_state || a[parser->chars_in_state - 1] != c)
    {
        return 0;
    }
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special_tuesday(date_parser *parser, char c)
{
    char *a = "esday";
    if (7 < parser->chars_in_state || a[parser->chars_in_state - 2] != c)
    {
        return 0;
    }
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special_wednesday(date_parser *parser, char c)
{
    char *a = "ednesday";
    if (8 < parser->chars_in_state || a[parser->chars_in_state - 1] != c)
    {
        return 0;
    }
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special_thursday(date_parser *parser, char c)
{
    char *a = "ursday";
    if (7 < parser->chars_in_state || a[parser->chars_in_state - 2] != c)
    {
        return 0;
    }
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special_friday(date_parser *parser, char c)
{
    char *a = "riday";
    if (5 < parser->chars_in_state || a[parser->chars_in_state - 1] != c)
    {
        return 0;
    }
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special_saturday(date_parser *parser, char c)
{
    char *a = "turday";
    if (7 < parser->chars_in_state || a[parser->chars_in_state - 2] != c)
    {
        return 0;
    }
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special_sunday(date_parser *parser, char c)
{
    char *a = "nday";
    if (5 < parser->chars_in_state || a[parser->chars_in_state - 2] != c)
    {
        return 0;
    }
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special_now(date_parser *parser, char c)
{
    char *a = "ow";
    if (2 < parser->chars_in_state || a[parser->chars_in_state - 1] != c)
    {
        return 0;
    }
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special_s(date_parser *parser, char c)
{
    if (1 != parser->chars_in_state)
    {
        return 0;
    }

    switch (c)
    {
        case 'a':
            parser->special = DPT_SATURDAY;
            break;
        case 'u':
            parser->special = DPT_SUNDAY;
            break;
        default:
            return 0;
    }
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special_t(date_parser *parser, char c)
{
    if (1 != parser->chars_in_state)
    {
        return 0;
    }

    switch (c)
    {
        case 'h':
            parser->special = DPT_THURSDAY;
            break;
        case 'o':
            parser->special = DPT_TO;
            break;
        case 'u':
            parser->special = DPT_TUESDAY;
            break;
        default:
            return 0;
    }
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special_to(date_parser *parser, char c)
{
    if (2 != parser->chars_in_state)
    {
        return 0;
    }

    switch (c)
    {
        case 'd':
            parser->special = DPT_TODAY;
            break;
        case 'm':
            parser->special = DPT_TOMORROW;
            break;
        default:
            return 0;
    }
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special_today(date_parser *parser, char c)
{
    char *a = "ay";
    if (4 < parser->chars_in_state || a[parser->chars_in_state - 3] != c)
    {
        return 0;
    }
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special_tomorrow(date_parser *parser, char c)
{
    char *a = "orrow";
    if (8 < parser->chars_in_state || a[parser->chars_in_state - 3] != c)
    {
        return 0;
    }
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special_yesterday(date_parser *parser, char c)
{
    char *a = "esterday";
    if (8 < parser->chars_in_state || a[parser->chars_in_state - 1] != c)
    {
        return 0;
    }
    parser->chars_in_state++;
    return 1;
}

static int parse_char_special(date_parser *parser, char c)
{
    switch (parser->special)
    {
        case DPT_FRIDAY:
            return parse_char_special_friday(parser, c);
        case DPT_MONDAY:
            return parse_char_special_monday(parser, c);
        case DPT_NOW:
            return parse_char_special_now(parser, c);
        case DPT_S:
            return parse_char_special_s(parser, c);
        case DPT_SATURDAY:
            return parse_char_special_saturday(parser, c);
        case DPT_SUNDAY:
            return parse_char_special_sunday(parser, c);
        case DPT_THURSDAY:
            return parse_char_special_thursday(parser, c);
        case DPT_T:
            return parse_char_special_t(parser, c);
        case DPT_TO:
            return parse_char_special_to(parser, c);
        case DPT_TODAY:
            return parse_char_special_today(parser, c);
        case DPT_TOMORROW:
            return parse_char_special_tomorrow(parser, c);
        case DPT_TUESDAY:
            return parse_char_special_tuesday(parser, c);
        case DPT_WEDNESDAY:
            return parse_char_special_wednesday(parser, c);
        case DPT_YESTERDAY:
            return parse_char_special_yesterday(parser, c);
        default:
            return 0;
    }
}

static int parse_char(date_parser *parser, char c)
{
    int ret = 0;

    switch (parser->state)
    {
        case DPS_NULL:
            ret = parse_char_null(parser, c);
            break;

        case DPS_YEAR:
            ret = parse_char_year(parser, c);
            break;

        case DPS_MONTH:
            ret = parse_char_month(parser, c);
            break;

        case DPS_DAY:
            ret = parse_char_day(parser, c);
            break;

        case DPS_HOUR:
            ret = parse_char_hour(parser, c);
            break;

        case DPS_MINUTE:
            ret = parse_char_minute(parser, c);
            break;

        case DPS_SECOND:
            ret = parse_char_second(parser, c);
            break;

        case DPS_SPECIAL:
            ret = parse_char_special(parser, c);
            break;

        default:
            break;
    }
    return ret;
}

static int make_date(date_parser *parser, int is_after, time_t *pdate)
{
    int ret = 0;
    struct tm atime;

    memset(&atime, 0, sizeof(atime));

    switch (parser->state)
    {
        case DPS_YEAR:
            ret = make_date_year(parser, is_after, &atime);
            break;

        case DPS_MONTH:
            ret = make_date_year_month(parser, is_after, &atime);
            break;

        case DPS_DAY:
            ret = make_date_year_month_day(parser, is_after, &atime);
            break;

        case DPS_HOUR:
            ret = make_date_year_month_day_hour(parser, is_after, &atime);
            break;

        case DPS_MINUTE:
            ret = make_date_year_month_day_hour_minute(parser, is_after, &atime);
            break;

        case DPS_SECOND:
            ret = make_date_year_month_day_hour_minute_second(parser, is_after, &atime);
            break;

        case DPS_SPECIAL:
            ret = make_date_special(parser, is_after, &atime);
            break;

        default:
            break;
    }
    if (1 == ret)
    {
        (*pdate) = mktime(&atime);
    }
    return ret;
}

int tm_date_parser_parse(unsigned int len, char *str, enum tm_date_parser_mode mode, time_t *pdate)
{
    date_parser parser;
    memset(&parser, 0, sizeof(date_parser));
    parser.state = DPS_NULL;
    parser.chars_in_state = 0;
    int is_after = 0;

    switch (mode)
    {
        case TDPM_WANT_LATER_DATES:
            is_after = 1;
            break;
        case TDPM_WANT_EARLIER_DATES:
            break;
        default:
            return 0;
    }

    for (unsigned int i = 0; i < len; i++)
    {
        if (0 == parse_char(&parser, str[i]))
        {
            return 0;
        }
    }
    return make_date(&parser, is_after, pdate);
}
