/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <stdio.h>
#include <time.h>

#include <criterion/criterion.h>

#include <tm_dateparser.h>

static char *tz;

void cache_timezone(void)
{
    tz = getenv("TZ");
    cr_log_info("Caching time zone: <%s>", tz);
    setenv("TZ", "", 1);
    tzset();
}

void restore_timezone(void)
{
    cr_log_info("Restoring time zone: <%s>", tz);
    if (0 != tz)
    {
        setenv("TZ", tz, 1);
    }
    else
    {
        unsetenv("TZ");
    }
    tzset();
}

TestSuite(dateparser, .init = cache_timezone, .fini = restore_timezone);

static void print_date(time_t date)
{
    struct tm atime;
    char tmp[64];
    memset(&atime, 0, sizeof(atime));
    localtime_r(&date, &atime);
    strftime(tmp, sizeof(tmp), "%c", &atime);
    cr_log_info("Date: %s", tmp);
}

static void dategt(char *a, time_t b)
{
    time_t ad = 0;

    cr_log_info("Parse date for comparison EARLIER: %s", a);
    cr_expect_eq(1, tm_date_parser_parse(strlen(a), a, TDPM_WANT_EARLIER_DATES, &ad));

    cr_log_info("Second date is earlier than first:");
    print_date(ad);
    print_date(b);
    cr_expect_geq(ad, b, "\"%s\" (timestamp %ld) >= timestamp %ld", a, ad, b);
}

static void datelt(char *a, time_t b)
{
    time_t ad = 0;

    cr_log_info("Parse date for comparison LATER: %s", a);
    cr_expect_eq(1, tm_date_parser_parse(strlen(a), a, TDPM_WANT_LATER_DATES, &ad));

    cr_log_info("Second date is later than first:");
    print_date(ad);
    print_date(b);
    cr_expect_leq(ad, b, "\"%s\" (timestamp %ld) <= timestamp %ld", a, ad, b);
}

static void invalid(char *a)
{
    time_t ad = 0;

    cr_expect_eq(0, tm_date_parser_parse(strlen(a), a, TDPM_WANT_LATER_DATES, &ad));
}

static time_t timeof(int year, int month, int day, int hour, int min, int sec)
{
    struct tm atime;
    memset(&atime, 0, sizeof(atime));
    atime.tm_year = year - 1900;
    atime.tm_mon = month - 1;
    atime.tm_mday = day;
    atime.tm_hour = hour;
    atime.tm_min = min;
    atime.tm_sec = sec;
    return mktime(&atime);
}

Test(dateparser, am_i_going_insane)
{
    time_t epoch = timeof(1970, 1, 1, 0, 0, 0);
    cr_expect_eq(0, epoch, "0 is not %ld. Are we in UTC timezone???", epoch);
}

Test(dateparser, d_1970_gt_1)
{
    datelt("1", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_gt_1900)
{
    datelt("1900", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_gt_1950)
{
    datelt("1950", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_gt_1969)
{
    datelt("1969", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_gt_1969_12)
{
    datelt("1969-12", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_gt_1969_12_31)
{
    datelt("1969-12-31", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_eq_self)
{
    datelt("1970", timeof(1970, 1, 1, 0, 0, 0));
    dategt("1970", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_eq_1970_01)
{
    datelt("1970-01", timeof(1970, 1, 1, 0, 0, 0));
    dategt("1970-01", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_eq_1970_01_01)
{
    datelt("1970-01-01", timeof(1970, 1, 1, 0, 0, 0));
    dategt("1970-01-01", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_lt_1970_02)
{
    dategt("1970-02", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_lt_1970_02_01)
{
    dategt("1970-02-01", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_lt_1971)
{
    dategt("1971", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_lt_2000)
{
    dategt("2000", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_lt_12345)
{
    dategt("12345", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_02_28_gt_1970_02_27)
{
    datelt("1970-02-27", timeof(1970, 2, 28, 0, 0, 0));
}

Test(dateparser, d_1970_02_28_eq_self)
{
    datelt("1970-02-28", timeof(1970, 2, 28, 0, 0, 0));
    dategt("1970-02-28", timeof(1970, 2, 28, 0, 0, 0));
}

Test(dateparser, d_1970_02_28_lt_1970_03_01)
{
    dategt("1970-03-01", timeof(1970, 2, 28, 0, 0, 0));
}

Test(dateparser, d_1972_02_29_gt_1972_02_28)
{
    dategt("1972-02-28", timeof(1972, 2, 29, 0, 0, 0));
}

Test(dateparser, d_1972_02_29_eq_self)
{
    datelt("1972-02-29", timeof(1972, 2, 29, 0, 0, 0));
    dategt("1972-02-29", timeof(1972, 2, 29, 0, 0, 0));
}

Test(dateparser, d_1972_02_29_lt_1972_03_01)
{
    dategt("1972-03-01", timeof(1972, 2, 29, 0, 0, 0));
}

Test(dateparser, d_1970_01_31_gt_1970_01_30)
{
    dategt("1970-01-30", timeof(1970, 1, 31, 0, 0, 0));
}

Test(dateparser, d_1970_01_31_eq_self)
{
    datelt("1970-01-31", timeof(1970, 1, 31, 0, 0, 0));
    dategt("1970-01-31", timeof(1970, 1, 31, 0, 0, 0));
}

Test(dateparser, d_1970_01_31_lt_1970_02_01)
{
    dategt("1970-02-01", timeof(1970, 1, 31, 0, 0, 0));
}

Test(dateparser, d_1970_04_30_lt_1970_04_29)
{
    dategt("1970-04-29", timeof(1970, 4, 30, 0, 0, 0));
}

Test(dateparser, d_1970_04_30_eq_self)
{
    datelt("1970-04-30", timeof(1970, 4, 30, 0, 0, 0));
    dategt("1970-04-30", timeof(1970, 4, 30, 0, 0, 0));
}

Test(dateparser, d_1970_04_30_lt_1970_05_01)
{
    dategt("1970-05-01", timeof(1970, 4, 30, 0, 0, 0));
}

Test(dateparser, d_1970_01_01_00_gt_1969_12_31_23)
{
    datelt("1969-12-31 23", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_01_01_00_eq_self)
{
    datelt("1970-01-01 00", timeof(1970, 1, 1, 0, 0, 0));
    dategt("1970-01-01 00", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_01_01_00_lt_1970_01_01_01)
{
    dategt("1970-01-01 01", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_01_01_00_00_gt_1969_12_31_59)
{
    datelt("1969-12-31 23:59", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_01_01_00_00_eq_self)
{
    datelt("1970-01-01 00:00", timeof(1970, 1, 1, 0, 0, 0));
    dategt("1970-01-01 00:00", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_01_01_00_00_lt_1970_01_01_00_01)
{
    dategt("1970-01-01 00:01", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_01_01_00_00_00_gt_1969_12_31_59_59)
{
    datelt("1969-12-31 23:59:59", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_01_01_00_00_00_eq_self)
{
    datelt("1970-01-01 00:00:00", timeof(1970, 1, 1, 0, 0, 0));
    dategt("1970-01-01 00:00:00", timeof(1970, 1, 1, 0, 0, 0));
}

Test(dateparser, d_1970_01_01_00_00_00_lt_1970_01_01_00_00_01)
{
    dategt("1970-01-01 00:00:01", timeof(1970, 1, 1, 0, 0, 0));
}

/*
 * @todo weekdays, months, 1st-31th
 */

Test(dateparser, d_now)
{
    time_t now = time(0);
    time_t now_minus_5_seconds = now - 5;
    time_t now_plus_5_seconds = now + 5;

    dategt("now", now_minus_5_seconds);
    datelt("now", now_plus_5_seconds);
}

Test(dateparser, d_today)
{
    time_t now = time(0);
    struct tm nowtm;

    memset(&nowtm, 0, sizeof(struct tm));

    localtime_r(&now, &nowtm);
    nowtm.tm_hour = 0;
    nowtm.tm_min = 0;
    nowtm.tm_sec = 0;

    dategt("today", mktime(&nowtm));

    nowtm.tm_hour = 23;
    nowtm.tm_min = 59;
    nowtm.tm_sec = 59;
    datelt("today", mktime(&nowtm));
}

static void subtract_day(struct tm *atime)
{
    if (1 == atime->tm_mday)
    {
        if (0 == atime->tm_mon)
        {
            atime->tm_year--;
            atime->tm_mon = 11;
            atime->tm_mday = 31;
        }
        else if (2 == atime->tm_mon)
        {
            atime->tm_mon--;
            /* @todo leap year? */
            atime->tm_mday = 28;
        }
        else if (3 == atime->tm_mon || 5 == atime->tm_mon || 8 == atime->tm_mon ||
            10 == atime->tm_mon)
        {
            atime->tm_mon--;
            atime->tm_mday = 31;
        }
        else
        {
            atime->tm_mon--;
            atime->tm_mday = 30;
        }
    }
    else
    {
        atime->tm_mday--;
    }
    mktime(atime);
}

static void add_day(struct tm *atime)
{
    if (11 == atime->tm_mon && 31 == atime->tm_mday)
    {
        atime->tm_year++;
        atime->tm_mon = 0;
        atime->tm_mday = 1;
    }
    else if (1 == atime->tm_mon && 28 <= atime->tm_mday)
    {
        atime->tm_mon++;
        atime->tm_mday = 1;
    }
    else if (30 == atime->tm_mday && (
        3 == atime->tm_mon || 5 == atime->tm_mon || 8 == atime->tm_mon ||
            10 == atime->tm_mon))
    {
        atime->tm_mon++;
        atime->tm_mday = 1;
    }
    else if (31 == atime->tm_mday)
    {
        atime->tm_mon++;
        atime->tm_mday = 1;
    }
    else
    {
        atime->tm_mday++;
    }
    mktime(atime);
}

Test(dateparser, d_yesterday)
{
    time_t now = time(0);
    struct tm nowtm;

    memset(&nowtm, 0, sizeof(struct tm));

    localtime_r(&now, &nowtm);
    nowtm.tm_hour = 0;
    nowtm.tm_min = 0;
    nowtm.tm_sec = 0;

    datelt("yesterday", mktime(&nowtm));

    subtract_day(&nowtm);
    dategt("yesterday", mktime(&nowtm));
}

Test(dateparser, d_tomorrow)
{
    time_t now = time(0);
    struct tm nowtm;

    memset(&nowtm, 0, sizeof(struct tm));

    localtime_r(&now, &nowtm);
    nowtm.tm_hour = 0;
    nowtm.tm_min = 0;
    nowtm.tm_sec = 0;

    add_day(&nowtm);
    datelt("tomorrow", mktime(&nowtm));

    add_day(&nowtm);
    dategt("tomorrow", mktime(&nowtm));
}

static void add_days_till_it_is_wday(struct tm *atime, unsigned int wday)
{
    mktime(atime);
    while (wday != atime->tm_wday)
    {
        add_day(atime);
    }
}

static void datewday(char *name, unsigned int wday)
{
    time_t now = time(0);
    struct tm nowtm;

    memset(&nowtm, 0, sizeof(struct tm));

    localtime_r(&now, &nowtm);
    nowtm.tm_hour = 0;
    nowtm.tm_min = 0;
    nowtm.tm_sec = 0;

    add_days_till_it_is_wday(&nowtm, wday);
    datelt(name, mktime(&nowtm));

    add_day(&nowtm);
    dategt(name, mktime(&nowtm));
}

Test(dateparser, d_monday)
{
    datewday("monday", 1);
}

Test(dateparser, d_tuesday)
{
    datewday("tuesday", 2);
}

Test(dateparser, d_wednesday)
{
    datewday("wednesday", 3);
}

Test(dateparser, d_thursday)
{
    datewday("thursday", 4);
}

Test(dateparser, d_friday)
{
    datewday("friday", 5);
}

Test(dateparser, d_saturday)
{
    datewday("saturday", 6);
}

Test(dateparser, d_sunday)
{
    datewday("sunday", 0);
}

Test(dateparser, invalid_empty)
{
    invalid("");
}

Test(dateparser, invalid_x)
{
    invalid("x");
}

Test(dateparser, invalid_0x)
{
    invalid("0x");
}

Test(dateparser, invalid_1970_minus)
{
    invalid("1970-");
}

Test(dateparser, invalid_1970_x)
{
    invalid("1970-x");
}

Test(dateparser, invalid_1970_0x)
{
    invalid("1970-0x");
}

Test(dateparser, invalid_1970_00)
{
    invalid("1970-00");
}

Test(dateparser, invalid_1970_123)
{
    invalid("1970-123");
}

Test(dateparser, invalid_1970_13)
{
    invalid("1970-13");
}

Test(dateparser, invalid_1970_20)
{
    invalid("1970-20");
}

Test(dateparser, invalid_1970_01_minus)
{
    invalid("1970-01-");
}

Test(dateparser, invalid_1970_01_x)
{
    invalid("1970-01-x");
}

Test(dateparser, invalid_1970_01_0)
{
    invalid("1970-01-0");
}

Test(dateparser, invalid_1970_01_00)
{
    invalid("1970-01-00");
}

Test(dateparser, invalid_1970_01_123)
{
    invalid("1970-01-123");
}

Test(dateparser, invalid_1970_01_32)
{
    invalid("1970-01-32");
}

Test(dateparser, invalid_1970_02_29)
{
    invalid("1970-02-29");
}

Test(dateparser, invalid_1972_02_30)
{
    invalid("1970-02-30");
}

Test(dateparser, invalid_1970_03_32)
{
    invalid("1970-03-32");
}

Test(dateparser, invalid_1970_04_31)
{
    invalid("1970-04-31");
}

Test(dateparser, invalid_1970_05_32)
{
    invalid("1970-05-32");
}

Test(dateparser, invalid_1970_06_31)
{
    invalid("1970-06-31");
}

Test(dateparser, invalid_1970_07_32)
{
    invalid("1970-07-32");
}

Test(dateparser, invalid_1970_08_32)
{
    invalid("1970-08-32");
}

Test(dateparser, invalid_1970_09_31)
{
    invalid("1970-09-31");
}

Test(dateparser, invalid_1970_10_32)
{
    invalid("1970-10-32");
}

Test(dateparser, invalid_1970_11_31)
{
    invalid("1970-11-31");
}

Test(dateparser, invalid_1970_12_32)
{
    invalid("1970-12-32");
}

Test(dateparser, invalid_1970_01_01_space)
{
    invalid("1970-01-01 ");
}

Test(dateparser, invalid_1970_01_01_x)
{
    invalid("1970-01-01 x");
}

Test(dateparser, invalid_1970_01_01_0x)
{
    invalid("1970-01-01 0x");
}

Test(dateparser, invalid_1970_01_01_00x)
{
    invalid("1970-01-01 00x");
}

Test(dateparser, invalid_1970_01_01_001)
{
    invalid("1970-01-01 001");
}

Test(dateparser, invalid_1970_01_01_24)
{
    invalid("1970-01-01 24");
}

Test(dateparser, invalid_1970_01_01_invalid_hours)
{
    for (unsigned int i = 24; i < 100; i++)
    {
        char tmp[15];
        sprintf(tmp, "1970-01-01 %u", i);
        invalid(tmp);
    }
}

Test(dateparser, invalid_1970_01_01_00_colon)
{
    invalid("1970-01-01 00:");
}

Test(dateparser, invalid_1970_01_01_00_x)
{
    invalid("1970-01-01 00:x");
}

Test(dateparser, invalid_1970_01_01_00_0x)
{
    invalid("1970-01-01 00:0x");
}

Test(dateparser, invalid_1970_01_01_00_00x)
{
    invalid("1970-01-01 00:00x");
}

Test(dateparser, invalid_1970_01_01_00_001)
{
    invalid("1970-01-01 00:001");
}

Test(dateparser, invalid_1970_01_01_00_60)
{
    invalid("1970-01-01 00:60");
}

Test(dateparser, invalid_1970_01_01_00_invalid_minutes)
{
    for (unsigned int i = 60; i < 100; i++)
    {
        char tmp[18];
        sprintf(tmp, "1970-01-01 00:%u", i);
        invalid(tmp);
    }
}

Test(dateparser, invalid_1970_01_01_00_00_colon)
{
    invalid("1970-01-01 00:00:");
}

Test(dateparser, invalid_1970_01_01_00_00_x)
{
    invalid("1970-01-01 00:00:x");
}

Test(dateparser, invalid_1970_01_01_00_00_0x)
{
    invalid("1970-01-01 00:00:0x");
}

Test(dateparser, invalid_1970_01_01_00_00_00x)
{
    invalid("1970-01-01 00:00:00x");
}

Test(dateparser, invalid_1970_01_01_00_00_001)
{
    invalid("1970-01-01 00:00:001");
}

Test(dateparser, invalid_1970_01_01_00_00_60)
{
    invalid("1970-01-01 00:00:60");
}

Test(dateparser, invalid_1970_01_01_00_00_invalid_seconds)
{
    for (unsigned int i = 60; i < 100; i++)
    {
        char tmp[21];
        sprintf(tmp, "1970-01-01 00:00:%u", i);
        invalid(tmp);
    }
}
