/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

const request = require('browser-request');

const doQuery = (opts) => {
  return new Promise((resolve, reject) => {
    const args = {
      url: '/api/query',
    };
    if (opts !== undefined && Object.keys(opts).length > 0) {
      args.method = 'POST';
      args.body = JSON.stringify(opts);
      args.headers = {'Content-Type': 'application/json'};
    }
    request(args, (err, res) => {
      if (err) {
        reject(err);
      } else if (res.statusCode !== 200) {
        reject(new Error('Server replied with wrong code ' + res.statusCode));
      } else {
        try {
          resolve(JSON.parse(res.body));
        } catch (e) {
          reject(e);
        }
      }
    });
  });
};

const doTask = (id) => {
  return new Promise((resolve, reject) => {
    const args = {
      url: '/api/task/' + id,
    };
    request(args, (err, res) => {
      if (err) {
        reject(err);
      } else if (res.statusCode === 404) {
        resolve(null);
      } else {
        try {
          resolve(JSON.parse(res.body));
        } catch (e) {
          reject(e);
        }
      }
    });
  });
};

const doAdd = (opts) => {
  return new Promise((resolve, reject) => {
    if (opts === undefined || Object.keys(opts).length == 0) {
      reject(new Error('opts must be non-empty object'));
      return;
    }
    const args = {
      url: '/api/add',
      headers: {'Content-Type': 'application/json'},
      method: 'POST',
      body: JSON.stringify(opts),
    };
    request(args, (err, res) => {
      if (err || res.statusCode !== 200) {
        reject(err);
      } else {
        resolve(JSON.parse(res.body));
      }
    });
  });
};

const doRemove = (opts) => {
  return new Promise((resolve, reject) => {
    if (opts === undefined || Object.keys(opts).length == 0) {
      reject(new Error('opts must be non-empty object'));
      return;
    }
    const args = {
      url: '/api/remove',
      headers: {'Content-Type': 'application/json'},
      method: 'POST',
      body: JSON.stringify(opts),
    };
    request(args, (err, res) => {
      if (err || res.statusCode !== 200) {
        reject(err);
      } else {
        resolve();
      }
    });
  });
};

const doModify = (filters, modifiers) => {
  return new Promise((resolve, reject) => {
    if (filters === undefined || Object.keys(filters).length == 0) {
      reject(new Error('filters must be non-empty object'));
      return;
    }
    if (modifiers === undefined || Object.keys(modifiers).length == 0) {
      reject(new Error('modifiers must be non-empty object'));
      return;
    }
    const requestArgs = {
      url: '/api/modify',
      headers: {'Content-Type': 'application/json'},
      method: 'POST',
      body: JSON.stringify({'filters': filters, 'modifiers': modifiers}),
    };
    request(requestArgs, (err, res) => {
      if (err || res.statusCode !== 200) {
        reject(err);
      } else {
        resolve();
      }
    });
  });
};

const convertNameToHumanReadable = (name) => {
  const lut = {
    name: 'Name',
    description: 'Description',
    status: 'Status',
  };
  return lut[name] !== undefined ? lut[name] : name;
};

const addColumns = (theDocument, tr, columns) => {
  for (let i = 0; i < columns.length; i++) {
    const td = theDocument.createElement('td');
    tr.appendChild(td);

    const humanName = convertNameToHumanReadable(columns[i]);
    td.appendChild(theDocument.createTextNode(humanName));
  }
};

const addRow = (theDocument, table, columns, item) => {
  const tr = theDocument.createElement('tr');
  table.appendChild(tr);

  let td;
  let a;

  td = theDocument.createElement('td');
  tr.appendChild(td);
  a = theDocument.createElement('a');
  td.appendChild(a);
  a.setAttribute('href', '/task/' + item.id + '.html');
  a.setAttribute('onclick', 'uiShowTask(' + item.id + ', window);return false;');
  a.appendChild(theDocument.createTextNode('#' + item.id));

  td = theDocument.createElement('td');
  tr.appendChild(td);
  a = theDocument.createElement('a');
  td.appendChild(a);
  a.setAttribute('href', '/task/' + item.id + '.html');
  a.setAttribute('onclick', 'uiShowTask(' + item.id + ', window);return false;');
  a.appendChild(theDocument.createTextNode(item.name));

  for (let i = 0; i < columns.length; i++) {
    td = theDocument.createElement('td');
    tr.appendChild(td);

    const value = item[columns[i]];
    td.appendChild(theDocument.createTextNode(value));
  }
};

const removeChildNodes = (node) => {
  while (node.childNodes.length > 0) {
    node.removeChild(node.childNodes[node.childNodes.length - 1]);
  }
};


const removeChildren = (node, min) => {
  while (node.children.length > min) {
    node.removeChild(node.children[node.children.length - 1]);
  }
};

const getChildById = (parent, id) => {
  for (let i = 0; i < parent.children.length; i++) {
    let child = parent.children[i];
    if (child.getAttribute('id') == id) {
      return child;
    }
    if (child.children !== undefined && child.children.length > 0 &&
      (child = getChildById(child, id))) {
      return child;
    }
  }
  return undefined;
};

const clearDivs = (theDocument) => {
  const addDiv = theDocument.getElementById('add');
  const editDiv = theDocument.getElementById('edit');
  const indexDiv = theDocument.getElementById('index');
  const taskDiv = theDocument.getElementById('task');

  const errorDiv = theDocument.getElementById('error');
  const warningDiv = theDocument.getElementById('warning');

  addDiv.style.display = 'none';
  editDiv.style.display = 'none';
  indexDiv.style.display = 'none';
  taskDiv.style.display = 'none';
  errorDiv.style.display = 'none';
  warningDiv.style.display = 'none';

  removeChildNodes(errorDiv);
  removeChildNodes(warningDiv);
};

const setError = (theDocument, closure) => {
  const errorDiv = theDocument.getElementById('error');
  errorDiv.style.display = '';
  closure(errorDiv);
};

const setWarning = (theDocument, closure) => {
  const warningDiv = theDocument.getElementById('warning');
  warningDiv.style.display = '';
  closure(warningDiv);
};

const uiPaintIndex = (ret, theWindow) => {
  const theDocument = theWindow.document;
  const indexDiv = theDocument.getElementById('index');

  const resultsTable = getChildById(indexDiv, 'results');
  const columnsTr = getChildById(indexDiv, 'columns');
  const rowsTBody = getChildById(indexDiv, 'rows');

  clearDivs(theDocument);
  indexDiv.style.display = '';
  removeChildren(columnsTr, 2);
  removeChildNodes(rowsTBody);

  const filters = ret.filters;
  if (filters) {
    const queryForm = getChildById(indexDiv, 'query');
    const formElements = queryForm.elements;

    [
      'id-equals',
      'name-contains',
      'description-contains',
      'status-contains',
      'date-created-after',
      'date-created-before',
      'date-updated-after',
      'date-updated-before',
    ].forEach((name) => {
      formElements.namedItem('filter-' + name).value = filters[name] !== undefined ?
        filters[name] : '';
    });
  }

  if (ret.error !== undefined) {
    resultsTable.style.display = 'none';
    setError(theDocument, (div) => {
      div.appendChild(theDocument.createTextNode(ret.error));
    });
  } else {
    resultsTable.style.display = '';

    const columns = ret.columns;
    const rows = ret.rows;

    addColumns(theDocument, columnsTr, columns);

    for (let i = 0; i < rows.length; i++) {
      const row = rows[i];
      addRow(theDocument, rowsTBody, columns, row);
    }
  }
};

const setTdTo = (theDocument, td, value) => {
  removeChildNodes(td);
  if (value !== undefined && value !== '') {
    td.appendChild(theDocument.createTextNode(value));
  }
};

const setTdToDate = (theDocument, td, date) => {
  const dateObj = new Date(1000 * date);
  const dateStr = '' + dateObj.getFullYear() + '-' +
    twoDigit(dateObj.getMonth() + 1) + '-' +
    twoDigit(dateObj.getDate()) + ' ' +
    twoDigit(dateObj.getHours()) + ':' +
    twoDigit(dateObj.getMinutes()) + ':' +
    twoDigit(dateObj.getSeconds());
  setTdTo(theDocument, td, dateStr);
};

const twoDigit = (number) => {
  return (number < 10 ? '0' : '') + number;
};

const appendHistory = (theDocument, historyTBody, history) => {
  if (history === undefined || history.length === 0) {
    return;
  }

  for (let i = 0; i < history.length; i++) {
    const item = history[i];

    const tr = theDocument.createElement('tr');
    historyTBody.appendChild(tr);

    const date = new Date(1000 * item.date);
    const dateStr = '' + date.getFullYear() + '-' +
      twoDigit(date.getMonth() + 1) + '-' +
      twoDigit(date.getDate()) + ' ' +
      twoDigit(date.getHours()) + ':' +
      twoDigit(date.getMinutes()) + ':' +
      twoDigit(date.getSeconds());
    let td = theDocument.createElement('td');
    tr.appendChild(td);
    td.appendChild(theDocument.createTextNode(dateStr));

    const typeInPast = item.type == 'add' ? 'added' : item.type + 'd';
    td = theDocument.createElement('td');
    tr.appendChild(td);
    td.appendChild(theDocument.createTextNode(typeInPast));

    const bName = theDocument.createElement('b');
    const humanName = convertNameToHumanReadable(item.name);
    td = theDocument.createElement('td');
    tr.appendChild(td);
    td.appendChild(bName);
    bName.appendChild(theDocument.createTextNode(humanName));

    const bValue = theDocument.createElement('u');
    td = theDocument.createElement('td');
    tr.appendChild(td);
    td.appendChild(bValue);
    bValue.appendChild(theDocument.createTextNode(item.value));

    if (item.type === 'change') {
      td.appendChild(theDocument.createTextNode(' -> '));

      const bNewValue = theDocument.createElement('u');
      td.appendChild(bNewValue);
      bNewValue.appendChild(theDocument.createTextNode(item['new-value']));
    }
  }
};

const uiPaintTask = (ret, theWindow) => {
  const theDocument = theWindow.document;
  const taskDiv = theDocument.getElementById('task');

  const dateCreatedTd = getChildById(taskDiv, 'date-created');
  const dateUpdatedTd = getChildById(taskDiv, 'date-updated');
  const descriptionTd = getChildById(taskDiv, 'description');
  const editA = getChildById(taskDiv, 'edit');
  const historyTBody = getChildById(taskDiv, 'history');
  const titleH1 = getChildById(taskDiv, 'title');

  const statusTd = getChildById(taskDiv, 'status');

  clearDivs(theDocument);
  removeChildNodes(titleH1);
  removeChildNodes(historyTBody);

  if (ret.error !== undefined) {
    taskDiv.style.display = 'none';
    setError(theDocument, (div) => {
      div.appendChild(theDocument.createTextNode(ret.error));
    });
  } else {
    taskDiv.style.display = '';

    const task = ret.task;

    const a = theDocument.createElement('a');
    a.setAttribute('href', '/task/' + task.id + '.html');
    a.setAttribute('onclick', 'uiShowTask(' + task.id + ', window);return false;');
    a.appendChild(theDocument.createTextNode('#' + task.id));

    titleH1.appendChild(a);
    titleH1.appendChild(theDocument.createTextNode(' - ' + task.name));

    editA.setAttribute('href', '/edit/' + task.id + '.html');
    editA.setAttribute('onclick', 'uiShowEdit(' + task.id + ', window);return false;');

    setTdToDate(theDocument, dateCreatedTd, task['date-created']);
    setTdToDate(theDocument, dateUpdatedTd, task['date-updated']);
    setTdTo(theDocument, descriptionTd, task.description);
    setTdTo(theDocument, statusTd, task.status);

    appendHistory(theDocument, historyTBody, task.history);
  }
};

const uiShowAdd = (theWindow) => {
  const opts = {
    status: 'new',
  };
  uiPaintAdd(opts, theWindow);
  return uiPushHistory({
    add: opts,
  }, '/add.html', theWindow);
};

const uiPaintAdd = (opts, theWindow) => {
  const theDocument = theWindow.document;
  const addDiv = theDocument.getElementById('add');

  const addForm = getChildById(addDiv, 'add');
  const formElements = addForm.elements;

  const name = formElements.namedItem('name');
  const description = formElements.namedItem('description');
  const status = formElements.namedItem('status');

  clearDivs(theDocument);
  addDiv.style.display = '';

  if (opts.error !== undefined) {
    setError(theDocument, (div) => {
      div.appendChild(theDocument.createTextNode(opts.error));
    });
  }

  if (opts.warning !== undefined) {
    switch (opts.warning.type) {
      case 'already-added':
        const task = opts.warning.task;
        setWarning(theDocument, (div) => {
          div.appendChild(theDocument.createTextNode('This task was already added as '));

          const a = theDocument.createElement('a');
          div.appendChild(a);
          a.setAttribute('href', '/task/' + task + '.html');
          a.setAttribute('onclick', 'uiShowTask(' + task + ', window);return false;');
          a.appendChild(theDocument.createTextNode('#' + task));
        });
        break;
      default:
        break;
    }
  }

  name.value = opts.name !== undefined ? opts.name : '';
  description.value = opts.description !== undefined ? opts.description : '';
  status.value = opts.status !== undefined ? opts.status : '';
};

const uiShowEdit = (id, theWindow) => {
  const url = '/edit/' + id + '.html';
  return doTask(id).then((ret) => {
    if (ret === null) {
      const err = 'No such task: ' + id;
      uiPaintEdit({error: err}, theWindow);
      return uiPushHistory({
        edit: {
          error: err,
        },
      }, url, theWindow);
    }

    const opts = ret;
    opts.id = id;
    uiPaintEdit(opts, theWindow);
    return uiPushHistory({
      edit: opts,
    }, url, theWindow);
  }, (err) => {
    uiPaintTask({error: err.message}, theWindow);
    return uiPushHistory({
      edit: {
        error: err.message,
      },
    }, url, theWindow);
  });
};

const uiPaintEdit = (opts, theWindow) => {
  const theDocument = theWindow.document;
  const editDiv = theDocument.getElementById('edit');

  clearDivs(theDocument);

  if (opts.error !== undefined) {
    setError(theDocument, (div) => {
      div.appendChild(theDocument.createTextNode(opts.error));
    });
  }

  if (opts.id === undefined) {
    return;
  }

  const titleH1 = getChildById(editDiv, 'title');
  const editForm = getChildById(editDiv, 'edit');
  const formElements = editForm.elements;

  const name = formElements.namedItem('name');
  const description = formElements.namedItem('description');
  const status = formElements.namedItem('status');

  editDiv.style.display = '';
  removeChildNodes(titleH1);

  const a = theDocument.createElement('a');
  a.setAttribute('href', '/task/' + opts.id + '.html');
  a.setAttribute('onclick', 'uiShowTask(' + opts.id + ', window);return false;');
  a.appendChild(theDocument.createTextNode('#' + opts.id));

  titleH1.appendChild(theDocument.createTextNode('Edit '));
  titleH1.appendChild(a);
  titleH1.appendChild(theDocument.createTextNode(' - ' + opts.name));

  editForm.setAttribute('action', '/edit/' + opts.id + '.html');
  editForm.setAttribute('onsubmit', 'uiSubmitEdit(' + opts.id + ', window);return false;');

  if (opts.warning !== undefined) {
    switch (opts.warning.type) {
      case 'already-edited':
        setWarning(theDocument, (div) => {
          div.appendChild(theDocument.createTextNode('This task was already edited'));
        });
        break;
      default:
        break;
    }
  }

  name.value = opts.name !== undefined ? opts.name : '';
  description.value = opts.description !== undefined ? opts.description : '';
  status.value = opts.status !== undefined ? opts.status : '';
};

const throwIfNotIntegerOrPushHistoryError = (theWindow, filters, name, humanName) => {
  if (filters[name] !== undefined) {
    const intized = parseInt(filters[name]);
    if (isNaN(intized)) {
      throw new Error('' + humanName + ' must be an integer!');
    }
    filters[name] = intized;
  }
};

const uiShowIndex = (theWindow) => {
  const theDocument = theWindow.document;
  const indexDiv = theDocument.getElementById('index');
  const queryForm = getChildById(indexDiv, 'query');
  const formElements = queryForm.elements;

  const idEquals = formElements.namedItem('filter-id-equals').value;
  const nameContains = formElements.namedItem('filter-name-contains').value;
  const descriptionContains = formElements.namedItem('filter-description-contains').value;
  const statusContains = formElements.namedItem('filter-status-contains').value;
  const dateCreatedAfter = formElements.namedItem('filter-date-created-after').value;
  const dateCreatedBefore = formElements.namedItem('filter-date-created-before').value;
  const dateUpdatedAfter = formElements.namedItem('filter-date-updated-after').value;
  const dateUpdatedBefore = formElements.namedItem('filter-date-updated-before').value;

  const filters = {};
  const opts = {
    'show-column-status': true,
  };
  const body = {
    opts: opts,
  };

  if (idEquals !== undefined && idEquals !== '') {
    filters['id-equals'] = idEquals;
  }
  if (nameContains !== undefined && nameContains !== '') {
    filters['name-contains'] = nameContains;
  }
  if (descriptionContains !== undefined && descriptionContains !== '') {
    filters['description-contains'] = descriptionContains;
  }
  if (statusContains !== undefined && statusContains !== '') {
    filters['status-contains'] = statusContains;
  }
  if (dateCreatedAfter !== undefined && dateCreatedAfter != '') {
    filters['date-created-after'] = dateCreatedAfter;
  }
  if (dateCreatedBefore !== undefined && dateCreatedBefore != '') {
    filters['date-created-before'] = dateCreatedBefore;
  }
  if (dateUpdatedAfter !== undefined && dateUpdatedAfter != '') {
    filters['date-updated-after'] = dateUpdatedAfter;
  }
  if (dateUpdatedBefore !== undefined && dateUpdatedBefore != '') {
    filters['date-updated-before'] = dateUpdatedBefore;
  }

  try {
    throwIfNotIntegerOrPushHistoryError(theWindow, filters, 'id-equals', 'ID');
  } catch (err) {
    uiPaintIndex({error: err.message}, theWindow);
    return uiPushHistory({
      index: {
        error: err.message,
        filters: filters,
      },
    }, '/index.html', theWindow);
  }

  if (Object.keys(filters).length > 0) {
    body.filters = filters;
  }

  return doQuery(body).then((ret) => {
    uiPaintIndex({
      columns: ret.columns,
      rows: ret.items,
      filters: filters,
    }, theWindow);
    return uiPushHistory({
      index: {
        columns: ret.columns,
        rows: ret.items,
        filters: filters,
      },
    }, '/index.html', theWindow);
  }, (err) => {
    uiPaintIndex({error: err.message}, theWindow);
    return uiPushHistory({
      index: {
        error: err.message,
        filters: filters,
      },
    }, '/index.html', theWindow);
  });
};

const uiShowTask = (id, theWindow) => {
  const url = '/task/' + id + '.html';
  return doTask(id).then((ret) => {
    if (ret === null) {
      const err = 'No such task: ' + id;
      uiPaintTask({error: err}, theWindow);
      return uiPushHistory({
        task: {
          id: id,
          error: err,
        },
      }, url, theWindow);
    }

    const theTask = ret;
    theTask.id = id;
    uiPaintTask({
      task: theTask,
    }, theWindow);
    return uiPushHistory({
      task: {
        id: id,
        task: theTask,
      },
    }, url, theWindow);
  }, (err) => {
    uiPaintTask({error: err.message}, theWindow);
    return uiPushHistory({
      task: {
        id: id,
        error: err.message,
      },
    }, url, theWindow);
  });
};

const uiSubmitAdd = (theWindow) => {
  const theDocument = theWindow.document;
  const addDiv = theDocument.getElementById('add');
  const addForm = getChildById(addDiv, 'add');
  const formElements = addForm.elements;

  const name = formElements.namedItem('name').value;
  const description = formElements.namedItem('description').value;
  const status = formElements.namedItem('status').value;

  const opts = {
    'name': name,
    'status': status,
  };
  if (description !== undefined && description !== '') {
    opts['description'] = description;
  }
  return doAdd(opts).then((ret) => {
    const id = ret.id;
    opts.warning = {
      type: 'already-added',
      task: id,
    };
    return uiReplaceHistory({
      add: opts,
    }, '/add.html', theWindow)
        .then(() => uiShowTask(id, theWindow));
  }, (err) => {
    opts.error = err.message;
    uiPaintAdd(opts, theWindow);
    return uiReplaceHistory({
      add: opts,
    }, '/add.html', theWindow);
  });
};

const uiSubmitEdit = (id, theWindow) => {
  const theDocument = theWindow.document;
  const editDiv = theDocument.getElementById('edit');
  const editForm = getChildById(editDiv, 'edit');
  const formElements = editForm.elements;

  const name = formElements.namedItem('name').value;
  const description = formElements.namedItem('description').value;
  const status = formElements.namedItem('status').value;

  const opts = {
    name: name,
    status: status,
  };
  if (description !== undefined && description !== '') {
    opts.description = description;
  }
  return doModify({
    'id-equals': id,
  }, opts).then((ret) => {
    opts.id = id;
    opts.warning = {
      type: 'already-edited',
    };
    return uiReplaceHistory({
      edit: opts,
    }, '/edit/' + id + '.html', theWindow)
        .then(() => uiShowTask(id, theWindow));
  }, (err) => {
    opts.id = id;
    opts.error = err.message;
    uiPaintEdit(opts, theWindow);
    return uiReplaceHistory({
      edit: opts,
    }, '/edit/' + id + '.html', theWindow);
  });
};

const popState = (theWindow, event) => {
  if (event && event.state && event.state != null) {
    const state = event.state;
    const theDocument = theWindow.document;

    if (state.add) {
      uiPaintAdd(state.add, theWindow);
      return Promise.resolve();
    }

    if (state.edit) {
      uiPaintEdit(state.edit, theWindow);
      return Promise.resolve();
    }

    if (state.index) {
      const indexDiv = theDocument.getElementById('index');
      const queryForm = getChildById(indexDiv, 'query');
      const formElements = queryForm.elements;

      const filters = state.index.filters;

      let item;
      let value;

      item = formElements.namedItem('filter-id-equals');
      value = filters['id-equals'];
      item.value = value !== undefined ? value : '';

      item = formElements.namedItem('filter-name-contains');
      value = filters['name-contains'];
      item.value = value !== undefined ? value : '';

      item = formElements.namedItem('filter-description-contains');
      value = filters['description-contains'];
      item.value = value !== undefined ? value : '';

      item = formElements.namedItem('filter-status-contains');
      value = filters['status-contains'];
      item.value = value !== undefined ? value : '';

      uiPaintIndex(state.index, theWindow);
      return Promise.resolve();
    }

    if (state.task) {
      uiPaintTask(state.task, theWindow);
      return Promise.resolve();
    }
    return Promise.reject(new Error('Cannot recognize event from onpopstate'));
  }

  theWindow.history.back();
  return Promise.resolve();
};

const uiPushHistory = (data, url, theWindow) => {
  theWindow.history.pushState(data, null, url);
  return Promise.resolve();
};

const uiReplaceHistory = (data, url, theWindow) => {
  theWindow.history.replaceState(data, null, url);
  return Promise.resolve();
};

const uiListenForPopState = (theWindow) => {
  theWindow.onpopstate = (event) => {
    return popState(theWindow, event);
  };
  return Promise.resolve();
};

module.exports = {
  doAdd: doAdd,
  doQuery: doQuery,
  doTask: doTask,
  doRemove: doRemove,
  doModify: doModify,
  uiShowAdd: uiShowAdd,
  uiShowEdit: uiShowEdit,
  uiShowIndex: uiShowIndex,
  uiShowTask: uiShowTask,
  uiSubmitAdd: uiSubmitAdd,
  uiSubmitEdit: uiSubmitEdit,
  uiListenForPopState: uiListenForPopState,
  uiPushHistory: uiPushHistory,
};

if (window !== undefined) {
  window.uiListenForPopState = uiListenForPopState;
  window.uiPaintIndex = uiPaintIndex;
  window.uiPushHistory = uiPushHistory;
  window.uiShowAdd = uiShowAdd;
  window.uiShowEdit = uiShowEdit;
  window.uiShowIndex = uiShowIndex;
  window.uiShowTask = uiShowTask;
  window.uiSubmitAdd = uiSubmitAdd;
  window.uiSubmitEdit = uiSubmitEdit;
}
