/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

const request = require('browser-request');

const doQuery = (opts) => {
  return new Promise((resolve, reject) => {
    const args = {
      url: '/api/query',
      headers: {'Content-Type': 'application/json'},
    };
    if (opts !== undefined && Object.keys(opts).length > 0) {
      args.method = 'POST';
      args.body = JSON.stringify(opts);
    }
    request(args, (err, res) => {
      if (err) {
        reject(err);
      } else {
        try {
          resolve(JSON.parse(res.body));
        } catch (e) {
          reject(e);
        }
      }
    });
  });
};

const doAdd = (opts) => {
  return new Promise((resolve, reject) => {
    if (opts === undefined || Object.keys(opts).length == 0) {
      reject(new Error('opts must be non-empty object'));
      return;
    }
    const args = {
      url: '/api/add',
      headers: {'Content-Type': 'application/json'},
      method: 'POST',
      body: JSON.stringify(opts),
    };
    request(args, (err, res) => {
      if (err) {
        reject(err);
      } else {
        resolve();
      }
    });
  });
};

const doRemove = (opts) => {
  return new Promise((resolve, reject) => {
    if (opts === undefined || Object.keys(opts).length == 0) {
      reject(new Error('opts must be non-empty object'));
      return;
    }
    const args = {
      url: '/api/remove',
      headers: {'Content-Type': 'application/json'},
      method: 'POST',
      body: JSON.stringify(opts),
    };
    request(args, (err, res) => {
      if (err) {
        reject(err);
      } else {
        resolve();
      }
    });
  });
};

const doModify = (filters, modifiers) => {
  return new Promise((resolve, reject) => {
    if (filters === undefined || Object.keys(filters).length == 0) {
      reject(new Error('filters must be non-empty object'));
      return;
    }
    if (modifiers === undefined || Object.keys(modifiers).length == 0) {
      reject(new Error('modifiers must be non-empty object'));
      return;
    }
    const requestArgs = {
      url: '/api/modify',
      headers: {'Content-Type': 'application/json'},
      method: 'POST',
      body: JSON.stringify({'filters': filters, 'modifiers': modifiers}),
    };
    request(requestArgs, (err, res) => {
      if (err) {
        reject(err);
      } else {
        resolve();
      }
    });
  });
};

const addColumnTr = (table, columns) => {
  const tr = document.createElement('tr');
  table.appendChild(tr);

  /* For remove */
  tr.appendChild(document.createElement('td'));
  /* For modify */
  tr.appendChild(document.createElement('td'));

  for (let i = 0; i < columns.length; i++) {
    const td = document.createElement('td');
    tr.appendChild(td);

    const column = columns[i];
    td.appendChild(document.createTextNode(column));
  }
};

const addRemoveButton = (tr, itemId) => {
  const td = document.createElement('td');
  tr.appendChild(td);

  const button = document.createElement('button');
  td.appendChild(button);
  button.appendChild(document.createTextNode('Remove'));
  button.onclick = () => {
    window.uiDoRemove2(itemId, td);
  };
};

const addModifyDiv = (tr, itemId) => {
  const td = document.createElement('td');
  tr.appendChild(td);

  const div = document.createElement('div');
  td.appendChild(div);

  const pButton = document.createElement('p');
  div.appendChild(pButton);

  const button = document.createElement('button');
  pButton.appendChild(button);
  button.appendChild(document.createTextNode('Modify'));

  const modifiers = [
    ['Name', 'name'],
    ['Description', 'description'],
  ];
  const modifierInputs = {};

  for (let i = 0; i < modifiers.length; i++) {
    const modifierData = modifiers[i];
    const labelName = modifierData[0];
    const modifierName = modifierData[1];

    const pName = document.createElement('p');
    div.appendChild(pName);

    pName.appendChild(document.createTextNode('' + labelName + ':'));

    const inputName = document.createElement('input');
    pName.appendChild(inputName);
    inputName.setAttribute('id', 'modify-' + itemId + '-' + modifierName);

    modifierInputs[modifierName] = inputName;
  }

  button.onclick = () => {
    window.uiDoModify2(itemId, pButton, modifierInputs);
  };
};

const addItemTr = (table, columns, item) => {
  const tr = document.createElement('tr');
  table.appendChild(tr);

  addRemoveButton(tr, item['id']);
  addModifyDiv(tr, item['id']);

  for (let i = 0; i < columns.length; i++) {
    const td = document.createElement('td');
    tr.appendChild(td);

    const column = columns[i];
    let value = '-';
    if (item.hasOwnProperty(column)) {
      value = item[column];
    }
    td.appendChild(document.createTextNode(value));
  }
};

const uiDoQuery = () => {
  return new Promise((resolve) => {
    const idEquals = document.getElementById('query-id-equals').value;
    const nameContains = document.getElementById('query-name-contains').value;
    const descriptionContains = document.getElementById('query-description-contains').value;

    const queryResultDiv = document.getElementById('query-result');
    const resultsLenDiv = document.getElementById('results-len');
    const resultsDiv = document.getElementById('results');
    const filters = {};

    queryResultDiv.innerHTML = '';
    resultsLenDiv.innerHTML = '';
    resultsDiv.innerHTML = '';

    if (idEquals !== undefined && idEquals !== '') {
      filters['id-equals'] = parseInt(idEquals);
      if (isNaN(filters['id-equals'])) {
        queryResultDiv.appendChild(document.createTextNode('ID must be an integer!'));
        resolve();
        return;
      }
    }
    if (nameContains !== undefined && nameContains !== '') {
      filters['name-contains'] = nameContains;
    }
    if (descriptionContains !== undefined && descriptionContains !== '') {
      filters['description-contains'] = descriptionContains;
    }

    doQuery(filters).then((ret) => {
      resultsLenDiv.appendChild(document.createTextNode('Results: ' + ret.length));
      queryResultDiv.appendChild(document.createTextNode('OK'));

      let columns = {};
      for (let i = 0; i < ret.length; i++) {
        const item = ret[i];
        const itemColumns = Object.keys(item);
        for (let j = 0; j < itemColumns.length; j++) {
          const column = itemColumns[j];
          columns[column] = 0;
        }
      }
      columns = Object.keys(columns);

      const table = document.createElement('table');
      resultsDiv.appendChild(table);
      table.setAttribute('border', '1');

      if (columns.length > 0) {
        addColumnTr(table, columns);
      }
      for (let i = 0; i < ret.length; i++) {
        const item = ret[i];
        addItemTr(table, columns, item);
      }
      resolve();
    }, (err) => {
      queryResultDiv.appendChild(document.createTextNode('ERR ' + err));
      resolve();
    });
  });
};

const uiDoAdd = () => {
  return new Promise((resolve) => {
    const name = document.getElementById('add-name').value;
    const description = document.getElementById('add-description').value;
    const addResultDiv = document.getElementById('add-result');
    const modifiers = {};

    addResultDiv.innerHTML = '';
    if (name !== undefined && name !== '') {
      modifiers['name'] = name;
    }
    if (description !== undefined && description !== '') {
      modifiers['description'] = description;
    }

    doAdd(modifiers).then(() => {
      addResultDiv.appendChild(document.createTextNode('OK'));
      resolve();
    }, (err) => {
      addResultDiv.appendChild(document.createTextNode('ERROR ' + err));
      resolve();
    });
  });
};

const uiDoRemove = (anId, anTd) => {
  return new Promise((resolve) => {
    const filters = {'id-equals': anId};

    doRemove(filters).then(() => {
      anTd.innerHTML = '';
      anTd.appendChild(document.createTextNode('OK'));
      resolve();
    }, (err) => {
      const button = anTd.children[0];
      anTd.innerHTML = '';
      anTd.appendChild(button);
      anTd.appendChild(document.createTextNode('ERROR ' + err));
      resolve();
    });
  });
};

const uiDoModify = (anId, anTd, modifierInputs) => {
  return new Promise((resolve) => {
    const filters = {'id-equals': anId};
    const modifiers = {};

    const modifierInputsKeys = Object.keys(modifierInputs);
    for (let i = 0; i < modifierInputsKeys.length; i++) {
      const key = modifierInputsKeys[i];
      const value = modifierInputs[key].value;
      if (value !== undefined && value !== '') {
        modifiers[key] = value;
      }
    }

    doModify(filters, modifiers).then(() => {
      anTd.innerHTML = '';
      anTd.appendChild(document.createTextNode('OK'));
      resolve();
    }, (err) => {
      const button = anTd.children[0];
      anTd.innerHTML = '';
      anTd.appendChild(button);
      anTd.appendChild(document.createTextNode('ERROR ' + err));
      resolve();
    });
  });
};

module.exports = {
  doAdd2: doAdd,
  doQuery2: doQuery,
  doRemove2: doRemove,
  doModify2: doModify,
  uiDoQuery2: uiDoQuery,
  uiDoAdd2: uiDoAdd,
  uiDoRemove2: uiDoRemove,
};

if (window !== undefined) {
  window.uiDoQuery2 = uiDoQuery;
  window.uiDoAdd2 = uiDoAdd;
  window.uiDoRemove2 = uiDoRemove;
  window.uiDoModify2 = uiDoModify;
}
