/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

let mockWindow;
let mockDocument;
let mockRequestExpects;

let errorDiv;
let warningDiv;

let addDiv;
let addForm;
let addFormName;
let addFormDescription;
let addFormStatus;

let editDiv;
let editTitleH1;
let editForm;
let editFormName;
let editFormDescription;
let editFormStatus;

let indexDiv;
let indexQueryForm;
let indexQueryIdEquals;
let indexQueryNameContains;
let indexQueryDescriptionContains;
let indexQueryStatusContains;
let indexQueryDateCreatedAfter;
let indexQueryDateCreatedBefore;
let indexQueryDateUpdatedAfter;
let indexQueryDateUpdatedBefore;

let indexResultsTable;
let indexResultsColumnsTr;
let indexResultsRowsTBody;

let taskDiv;
let taskEditA;
let taskTitleH1;
let taskPropertiesDateCreatedTd;
let taskPropertiesDateUpdatedTd;
let taskPropertiesDescriptionTd;
let taskPropertiesStatusTd;
let taskHistoryTBody;

const verifyDivVisible = (div) => {
  const mapping = {
    'add': addDiv,
    'edit': editDiv,
    'index': indexDiv,
    'task': taskDiv,
  };
  Object.keys(mapping).forEach((name) => {
    const aDiv = mapping[name];
    const actual = aDiv.style.display;
    const expected = aDiv == div ? '' : 'none';
    if (actual != expected) {
      throw new Error('Invalid style.display of ' + name +
        ': expected: "' + expected + '", actual: "' + actual + '"');
    }
  });
};

const verifyIndexColumns = (columnsTr, expectedColumns) => {
  expect(columnsTr.children.length).toEqual(expectedColumns.length + 2);
  for (let i = 0; i < columnsTr.children.length; i++) {
    const td = columnsTr.children[i];
    expect(td.childNodes.length).toEqual(1);

    let expectedColumn;
    switch (i) {
      case 0: expectedColumn = 'ID'; break;
      case 1: expectedColumn = 'Name'; break;
      default: expectedColumn = expectedColumns[i - 2];
    }

    expect('' + td.childNodes[0].text).toEqual('' + expectedColumn);
  }
};

const verifyIndexResults = (resultsTBody, items) => {
  expect(resultsTBody.children.length).toEqual(items.length);

  for (let i = 0; i < items.length; i++) {
    const tr = resultsTBody.children[i];
    const item = items[i];
    expect(tr.children.length).toEqual(item.length);

    const itemId = item[0];

    for (let j = 0; j < item.length; j++) {
      const td = tr.children[j];
      expect(td.childNodes.length).toEqual(1);

      if (j < 2) {
        const a = td.childNodes[0];

        expect(Object.keys(a.attributes).length).toEqual(2);
        expect(a.attributes['href']).toEqual('/task/' + itemId + '.html');
        expect(a.attributes['onclick']).toEqual('uiShowTask(' + itemId + ', window);return false;');

        expect(a.childNodes.length).toEqual(1);
        expect('' + a.childNodes[0].text).toEqual((j == 0 ? '#' : '') + item[j]);
      } else {
        expect('' + td.childNodes[0].text).toEqual('' + item[j]);
      }
    }
  }
};

const verifyTaskTitle = (titleH1, id, name) => {
  expect(titleH1.childNodes.length).toEqual(2);

  const a = titleH1.childNodes[0];
  expect(a.tag).toEqual('a');
  expect(Object.keys(a.attributes).length).toEqual(2);
  expect(a.attributes['href']).toEqual('/task/' + id + '.html');
  expect(a.attributes['onclick']).toEqual('uiShowTask(' + id + ', window);return false;');
  expect(a.childNodes.length).toEqual(1);
  expect('' + a.childNodes[0].text).toEqual('#' + id);

  expect('' + titleH1.childNodes[1].text).toEqual(' - ' + name);
};

const verifyTdText = (td, value) => {
  expect(td.childNodes.length).toEqual(1);
  expect(td.childNodes[0].text).toEqual(value);
};

const verifyEditTitle = (titleH1, id, name) => {
  expect(titleH1.childNodes.length).toEqual(3);

  expect('' + titleH1.childNodes[0].text).toEqual('Edit ');

  const a = titleH1.childNodes[1];
  expect(a.tag).toEqual('a');
  expect(Object.keys(a.attributes).length).toEqual(2);
  expect(a.attributes['href']).toEqual('/task/' + id + '.html');
  expect(a.attributes['onclick']).toEqual('uiShowTask(' + id + ', window);return false;');
  expect(a.childNodes.length).toEqual(1);
  expect('' + a.childNodes[0].text).toEqual('#' + id);

  expect('' + titleH1.childNodes[2].text).toEqual(' - ' + name);
};

const verifyEditForm = (form, id) => {
  expect(form.attributes['action']).toEqual('/edit/' + id + '.html');
  expect(form.attributes['onsubmit']).toEqual('uiSubmitEdit(' + id + ', window);return false;');
};

const verifyTaskEdit = (editA, id) => {
  expect(editA.attributes['href']).toEqual('/edit/' + id + '.html');
  expect(editA.attributes['onclick']).toEqual('uiShowEdit(' + id + ', window);return false;');
};

const verifyTaskHistory = (historyTBody, items) => {
  expect(historyTBody.children.length).toEqual(items.length);

  for (let i = 0; i < historyTBody.children.length; i++) {
    const item = items[i];
    const tr = historyTBody.children[i];
    expect(tr.children.length).toEqual(4);

    let td;
    td = tr.children[0];
    expect(td.childNodes.length).toEqual(1);
    expect(td.childNodes[0].text).toEqual(item.date);

    td = tr.children[1];
    expect(td.childNodes.length).toEqual(1);
    const typeInPast = item.type === 'add' ? 'added' : item.type + 'd';
    expect(td.childNodes[0].text).toEqual(typeInPast);

    td = tr.children[2];
    expect(td.childNodes.length).toEqual(1);
    expect(td.childNodes[0].tag).toEqual('b');
    expect(td.childNodes[0].childNodes.length).toEqual(1);
    expect(td.childNodes[0].childNodes[0].text).toEqual(item.name);

    td = tr.children[3];
    expect(td.childNodes.length).toEqual(1 + (item.type === 'change' ? 2 : 0));

    expect(td.childNodes[0].tag).toEqual('u');
    expect(td.childNodes[0].childNodes.length).toEqual(1);
    expect(td.childNodes[0].childNodes[0].text).toEqual(item.value);

    if (item.type === 'change') {
      expect(td.childNodes[1].text).toEqual(' -> ');

      expect(td.childNodes[2].tag).toEqual('u');
      expect(td.childNodes[2].childNodes.length).toEqual(1);
      expect(td.childNodes[2].childNodes[0].text).toEqual(item['new-value']);
    }
  }
};

const expectError = (expected) => {
  expect(errorDiv.childNodes.length).toEqual(1, 'No error occured');
  expect(errorDiv.childNodes[0].text).toEqual(expected);
};

const expectNoError = () => {
  if (errorDiv.childNodes.length != 0) {
    throw new Error(errorDiv.childNodes[0].text);
  }
};

const expectNoWarning = () => {
  if (warningDiv.childNodes.length != 0) {
    throw new Error(warningDiv.childNodes[0].text);
  }
};

const expectWarningAlreadyAdded = (task) => {
  expect(warningDiv.childNodes.length).toEqual(2);
  expect(warningDiv.childNodes[0].text).toEqual('This task was already added as ');

  const a = warningDiv.childNodes[1];
  expect(a.tag).toEqual('a');
  expect(Object.keys(a.attributes).length).toEqual(2);
  expect(a.attributes['href']).toEqual('/task/987.html');
  expect(a.attributes['onclick']).toEqual('uiShowTask(987, window);return false;');
  expect(a.childNodes.length).toEqual(1);
  expect('' + a.childNodes[0].text).toEqual('#987');
};

const expectWarningAlreadyEdited = () => {
  expect(warningDiv.childNodes.length).toEqual(1);
  expect(warningDiv.childNodes[0].text).toEqual('This task was already edited');
};

jest.mock('browser-request', () => {
  return jest.fn((opts, callback) => {
    if (mockRequestExpects === undefined) {
      throw new Error('Unexpected request: ' + JSON.stringify(opts));
    }

    const anExpect = mockRequestExpects.shift();
    if (anExpect === undefined) {
      throw new Error('Unexpected request: ' + JSON.stringify(opts));
    }

    expect(opts).toEqual(anExpect.opts);

    if (anExpect.err !== undefined) {
      callback(new Error(anExpect.err));
    } else {
      let statusCode = anExpect.statusCode;
      if (statusCode === undefined) {
        statusCode = 200;
      }

      callback(undefined, {
        statusCode: statusCode,
        body: anExpect.body,
      });
    }
  });
});

require('browser-request');

const libJs = require('../src/lib.js');

describe('lib.js', () => {
  beforeEach(() => {
    mockDocument = {
      createElement: (tag) => {
        const self = {
          attributes: {},
          children: [],
          childNodes: [],
          style: {
            display: '',
          },

          appendChild: (child) => {
            self.childNodes.push(child);
            if (child.tag !== 'text') {
              self.children.push(child);
            }
          },
          getAttribute: (key) => {
            return self.attributes[key];
          },
          removeAttribute: (key) => {
            delete self.attributes[key];
          },
          removeChild: (child) => {
            let index;

            index = self.childNodes.indexOf(child);
            if (index > -1) {
              self.childNodes.splice(index, 1);
            }

            index = self.children.indexOf(child);
            if (index > -1) {
              self.children.splice(index, 1);
            }
          },
          setAttribute: (key, value) => {
            self.attributes[key] = value;
          },
          tag: tag,
        };
        return self;
      },
      createTextNode: (text) => {
        const node = mockDocument.createElement('text');
        node.text = text;
        return node;
      },
      getElementById: (id) => {
        if (id == 'add') {
          return addDiv;
        }
        if (id == 'edit') {
          return editDiv;
        }
        if (id == 'error') {
          return errorDiv;
        }
        if (id == 'index') {
          return indexDiv;
        }
        if (id == 'task') {
          return taskDiv;
        }
        if (id == 'warning') {
          return warningDiv;
        }
        return undefined;
      },
    };

    mockWindow = {
      document: mockDocument,
      history: {
        back: jest.fn(),
        pushState: jest.fn(),
        replaceState: jest.fn(),
      },
      location: {
        href: '/',
      },
    };

    let td;

    errorDiv = mockDocument.createElement('div');
    errorDiv.setAttribute('id', 'error');

    warningDiv = mockDocument.createElement('div');
    warningDiv.setAttribute('id', 'warning');

    indexResultsColumnsTr = mockDocument.createElement('tr');
    indexResultsColumnsTr.setAttribute('id', 'columns');

    td = mockDocument.createElement('td');
    indexResultsColumnsTr.appendChild(td);
    td.appendChild(mockDocument.createTextNode('ID'));

    td = mockDocument.createElement('td');
    indexResultsColumnsTr.appendChild(td);
    td.appendChild(mockDocument.createTextNode('Name'));

    const thead = mockDocument.createElement('thead');
    thead.appendChild(indexResultsColumnsTr);

    indexResultsRowsTBody = mockDocument.createElement('tbody');
    indexResultsRowsTBody.setAttribute('id', 'rows');

    indexResultsTable = mockDocument.createElement('table');
    indexResultsTable.setAttribute('id', 'results');
    indexResultsTable.appendChild(thead);
    indexResultsTable.appendChild(indexResultsRowsTBody);

    indexQueryIdEquals = mockDocument.createElement('input');
    indexQueryNameContains = mockDocument.createElement('input');
    indexQueryDescriptionContains = mockDocument.createElement('input');
    indexQueryStatusContains = mockDocument.createElement('input');
    indexQueryDateCreatedAfter = mockDocument.createElement('input');
    indexQueryDateCreatedBefore = mockDocument.createElement('input');
    indexQueryDateUpdatedAfter = mockDocument.createElement('input');
    indexQueryDateUpdatedBefore = mockDocument.createElement('input');

    indexQueryForm = mockDocument.createElement('form');
    indexQueryForm.setAttribute('id', 'query');
    indexQueryForm.elements = {
      namedItem: (id) => {
        if (id == 'filter-id-equals') {
          return indexQueryIdEquals;
        }
        if (id == 'filter-name-contains') {
          return indexQueryNameContains;
        }
        if (id == 'filter-description-contains') {
          return indexQueryDescriptionContains;
        }
        if (id == 'filter-status-contains') {
          return indexQueryStatusContains;
        }
        if (id == 'filter-date-created-after') {
          return indexQueryDateCreatedAfter;
        }
        if (id == 'filter-date-created-before') {
          return indexQueryDateCreatedBefore;
        }
        if (id == 'filter-date-updated-after') {
          return indexQueryDateUpdatedAfter;
        }
        if (id == 'filter-date-updated-before') {
          return indexQueryDateUpdatedBefore;
        }
      },
    };

    indexDiv = mockDocument.createElement('div');
    indexDiv.appendChild(indexQueryForm);
    indexDiv.appendChild(indexResultsTable);

    addFormName = mockDocument.createElement('input');
    addFormDescription = mockDocument.createElement('input');
    addFormStatus = mockDocument.createElement('input');

    addForm = mockDocument.createElement('form');
    addForm.setAttribute('id', 'add');
    addForm.elements = {
      namedItem: (id) => {
        if (id == 'name') {
          return addFormName;
        }
        if (id == 'description') {
          return addFormDescription;
        }
        if (id == 'status') {
          return addFormStatus;
        }
      },
    };

    addDiv = mockDocument.createElement('div');
    addDiv.appendChild(addForm);

    editTitleH1 = mockDocument.createElement('h1');
    editTitleH1.setAttribute('id', 'title');

    editFormName = mockDocument.createElement('input');
    editFormDescription = mockDocument.createElement('input');
    editFormStatus = mockDocument.createElement('input');

    editForm = mockDocument.createElement('form');
    editForm.setAttribute('id', 'edit');
    editForm.setAttribute('action', '');
    editForm.setAttribute('onsubmit', '');
    editForm.elements = {
      namedItem: (id) => {
        if (id == 'name') {
          return editFormName;
        }
        if (id == 'description') {
          return editFormDescription;
        }
        if (id == 'status') {
          return editFormStatus;
        }
      },
    };

    editDiv = mockDocument.createElement('div');
    editDiv.appendChild(editTitleH1);
    editDiv.appendChild(editForm);

    taskTitleH1 = mockDocument.createElement('h1');
    taskTitleH1.setAttribute('id', 'title');

    taskEditA = mockDocument.createElement('a');
    taskEditA.setAttribute('id', 'edit');

    const taskPropertiesTBody = mockDocument.createElement('tbody');
    taskPropertiesTBody.setAttribute('id', 'properties');

    let tr;

    tr = mockDocument.createElement('tr');
    taskPropertiesTBody.appendChild(tr);
    taskPropertiesDateCreatedTd = mockDocument.createElement('td');
    tr.appendChild(taskPropertiesDateCreatedTd);
    taskPropertiesDateCreatedTd.setAttribute('id', 'date-created');

    tr = mockDocument.createElement('tr');
    taskPropertiesTBody.appendChild(tr);
    taskPropertiesDateUpdatedTd = mockDocument.createElement('td');
    tr.appendChild(taskPropertiesDateUpdatedTd);
    taskPropertiesDateUpdatedTd.setAttribute('id', 'date-updated');

    tr = mockDocument.createElement('tr');
    taskPropertiesTBody.appendChild(tr);
    taskPropertiesDescriptionTd = mockDocument.createElement('td');
    tr.appendChild(taskPropertiesDescriptionTd);
    taskPropertiesDescriptionTd.setAttribute('id', 'description');

    tr = mockDocument.createElement('tr');
    taskPropertiesTBody.appendChild(tr);
    taskPropertiesStatusTd = mockDocument.createElement('td');
    tr.appendChild(taskPropertiesStatusTd);
    taskPropertiesStatusTd.setAttribute('id', 'status');

    const tableProperties = mockDocument.createElement('table');
    tableProperties.appendChild(taskPropertiesTBody);

    taskHistoryTBody = mockDocument.createElement('tbody');
    taskHistoryTBody.setAttribute('id', 'history');

    const tableHistory = mockDocument.createElement('table');
    tableHistory.appendChild(taskHistoryTBody);

    taskDiv = mockDocument.createElement('div');
    taskDiv.appendChild(taskTitleH1);
    taskDiv.appendChild(taskEditA);
    taskDiv.appendChild(tableProperties);
    taskDiv.appendChild(tableHistory);
  });
  describe('doQuery', () => {
    it('is defined', () => {
      expect(libJs.doQuery).toBeDefined();
    });
    it('no args', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/query',
          },
          statusCode: 200,
          body: JSON.stringify({
            columns: [],
            items: [],
          }),
        },
      ];
      libJs.doQuery().then((ret) => {
        expect(ret).toEqual({
          columns: [],
          items: [],
        });
        done();
      }, done.fail);
    });
    it('generic error', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/query',
          },
          err: 'error',
        },
      ];

      libJs.doQuery().then(done.fail, (err) => {
        expect(err.message).toEqual('error');
        done();
      });
    });
    it('response not JSON', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/query',
            headers: {'Content-Type': 'application/json'},
          },
          body: 'anything};3',
        },
      ];
      libJs.doQuery().then(done.fail, (err) => {
        expect(err.message).toBeDefined();
        done();
      });
    });
    it('empty args', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/query',
          },
          body: JSON.stringify({
            columns: [],
            items: [],
          }),
        },
      ];
      libJs.doQuery({}).then((ret) => {
        expect(ret).toEqual({
          columns: [],
          items: [],
        });
        done();
      }, done.fail);
    });
    it('some args', (done) => {
      const body = {
        filters: {
          'name-contains': 3,
        },
      };
      mockRequestExpects = [
        {
          opts: {
            url: '/api/query',
            method: 'POST',
            body: JSON.stringify(body),
            headers: {'Content-Type': 'application/json'},
          },
          body: JSON.stringify({
            columns: [],
            items: [],
          }),
        },
      ];
      libJs.doQuery(body).then((ret) => {
        expect(ret).toEqual({
          columns: [],
          items: [],
        });
        done();
      }, done.fail);
    });
  });
  describe('doTask', () => {
    it('is defined', () => {
      expect(libJs.doTask).toBeDefined();
    });
    it('gets the task', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/task/1',
          },
          body: JSON.stringify({}),
        },
      ];
      libJs.doTask(1).then((ret) => {
        expect(ret).toEqual({});
        done();
      }, done.fail);
    });
    it('generic error', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/task/1',
          },
          err: 'error',
        },
      ];
      libJs.doTask(1).then(done.fail, (err) => {
        expect(err.message).toEqual('error');
        done();
      });
    });
    it('response not JSON', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/task/1',
            headers: {'Content-Type': 'application/json'},
          },
          body: 'anything};3',
        },
      ];
      libJs.doTask(1).then(done.fail, (err) => {
        expect(err.message).toBeDefined();
        done();
      });
    });
  });
  describe('doAdd', () => {
    it('is defined', () => {
      expect(libJs.doAdd).toBeDefined();
    });
    it('error when no args', (done) => {
      libJs.doAdd().then(done.fail, (err) => {
        expect(err.message).toEqual('opts must be non-empty object');
        done();
      });
    });
    it('error when empty args', (done) => {
      libJs.doAdd({}).then(done.fail, (err) => {
        expect(err.message).toEqual('opts must be non-empty object');
        done();
      });
    });
    it('request error', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/add',
            headers: {'Content-Type': 'application/json'},
            method: 'POST',
            body: JSON.stringify({'name': 'Buy milk'}),
          },
          err: 'opts must be non-empty object',
        },
      ];
      libJs.doAdd({'name': 'Buy milk'}).then(done.fail, (err) => {
        expect(err.message).toEqual('opts must be non-empty object');
        done();
      });
    });
    it('some args', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/add',
            headers: {'Content-Type': 'application/json'},
            method: 'POST',
            body: JSON.stringify({'name': 'Buy milk'}),
          },
          body: JSON.stringify({
            id: 1,
          }),
        },
      ];
      libJs.doAdd({'name': 'Buy milk'}).then((ret) => {
        expect(ret.id).toEqual(1);
        done();
      }, done.fail);
    });
  });
  describe('doRemove', () => {
    it('error when no args', (done) => {
      libJs.doRemove().then(done.fail, (err) => {
        expect(err.message).toEqual('opts must be non-empty object');
        done();
      });
    });
    it('error when empty args', (done) => {
      libJs.doRemove({}).then(done.fail, (err) => {
        expect(err.message).toEqual('opts must be non-empty object');
        done();
      });
    });
    it('request error', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/remove',
            headers: {'Content-Type': 'application/json'},
            method: 'POST',
            body: JSON.stringify({'name': 'Buy milk'}),
          },
          err: 'opts must be non-empty object',
        },
      ];
      libJs.doRemove({'name': 'Buy milk'}).then(done.fail, (err) => {
        expect(err.message).toEqual('opts must be non-empty object');
        done();
      });
    });
    it('some args', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/remove',
            headers: {'Content-Type': 'application/json'},
            method: 'POST',
            body: JSON.stringify({'name': 'Buy milk'}),
          },
        },
      ];
      libJs.doRemove({'name': 'Buy milk'}).then(done, done.fail);
    });
  });
  describe('doModify', () => {
    it('error when no filters', (done) => {
      libJs.doModify().then(done.fail, (err) => {
        expect(err.message).toEqual('filters must be non-empty object');
        done();
      });
    });
    it('error when empty filters', (done) => {
      libJs.doModify({}).then(done.fail, (err) => {
        expect(err.message).toEqual('filters must be non-empty object');
        done();
      });
    });
    it('error when no modifiers', (done) => {
      const filters = {'id-equals': 1};
      libJs.doModify(filters).then(done.fail, (err) => {
        expect(err.message).toEqual('modifiers must be non-empty object');
        done();
      });
    });
    it('error when empty modifiers', (done) => {
      const filters = {'id-equals': 1};
      libJs.doModify(filters, {}).then(done.fail, (err) => {
        expect(err.message).toEqual('modifiers must be non-empty object');
        done();
      });
    });
    it('request error', (done) => {
      const filters = {'id-equals': 1};
      const modifiers = {'name': 'Buy beer'};
      mockRequestExpects = [
        {
          opts: {
            url: '/api/modify',
            headers: {'Content-Type': 'application/json'},
            method: 'POST',
            body: JSON.stringify({'filters': filters, 'modifiers': modifiers}),
          },
          err: 'opts must be non-empty object',
        },
      ];
      libJs.doModify(filters, modifiers).then(done.fail, (err) => {
        expect(err.message).toEqual('opts must be non-empty object');
        done();
      });
    });
    it('some modifiers', (done) => {
      const filters = {'id-equals': 1};
      const modifiers = {'name': 'Buy beer'};
      mockRequestExpects = [
        {
          opts: {
            url: '/api/modify',
            headers: {'Content-Type': 'application/json'},
            method: 'POST',
            body: JSON.stringify({'filters': filters, 'modifiers': modifiers}),
          },
        },
      ];
      libJs.doModify(filters, modifiers).then(done, done.fail);
    });
  });
  describe('uiShowAdd', () => {
    it('is defined', () => {
      expect(libJs.uiShowAdd).toBeDefined();
    });
    it('shows empty add task page', (done) => {
      libJs.uiShowAdd(mockWindow).then(() => {
        verifyDivVisible(addDiv);
        expectNoError();
        expectNoWarning();
        expect(addFormName.value).toEqual('');
        expect(addFormDescription.value).toEqual('');
        expect(addFormStatus.value).toEqual('new');
        expect(mockWindow.history.pushState.mock.calls).toEqual([
          [
            {
              add: {
                status: 'new',
              },
            },
            null,
            '/add.html',
          ],
        ]);

        done();
      }, done.fail);
    });
  });
  describe('uiSubmitAdd', () => {
    it('is defined', () => {
      expect(libJs.uiSubmitAdd).toBeDefined();
    });
    it('adds a task and opens its page', (done) => {
      addFormName.value = 'Buy milk';
      addFormDescription.value = 'Goat milk';
      addFormStatus.value = 'open';

      const requestAdd = {
        opts: {
          url: '/api/add',
          headers: {'Content-Type': 'application/json'},
          method: 'POST',
          body: JSON.stringify({
            name: 'Buy milk',
            status: 'open',
            description: 'Goat milk',
          }),
        },
        body: JSON.stringify({
          id: 1,
        }),
      };
      const requestTask = {
        opts: {
          url: '/api/task/1',
        },
        body: JSON.stringify({
          'id': 1,
          'name': 'Buy milk',
          'description': 'Goat milk',
          'status': 'open',
          'date-created': 1,
          'date-updated': 1,
          'history': [
            {
              'type': 'add',
              'name': 'name',
              'value': 'Buy milk',
              'date': 1,
            },
            {
              'type': 'add',
              'name': 'description',
              'value': 'Goat milk',
              'date': 1,
            },
            {
              'type': 'add',
              'name': 'status',
              'value': 'open',
              'date': 1,
            },
          ],
        }),
      };
      mockRequestExpects = [requestAdd, requestTask];

      libJs.uiSubmitAdd(mockWindow).then(() => {
        verifyDivVisible(taskDiv);
        expectNoError();
        expectNoWarning();
        verifyTaskTitle(taskTitleH1, 1, 'Buy milk');
        verifyTaskEdit(taskEditA, 1);
        verifyTdText(taskPropertiesDateCreatedTd, '1970-01-01 00:00:01');
        verifyTdText(taskPropertiesDateUpdatedTd, '1970-01-01 00:00:01');
        verifyTdText(taskPropertiesDescriptionTd, 'Goat milk');
        verifyTdText(taskPropertiesStatusTd, 'open');
        verifyTaskHistory(taskHistoryTBody, [
          {
            'type': 'add',
            'name': 'Name',
            'value': 'Buy milk',
            'date': '1970-01-01 00:00:01',
          },
          {
            'type': 'add',
            'name': 'Description',
            'value': 'Goat milk',
            'date': '1970-01-01 00:00:01',
          },
          {
            'type': 'add',
            'name': 'Status',
            'value': 'open',
            'date': '1970-01-01 00:00:01',
          },
        ]);
        expect(mockWindow.history.replaceState.mock.calls).toEqual([
          [
            {
              add: {
                name: 'Buy milk',
                description: 'Goat milk',
                status: 'open',
                warning: {
                  type: 'already-added',
                  task: 1,
                },
              },
            },
            null,
            '/add.html',
          ],
        ]);
        expect(mockWindow.history.pushState.mock.calls).toEqual([
          [
            {
              task: {
                id: 1,
                task: {
                  'id': 1,
                  'name': 'Buy milk',
                  'description': 'Goat milk',
                  'status': 'open',
                  'date-created': 1,
                  'date-updated': 1,
                  'history': [
                    {
                      'type': 'add',
                      'name': 'name',
                      'value': 'Buy milk',
                      'date': 1,
                    },
                    {
                      'type': 'add',
                      'name': 'description',
                      'value': 'Goat milk',
                      'date': 1,
                    },
                    {
                      'type': 'add',
                      'name': 'status',
                      'value': 'open',
                      'date': 1,
                    },
                  ],
                },
              },
            },
            null,
            '/task/1.html',
          ],
        ]);

        done();
      }, done.fail);
    });
  });
  describe('uiShowEdit', () => {
    it('is defined', () => {
      expect(libJs.uiShowEdit).toBeDefined();
    });
    it('got task', (done) => {
      const body = JSON.stringify({
        'name': 'Buy milk',
        'description': 'Goat milk',
        'status': 'open',
      });
      mockRequestExpects = [
        {
          opts: {
            url: '/api/task/1',
          },
          body: body,
        },
      ];
      libJs.uiShowEdit(1, mockWindow).then(() => {
        verifyDivVisible(editDiv);

        expectNoError();
        expectNoWarning();
        verifyEditTitle(editTitleH1, 1, 'Buy milk');
        verifyEditForm(editForm, 1);
        expect(editFormName.value).toEqual('Buy milk');
        expect(editFormDescription.value).toEqual('Goat milk');
        expect(editFormStatus.value).toEqual('open');
        expect(mockWindow.history.pushState.mock.calls).toEqual([
          [
            {
              edit: {
                id: 1,
                name: 'Buy milk',
                description: 'Goat milk',
                status: 'open',
              },
            },
            null,
            '/edit/1.html',
          ],
        ]);

        done();
      }, done.fail);
    });
    it('task not found', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/task/1',
          },
          statusCode: 404,
        },
      ];
      libJs.uiShowEdit(1, mockWindow).then(() => {
        verifyDivVisible();
        expectError('No such task: 1');
        expect(mockWindow.history.pushState.mock.calls).toEqual([
          [
            {
              edit: {
                error: 'No such task: 1',
              },
            },
            null,
            '/edit/1.html',
          ],
        ]);

        done();
      }, done.fail);
    });
    it('error', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/task/1',
          },
          err: 'Errur',
        },
      ];
      libJs.uiShowEdit(1, mockWindow).then(() => {
        verifyDivVisible();
        expectError('Errur');
        expect(mockWindow.history.pushState.mock.calls).toEqual([
          [
            {
              edit: {
                error: 'Errur',
              },
            },
            null,
            '/edit/1.html',
          ],
        ]);

        done();
      }, done.fail);
    });
  });
  describe('uiSubmitEdit', () => {
    it('is defined', () => {
      expect(libJs.uiSubmitEdit).toBeDefined();
    });
    it('edits a task and opens its page', (done) => {
      editFormName.value = 'Buy beer';
      editFormDescription.value = 'Lager';
      editFormStatus.value = 'open';

      const requestModify = {
        opts: {
          url: '/api/modify',
          headers: {'Content-Type': 'application/json'},
          method: 'POST',
          body: JSON.stringify({
            filters: {
              'id-equals': 1,
            },
            modifiers: {
              name: 'Buy beer',
              status: 'open',
              description: 'Lager',
            },
          }),
        },
        body: JSON.stringify({
          id: 1,
        }),
      };
      const requestTask = {
        opts: {
          url: '/api/task/1',
        },
        body: JSON.stringify({
          'id': 1,
          'name': 'Buy beer',
          'description': 'Lager',
          'status': 'open',
          'date-created': 1,
          'date-updated': 2,
          'history': [
            {
              'type': 'change',
              'name': 'name',
              'value': 'Buy milk',
              'new-value': 'Buy beer',
              'date': 2,
            },
            {
              'type': 'change',
              'name': 'description',
              'value': 'Goat milk',
              'new-value': 'Lager',
              'date': 2,
            },
            {
              'type': 'change',
              'name': 'status',
              'value': 'new',
              'new-value': 'open',
              'date': 2,
            },
            {
              'type': 'add',
              'name': 'name',
              'value': 'Buy milk',
              'date': 1,
            },
            {
              'type': 'add',
              'name': 'description',
              'value': 'Goat milk',
              'date': 1,
            },
            {
              'type': 'add',
              'name': 'status',
              'value': 'open',
              'date': 1,
            },
          ],
        }),
      };
      mockRequestExpects = [requestModify, requestTask];

      libJs.uiSubmitEdit(1, mockWindow).then(() => {
        verifyDivVisible(taskDiv);
        expectNoError();
        expectNoWarning();
        verifyTaskTitle(taskTitleH1, 1, 'Buy beer');
        verifyTaskEdit(taskEditA, 1);
        verifyTdText(taskPropertiesDateCreatedTd, '1970-01-01 00:00:01');
        verifyTdText(taskPropertiesDateUpdatedTd, '1970-01-01 00:00:02');
        verifyTdText(taskPropertiesDescriptionTd, 'Lager');
        verifyTdText(taskPropertiesStatusTd, 'open');
        verifyTaskHistory(taskHistoryTBody, [
          {
            'type': 'change',
            'name': 'Name',
            'value': 'Buy milk',
            'new-value': 'Buy beer',
            'date': '1970-01-01 00:00:02',
          },
          {
            'type': 'change',
            'name': 'Description',
            'value': 'Goat milk',
            'new-value': 'Lager',
            'date': '1970-01-01 00:00:02',
          },
          {
            'type': 'change',
            'name': 'Status',
            'value': 'new',
            'new-value': 'open',
            'date': '1970-01-01 00:00:02',
          },
          {
            'type': 'add',
            'name': 'Name',
            'value': 'Buy milk',
            'date': '1970-01-01 00:00:01',
          },
          {
            'type': 'add',
            'name': 'Description',
            'value': 'Goat milk',
            'date': '1970-01-01 00:00:01',
          },
          {
            'type': 'add',
            'name': 'Status',
            'value': 'open',
            'date': '1970-01-01 00:00:01',
          },
        ]);
        expect(mockWindow.history.replaceState.mock.calls).toEqual([
          [
            {
              edit: {
                id: 1,
                name: 'Buy beer',
                description: 'Lager',
                status: 'open',
                warning: {
                  type: 'already-edited',
                },
              },
            },
            null,
            '/edit/1.html',
          ],
        ]);
        expect(mockWindow.history.pushState.mock.calls).toEqual([
          [
            {
              task: {
                id: 1,
                task: {
                  'id': 1,
                  'name': 'Buy beer',
                  'description': 'Lager',
                  'status': 'open',
                  'date-created': 1,
                  'date-updated': 2,
                  'history': [
                    {
                      'type': 'change',
                      'name': 'name',
                      'value': 'Buy milk',
                      'new-value': 'Buy beer',
                      'date': 2,
                    },
                    {
                      'type': 'change',
                      'name': 'description',
                      'value': 'Goat milk',
                      'new-value': 'Lager',
                      'date': 2,
                    },
                    {
                      'type': 'change',
                      'name': 'status',
                      'value': 'new',
                      'new-value': 'open',
                      'date': 2,
                    },
                    {
                      'type': 'add',
                      'name': 'name',
                      'value': 'Buy milk',
                      'date': 1,
                    },
                    {
                      'type': 'add',
                      'name': 'description',
                      'value': 'Goat milk',
                      'date': 1,
                    },
                    {
                      'type': 'add',
                      'name': 'status',
                      'value': 'open',
                      'date': 1,
                    },
                  ],
                },
              },
            },
            null,
            '/task/1.html',
          ],
        ]);

        done();
      }, done.fail);
    });
  });
  describe('uiShowIndex', () => {
    it('is defined', () => {
      expect(libJs.uiShowIndex).toBeDefined();
    });
    it('empty query, empty results', (done) => {
      const body = {
        opts: {
          'show-column-status': true,
        },
      };
      mockRequestExpects = [
        {
          opts: {
            url: '/api/query',
            headers: {'Content-Type': 'application/json'},
            method: 'POST',
            body: JSON.stringify(body),
          },
          body: JSON.stringify({
            columns: [],
            items: [],
          }),
        },
      ];
      libJs.uiShowIndex(mockWindow).then(() => {
        verifyDivVisible(indexDiv);
        expectNoError();
        expectNoWarning();
        expect(indexResultsTable.style.display).toEqual('');
        verifyIndexColumns(indexResultsColumnsTr, []);
        verifyIndexResults(indexResultsRowsTBody, []);
        expect(mockWindow.history.pushState.mock.calls).toEqual([
          [
            {
              index: {
                columns: [],
                filters: {},
                rows: [],
              },
            },
            null,
            '/index.html',
          ],
        ]);

        done();
      }, done.fail);
    });
    it('repeated query cleans previous results', (done) => {
      const body = {
        opts: {
          'show-column-status': true,
        },
      };
      mockRequestExpects = [
        {
          opts: {
            url: '/api/query',
            headers: {'Content-Type': 'application/json'},
            method: 'POST',
            body: JSON.stringify(body),
          },
          body: JSON.stringify({
            columns: [],
            items: [
              {'id': 2, 'name': 'Buy milk'},
            ],
          }),
        },
      ];
      libJs.uiShowIndex(mockWindow).then(() => {
        verifyDivVisible(indexDiv);
        expectNoError();
        expectNoWarning();
        expect(indexResultsTable.style.display).toEqual('');
        verifyIndexColumns(indexResultsColumnsTr, []);
        verifyIndexResults(indexResultsRowsTBody, [
          ['2', 'Buy milk'],
        ]);
        expect(mockWindow.history.pushState.mock.calls).toEqual([
          [
            {
              index: {
                columns: [],
                filters: {},
                rows: [{'id': 2, 'name': 'Buy milk'}],
              },
            },
            null,
            '/index.html',
          ],
        ]);

        mockRequestExpects = [
          {
            opts: {
              url: '/api/query',
              headers: {'Content-Type': 'application/json'},
              method: 'POST',
              body: JSON.stringify(body),
            },
            body: JSON.stringify({
              columns: [],
              items: [
                {'id': 3, 'name': 'Buy beer'},
              ],
            }),
          },
        ];
        return libJs.uiShowIndex(mockWindow).then(() => {
          verifyDivVisible(indexDiv);
          expectNoError();
          expectNoWarning();
          expect(indexResultsTable.style.display).toEqual('');
          verifyIndexColumns(indexResultsColumnsTr, []);
          verifyIndexResults(indexResultsRowsTBody, [
            ['3', 'Buy beer'],
          ]);
          done();
        });
      }).catch(done.fail);
    });
    it('empty query, long results', (done) => {
      const body = {
        opts: {
          'show-column-status': true,
        },
      };
      mockRequestExpects = [
        {
          opts: {
            url: '/api/query',
            headers: {'Content-Type': 'application/json'},
            method: 'POST',
            body: JSON.stringify(body),
          },
          body: JSON.stringify({
            columns: [],
            items: [
              {'id': 1, 'name': 'Buy milk'},
              {'id': 2, 'name': 'Buy beer'},
              {'id': 3, 'name': 'Buy gas'},
            ],
          }),
        },
      ];
      libJs.uiShowIndex(mockWindow).then(() => {
        verifyDivVisible(indexDiv);
        expectNoError();
        expectNoWarning();
        expect(indexResultsTable.style.display).toEqual('');
        verifyIndexColumns(indexResultsColumnsTr, []);
        verifyIndexResults(indexResultsRowsTBody, [
          ['1', 'Buy milk'],
          ['2', 'Buy beer'],
          ['3', 'Buy gas'],
        ]);
        expect(mockWindow.history.pushState.mock.calls).toEqual([
          [
            {
              index: {
                columns: [],
                filters: {},
                rows: [
                  {'id': 1, 'name': 'Buy milk'},
                  {'id': 2, 'name': 'Buy beer'},
                  {'id': 3, 'name': 'Buy gas'},
                ],
              },
            },
            null,
            '/index.html',
          ],
        ]);

        done();
      }, done.fail);
    });
    describe('single queries', () => {
      const doIt = (filterName, filterValue, inputCallback, inputValue, expectedDate) => {
        it(filterName, (done) => {
          const body = {
            opts: {
              'show-column-status': true,
            },
            filters: {},
          };
          body.filters[filterName] = filterValue;
          mockRequestExpects = [
            {
              opts: {
                url: '/api/query',
                headers: {'Content-Type': 'application/json'},
                method: 'POST',
                body: JSON.stringify(body),
              },
              body: JSON.stringify({
                columns: [],
                items: [],
              }),
            },
          ];

          const input = inputCallback();
          input.value = inputValue;

          libJs.uiShowIndex(mockWindow).then(() => {
            verifyDivVisible(indexDiv);
            expectNoError();
            expectNoWarning();
            expect(indexResultsTable.style.display).toEqual('');
            verifyIndexColumns(indexResultsColumnsTr, []);
            verifyIndexResults(indexResultsRowsTBody, []);

            done();
          }, done.fail);
        });
      };

      doIt('id-equals', 1, () => indexQueryIdEquals, '1');
      doIt('name-contains', 'beer', () => indexQueryNameContains, 'beer');
      doIt('description-contains', 'beer', () => indexQueryDescriptionContains, 'beer');
      doIt('status-contains', 'open', () => indexQueryStatusContains, 'open');
      doIt('date-created-after', '1', () => indexQueryDateCreatedAfter, '1', new Date(1000));
      doIt('date-created-before', '1', () => indexQueryDateCreatedBefore, '1', new Date(1000));
      doIt('date-updated-after', '1', () => indexQueryDateUpdatedAfter, '1', new Date(1000));
      doIt('date-updated-before', '1', () => indexQueryDateUpdatedBefore, '1', new Date(1000));
    });
    describe('single queries with invalid value', () => {
      describe('not integer', () => {
        const doIt = (filterName, inputCallback, inputValue) => {
          it(filterName, (done) => {
            const input = inputCallback();
            input.value = inputValue;

            libJs.uiShowIndex(mockWindow).then(() => {
              expectError('ID must be an integer!');
              verifyDivVisible(indexDiv);
              expect(indexResultsTable.style.display).toEqual('none');
              expect(indexResultsRowsTBody.children.length).toEqual(0);
              done();
            }, done.fail);
          });
        };

        doIt('id-equals', () => indexQueryIdEquals, 'invalid');
      });
    });
    it('error query', (done) => {
      const body = {
        opts: {
          'show-column-status': true,
        },
      };
      mockRequestExpects = [
        {
          opts: {
            url: '/api/query',
            headers: {'Content-Type': 'application/json'},
            method: 'POST',
            body: JSON.stringify(body),
          },
          err: 'Network unavailable',
        },
      ];
      libJs.uiShowIndex(mockWindow).then(() => {
        expectError('Network unavailable');
        expect(indexResultsRowsTBody.children.length).toEqual(0);
        done();
      }, done.fail);
    });
  });
  describe('uiShowTask', () => {
    it('is defined', () => {
      expect(libJs.uiShowTask).toBeDefined();
    });
    it('got task', (done) => {
      const body = JSON.stringify({
        'name': 'Buy milk',
        'description': 'Goat milk',
        'status': 'open',
        'date-created': 1,
        'date-updated': 7,
        'history': [
          {
            'type': 'remove',
            'name': 'name',
            'value': 'Buy beer',
            'date': 7,
          },
          {
            'type': 'remove',
            'name': 'description',
            'value': 'Lager',
            'date': 6,
          },
          {
            'type': 'remove',
            'name': 'status',
            'value': 'open',
            'date': 5,
          },
          {
            'type': 'change',
            'name': 'name',
            'value': 'Buy milk',
            'new-value': 'Buy beer',
            'date': 4,
          },
          {
            'type': 'change',
            'name': 'description',
            'value': 'Goat milk',
            'new-value': 'Lager',
            'date': 3,
          },
          {
            'type': 'change',
            'name': 'status',
            'value': 'new',
            'new-value': 'open',
            'date': 2,
          },
          {
            'type': 'add',
            'name': 'name',
            'value': 'Buy milk',
            'date': 1,
          },
          {
            'type': 'add',
            'name': 'description',
            'value': 'Goat milk',
            'date': 1,
          },
          {
            'type': 'add',
            'name': 'status',
            'value': 'new',
            'date': 1,
          },
        ],
      });
      mockRequestExpects = [
        {
          opts: {
            url: '/api/task/1',
          },
          body: body,
        },
      ];
      libJs.uiShowTask(1, mockWindow).then(() => {
        verifyDivVisible(taskDiv);
        expectNoError();
        expectNoWarning();
        verifyTaskTitle(taskTitleH1, 1, 'Buy milk');
        verifyTaskEdit(taskEditA, 1);
        verifyTdText(taskPropertiesDateCreatedTd, '1970-01-01 00:00:01');
        verifyTdText(taskPropertiesDateUpdatedTd, '1970-01-01 00:00:07');
        verifyTdText(taskPropertiesDescriptionTd, 'Goat milk');
        verifyTdText(taskPropertiesStatusTd, 'open');
        verifyTaskHistory(taskHistoryTBody, [
          {
            'type': 'remove',
            'name': 'Name',
            'value': 'Buy beer',
            'date': '1970-01-01 00:00:07',
          },
          {
            'type': 'remove',
            'name': 'Description',
            'value': 'Lager',
            'date': '1970-01-01 00:00:06',
          },
          {
            'type': 'remove',
            'name': 'Status',
            'value': 'open',
            'date': '1970-01-01 00:00:05',
          },
          {
            'type': 'change',
            'name': 'Name',
            'value': 'Buy milk',
            'new-value': 'Buy beer',
            'date': '1970-01-01 00:00:04',
          },
          {
            'type': 'change',
            'name': 'Description',
            'value': 'Goat milk',
            'new-value': 'Lager',
            'date': '1970-01-01 00:00:03',
          },
          {
            'type': 'change',
            'name': 'Status',
            'value': 'new',
            'new-value': 'open',
            'date': '1970-01-01 00:00:02',
          },
          {
            'type': 'add',
            'name': 'Name',
            'value': 'Buy milk',
            'date': '1970-01-01 00:00:01',
          },
          {
            'type': 'add',
            'name': 'Description',
            'value': 'Goat milk',
            'date': '1970-01-01 00:00:01',
          },
          {
            'type': 'add',
            'name': 'Status',
            'value': 'new',
            'date': '1970-01-01 00:00:01',
          },
        ]);
        expect(mockWindow.history.pushState.mock.calls).toEqual([
          [
            {
              task: {
                id: 1,
                task: {
                  'id': 1,
                  'name': 'Buy milk',
                  'description': 'Goat milk',
                  'status': 'open',
                  'date-created': 1,
                  'date-updated': 7,
                  'history': [
                    {
                      'type': 'remove',
                      'name': 'name',
                      'value': 'Buy beer',
                      'date': 7,
                    },
                    {
                      'type': 'remove',
                      'name': 'description',
                      'value': 'Lager',
                      'date': 6,
                    },
                    {
                      'type': 'remove',
                      'name': 'status',
                      'value': 'open',
                      'date': 5,
                    },
                    {
                      'type': 'change',
                      'name': 'name',
                      'value': 'Buy milk',
                      'new-value': 'Buy beer',
                      'date': 4,
                    },
                    {
                      'type': 'change',
                      'name': 'description',
                      'value': 'Goat milk',
                      'new-value': 'Lager',
                      'date': 3,
                    },
                    {
                      'type': 'change',
                      'name': 'status',
                      'value': 'new',
                      'new-value': 'open',
                      'date': 2,
                    },
                    {
                      'type': 'add',
                      'name': 'name',
                      'value': 'Buy milk',
                      'date': 1,
                    },
                    {
                      'type': 'add',
                      'name': 'description',
                      'value': 'Goat milk',
                      'date': 1,
                    },
                    {
                      'type': 'add',
                      'name': 'status',
                      'value': 'new',
                      'date': 1,
                    },
                  ],
                },
              },
            },
            null,
            '/task/1.html',
          ],
        ]);

        done();
      }, done.fail);
    });
    it('task not found', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/task/1',
          },
          statusCode: 404,
        },
      ];
      libJs.uiShowTask(1, mockWindow).then(() => {
        verifyDivVisible();
        expectError('No such task: 1');
        expect(taskTitleH1.childNodes.length).toEqual(0);
        expect(mockWindow.history.pushState.mock.calls).toEqual([
          [
            {
              task: {
                id: 1,
                error: 'No such task: 1',
              },
            },
            null,
            '/task/1.html',
          ],
        ]);

        done();
      }, done.fail);
    });
    it('error', (done) => {
      mockRequestExpects = [
        {
          opts: {
            url: '/api/task/1',
          },
          err: 'Errur',
        },
      ];
      libJs.uiShowTask(1, mockWindow).then(() => {
        verifyDivVisible();
        expectError('Errur');
        expect(taskTitleH1.childNodes.length).toEqual(0);
        expect(mockWindow.history.pushState.mock.calls).toEqual([
          [
            {
              task: {
                id: 1,
                error: 'Errur',
              },
            },
            null,
            '/task/1.html',
          ],
        ]);

        done();
      }, done.fail);
    });
  });
  describe('History API', () => {
    it('API-s are defined', () => {
      expect(libJs.uiPushHistory).toBeDefined();
      expect(libJs.uiListenForPopState).toBeDefined();
    });
    it('pushes state', () => {
      const data = {a: 1, b: [], c: {}};
      libJs.uiPushHistory(data, 'url', mockWindow);
      expect(mockWindow.history.pushState.mock.calls).toEqual([
        [
          data,
          null,
          'url',
        ],
      ]);
    });
    it('goes back when no event', (done) => {
      libJs.uiListenForPopState(mockWindow);
      expect(mockWindow.onpopstate).toBeDefined();
      mockWindow.onpopstate()
          .then(() => {
            expect(mockWindow.history.back.mock.calls).toEqual([[]]);
            done();
          }).catch(done.fail);
    });
    it('paints empty add', (done) => {
      const data = {
        state: {
          add: {
            name: 'Abecadlo',
            description: 'S pieca spadlo',
            status: 'new',
          },
        },
      };
      libJs.uiListenForPopState(mockWindow);
      expect(mockWindow.onpopstate).toBeDefined();

      mockWindow.onpopstate(data).then(() => {
        verifyDivVisible(addDiv);
        expectNoError();
        expectNoWarning();
        expect(addFormName.value).toEqual('Abecadlo');
        expect(addFormDescription.value).toEqual('S pieca spadlo');
        expect(addFormStatus.value).toEqual('new');
        done();
      }, done.fail);
    });
    it('paints add with error', (done) => {
      const data = {
        state: {
          add: {
            status: 'new',
            error: 'Krowa',
          },
        },
      };
      libJs.uiListenForPopState(mockWindow);
      expect(mockWindow.onpopstate).toBeDefined();

      mockWindow.onpopstate(data).then(() => {
        verifyDivVisible(addDiv);
        expectError('Krowa');
        expectNoWarning();
        expect(addFormName.value).toEqual('');
        expect(addFormDescription.value).toEqual('');
        expect(addFormStatus.value).toEqual('new');
        done();
      }, done.fail);
    });
    it('paints add with already added warning', (done) => {
      const data = {
        state: {
          add: {
            status: 'new',
            warning: {
              type: 'already-added',
              task: 987,
            },
          },
        },
      };
      libJs.uiListenForPopState(mockWindow);
      expect(mockWindow.onpopstate).toBeDefined();

      mockWindow.onpopstate(data).then(() => {
        verifyDivVisible(addDiv);
        expectNoError();
        expectWarningAlreadyAdded(987);
        expect(addFormName.value).toEqual('');
        expect(addFormDescription.value).toEqual('');
        expect(addFormStatus.value).toEqual('new');
        done();
      }, done.fail);
    });
    it('paints empty edit', (done) => {
      const data = {
        state: {
          edit: {
            id: 1,
            name: 'Buy milk',
            description: 'Goat milk',
            status: 'open',
          },
        },
      };
      libJs.uiListenForPopState(mockWindow);
      expect(mockWindow.onpopstate).toBeDefined();

      mockWindow.onpopstate(data).then(() => {
        verifyDivVisible(editDiv);
        verifyEditTitle(editTitleH1, 1, 'Buy milk');
        verifyEditForm(editForm, 1);
        expect(editFormName.value).toEqual('Buy milk');
        expect(editFormDescription.value).toEqual('Goat milk');
        expect(editFormStatus.value).toEqual('open');
        done();
      }, done.fail);
    });
    it('paints edit with error', (done) => {
      const data = {
        state: {
          edit: {
            id: 1,
            name: 'Buy milk',
            description: 'Goat milk',
            status: 'open',
            error: 'Krowa',
          },
        },
      };
      libJs.uiListenForPopState(mockWindow);
      expect(mockWindow.onpopstate).toBeDefined();

      mockWindow.onpopstate(data).then(() => {
        verifyDivVisible(editDiv);
        expectError('Krowa');
        expectNoWarning();

        verifyEditTitle(editTitleH1, 1, 'Buy milk');
        verifyEditForm(editForm, 1);
        expect(editFormName.value).toEqual('Buy milk');
        expect(editFormDescription.value).toEqual('Goat milk');
        expect(editFormStatus.value).toEqual('open');
        done();
      }, done.fail);
    });
    it('paints edit with already edited warning', (done) => {
      const data = {
        state: {
          edit: {
            id: 1,
            name: 'Buy milk',
            description: 'Goat milk',
            status: 'open',
            warning: {
              type: 'already-edited',
            },
          },
        },
      };
      libJs.uiListenForPopState(mockWindow);
      expect(mockWindow.onpopstate).toBeDefined();

      mockWindow.onpopstate(data).then(() => {
        verifyDivVisible(editDiv);
        expectNoError();
        expectWarningAlreadyEdited();

        verifyEditTitle(editTitleH1, 1, 'Buy milk');
        verifyEditForm(editForm, 1);
        expect(editFormName.value).toEqual('Buy milk');
        expect(editFormDescription.value).toEqual('Goat milk');
        expect(editFormStatus.value).toEqual('open');
        done();
      }, done.fail);
    });
    it('paints results index', (done) => {
      const data = {
        state: {
          index: {
            filters: {},
            columns: ['status'],
            rows: [
              {
                id: 1,
                name: 'Buy milk',
                status: 'open',
                columns: ['open'],
              },
            ],
          },
        },
      };
      libJs.uiListenForPopState(mockWindow);
      expect(mockWindow.onpopstate).toBeDefined();

      mockWindow.onpopstate(data).then(() => {
        verifyDivVisible(indexDiv);
        expectNoError();
        expectNoWarning();
        expect(indexResultsTable.style.display).toEqual('');
        verifyIndexColumns(indexResultsColumnsTr, ['Status']);
        verifyIndexResults(indexResultsRowsTBody, [
          ['1', 'Buy milk', 'open'],
        ]);
        done();
      }, done.fail);
    });
    it('paints error index', (done) => {
      const data = {
        state: {
          index: {
            filters: {},
            error: 'Szatan',
          },
        },
      };
      libJs.uiListenForPopState(mockWindow);
      expect(mockWindow.onpopstate).toBeDefined();

      mockWindow.onpopstate(data).then(() => {
        expectError('Szatan');
        verifyDivVisible(indexDiv);
        expect(indexResultsTable.style.display).toEqual('none');
        expect(indexResultsRowsTBody.children.length).toEqual(0);
        done();
      }, done.fail);
    });
    it('paints task', (done) => {
      const data = {
        state: {
          task: {
            id: 1,
            task: {
              'id': 1,
              'name': 'Buy milk',
              'description': 'Goat milk',
              'status': 'open',
              'date-created': 1,
              'date-updated': 1,
              'history': [
                {'type': 'add', 'name': 'name', 'value': 'Buy milk', 'date': 1},
                {'type': 'add', 'name': 'description', 'value': 'Goat milk', 'date': 1},
                {'type': 'add', 'name': 'status', 'value': 'open', 'date': 1},
              ],
            },
          },
        },
      };
      libJs.uiListenForPopState(mockWindow);
      expect(mockWindow.onpopstate).toBeDefined();

      mockWindow.onpopstate(data).then(() => {
        expectNoError();
        expectNoWarning();
        verifyDivVisible(taskDiv);
        verifyTaskTitle(taskTitleH1, 1, 'Buy milk');
        verifyTaskEdit(taskEditA, 1);
        verifyTdText(taskPropertiesDateCreatedTd, '1970-01-01 00:00:01');
        verifyTdText(taskPropertiesDateUpdatedTd, '1970-01-01 00:00:01');
        verifyTdText(taskPropertiesDescriptionTd, 'Goat milk');
        verifyTdText(taskPropertiesStatusTd, 'open');
        verifyTaskHistory(taskHistoryTBody, [
          {
            'type': 'add',
            'name': 'Name',
            'value': 'Buy milk',
            'date': '1970-01-01 00:00:01',
          },
          {
            'type': 'add',
            'name': 'Description',
            'value': 'Goat milk',
            'date': '1970-01-01 00:00:01',
          },
          {
            'type': 'add',
            'name': 'Status',
            'value': 'open',
            'date': '1970-01-01 00:00:01',
          },
        ]);
        done();
      }, done.fail);
    });
  });
});
