/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

let mockRequestOpts;
let mockRequestErr;
let mockRequestRet;

let addName;
let addResult;

let queryIdEquals;
let queryNameContains;
let queryDescriptionContains;
let queryResult;

let resultsLen;
let results;

const verifyRemoveTd = (td) => {
  expect(td.children.length).toEqual(1);

  const button = td.children[0];
  expect(button.tagName).toEqual('BUTTON');
  expect(button.onclick).toBeDefined();
  expect(button.innerHTML).toEqual('Remove');
};


const verifyModifyTd = (td) => {
  expect(td.children.length).toEqual(1);
  const div = td.children[0];
  expect(div.tagName).toEqual('DIV');
  expect(div.children.length).toEqual(3);

  const pButton = div.children[0];
  expect(pButton.tagName).toEqual('P');
  expect(pButton.children.length).toEqual(1);

  const button = pButton.children[0];
  expect(button.tagName).toEqual('BUTTON');
  expect(button.onclick).toBeDefined();
  expect(button.innerHTML).toEqual('Modify');

  let pName;
  let inputName;

  pName = div.children[1];
  expect(pName.tagName).toEqual('P');
  expect(pName.textContent).toEqual('Name:');
  expect(pName.children.length).toEqual(1);

  inputName = pName.children[0];
  expect(inputName.tagName).toEqual('INPUT');
  expect(inputName.id).toBeDefined();

  pName = div.children[2];
  expect(pName.tagName).toEqual('P');
  expect(pName.textContent).toEqual('Description:');
  expect(pName.children.length).toEqual(1);

  inputName = pName.children[0];
  expect(inputName.tagName).toEqual('INPUT');
  expect(inputName.id).toBeDefined();
};

const verifyTable = (table, items) => {
  const headerOffset = (items.length == 0 ? 0 : 1);
  expect(table.children.length).toEqual(items.length + headerOffset);

  for (let i = 0; i < items.length; i++) {
    const tr = table.children[i + headerOffset];
    const item = items[i];
    expect(tr.children.length).toEqual(item.length + 2);

    verifyRemoveTd(tr.children[0]);
    verifyModifyTd(tr.children[1]);

    for (let j = 0; j < item.length; j++) {
      const td = tr.children[j + 2];
      expect(td.innerHTML).toEqual(item[j]);
    }
  }
};

jest.mock('browser-request', () => {
  return jest.fn((opts, callback) => {
    expect(opts).toEqual(mockRequestOpts);

    if (mockRequestErr !== undefined) {
      callback(mockRequestErr);
    } else {
      callback(undefined, {body: mockRequestRet});
    }
  });
});

require('browser-request');

document.getElementById = (id) => {
  if (id == 'query-id-equals') {
    return {'value': queryIdEquals};
  }
  if (id == 'query-name-contains') {
    return {'value': queryNameContains};
  }
  if (id == 'query-description-contains') {
    return {'value': queryDescriptionContains};
  }
  if (id == 'add-name') {
    return {'value': addName};
  }
  if (id == 'add-result') {
    return addResult;
  }
  if (id == 'query-result') {
    return queryResult;
  }
  if (id == 'results') {
    return results;
  }
  if (id == 'results-len') {
    return resultsLen;
  }
  return undefined;
};

const otherJs = require('../src/other.js');

describe('other.js', () => {
  beforeEach(() => {
    mockRequestOpts = {};
    mockRequestErr = undefined;
    mockRequestRet = undefined;
    addResult = document.createElement('div');
    queryResult = document.createElement('div');
    results = document.createElement('div');
    resultsLen = document.createElement('div');

    queryIdEquals = '';
    queryNameContains = '';
    queryDescriptionContains = '';

    addName = '';
  });
  describe('doQuery', () => {
    it('is defined', () => {
      expect(otherJs.doQuery2).toBeDefined();
    });
    it('no args', (done) => {
      mockRequestOpts = {
        url: '/api/query',
        headers: {'Content-Type': 'application/json'},
      };
      mockRequestRet = '[]';
      otherJs.doQuery2().then((ret) => {
        expect(ret).toEqual([]);
        done();
      }, done.fail);
    });
    it('generic error', (done) => {
      mockRequestOpts = {
        url: '/api/query',
        headers: {'Content-Type': 'application/json'},
      };
      mockRequestErr = 'error';
      otherJs.doQuery2().then(done.fail, (err) => {
        expect(err).toEqual('error');
        done();
      });
    });
    it('response not JSON', (done) => {
      mockRequestOpts = {
        url: '/api/query',
        headers: {'Content-Type': 'application/json'},
      };
      mockRequestRet = 'anything};3';
      otherJs.doQuery2().then(done.fail, (err) => {
        expect(err.message).toBeDefined();
        done();
      });
    });
    it('empty args', (done) => {
      mockRequestOpts = {
        url: '/api/query',
        headers: {'Content-Type': 'application/json'},
      };
      mockRequestRet = '[]';
      otherJs.doQuery2({}).then((ret) => {
        expect(ret).toEqual([]);
        done();
      }, done.fail);
    });
    it('some args', (done) => {
      const args = {'name-contains': 3};
      mockRequestOpts = {
        url: '/api/query',
        method: 'POST',
        body: JSON.stringify(args),
        headers: {'Content-Type': 'application/json'},
      };
      mockRequestRet = '[]';
      otherJs.doQuery2(args).then((ret) => {
        expect(ret).toEqual([]);
        done();
      }, done.fail);
    });
  });
  describe('doAdd', () => {
    it('is defined', () => {
      expect(otherJs.doAdd2).toBeDefined();
    });
    it('error when no args', (done) => {
      otherJs.doAdd2().then(done.fail, (err) => {
        expect(err.message).toEqual('opts must be non-empty object');
        done();
      });
    });
    it('error when empty args', (done) => {
      otherJs.doAdd2({}).then(done.fail, (err) => {
        expect(err.message).toEqual('opts must be non-empty object');
        done();
      });
    });
    it('request error', (done) => {
      mockRequestOpts = {
        url: '/api/add',
        headers: {'Content-Type': 'application/json'},
        method: 'POST',
        body: JSON.stringify({'name': 'Buy milk'}),
      };
      mockRequestErr = 'opts must be non-empty object';
      otherJs.doAdd2({'name': 'Buy milk'}).then(done.fail, (err) => {
        expect(err).toEqual('opts must be non-empty object');
        done();
      });
    });
    it('some args', (done) => {
      mockRequestOpts = {
        url: '/api/add',
        headers: {'Content-Type': 'application/json'},
        method: 'POST',
        body: JSON.stringify({'name': 'Buy milk'}),
      };
      mockRequestRet = '[]';
      otherJs.doAdd2({'name': 'Buy milk'}).then(done, done.fail);
    });
  });
  describe('doRemove', () => {
    it('error when no args', (done) => {
      otherJs.doRemove2().then(done.fail, (err) => {
        expect(err.message).toEqual('opts must be non-empty object');
        done();
      });
    });
    it('error when empty args', (done) => {
      otherJs.doRemove2({}).then(done.fail, (err) => {
        expect(err.message).toEqual('opts must be non-empty object');
        done();
      });
    });
    it('request error', (done) => {
      mockRequestOpts = {
        url: '/api/remove',
        headers: {'Content-Type': 'application/json'},
        method: 'POST',
        body: JSON.stringify({'name': 'Buy milk'}),
      };
      mockRequestErr = 'opts must be non-empty object';
      otherJs.doRemove2({'name': 'Buy milk'}).then(done.fail, (err) => {
        expect(err).toEqual('opts must be non-empty object');
        done();
      });
    });
    it('some args', (done) => {
      mockRequestOpts = {
        url: '/api/remove',
        headers: {'Content-Type': 'application/json'},
        method: 'POST',
        body: JSON.stringify({'name': 'Buy milk'}),
      };
      otherJs.doRemove2({'name': 'Buy milk'}).then(done, done.fail);
    });
  });
  describe('doModify', () => {
    it('error when no filters', (done) => {
      otherJs.doModify2().then(done.fail, (err) => {
        expect(err.message).toEqual('filters must be non-empty object');
        done();
      });
    });
    it('error when empty filters', (done) => {
      otherJs.doModify2({}).then(done.fail, (err) => {
        expect(err.message).toEqual('filters must be non-empty object');
        done();
      });
    });
    it('error when no modifiers', (done) => {
      const filters = {'id-equals': 1};
      otherJs.doModify2(filters).then(done.fail, (err) => {
        expect(err.message).toEqual('modifiers must be non-empty object');
        done();
      });
    });
    it('error when empty modifiers', (done) => {
      const filters = {'id-equals': 1};
      otherJs.doModify2(filters, {}).then(done.fail, (err) => {
        expect(err.message).toEqual('modifiers must be non-empty object');
        done();
      });
    });
    it('request error', (done) => {
      const filters = {'id-equals': 1};
      const modifiers = {'name': 'Buy beer'};
      mockRequestOpts = {
        url: '/api/modify',
        headers: {'Content-Type': 'application/json'},
        method: 'POST',
        body: JSON.stringify({'filters': filters, 'modifiers': modifiers}),
      };
      mockRequestErr = 'opts must be non-empty object';
      otherJs.doModify2(filters, modifiers).then(done.fail, (err) => {
        expect(err).toEqual('opts must be non-empty object');
        done();
      });
    });
    it('some modifiers', (done) => {
      const filters = {'id-equals': 1};
      const modifiers = {'name': 'Buy beer'};
      mockRequestOpts = {
        url: '/api/modify',
        headers: {'Content-Type': 'application/json'},
        method: 'POST',
        body: JSON.stringify({'filters': filters, 'modifiers': modifiers}),
      };
      otherJs.doModify2(filters, modifiers).then(done, done.fail);
    });
  });
  describe('uiDoQuery', () => {
    it('is defined', () => {
      expect(otherJs.uiDoQuery2).toBeDefined();
    });
    it('empty query, empty results', (done) => {
      mockRequestOpts = {
        url: '/api/query',
        headers: {'Content-Type': 'application/json'},
      };
      mockRequestRet = '[]';
      otherJs.uiDoQuery2().then(() => {
        expect(queryResult.innerHTML).toEqual('OK');
        expect(resultsLen.innerHTML).toEqual('Results: 0');
        expect(results.children.length).toEqual(1);
        verifyTable(results.children[0], []);
        done();
      }, done.fail);
    });
    it('repeated query cleans previous results', (done) => {
      mockRequestOpts = {
        url: '/api/query',
        headers: {'Content-Type': 'application/json'},
      };
      mockRequestRet = JSON.stringify([
        {'id': 2, 'name': 'Buy milk'},
      ]);
      otherJs.uiDoQuery2().then(() => {
        expect(queryResult.innerHTML).toEqual('OK');
        expect(resultsLen.innerHTML).toEqual('Results: 1');
        expect(results.children.length).toEqual(1);
        verifyTable(results.children[0], [
          ['2', 'Buy milk'],
        ]);

        mockRequestRet = JSON.stringify([
          {'id': 2, 'name': 'Buy beer'},
        ]);
        return otherJs.uiDoQuery2().then(() => {
          expect(queryResult.innerHTML).toEqual('OK');
          expect(resultsLen.innerHTML).toEqual('Results: 1');
          expect(results.children.length).toEqual(1);
          verifyTable(results.children[0], [
            ['2', 'Buy beer'],
          ]);
          done();
        });
      }).catch(done.fail);
    });
    it('empty query, long results', (done) => {
      mockRequestOpts = {
        url: '/api/query',
        headers: {'Content-Type': 'application/json'},
      };
      mockRequestRet = JSON.stringify([
        {'id': 1, 'name': 'Buy milk'},
        {'id': 2, 'name': 'Buy beer'},
        {'id': 3, 'name': 'Buy gas'},
      ]);
      otherJs.uiDoQuery2().then(() => {
        expect(queryResult.innerHTML).toEqual('OK');
        expect(results.children.length).toEqual(1);
        verifyTable(results.children[0], [
          ['1', 'Buy milk'],
          ['2', 'Buy beer'],
          ['3', 'Buy gas'],
        ]);
        done();
      }, done.fail);
    });
    it('id equals', (done) => {
      mockRequestOpts = {
        url: '/api/query',
        headers: {'Content-Type': 'application/json'},
        method: 'POST',
        body: JSON.stringify({'id-equals': 1}),
      };
      mockRequestRet = '[]';
      queryIdEquals = '1';
      otherJs.uiDoQuery2().then(() => {
        expect(queryResult.innerHTML).toEqual('OK');
        expect(results.children.length).toEqual(1);
        verifyTable(results.children[0], []);
        done();
      }, done.fail);
    });
    it('id equals not integer', (done) => {
      mockRequestOpts = {};
      mockRequestErr = new Error('Should not happen');
      queryIdEquals = 'invalid';
      otherJs.uiDoQuery2().then(() => {
        expect(queryResult.innerHTML).toEqual('ID must be an integer!');
        expect(results.children.length).toEqual(0);
        done();
      }, done.fail);
    });
    it('name contains', (done) => {
      mockRequestOpts = {
        url: '/api/query',
        headers: {'Content-Type': 'application/json'},
        method: 'POST',
        body: JSON.stringify({'name-contains': 'beer'}),
      };
      mockRequestRet = '[]';
      queryNameContains = 'beer';
      otherJs.uiDoQuery2().then(() => {
        expect(queryResult.innerHTML).toEqual('OK');
        expect(results.children.length).toEqual(1);
        verifyTable(results.children[0], []);
        done();
      }, done.fail);
    });
    it('description contains', (done) => {
      mockRequestOpts = {
        url: '/api/query',
        headers: {'Content-Type': 'application/json'},
        method: 'POST',
        body: JSON.stringify({'description-contains': 'beer'}),
      };
      mockRequestRet = '[]';
      queryDescriptionContains = 'beer';
      otherJs.uiDoQuery2().then(() => {
        expect(queryResult.innerHTML).toEqual('OK');
        expect(results.children.length).toEqual(1);
        verifyTable(results.children[0], []);
        done();
      }, done.fail);
    });
    it('error query', (done) => {
      mockRequestOpts = {
        url: '/api/query',
        headers: {'Content-Type': 'application/json'},
      };
      mockRequestErr = new Error('Network unavailable');
      otherJs.uiDoQuery2().then(() => {
        expect(queryResult.innerHTML).toEqual('ERR Error: Network unavailable');
        expect(results.children.length).toEqual(0);
        done();
      }, done.fail);
    });
  });
});
