'use strict';

const fs = require('fs');
const {spawn} = require('child_process');
const puppeteer = require('puppeteer-core');

const expectLength = async (arr, length) => {
  if (arr.length != length) {
    throw new Error('Expected length ' + length + ', actual ' + arr.length);
  }
  return arr;
};

const expectToBeEqual = async (left, right) => {
  if (JSON.stringify(left) != JSON.stringify(right)) {
    throw new Error('Expected ' + left + ' to equal ' + right);
  }
};

const expectNotToBeEmpty = async (left) => {
  if (left == '') {
    throw new Error('Expected ' + left + ' not to be empty');
  }
};

const all = (parent, xp) => parent.$x(xp);

const first = async (parent, xp) => {
  const ret = await all(parent, xp);
  await expectLength(ret, 1);
  return ret[0];
};

const getAddNewTask = async (page) => {
  const XPATH_DESCRIPTION = './/textarea[@name="description"]';
  const XPATH_NAME = './/input[@name="name"]';
  const XPATH_STATUS = './/input[@name="status"]';
  const XPATH_SUBMIT = './/input[@type="submit"]';
  const obj = await first(page, './/div[@id="add"]');

  const clickSubmit = async () => {
    const input = await first(obj, XPATH_SUBMIT);
    await input.click();
  };

  const expectToBeVisible = async () => {
    const ret = await page.evaluate((obj) => obj.style.display, obj);
    await expectToBeEqual(ret, '');
  };

  const setDescription = async (value) => {
    const input = await first(obj, XPATH_DESCRIPTION);
    await page.evaluate((input, value) => {
      input.value = value;
    }, input, value);
  };

  const setName = async (value) => {
    const input = await first(obj, XPATH_NAME);
    await page.evaluate((input, value) => {
      input.value = value;
    }, input, value);
  };

  const setStatus = async (value) => {
    const input = await first(obj, XPATH_STATUS);
    await page.evaluate((input, value) => {
      input.value = value;
    }, input, value);
  };

  return {
    clickSubmit: clickSubmit,
    expectToBeVisible: expectToBeVisible,
    setDescription: setDescription,
    setName: setName,
    setStatus: setStatus,
  };
};


const getEditTask = async (page) => {
  const XPATH_DESCRIPTION = './/textarea[@name="description"]';
  const XPATH_NAME = './/input[@name="name"]';
  const XPATH_STATUS = './/input[@name="status"]';
  const XPATH_SUBMIT = './/input[@type="submit"]';
  const obj = await first(page, './/div[@id="edit"]');

  const clickSubmit = async () => {
    const input = await first(obj, XPATH_SUBMIT);
    await input.click();
  };

  const expectToBeVisible = async () => {
    const ret = await page.evaluate((obj) => obj.style.display, obj);
    await expectToBeEqual(ret, '');
  };

  const setDescription = async (value) => {
    const input = await first(obj, XPATH_DESCRIPTION);
    await page.evaluate((input, value) => {
      input.value = value;
    }, input, value);
  };

  const setName = async (value) => {
    const input = await first(obj, XPATH_NAME);
    await page.evaluate((input, value) => {
      input.value = value;
    }, input, value);
  };

  const setStatus = async (value) => {
    const input = await first(obj, XPATH_STATUS);
    await page.evaluate((input, value) => {
      input.value = value;
    }, input, value);
  };

  return {
    clickSubmit: clickSubmit,
    expectToBeVisible: expectToBeVisible,
    setDescription: setDescription,
    setName: setName,
    setStatus: setStatus,
  };
};

const getIndex = async (page) => {
  const XPATH_QUERY_ID_EQUALS = './/input[@name="filter-id-equals"]';
  const XPATH_QUERY_SUBMIT = './/input[@type="submit"]';
  const obj = await first(page, './/div[@id="index"]');

  const clickResult = async (idx) => {
    const a = await first(obj, './/table[@id="results"]/tbody/tr[' + (idx + 1) + ']/td[1]/a');
    await a.click();
  };

  const clickQuerySubmit = async () => {
    const input = await first(obj, XPATH_QUERY_SUBMIT);
    await input.click();
  };

  const expectResultsToEqual = async (rows) => {
    const trs = await all(obj, './/table[@id="results"]/tbody/tr');
    await expectLength(trs, rows.length);

    await Promise.all(trs.map(async (tr, idx) => {
      const row = rows[idx];

      const tds = await all(tr, './/td');
      await expectLength(tds, 3);
      const values = await Promise.all(tds.map(async (td) =>
        await page.evaluate((td) => td.textContent.trim(), td)
      ));
      await expectToBeEqual(values, row);
    }));
  };

  const expectToBeVisible = async () => {
    const value = await page.evaluate((obj) => obj.style.display, obj);
    await expectToBeEqual(value, '');
  };

  const setQueryIdEquals = async (value) => {
    const input = await first(obj, XPATH_QUERY_ID_EQUALS);
    await page.evaluate((input, value) => {
      input.value = value;
    }, input, value);
  };

  return {
    clickResult: clickResult,
    clickQuerySubmit: clickQuerySubmit,
    expectResultsToEqual: expectResultsToEqual,
    expectToBeVisible: expectToBeVisible,
    setQueryIdEquals: setQueryIdEquals,
  };
};

const getNavBar = async (page) => {
  const XPATH_ADD_NEW_TASK = './/a[@href="/add.html" and text()="Add new task"]';
  const XPATH_INDEX = './/a[@href="/index.html" and text()="Query tasks"]';

  const obj = await first(page, './/p[@id="nav"]');

  const clickAddNewTask = async () => {
    const a = await first(obj, XPATH_ADD_NEW_TASK);
    await a.click();
  };

  const clickIndex = async () => {
    const a = await first(obj, XPATH_INDEX);
    await a.click();
  };

  return {
    clickAddNewTask: clickAddNewTask,
    clickIndex: clickIndex,
  };
};

const getTask = async (page) => {
  const XPATH_DATE_CREATED = './/td[@id="date-created"]';
  const XPATH_DATE_UPDATED = './/td[@id="date-updated"]';
  const XPATH_DESCRIPTION = './/td[@id="description"]';
  const XPATH_STATUS = './/td[@id="status"]';
  const XPATH_TITLE = './/h1[@id="title"]';
  const obj = await first(page, './/div[@id="task"]');

  const clickEdit = async (idx) => {
    const a = await first(obj, './/a[@id="edit"]');
    await a.click();
  };

  const expectDateCreatedNotToBeEmpty = async () => {
    const td = await first(obj, XPATH_DATE_CREATED);
    const val = await page.evaluate((td) => td.textContent.trim(), td);
    await expectNotToBeEmpty(val);
  };

  const expectDateUpdatedNotToBeEmpty = async () => {
    const td = await first(obj, XPATH_DATE_UPDATED);
    const val = await page.evaluate((td) => td.textContent.trim(), td);
    await expectNotToBeEmpty(val);
  };

  const expectDescriptionToEqual = async (description) => {
    const td = await first(obj, XPATH_DESCRIPTION);
    const val = await page.evaluate((td) => td.textContent.trim(), td);
    await expectToBeEqual(val, description);
  };

  const expectNameToEqual = async (id, name) => {
    const h1 = await first(obj, XPATH_TITLE);
    const val = await page.evaluate((h1) => h1.textContent.trim(), h1);
    await expectToBeEqual(val, '#' + id + ' - ' + name);
  };

  const expectStatusToEqual = async (status) => {
    const td = await first(obj, XPATH_STATUS);
    const val = await page.evaluate((td) => td.textContent.trim(), td);
    await expectToBeEqual(val, status);
  };

  const expectToBeVisible = async () => {
    const value = await page.evaluate((obj) => obj.style.display, obj);
    await expectToBeEqual(value, '');
  };

  return {
    clickEdit: clickEdit,
    expectDateCreatedNotToBeEmpty: expectDateCreatedNotToBeEmpty,
    expectDateUpdatedNotToBeEmpty: expectDateUpdatedNotToBeEmpty,
    expectDescriptionToEqual: expectDescriptionToEqual,
    expectNameToEqual: expectNameToEqual,
    expectStatusToEqual: expectStatusToEqual,
    expectToBeVisible: expectToBeVisible,
  };
};

const waitForTms = (tms) => {
  return new Promise((resolve, reject) => {
    const re = new RegExp('^Started on 0.0.0.0:([0-9]*)');
    tms.stdout.on('data', (data) => {
      const line = data.toString();
      const match = line.match(re);
      if (match) {
        const port = parseInt(match[1]);
        resolve(['127.0.0.1', port]);
      }
    });
    tms.stderr.on('data', (data) => {
      reject(new Error(data.toString()));
    });
  });
};

const removeDatabaseFile = () =>
  new Promise((resolve, reject) => {
    fs.unlink('tm.db', (err) => {
      if (err) {
        reject(err);
      } else {
        resolve();
      }
    });
  });

const withTmsAndPuppeteer = async (closure) => {
  const tms = await spawn('./tms', ['-p', '0']);
  const [ip, port] = await waitForTms(tms);

  const browser = await puppeteer.launch({
    executablePath: '/usr/bin/chromium',
    args: ['--no-sandbox'],
  });
  const page = await browser.newPage();
  page.on('console', (msg) => {
    console.log(msg.text());
  });

  try {
    await Promise.race([
      closure(ip, port, page),
      new Promise((resolve, reject) => {
        tms.on('close', (code) => {
          reject(new Error('Prematurily exited with code ' + code));
        });
      }),
      new Promise((resolve, reject) => {
        page.on('error', (msg) => {
          reject(new Error(msg));
        });
      }),
    ]);
  } finally {
    await browser.close();
    tms.kill();
    await removeDatabaseFile();
  }
};

module.exports = {
  getAddNewTask: getAddNewTask,
  getEditTask: getEditTask,
  getIndex: getIndex,
  getNavBar: getNavBar,
  getTask: getTask,
  withTmsAndPuppeteer: withTmsAndPuppeteer,
};
