/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <stdio.h>
#include <string.h>

#include <tmm.h>

typedef struct argz argz;
struct argz
{
    int ca;
    int cd;
    int cm;
    int cp;
    int cq;
    int cr;

    char *d;
    int h;

    char *arg;

    unsigned int filters_len;
    unsigned long *filter_i;
    char *filter_d;
    char *filter_n;
    char *filter_s;
    time_t *filter_dcgt;
    time_t *filter_dclt;
    time_t *filter_dugt;
    time_t *filter_dult;

    unsigned int modifiers_len;
    char *modifier_d;
    char *modifier_n;
    char *modifier_s;
};

enum filter_key
{
    FK_NONE,
    FK_DESCRIPTION,
    FK_DATE_CREATED,
    FK_DATE_CREATED_AFTER,
    FK_DATE_CREATED_BEFORE,
    FK_DATE_UPDATED,
    FK_DATE_UPDATED_AFTER,
    FK_DATE_UPDATED_BEFORE,
    FK_ID,
    FK_NAME,
    FK_STATUS
};

static void usage(char *argv0)
{
    fprintf(stderr, "Usage: %s [-h]\n", argv0);
    fprintf(stderr, "       %s [opts] a{dd} [name] {[modifiers]}\n", argv0);
    fprintf(stderr, "       %s [opts] p{rint} [id]\n", argv0);
    fprintf(stderr, "       %s [opts] q{uery} [filters]\n", argv0);
    fprintf(stderr, "       %s [opts] r{emove} [filters]\n", argv0);
    fprintf(stderr, "       %s [opts] m{odify} [filters] -- [modifiers]\n", argv0);
    fprintf(stderr, "\n");
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "       -d/--dbfile dbfile. Default: tm.db\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "Add/modify modifiers:\n");
    fprintf(stderr, "       d/description description\n");
    fprintf(stderr, "       n/name name (not add!)\n");
    fprintf(stderr, "       s/status status\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "Query/remove/modify filters:\n");
    fprintf(stderr, "       d/description description\n");
    fprintf(stderr, "       dc/date-created date\n");
    fprintf(stderr, "       dcgt/date-created-after date\n");
    fprintf(stderr, "       dclt/date-created-before date\n");
    fprintf(stderr, "       du/date-updated date\n");
    fprintf(stderr, "       dugt/date-updated-after date\n");
    fprintf(stderr, "       dult/date-updated-before date\n");
    fprintf(stderr, "       i/id id\n");
    fprintf(stderr, "       n/name name\n");
    fprintf(stderr, "       s/status status\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "Legal dates:\n");
    fprintf(stderr, "       2018\n");
    fprintf(stderr, "       2018-12\n");
    fprintf(stderr, "       2018-12-31\n");
    fprintf(stderr, "       2018-12-31 23\n");
    fprintf(stderr, "       2018-12-31 23:59\n");
    fprintf(stderr, "       2018-12-31 23:59:59\n");
    fprintf(stderr, "       yesterday/now/today/tomorrow\n");
    fprintf(stderr, "       monday/tuesday/wednesday/thursday/friday/saturday/sunday\n");
    fprintf(stderr, "\n");
}

enum parse_command_has
{
    PCH_NO,
    PCH_MAY,
    PCH_MUST
};

static int parse_command(char *command, enum parse_command_has *phas_filters,
    enum parse_command_has *phas_modifiers, enum parse_command_has *phas_arg,
    argz *arg)
{
    if (0 == strcmp("a", command) || 0 == strcmp("add", command))
    {
        *phas_arg = PCH_MUST;
        *phas_modifiers = PCH_MAY;
        arg->ca = 1;
    }
    else if (0 == strcmp("m", command) || 0 == strcmp("modify", command))
    {
        *phas_filters = PCH_MUST;
        *phas_modifiers = PCH_MUST;
        arg->cm = 1;
    }
    else if (0 == strcmp("q", command) || 0 == strcmp("query", command))
    {
        *phas_filters = PCH_MAY;
        arg->cq = 1;
    }
    else if (0 == strcmp("p", command) || 0 == strcmp("print", command))
    {
        *phas_arg = PCH_MUST;
        arg->cp = 1;
    }
    else if (0 == strcmp("r", command) || 0 == strcmp("remove", command))
    {
        *phas_filters = PCH_MUST;
        arg->cr = 1;
    }
    else
    {
        return 0;
    }
    return 1;
}

enum parse_ret
{
    PR_ERROR,
    PR_OK,
    PR_NO_MORE
};

static enum parse_ret parse_opt_key(unsigned int *pi, char *k, char *pl)
{
    if (0 == strcmp("-d", k) ||
        0 == strcmp("--dbfile", k))
    {
        *pl = 'd';
    }
    else if ('-' == k[0])
    {
        return PR_ERROR;
    }
    else
    {
        return PR_NO_MORE;
    }
    return PR_OK;
}

static enum parse_ret parse_filter_key(unsigned int *pi, char *k, enum filter_key *pl)
{
    if (0 == strcmp("d", k) ||
        0 == strcmp("description", k))
    {
        *pl = FK_DESCRIPTION;
    }
    else if (0 == strcmp("dc", k) ||
        0 == strcmp("date-created", k))
    {
        *pl = FK_DATE_CREATED;
    }
    else if (0 == strcmp("dcgt", k) ||
        0 == strcmp("date-created-after", k))
    {
        *pl = FK_DATE_CREATED_AFTER;
    }
    else if (0 == strcmp("dclt", k) ||
        0 == strcmp("date-created-before", k))
    {
        *pl = FK_DATE_CREATED_BEFORE;
    }
    else if (0 == strcmp("du", k) ||
        0 == strcmp("date-updated", k))
    {
        *pl = FK_DATE_UPDATED;
    }
    else if (0 == strcmp("dugt", k) ||
        0 == strcmp("date-updated-after", k))
    {
        *pl = FK_DATE_UPDATED_AFTER;
    }
    else if (0 == strcmp("dult", k) ||
        0 == strcmp("date-updated-before", k))
    {
        *pl = FK_DATE_UPDATED_BEFORE;
    }
    else if (0 == strcmp("i", k) ||
        0 == strcmp("id", k))
    {
        *pl = FK_ID;
    }
    else if (0 == strcmp("n", k) ||
        0 == strcmp("name", k))
    {
        *pl = FK_NAME;
    }
    else if (0 == strcmp("s", k) ||
        0 == strcmp("status", k))
    {
        *pl = FK_STATUS;
    }
    else if (0 == strcmp("--", k))
    {
        (*pi)++;
        return PR_NO_MORE;
    }
    else
    {
        return PR_ERROR;
    }
    return PR_OK;
}

static enum parse_ret parse_modifier_key(unsigned int *pi, char *k, int has_arg, char *pl)
{
    if (0 == strcmp("d", k) ||
        0 == strcmp("description", k))
    {
        *pl = 'd';
    }
    else if (0 == has_arg && (
        0 == strcmp("n", k) ||
        0 == strcmp("name", k)))
    {
        *pl = 'n';
    }
    else if (0 == strcmp("s", k) ||
        0 == strcmp("status", k))
    {
        *pl = 's';
    }
    else
    {
        return PR_ERROR;
    }
    return PR_OK;
}

static void parse_opt_value(char k, char *v, argz *arg)
{
    if ('d' == k)
    {
        free(arg->d);
        arg->d = strdup(v);
    }
}

static int parse_filter_value_date(enum filter_key *pk, char *v, argz *arg)
{
    int is_after = *pk == FK_DATE_CREATED_AFTER || *pk == FK_DATE_UPDATED_AFTER;
    time_t date = 0;

    if (0 == parse_date(v, is_after, &date))
    {
        return 0;
    }

    switch (*pk)
    {
        case FK_DATE_CREATED_AFTER:
            arg->filter_dcgt = (time_t *)malloc(sizeof(time_t));
            (*arg->filter_dcgt) = date;
            break;
        case FK_DATE_CREATED_BEFORE:
            arg->filter_dclt = (time_t *)malloc(sizeof(time_t));
            (*arg->filter_dclt) = date;
            break;
        case FK_DATE_UPDATED_AFTER:
            arg->filter_dugt = (time_t *)malloc(sizeof(time_t));
            (*arg->filter_dugt) = date;
            break;
        case FK_DATE_UPDATED_BEFORE:
            arg->filter_dult = (time_t *)malloc(sizeof(time_t));
            (*arg->filter_dult) = date;
            break;
        default:
            return 0;
    }
    *pk = FK_NONE;
    return 1;
}

static int parse_filter_value_date_created(enum filter_key *pk, char *v, argz *arg)
{
    enum filter_key pa = FK_DATE_CREATED_AFTER;
    enum filter_key pb = FK_DATE_CREATED_BEFORE;
    if (0 == parse_filter_value_date(&pa, v, arg) ||
        0 == parse_filter_value_date(&pb, v, arg))
    {
        return 0;
    }

    *pk = FK_NONE;
    return 1;
}


static int parse_filter_value_date_updated(enum filter_key *pk, char *v, argz *arg)
{
    enum filter_key pa = FK_DATE_UPDATED_AFTER;
    enum filter_key pb = FK_DATE_UPDATED_BEFORE;
    if (0 == parse_filter_value_date(&pa, v, arg) ||
        0 == parse_filter_value_date(&pb, v, arg))
    {
        return 0;
    }

    *pk = FK_NONE;
    return 1;
}

static int parse_filter_value(enum filter_key *pk, char *v, argz *arg)
{
    switch (*pk)
    {
        case FK_DESCRIPTION:
            arg->filter_d = v;
            break;
        case FK_DATE_CREATED:
            return parse_filter_value_date_created(pk, v, arg);
        case FK_DATE_UPDATED:
            return parse_filter_value_date_updated(pk, v, arg);
        case FK_DATE_CREATED_AFTER:
        case FK_DATE_CREATED_BEFORE:
        case FK_DATE_UPDATED_AFTER:
        case FK_DATE_UPDATED_BEFORE:
            return parse_filter_value_date(pk, v, arg);
        case FK_ID:
            if (0 != arg->filter_i)
            {
                free(arg->filter_i);
            }
            arg->filter_i = (unsigned long *)malloc(sizeof(unsigned long));
            *(arg->filter_i) = atol(v);
            break;
        case FK_NAME:
            arg->filter_n = v;
            break;
        case FK_STATUS:
            arg->filter_s = v;
            break;
        default:
            return 0;
    }
    *pk = FK_NONE;
    return 1;
}

static void parse_modifier_value(char k, char *v, argz *arg)
{
    if ('d' == k)
    {
        arg->modifier_d = v;
    }
    else if ('n' == k)
    {
        arg->modifier_n = v;
    }
    else if ('s' == k)
    {
        arg->modifier_s = v;
    }
}

static int parse_opts(unsigned int *pi, int argc, char **argv, argz *arg)
{
    char k = 0;
    for (; (*pi) < argc; (*pi)++)
    {
        char *argvi = argv[*pi];
        if (0 != k)
        {
            parse_opt_value(k, argvi, arg);
            k = 0;
        }
        else
        {
            enum parse_ret ret = parse_opt_key(pi, argvi, &k);
            if (PR_ERROR == ret)
            {
                return 0;
            }
            else if (PR_NO_MORE == ret)
            {
                break;
            }
        }
    }
    return 0 != k ? 0 : 1;
}

static int parse_filters(unsigned int *pi, int argc, char **argv, argz *arg)
{
    enum filter_key k = FK_NONE;
    for (; (*pi) < argc; (*pi)++)
    {
        char *argvi = argv[*pi];
        if (0 == k)
        {
            enum parse_ret ret = parse_filter_key(pi, argvi, &k);
            if (PR_ERROR == ret)
            {
                return 0;
            }
            else if (PR_NO_MORE == ret)
            {
                break;
            }
            arg->filters_len++;
        }
        else if (0 == parse_filter_value(&k, argvi, arg))
        {
            return 0;
        }
    }
    return FK_NONE != k ? 0 : 1;
}

static int parse_modifiers(unsigned int *pi, int argc, char **argv, int has_arg, argz *arg)
{
    char k = 0;
    for (; (*pi) < argc; (*pi)++)
    {
        char *argvi = argv[*pi];
        if (0 == k)
        {
            enum parse_ret ret = parse_modifier_key(pi, argvi, has_arg, &k);
            if (PR_ERROR == ret)
            {
                return 0;
            }
            arg->modifiers_len++;
        }
        else
        {
            parse_modifier_value(k, argvi, arg);
            k = 0;
        }
    }
    return 0 != k ? 0 : 1;
}

static int parse_args(int argc, char **argv, argz *arg)
{
    unsigned int i = 1;
    enum parse_command_has has_arg = PCH_NO;
    enum parse_command_has has_filters = PCH_NO;
    enum parse_command_has has_modifiers = PCH_NO;

    memset(arg, 0, sizeof(argz));
    arg->d = strdup("tm.db");

    if (argc == 1)
    {
        return 0;
    }

    if (0 == strcmp("-h", argv[1]) || 0 == strcmp("--help", argv[1]))
    {
        arg->h = 1;
        return 1;
    }
    if (0 == parse_opts(&i, argc, argv, arg))
    {
        return 0;
    }
    if (0 == parse_command(argv[i++], &has_filters, &has_modifiers,
      &has_arg, arg))
    {
        return 0;
    }
    if (PCH_NO != has_filters && (
        0 == parse_filters(&i, argc, argv, arg) || (
            PCH_MUST == has_filters && 0 == arg->filters_len )))
    {
        return 0;
    }


    if (PCH_NO != has_arg)
    {
        if (0 == argv[i])
        {
            return 0;
        }
        arg->arg = argv[i];
        i++;
    }

    if (PCH_NO != has_modifiers && (
        0 == parse_modifiers(&i, argc, argv, has_arg, arg) || (
        PCH_MUST == has_modifiers && 0 == arg->modifiers_len)))
    {
        return 0;
    }
    return 1;
}

static int do_add(argz *arg)
{
    tm_lib *tm = 0;
    tm_storage *storage = 0;
    int ret = 0;
    tm_task *task = 0;

    tm_storage_sqlite(arg->d, &storage);
    tm_new(storage, &tm);

    tm_task_new(&task);
    tm_task_set_name(task, arg->arg != 0 ? arg->arg : "");

    if (0 != arg->modifier_d)
    {
        tm_task_set_description(task, arg->modifier_d);
    }
    if (0 != arg->modifier_s)
    {
        tm_task_set_status(task, arg->modifier_s);
    }

    if (0 == tm_add_task(tm, task))
    {
        ret = 1;
        fprintf(stderr, "Cannot add task\n");
    }
    tm_task_free(&task);
    tm_free(&tm);
    return ret;
}

static void build_query(argz *arg, tm_lib *tm, tm_query **pquery)
{
    if (0 < arg->filters_len)
    {
        tm_query **children = 0;
        unsigned int i = 0;

        children = (tm_query **)malloc(sizeof(tm_query *) * arg->filters_len);
        memset(children, 0, sizeof(tm_query *) * arg->filters_len);

        if (0 != arg->filter_d)
        {
            tm_query_description_contains(tm, arg->filter_d, &(children[i++]));
        }
        if (0 != arg->filter_dcgt)
        {
            tm_query_date_created_after(tm, *(arg->filter_dcgt), &(children[i++]));
        }
        if (0 != arg->filter_dclt)
        {
            tm_query_date_created_before(tm, *(arg->filter_dclt), &(children[i++]));
        }
        if (0 != arg->filter_dugt)
        {
            tm_query_date_updated_after(tm, *(arg->filter_dugt), &(children[i++]));
        }
        if (0 != arg->filter_dult)
        {
            tm_query_date_updated_before(tm, *(arg->filter_dult), &(children[i++]));
        }
        if (0 != arg->filter_i)
        {
            tm_query_id_equals(tm, *(arg->filter_i), &(children[i++]));
        }
        if (0 != arg->filter_n)
        {
            tm_query_name_contains(tm, arg->filter_n, &(children[i++]));
        }
        if (0 != arg->filter_s)
        {
            tm_query_status_contains(tm, arg->filter_s, &(children[i++]));
        }

        tm_query_and(tm, arg->filters_len, children, pquery);
    }
}

static int alog10(unsigned long a)
{
    unsigned int i = 0;
    for (; a > 0; i++, a/=10)
    {
    }
    return i;
}

static unsigned int get_id_column_len(tm_task **arr, unsigned int arr_len)
{
    unsigned int id_column_len = 2;
    for (unsigned i = 0; i < arr_len; i++)
    {
        unsigned long id = -1;
        unsigned int id_len;

        tm_task_get_id(arr[i], &id);
        id_len = alog10(id);
        if (id_len > id_column_len)
        {
            id_column_len = id_len;
        }
    }
    return id_column_len;
}

static int do_query(argz *arg)
{
    tm_lib *tm = 0;
    tm_storage *storage = 0;
    tm_task **arr = 0;
    unsigned int arr_len = 0;
    tm_query *query = 0;
    unsigned int id_column_len;

    tm_storage_sqlite(arg->d, &storage);
    tm_new(storage, &tm);

    build_query(arg, tm, &query);
    tm_get_tasks(tm, query, &arr, &arr_len);

    id_column_len = get_id_column_len(arr, arr_len);

    printf("ID%*s | Name\n", id_column_len - 2, "");
    for (unsigned i = 0; i < arr_len; i++)
    {
        unsigned long id = -1;
        char *name = 0;
        unsigned int id_len;

        tm_task_get_id(arr[i], &id);
        tm_task_get_name(arr[i], &name);

        id_len = alog10(id);
        printf("%lu%*s | %s\n", id, id_column_len - id_len, "", name);
    }
    tm_task_arr_free(&arr, arr_len);
    tm_query_free(&query);
    tm_free(&tm);
    return 0;
}

static void print_date(time_t date)
{
    char buf[20];
    struct tm tinfo;
    localtime_r(&date, &tinfo);
    strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", &tinfo);
    printf("%s", buf);
}

static void print_task_diff(tm_task_diff *prev, tm_task_diff *next)
{
    if (0 != next)
    {
        time_t next_date = 0;
        char *name = 0;
        char *prev_value = 0;
        char *next_value = 0;

        tm_task_diff_get_name(next, &name);
        tm_task_diff_get_value(next, &next_value);
        tm_task_diff_get_date(next, &next_date);
        if (0 != prev)
        {
            tm_task_diff_get_value(prev, &prev_value);
        }

        printf("- [");
        print_date(next_date);

        if (0 != next_value)
        {
            if (0 != prev_value)
            {
                printf("] changed %s \"%s\" -> \"%s\"\n", name, prev_value, next_value);
            }
            else
            {
                printf("] added %s \"%s\"\n", name, next_value);
            }
        }
        else if (0 != prev_value)
        {
            printf("] removed %s \"%s\"\n", name, prev_value);
        }
    }
}

typedef struct task_diff_print_entry task_diff_print_entry;
struct task_diff_print_entry
{
    tm_task_diff *this;
    tm_task_diff *prev;
};

static void calculate_print_entries(tm_task_diff **diffs, unsigned int len,
    task_diff_print_entry ***pentries)
{
    task_diff_print_entry **entries = (task_diff_print_entry **)malloc(
        sizeof(task_diff_print_entry *) * len);
    task_diff_print_entry *next_name = 0;
    task_diff_print_entry *next_description = 0;
    task_diff_print_entry *next_status = 0;

    for (unsigned int i = 0; i < len; i++)
    {
        task_diff_print_entry *entry;
        char *name = 0;
        task_diff_print_entry **pnext = 0;

        entry = (task_diff_print_entry *)malloc(sizeof(task_diff_print_entry));
        entry->this = diffs[i];
        entry->prev = 0;

        entries[i] = entry;

        tm_task_diff_get_name(diffs[i], &name);
        if (0 == strcmp("name", name))
        {
            pnext = &next_name;
        }
        else if (0 == strcmp("description", name))
        {
            pnext = &next_description;
        }
        else if (0 == strcmp("status", name))
        {
            pnext = &next_status;
        }
        else
        {
            continue;
        }

        if (0 != *pnext)
        {
            (*pnext)->prev = diffs[i];
        }

        *pnext = entry;
    }

    *pentries = entries;
}

static void free_print_entries(task_diff_print_entry ***pentries, unsigned int len)
{
    task_diff_print_entry **entries = *pentries;
    for (unsigned int i = 0; i < len; i++)
    {
        free(entries[i]);
    }
    free(entries);
    *pentries = 0;
}

static void print_task_history(tm_lib *tm, tm_task *task)
{
    task_diff_print_entry **entries = 0;
    tm_task_diff **diffs = 0;
    unsigned int len = 0;

    tm_get_task_diffs(tm, task, &diffs, &len);
    calculate_print_entries(diffs, len, &entries);

    printf("History:\n");
    for (unsigned int i = 0; i < len; i++)
    {
        task_diff_print_entry *entry = entries[i];
        print_task_diff(entry->prev, entry->this);
    }

    free_print_entries(&entries, len);
    tm_task_diff_arr_free(&diffs, len);
}

static void print_name(tm_task *task)
{
    char *name = 0;

    tm_task_get_name(task, &name);
    if (0 != name && 0 != name[0])
    {
        printf("Name: %s\n", name);
    }
}

static void print_description(tm_task *task)
{
    char *description = 0;

    tm_task_get_description(task, &description);
    if (0 != description && 0 != description[0])
    {
        printf("Description: %s\n", description);
    }
}

static void print_status(tm_task *task)
{
    char *status = 0;

    tm_task_get_status(task, &status);
    if (0 != status && 0 != status[0])
    {
        printf("Status: %s\n", status);
    }
}

static int do_print(argz *arg)
{
    tm_lib *tm = 0;
    tm_storage *storage = 0;
    unsigned int id;
    tm_task *task = 0;
    int ret = 0;

    tm_storage_sqlite(arg->d, &storage);
    tm_new(storage, &tm);

    id = atol(arg->arg);
    tm_get_task(tm, id, &task);

    if (0 == task)
    {
        fprintf(stderr, "No such task\n");
        ret = 1;
    }
    else
    {
        time_t date_created = 0;
        time_t date_updated = 0;

        print_name(task);
        print_description(task);
        print_status(task);

        tm_task_get_date_created(task, &date_created);
        printf("Date created: ");
        print_date(date_created);
        printf("\n");

        tm_task_get_date_updated(task, &date_updated);
        printf("Date updated: ");
        print_date(date_updated);
        printf("\n");

        print_task_history(tm, task);
        tm_task_free(&task);
    }
    tm_free(&tm);
    return ret;
}

static int do_remove(argz *arg)
{
    tm_lib *tm = 0;
    tm_storage *storage = 0;
    tm_task **arr = 0;
    unsigned int arr_len = 0;
    tm_query *query = 0;
    int ret = 1;

    tm_storage_sqlite(arg->d, &storage);
    tm_new(storage, &tm);

    build_query(arg, tm, &query);
    if (0 == query)
    {
        fprintf(stderr, "Please provide at least one filter\n");
    }
    else
    {
        tm_get_tasks(tm, query, &arr, &arr_len);
        if (0 == arr_len)
        {
            fprintf(stderr, "No tasks to remove\n");
        }
        else if (0 == tm_remove_tasks(tm, arr, arr_len))
        {
            fprintf(stderr, "Cannot remove %u tasks\n", arr_len);
        }
        else
        {
            ret = 0;
        }
        tm_task_arr_free(&arr, arr_len);
        tm_query_free(&query);
    }
    tm_free(&tm);
    return ret;
}

static int modify_the_tasks(tm_lib *tm, argz *arg, tm_task **arr, unsigned int arr_len)
{
    if (0 != arg->modifier_n)
    {
        tm_task_arr_set_name(arr, arr_len, arg->modifier_n);
    }
    if (0 != arg->modifier_d)
    {
        tm_task_arr_set_description(arr, arr_len, arg->modifier_d);
    }
    if (0 != arg->modifier_s)
    {
        tm_task_arr_set_status(arr, arr_len, arg->modifier_s);
    }

    if (0 == tm_modify_tasks(tm, arr, arr_len))
    {
        fprintf(stderr, "Cannot modify %u tasks\n", arr_len);
    }
    else
    {
        return 0;
    }
    return 1;
}

static int do_modify(argz *arg)
{
    tm_lib *tm = 0;
    tm_storage *storage = 0;
    tm_task **arr = 0;
    unsigned int arr_len = 0;
    tm_query *query = 0;
    int ret = 1;

    tm_storage_sqlite(arg->d, &storage);
    tm_new(storage, &tm);

    build_query(arg, tm, &query);
    if (0 == query)
    {
        fprintf(stderr, "Please provide at least one filter\n");
    }
    else
    {
        tm_get_tasks(tm, query, &arr, &arr_len);
        if (0 == arr_len)
        {
            fprintf(stderr, "No tasks to modify\n");
        }
        else
        {
            ret = modify_the_tasks(tm, arg, arr, arr_len);
        }
        tm_task_arr_free(&arr, arr_len);
        tm_query_free(&query);
    }
    tm_free(&tm);
    return ret;
}

static void argz_free(argz *arg)
{
    if (0 != arg->d)
    {
        free(arg->d);
    }
    if (0 != arg->filter_i)
    {
        free(arg->filter_i);
    }
    if (0 != arg->filter_dcgt)
    {
        free(arg->filter_dcgt);
    }
    if (0 != arg->filter_dclt)
    {
        free(arg->filter_dclt);
    }
    if (0 != arg->filter_dugt)
    {
        free(arg->filter_dugt);
    }
    if (0 != arg->filter_dult)
    {
        free(arg->filter_dult);
    }
}

int main(int argc, char **argv)
{
    argz arg;
    int ret = 1;

    if (0 == parse_args(argc, argv, &arg))
    {
        usage(argv[0]);
    }
    else if (1 == arg.h)
    {
        usage(argv[0]);
        ret = 0;
    }
    else if (1 == arg.ca)
    {
        ret = do_add(&arg);
    }
    else if (1 == arg.cm)
    {
        ret = do_modify(&arg);
    }
    else if (1 == arg.cp)
    {
        ret = do_print(&arg);
    }
    else if (1 == arg.cq)
    {
        ret = do_query(&arg);
    }
    else if (1 == arg.cr)
    {
        ret = do_remove(&arg);
    }
    argz_free(&arg);
    return ret;
}
