/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <errno.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <criterion/criterion.h>

#include "common.h"

TestSuite(add, .init = cache_timezone, .fini = restore_timezone);

Test(add, one)
{
    char *expected_output = "ID | Name\n"
        "1  | Buy milk\n";

    init_folder();
    do_add("Buy milk", 0);
    do_query(expected_output, 0);

    do_query(expected_output, 2, "n", "Buy milk");
    do_query(expected_output, 2, "name", "Buy milk");

    do_query(expected_output, 2, "s", "new");
    do_query(expected_output, 2, "status", "new");

    do_print(1, 1, 5,
        "Name", "Buy milk",
        "Status", "new",
        "Date created", LAST_1_MINUTE,
        "Date updated", LAST_1_MINUTE,
        "History", 2,
            LAST_1_MINUTE, "added", "status", "new",
            LAST_1_MINUTE, "added", "name", "Buy milk");
    clear_folder();
}

Test(add, dates_year)
{
    time_t now = time(0);
    struct tm nowtm;
    char tmp[64];
    char *expected_output = "ID | Name\n"
        "1  | Buy milk\n";

    localtime_r(&now, &nowtm);

    init_folder();
    do_add("Buy milk", 0);

    sprintf(tmp, "%u", nowtm.tm_year + 1900);
    do_query(expected_output, 2, "dc", tmp);
    do_query(expected_output, 2, "date-created", tmp);

    do_query(expected_output, 2, "dcgt", tmp);
    do_query(expected_output, 2, "date-created-after", tmp);

    do_query(expected_output, 2, "dclt", tmp);
    do_query(expected_output, 2, "date-created-before", tmp);

    do_query(expected_output, 2, "du", tmp);
    do_query(expected_output, 2, "date-updated", tmp);

    do_query(expected_output, 2, "dugt", tmp);
    do_query(expected_output, 2, "date-updated-after", tmp);

    do_query(expected_output, 2, "dult", tmp);
    do_query(expected_output, 2, "date-updated-before", tmp);

    clear_folder();
}

Test(add, dates_year_month)
{
    time_t now = time(0);
    struct tm nowtm;
    char tmp[64];
    char *expected_output = "ID | Name\n"
        "1  | Buy milk\n";

    localtime_r(&now, &nowtm);

    init_folder();
    do_add("Buy milk", 0);

    sprintf(tmp, "%u-%u", nowtm.tm_year + 1900, nowtm.tm_mon + 1);
    do_query(expected_output, 2, "dc", tmp);
    do_query(expected_output, 2, "date-created", tmp);

    do_query(expected_output, 2, "dcgt", tmp);
    do_query(expected_output, 2, "date-created-after", tmp);

    do_query(expected_output, 2, "dclt", tmp);
    do_query(expected_output, 2, "date-created-before", tmp);

    do_query(expected_output, 2, "du", tmp);
    do_query(expected_output, 2, "date-updated", tmp);

    do_query(expected_output, 2, "dugt", tmp);
    do_query(expected_output, 2, "date-updated-after", tmp);

    do_query(expected_output, 2, "dult", tmp);
    do_query(expected_output, 2, "date-updated-before", tmp);

    clear_folder();
}

Test(add, dates_year_month_day)
{
    time_t now = time(0);
    struct tm nowtm;
    char tmp[64];
    char *expected_output = "ID | Name\n"
        "1  | Buy milk\n";

    localtime_r(&now, &nowtm);

    init_folder();
    do_add("Buy milk", 0);

    sprintf(tmp, "%u-%u-%u", nowtm.tm_year + 1900, nowtm.tm_mon + 1, nowtm.tm_mday);
    do_query(expected_output, 2, "dc", tmp);
    do_query(expected_output, 2, "date-created", tmp);

    do_query(expected_output, 2, "dcgt", tmp);
    do_query(expected_output, 2, "date-created-after", tmp);

    do_query(expected_output, 2, "dclt", tmp);
    do_query(expected_output, 2, "date-created-before", tmp);

    do_query(expected_output, 2, "du", tmp);
    do_query(expected_output, 2, "date-updated", tmp);

    do_query(expected_output, 2, "dugt", tmp);
    do_query(expected_output, 2, "date-updated-after", tmp);

    do_query(expected_output, 2, "dult", tmp);
    do_query(expected_output, 2, "date-updated-before", tmp);

    clear_folder();
}

Test(add, modifier_description_short)
{
    init_folder();
    do_add("Buy milk", 2, "d", "Goat milk");
    do_query("ID | Name\n"
        "1  | Buy milk\n", 0);
    do_query("ID | Name\n"
        "1  | Buy milk\n",
        2, "d", "Goat milk");
    do_print(1, 1, 6,
        "Name", "Buy milk",
        "Description", "Goat milk",
        "Status", "new",
        "Date created", LAST_1_MINUTE,
        "Date updated", LAST_1_MINUTE,
        "History", 3,
            LAST_1_MINUTE, "added", "status", "new",
            LAST_1_MINUTE, "added", "description", "Goat milk",
            LAST_1_MINUTE, "added", "name", "Buy milk");
    clear_folder();
}

Test(add, modifier_description_long)
{
    init_folder();
    do_add("Buy milk", 2, "description", "Goat milk");
    do_query("ID | Name\n"
        "1  | Buy milk\n", 0);
    do_query("ID | Name\n"
        "1  | Buy milk\n",
        2, "description", "Goat milk");
    do_print(1, 1, 6,
        "Name", "Buy milk",
        "Description", "Goat milk",
        "Status", "new",
        "Date created", LAST_1_MINUTE,
        "Date updated", LAST_1_MINUTE,
        "History", 3,
            LAST_1_MINUTE, "added", "status", "new",
            LAST_1_MINUTE, "added", "description", "Goat milk",
            LAST_1_MINUTE, "added", "name", "Buy milk");
    clear_folder();
}

Test(add, modifier_status_short)
{
    init_folder();
    do_add("Buy milk", 2, "s", "open");
    do_query("ID | Name\n"
        "1  | Buy milk\n", 0);
    do_query("ID | Name\n"
        "1  | Buy milk\n",
        2, "s", "open");
    do_print(1, 1, 5,
        "Name", "Buy milk",
        "Status", "open",
        "Date created", LAST_1_MINUTE,
        "Date updated", LAST_1_MINUTE,
        "History", 2,
            LAST_1_MINUTE, "added", "status", "open",
            LAST_1_MINUTE, "added", "name", "Buy milk");
    clear_folder();
}

Test(add, modifier_status_long)
{
    init_folder();
    do_add("Buy milk", 2, "status", "open");
    do_query("ID | Name\n"
        "1  | Buy milk\n", 0);
    do_query("ID | Name\n"
        "1  | Buy milk\n",
        2, "status", "open");
    do_print(1, 1, 5,
        "Name", "Buy milk",
        "Status", "open",
        "Date created", LAST_1_MINUTE,
        "Date updated", LAST_1_MINUTE,
        "History", 2,
            LAST_1_MINUTE, "added", "status", "open",
            LAST_1_MINUTE, "added", "name", "Buy milk");
    clear_folder();
}

Test(add, many_dbfiles)
{
    init_folder();
    do_add("Buy milk", 0);
    do_add_dbfile("a.db", "Buy beer", 0);

    do_query("ID | Name\n1  | Buy milk\n", 0);
    do_query_dbfile("a.db", "ID | Name\n1  | Buy beer\n", 0);

    do_print(1, 1, 5,
        "Name", "Buy milk",
        "Status", "new",
        "Date created", LAST_1_MINUTE,
        "Date updated", LAST_1_MINUTE,
        "History", 2,
            LAST_1_MINUTE, "added", "status", "new",
            LAST_1_MINUTE, "added", "name", "Buy milk");
    do_print_dbfile("a.db", 1, 1, 5,
        "Name", "Buy beer",
        "Status", "new",
        "Date created", LAST_1_MINUTE,
        "Date updated", LAST_1_MINUTE,
        "History", 2,
            LAST_1_MINUTE, "added", "status", "new",
            LAST_1_MINUTE, "added", "name", "Buy beer");

    remove("a.db");
    clear_folder();
}

Test(add, same)
{
    init_folder();
    do_add("Buy milk", 0);
    do_add("Buy milk", 0);
    do_query("ID | Name\n1  | Buy milk\n2  | Buy milk\n", 0);

    do_print(1, 1, 5,
        "Name", "Buy milk",
        "Status", "new",
        "Date created", LAST_1_MINUTE,
        "Date updated", LAST_1_MINUTE,
        "History", 2,
            LAST_1_MINUTE, "added", "status", "new",
            LAST_1_MINUTE, "added", "name", "Buy milk");
    do_print(1, 2, 5,
        "Name", "Buy milk",
        "Status", "new",
        "Date created", LAST_1_MINUTE,
        "Date updated", LAST_1_MINUTE,
        "History", 2,
            LAST_1_MINUTE, "added", "status", "new",
            LAST_1_MINUTE, "added", "name", "Buy milk");

    clear_folder();
}

Test(add, three)
{
    init_folder();
    do_add("Buy milk", 0);
    do_add("Buy gas", 0);
    do_add("Buy beer", 0);
    do_query("ID | Name\n1  | Buy milk\n2  | Buy gas\n3  | Buy beer\n", 0);
    do_query("ID | Name\n1  | Buy milk\n2  | Buy gas\n3  | Buy beer\n", 2,
        "name", "Buy");
    do_query("ID | Name\n1  | Buy milk\n", 2, "name", "milk");
    do_query("ID | Name\n2  | Buy gas\n", 2, "name", "gas");
    do_query("ID | Name\n3  | Buy beer\n", 2, "name", "beer");

    do_print(1, 1, 5,
        "Name", "Buy milk",
        "Status", "new",
        "Date created", LAST_1_MINUTE,
        "Date updated", LAST_1_MINUTE,
        "History", 2,
            LAST_1_MINUTE, "added", "status", "new",
            LAST_1_MINUTE, "added", "name", "Buy milk");
    do_print(1, 2, 5,
        "Name", "Buy gas",
        "Status", "new",
        "Date created", LAST_1_MINUTE,
        "Date updated", LAST_1_MINUTE,
        "History", 2,
            LAST_1_MINUTE, "added", "status", "new",
            LAST_1_MINUTE, "added", "name", "Buy gas");
    do_print(1, 3, 5,
        "Name", "Buy beer",
        "Status", "new",
        "Date created", LAST_1_MINUTE,
        "Date updated", LAST_1_MINUTE,
        "History", 2,
            LAST_1_MINUTE, "added", "status", "new",
            LAST_1_MINUTE, "added", "name", "Buy beer");

    clear_folder();
}
