/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <errno.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <criterion/criterion.h>
#include <subprocess.h>

#include "common.h"

/* LCOV_EXCL_START */

static char *expected_usage = "Usage: ./tm [-h]\n"
                              "       ./tm [opts] a{dd} [name] {[modifiers]}\n"
                              "       ./tm [opts] p{rint} [id]\n"
                              "       ./tm [opts] q{uery} [filters]\n"
                              "       ./tm [opts] r{emove} [filters]\n"
                              "       ./tm [opts] m{odify} [filters] -- [modifiers]\n"
                              "\n"
                              "Options:\n"
                              "       -d/--dbfile dbfile. Default: tm.db\n"
                              "\n"
                              "Add/modify modifiers:\n"
                              "       d/description description\n"
                              "       n/name name (not add!)\n"
                              "       s/status status\n"
                              "\n"
                              "Query/remove/modify filters:\n"
                              "       d/description description\n"
                              "       dc/date-created date\n"
                              "       dcgt/date-created-after date\n"
                              "       dclt/date-created-before date\n"
                              "       du/date-updated date\n"
                              "       dugt/date-updated-after date\n"
                              "       dult/date-updated-before date\n"
                              "       i/id id\n"
                              "       n/name name\n"
                              "       s/status status\n"
                              "\n"
                              "Legal dates:\n"
                              "       2018\n"
                              "       2018-12\n"
                              "       2018-12-31\n"
                              "       2018-12-31 23\n"
                              "       2018-12-31 23:59\n"
                              "       2018-12-31 23:59:59\n"
                              "       yesterday/now/today/tomorrow\n"
                              "       monday/tuesday/wednesday/thursday/friday/saturday/sunday\n"
                              "\n";

void *LAST_1_MINUTE = &LAST_1_MINUTE;

void run_with_args(char **argv, unsigned int expected_exit, char *expected_stdout,
    unsigned int expected_stdout_len, char *expected_stderr, unsigned int expected_stderr_len)
{
    subprocess *sub = 0;
    subprocess_partial_stream *ps = 0;
    char *env[] = {"LC_ALL", "C", 0};

    if (0 == subprocess_new(argv, env, 1, 1, &sub))
    {
        return;
    }

    subprocess_read_output(sub, 10, &ps);
    subprocess_partial_stream_cmp(ps, expected_stdout_len, expected_stdout);
    subprocess_partial_stream_free(&ps);

    subprocess_read_error(sub, 0, &ps);
    subprocess_partial_stream_cmp(ps, expected_stderr_len, expected_stderr);
    subprocess_partial_stream_free(&ps);

    subprocess_wait(sub, expected_exit);
    subprocess_free(&sub);
}

void expect_usage(char **argv, unsigned int code)
{
    subprocess *sub = 0;
    subprocess_partial_stream *ps = 0;

    if (0 == subprocess_new(argv, 0, 1, 1, &sub))
    {
        return;
    }

    subprocess_read_output(sub, 10, &ps);
    subprocess_partial_stream_free(&ps);

    subprocess_read_error(sub, 0, &ps);
    subprocess_partial_stream_cmp(ps, strlen(expected_usage), expected_usage);
    subprocess_partial_stream_free(&ps);

    subprocess_wait(sub, code);
    subprocess_free(&sub);
}

static void do_add_dbfile_v(char *dbfile, char *name, unsigned int modifiers_len,
    va_list list)
{
    char **argv;
    unsigned int i = 0;
    unsigned int j;
    unsigned int argv_len = 4 + (0 != dbfile ? 2 : 0) + modifiers_len;

    argv = (char **)malloc(sizeof(char *) * argv_len);
    argv[i++] = "../tm";
    if (0 != dbfile)
    {
        argv[i++] = "-d";
        argv[i++] = dbfile;
    }
    argv[i++] = "a";
    argv[i++] = name;

    for (j = 0; j < modifiers_len; j++)
    {
        argv[i++] = va_arg(list, char *);
    }
    argv[i++] = 0;

    run_with_args(argv, 0, "", 0, "", 0);
    free(argv);
}

void do_add_dbfile(char *dbfile, char *name, unsigned int modifiers_len, ...)
{
    va_list list;
    va_start(list, modifiers_len);
    do_add_dbfile_v(dbfile, name, modifiers_len, list);
    va_end(list);
}

void do_add(char *name, unsigned int modifiers_len, ...)
{
    va_list list;
    va_start(list, modifiers_len);
    do_add_dbfile_v(0, name, modifiers_len, list);
    va_end(list);
}

static void do_remove_dbfile_v(char *dbfile, int found, unsigned int filters_len, va_list list)
{
    char **argv;
    unsigned int i = 0;
    unsigned int j;
    unsigned int argv_len = 3 + (0 != dbfile ? 2 : 0) + filters_len;

    argv = (char **)malloc(sizeof(char *) * argv_len);
    argv[i++] = "../tm";
    if (0 != dbfile)
    {
        argv[i++] = "-d";
        argv[i++] = dbfile;
    }
    argv[i++] = "r";

    for (j = 0; j < filters_len; j++)
    {
        argv[i++] = va_arg(list, char *);
    }
    argv[i++] = 0;

    if (1 == found)
    {
        run_with_args(argv, 0, "", 0, "", 0);
    }
    else
    {
        run_with_args(argv, 1, "", 0, "No tasks to remove\n", 19);
    }
    free(argv);
}

/*void do_remove_dbfile(char *dbfile, int found, unsigned int filters_len, ...)
{
    va_list list;
    va_start(list, filters_len);
    do_remove_dbfile_v(dbfile, found, filters_len, list);
    va_end(list);
}*/

void do_remove(int found, unsigned int filters_len, ...)
{
    va_list list;
    va_start(list, filters_len);
    do_remove_dbfile_v(0, found, filters_len, list);
    va_end(list);
}

static void do_modify_dbfile_v(char *dbfile, int found, unsigned int filters_len,
    unsigned int modifiers_len, va_list list)
{
    char **argv;
    unsigned int i = 0;
    unsigned int j;
    unsigned int argv_len = 4 + (0 != dbfile ? 2 : 0) + filters_len + modifiers_len;

    argv = (char **)malloc(sizeof(char *) * argv_len);
    argv[i++] = "../tm";
    if (0 != dbfile)
    {
        argv[i++] = "-d";
        argv[i++] = dbfile;
    }
    argv[i++] = "m";

    for (j = 0; j < filters_len; j++)
    {
        argv[i++] = va_arg(list, char *);
    }
    argv[i++] = "--";
    for (j = 0; j < modifiers_len; j++)
    {
        argv[i++] = va_arg(list, char *);
    }
    argv[i++] = 0;

    if (1 == found)
    {
        run_with_args(argv, 0, "", 0, "", 0);
    }
    else
    {
        run_with_args(argv, 1, "", 0, "No tasks to modify\n", 19);
    }
    free(argv);
}

/*void do_modify_dbfile(char *dbfile, int found, unsigned int filters_len,
    unsigned int modifiers_len, ...)
{
    va_list list;
    va_start(list, modifiers_len);
    do_modify_dbfile_v(dbfile, found, filters_len, modifiers_len, list);
    va_end(list);
}*/

void do_modify(int found, unsigned int filters_len, unsigned int modifiers_len, ...)
{
    va_list list;
    va_start(list, modifiers_len);
    do_modify_dbfile_v(0, found, filters_len, modifiers_len, list);
    va_end(list);
}

static void do_query_dbfile_v(char *dbfile, char *output, unsigned int filters_len,
    va_list list)
{
    char **argv;
    unsigned int i = 0;
    unsigned int j;
    unsigned int argv_len = 3 + (0 != dbfile ? 2 : 0) + filters_len;

    argv = (char **)malloc(sizeof(char *) * argv_len);
    argv[i++] = "../tm";
    if (0 != dbfile)
    {
        argv[i++] = "-d";
        argv[i++] = dbfile;
    }
    argv[i++] = "q";

    for (j = 0; j < filters_len; j++)
    {
        argv[i++] = va_arg(list, char *);
    }
    argv[i++] = 0;

    run_with_args(argv, 0, output, strlen(output), "", 0);
    free(argv);
}

void do_query_dbfile(char *dbfile, char *output, unsigned int filters_len, ...)
{
    va_list list;
    va_start(list, filters_len);
    do_query_dbfile_v(dbfile, output, filters_len, list);
    va_end(list);
}

void do_query(char *output, unsigned int filters_len, ...)
{
    va_list list;
    va_start(list, filters_len);
    do_query_dbfile_v(0, output, filters_len, list);
    va_end(list);
}

static void expect_date_in_range(time_t a, time_t b, time_t c)
{
    if (a > b || b > c)
    {
        char tmp[64];
        char tmp2[64];
        char tmp3[64];
        struct tm tma;
        struct tm tmb;
        struct tm tmc;

        localtime_r(&a, &tma);
        localtime_r(&b, &tmb);
        localtime_r(&c, &tmc);
        strftime(tmp, sizeof(tmp), "%c", &tma);
        strftime(tmp2, sizeof(tmp2), "%c", &tmb);
        strftime(tmp3, sizeof(tmp3), "%c", &tmc);

        cr_expect_fail("Date not in range\n"
            "Expected to be after: %s (%lu)\n"
            "Expected to be below: %s (%lu)\n"
            "Actually is         : %s (%lu)",
            tmp, a, tmp3, c, tmp2, b);
    }
}

static void do_print_found_date(subprocess_partial_stream *ps, void *date)
{
    cr_log_info("do_print_found_date");

    if (date == LAST_1_MINUTE)
    {
        struct tm atime;
        memset(&atime, 0, sizeof(struct tm));

        time_t now = time(0);
        time_t then = 0;
        char *tmp = 0;

        /* 2012 */
        subprocess_partial_stream_get_for(ps, 4, &tmp);
        if (0 != tmp && 0 != *tmp)
        {
            atime.tm_year = atoi(tmp) - 1900;
        }
        subprocess_partial_stream_cmp(ps, 1, "-");

        /* 08 */
        subprocess_partial_stream_get_for(ps, 2, &tmp);
        if (0 != tmp && 0 != *tmp)
        {
            atime.tm_mon = atoi(tmp) - 1;
        }
        subprocess_partial_stream_cmp(ps, 1, "-");

        /* 19 */
        subprocess_partial_stream_get_for(ps, 2, &tmp);
        if (0 != tmp && 0 != *tmp)
        {
            atime.tm_mday = atoi(tmp);
        }
        subprocess_partial_stream_cmp(ps, 1, " ");

        /* 02 */
        subprocess_partial_stream_get_for(ps, 2, &tmp);
        if (0 != tmp && 0 != *tmp)
        {
            atime.tm_hour = atoi(tmp);
        }
        subprocess_partial_stream_cmp(ps, 1, ":");

        /* 56 */
        subprocess_partial_stream_get_for(ps, 2, &tmp);
        if (0 != tmp && 0 != *tmp)
        {
            atime.tm_min = atoi(tmp);
        }
        subprocess_partial_stream_cmp(ps, 1, ":");

        /* 02 */
        subprocess_partial_stream_get_for(ps, 2, &tmp);
        if (0 != tmp && 0 != *tmp)
        {
            atime.tm_sec = atoi(tmp);
        }

        then = mktime(&atime);
        expect_date_in_range(now - 60, then, now);
    }
}

static void do_print_found_history(subprocess_partial_stream *ps, unsigned int history_len,
    va_list list)
{
    cr_log_info("do_print_found_history len=%u", history_len);

    for (unsigned int i = 0; i < history_len; i++)
    {
        void *date;
        char *type;
        char *name;
        char *value;

        date = va_arg(list, void *);
        type = va_arg(list, char *);
        name = va_arg(list, char *);
        value = va_arg(list, char *);

        subprocess_partial_stream_cmp(ps, 3, "- [");
        do_print_found_date(ps, date);
        subprocess_partial_stream_cmp(ps, 2, "] ");
        subprocess_partial_stream_cmp(ps, strlen(type), type);
        subprocess_partial_stream_cmp(ps, 1, " ");
        subprocess_partial_stream_cmp(ps, strlen(name), name);
        subprocess_partial_stream_cmp(ps, 2, " \"");
        subprocess_partial_stream_cmp(ps, strlen(value), value);
        subprocess_partial_stream_cmp(ps, 1, "\"");

        if (0 == strcmp("changed", type))
        {
            value = va_arg(list, char *);
            subprocess_partial_stream_cmp(ps, 5, " -> \"");
            subprocess_partial_stream_cmp(ps, strlen(value), value);
            subprocess_partial_stream_cmp(ps, 1, "\"");
        }
        subprocess_partial_stream_cmp(ps, 1, "\n");
    }
}

static void do_print_found(subprocess_partial_stream *ps, unsigned int props_len, va_list list)
{
    for (unsigned int i = 0; i < props_len; i++)
    {
        char *name = va_arg(list, char *);
        subprocess_partial_stream_cmp(ps, strlen(name), name);
        subprocess_partial_stream_cmp(ps, 1, ":");
        cr_log_info("do_print_found i=%u/%u", i, props_len);

        if (0 == strcmp("History", name))
        {
            unsigned int history_len = va_arg(list, unsigned int);
            subprocess_partial_stream_cmp(ps, 1, "\n");
            do_print_found_history(ps, history_len, list);
        }
        else if (0 == strcmp("Date created", name) ||
            0 == strcmp("Date updated", name))
        {
            void *date = va_arg(list, void *);

            subprocess_partial_stream_cmp(ps, 1, " ");
            do_print_found_date(ps, date);
            subprocess_partial_stream_cmp(ps, 1, "\n");
        }
        else if (0 == strcmp("Name", name) ||
            0 == strcmp("Description", name) ||
            0 == strcmp("Status", name))
        {
            char *value = va_arg(list, char *);

            subprocess_partial_stream_cmp(ps, 1, " ");
            subprocess_partial_stream_cmp(ps, strlen(value), value);
            subprocess_partial_stream_cmp(ps, 1, "\n");
        }
    }
}

void do_print_dbfile_v(char *dbfile, int found, unsigned int id,
    unsigned int props_len, va_list list)
{
    subprocess *sub = 0;
    char *argv[6];
    unsigned int i = 0;
    char *tmp;
    subprocess_partial_stream *out = 0;
    subprocess_partial_stream *err = 0;
    char *env[] = {"LC_ALL", "C", 0};
    int ret;

    argv[i++] = "../tm";
    if (0 != dbfile)
    {
        argv[i++] = "-d";
        argv[i++] = dbfile;
    }
    argv[i++] = "p";

    tmp = (char *)malloc(sizeof(char) * 128);
    sprintf(tmp, "%u", id);
    argv[i++] = tmp;
    argv[i++] = 0;

    ret = subprocess_new(argv, env, 1, 1, &sub);
    free(tmp);
    if (0 == ret)
    {
        return;
    }

    subprocess_read_output(sub, 10, &out);
    subprocess_read_error(sub, 0, &err);

    if (0 == found)
    {
        subprocess_partial_stream_cmp(err, 13, "No such task\n");
    }
    else
    {
        do_print_found(out, props_len, list);
    }
    subprocess_partial_stream_free(&out);
    subprocess_partial_stream_free(&err);
    subprocess_wait(sub, 1 - found);
    subprocess_free(&sub);
}

void do_print(int found, unsigned int id, unsigned int props_len, ...)
{
    va_list list;
    va_start(list, props_len);
    do_print_dbfile_v(0, found, id, props_len, list);
    va_end(list);
}

void do_print_dbfile(char *file, int found, unsigned int id, unsigned int props_len, ...)
{
    va_list list;
    va_start(list, props_len);
    do_print_dbfile_v(file, found, id, props_len, list);
    va_end(list);
}

void init_folder(void)
{
    sprintf(tm_dir, "tm-%d", getpid());
    if (0 != mkdir(tm_dir, 0700))
    {
        cr_expect_fail("Cannot mkdir() %s: %d %s", tm_dir, errno, strerror(errno));
    }
    else if (0 != chdir(tm_dir))
    {
        rmdir(tm_dir);
        cr_expect_fail("Cannot chdir() %s: %d %s", tm_dir, errno, strerror(errno));
    }
}

void clear_folder(void)
{
    remove("tm.db");
    chdir("..");
    rmdir(tm_dir);
}

static char *tz;

void cache_timezone(void)
{
    tz = getenv("TZ");
    cr_log_info("Caching time zone: %s", tz);
    setenv("TZ", "", 1);
    tzset();
}

void restore_timezone(void)
{
    cr_log_info("Restoring time zone: %s", tz);
    if (0 != tz)
    {
        setenv("TZ", tz, 1);
    }
    else
    {
        unsetenv("TZ");
    }
    tzset();
}

/* LCOV_EXCL_STOP */
