/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <stdio.h>
#include <string.h>

#include <jansson.h>
#include <tm_dateparser.h>

#include <tms.h>

typedef struct modifiers modifiers;
typedef struct query_opts query_opts;

struct modifiers
{
    char *name;
    char *description;
    char *status;
};

struct query_opts
{
    int show_column_description;
    int show_column_status;
};

static void modifiers_free(modifiers *the_modifiers)
{
    if (0 != the_modifiers->name)
    {
        free(the_modifiers->name);
    }
    if (0 != the_modifiers->description)
    {
        free(the_modifiers->description);
    }
    if (0 != the_modifiers->status)
    {
        free(the_modifiers->status);
    }
}

static int is_valid_string_date(char *str)
{
    time_t ad = 0;
    return 1 == tm_date_parser_parse(strlen(str), str, TDPM_WANT_EARLIER_DATES, &ad);
}

static int validate_filters(json_t *id_equals, json_t *name_contains,
    json_t *description_contains, json_t *status_contains, json_t *date_created_after,
    json_t *date_created_before, json_t *date_updated_after, json_t *date_updated_before)
{
    if (0 != id_equals && 0 == json_is_integer(id_equals))
    {
        return 0;
    }

    if (0 != name_contains && 0 == json_is_string(name_contains))
    {
        return 0;
    }

    if (0 != description_contains && 0 == json_is_string(description_contains))
    {
        return 0;
    }

    if (0 != status_contains && 0 == json_is_string(status_contains))
    {
        return 0;
    }
    if (0 != date_created_after && (0 == json_is_string(date_created_after) ||
        0 == is_valid_string_date((char *)json_string_value(date_created_after))))
    {
        return 0;
    }
    if (0 != date_created_before && (0 == json_is_string(date_created_before) ||
        0 == is_valid_string_date((char *)json_string_value(date_created_before))))
    {
        return 0;
    }
    if (0 != date_updated_after && (0 == json_is_string(date_updated_after) ||
        0 == is_valid_string_date((char *)json_string_value(date_updated_after))))
    {
        return 0;
    }
    if (0 != date_updated_before && (0 == json_is_string(date_updated_before) ||
        0 == is_valid_string_date((char *)json_string_value(date_updated_before))))
    {
        return 0;
    }
    return 1;
}


static void fill_filters_into_array(tm_query **children,
    json_t *id_equals, json_t *name_contains, json_t *description_contains,
    json_t *status_contains, json_t *date_created_after, json_t *date_created_before,
    json_t *date_updated_after, json_t *date_updated_before)
{
    unsigned int i = 0;
    if (0 != id_equals)
    {
        tm_query_id_equals(taskman, json_integer_value(id_equals), &(children[i++]));
    }
    if (0 != name_contains)
    {
        tm_query_name_contains(taskman, (char *)json_string_value(name_contains),
            &(children[i++]));
    }
    if (0 != description_contains)
    {
        tm_query_description_contains(taskman, (char *)json_string_value(description_contains),
            &(children[i++]));
    }
    if (0 != status_contains)
    {
        tm_query_status_contains(taskman, (char *)json_string_value(status_contains),
            &(children[i++]));
    }
    if (0 != date_created_after)
    {
        time_t date_time = 0;
        char *date_str = (char *)json_string_value(date_created_after);
        tm_date_parser_parse(strlen(date_str), date_str, TDPM_WANT_LATER_DATES, &date_time);
        tm_query_date_created_after(taskman, date_time, &(children[i++]));
    }
    if (0 != date_created_before)
    {
        time_t date_time = 0;
        char *date_str = (char *)json_string_value(date_created_before);
        tm_date_parser_parse(strlen(date_str), date_str, TDPM_WANT_EARLIER_DATES, &date_time);
        tm_query_date_created_before(taskman, date_time, &(children[i++]));
    }
    if (0 != date_updated_after)
    {
        time_t date_time = 0;
        char *date_str = (char *)json_string_value(date_updated_after);
        tm_date_parser_parse(strlen(date_str), date_str, TDPM_WANT_LATER_DATES, &date_time);
        tm_query_date_updated_after(taskman, date_time, &(children[i++]));
    }
    if (0 != date_updated_before)
    {
        time_t date_time = 0;
        char *date_str = (char *)json_string_value(date_updated_before);
        tm_date_parser_parse(strlen(date_str), date_str, TDPM_WANT_EARLIER_DATES, &date_time);
        tm_query_date_updated_before(taskman, date_time, &(children[i++]));
    }
}

static int extract_filters(json_t *root, tm_query **pquery)
{
    unsigned int len = 0;
    json_t *id_equals = 0;
    json_t *name_contains = 0;
    json_t *description_contains = 0;
    json_t *status_contains = 0;
    json_t *date_created_after = 0;
    json_t *date_created_before = 0;
    json_t *date_updated_after = 0;
    json_t *date_updated_before = 0;
    tm_query **children = 0;

    if (0 == json_object_size(root))
    {
        return 1;
    }

    id_equals = json_object_get(root, "id-equals");
    name_contains = json_object_get(root, "name-contains");
    description_contains = json_object_get(root, "description-contains");
    status_contains = json_object_get(root, "status-contains");
    date_created_after = json_object_get(root, "date-created-after");
    date_created_before = json_object_get(root, "date-created-before");
    date_updated_after = json_object_get(root, "date-updated-after");
    date_updated_before = json_object_get(root, "date-updated-before");

    if (0 == validate_filters(id_equals, name_contains, description_contains,
        status_contains, date_created_after, date_created_before, date_updated_after,
        date_updated_before))
    {
        return 0;
    }

    len += 0 != id_equals;
    len += 0 != name_contains;
    len += 0 != description_contains;
    len += 0 != status_contains;
    len += 0 != date_created_after;
    len += 0 != date_created_before;
    len += 0 != date_updated_after;
    len += 0 != date_updated_before;

    if (0 == len)
    {
        return 0;
    }

    children = (tm_query **)malloc(sizeof(tm_query *) * len);
    memset(children, 0, sizeof(tm_query *) * len);
    fill_filters_into_array(children, id_equals, name_contains, description_contains,
        status_contains, date_created_after, date_created_before, date_updated_after,
        date_updated_before);

    if (1 == len)
    {
        *pquery = children[0];
        free(children);
    }
    else
    {
        tm_query_and(taskman, len, children, pquery);
    }
    return 1;
}

static int extract_opts(json_t *root, query_opts *opts)
{
    unsigned int len = 0;
    json_t *show_column_description = 0;
    json_t *show_column_status = 0;

    if (0 == json_object_size(root))
    {
        return 1;
    }

    if (0 != (show_column_description = json_object_get(root, "show-column-description")))
    {
        if (0 == json_is_true(show_column_description))
        {
            return 0;
        }
        len++;
    }

    if (0 != (show_column_status = json_object_get(root, "show-column-status")))
    {
        if (0 == json_is_true(show_column_status))
        {
            return 0;
        }
        len++;
    }

    if (0 == len)
    {
        return 0;
    }

    if (0 != show_column_description)
    {
        opts->show_column_description = 1;
    }
    if (0 != show_column_status)
    {
        opts->show_column_status = 1;
    }
    return 1;
}

static void add_column(json_t *columns, char *name)
{
    json_t *column = json_string(name);

    json_array_append(columns, column);
    json_decref(column);
}

static void append_query_item(query_opts *opts, json_t *items, tm_task *task)
{
    unsigned long id = -1;
    char *tmpstr = 0;
    json_t *tmp = 0;
    json_t *entry;

    entry = json_object();

    tm_task_get_id(task, &id);
    tmp = json_integer(id);
    json_object_set(entry, "id", tmp);
    json_decref(tmp);

    tm_task_get_name(task, &tmpstr);
    tmp = 0 != tmpstr ? json_string(tmpstr) : json_null();
    json_object_set(entry, "name", tmp);
    json_decref(tmp);

    if (1 == opts->show_column_description)
    {
        tm_task_get_description(task, &tmpstr);
        tmp = 0 != tmpstr ? json_string(tmpstr) : json_null();
        json_object_set(entry, "description", tmp);
        json_decref(tmp);
    }

    if (1 == opts->show_column_status)
    {
        tm_task_get_status(task, &tmpstr);
        tmp = 0 != tmpstr ? json_string(tmpstr) : json_null();
        json_object_set(entry, "status", tmp);
        json_decref(tmp);
    }

    json_array_append(items, entry);
    json_decref(entry);
}

static void create_query_reply(query_opts *opts, tm_task **arr, unsigned int arr_len, char **pret)
{
    json_t *ret_json = json_object();
    json_t *columns = json_array();
    json_t *items = json_array();
    char *ret_str = 0;

    if (0 != opts->show_column_description)
    {
        add_column(columns, "description");
    }
    if (0 != opts->show_column_status)
    {
        add_column(columns, "status");
    }

    json_object_set(ret_json, "columns", columns);
    json_decref(columns);

    json_object_set(ret_json, "items", items);
    for (unsigned i = 0; i < arr_len; i++)
    {
        append_query_item(opts, items, arr[i]);
    }
    json_decref(items);

    ret_str = json_dumps(ret_json, JSON_COMPACT);
    json_decref(ret_json);
    *pret = ret_str;
}

static int extract_query(json_t *root, tm_query **pquery, query_opts *opts)
{
    json_t *filters_json = 0;
    json_t *opts_json = 0;
    unsigned int size = 0;

    filters_json = json_object_get(root, "filters");
    opts_json = json_object_get(root, "opts");

    size += (0 != filters_json ? 1 : 0);
    size += (0 != opts_json ? 1 : 0);

    if (size != json_object_size(root))
    {
        return 0;
    }

    if (0 != filters_json && 0 == extract_filters(filters_json, pquery))
    {
        return 0;
    }

    if (0 != opts_json && 0 == extract_opts(opts_json, opts))
    {
        return 0;
    }
    return 1;
}

void do_api_query(struct mg_connection *conn, struct http_message *hm)
{
    tm_query *query = 0;
    query_opts opts;
    tm_task **arr = 0;
    unsigned int arr_len = 0;
    char *reply = 0;
    int ret = 1;

    memset(&opts, 0, sizeof(opts));

    if (0 == mg_vcmp(&(hm->method), "POST"))
    {
        json_error_t err;
        json_t *root;

        root = json_loadb(hm->body.p, hm->body.len, 0, &err);
        if (0 == root)
        {
            send_error(conn, hm, 400);
            return;
        }

        ret = extract_query(root, &query, &opts);
        json_decref(root);

        if (0 == ret)
        {
            send_error(conn, hm, 400);
            return;
        }
    }
    else if (0 != mg_vcmp(&(hm->method), "GET"))
    {
        send_error(conn, hm, 405);
        return;
    }

    ret = tm_get_tasks(taskman, query, &arr, &arr_len);
    tm_query_free(&query);

    /* LCOV_EXCL_START */
    if (0 == ret)
    {
        send_error(conn, hm, 500);
        return;
    }
    /* LCOV_EXCL_STOP */

    create_query_reply(&opts, arr, arr_len, &reply);
    tm_task_arr_free(&arr, arr_len);

    send_ok(conn, hm, "Content-Type: application/json\r\n"
        "Cache-Control: no-store", strlen(reply), reply);
    free(reply);
}

static void add_task_history_diff_json(json_t *ret, char *type, char *name, time_t date,
    char *value, char *new_value)
{
    json_t *item;
    json_t *tmp;
    item = json_object();

    tmp = json_string(type);
    json_object_set(item, "type", tmp);
    json_decref(tmp);

    tmp = json_string(name);
    json_object_set(item, "name", tmp);
    json_decref(tmp);

    tmp = json_string(value);
    json_object_set(item, "value", tmp);
    json_decref(tmp);

    tmp = json_integer(date);
    json_object_set(item, "date", tmp);
    json_decref(tmp);

    if (0 != new_value)
    {
        tmp = json_string(new_value);
        json_object_set(item, "new-value", tmp);
        json_decref(tmp);
    }

    json_array_append(ret, item);
    json_decref(item);
}

static void add_task_history_diff(char *name, tm_task_diff *prev, tm_task_diff **plast,
    json_t *ret)
{
    tm_task_diff *last = *plast;

    if (0 != *plast)
    {
        char *type = 0;
        char *prev_value = 0;
        char *last_value = 0;
        char *value = 0;
        time_t date = 0;

        if (0 != prev)
        {
            tm_task_diff_get_value(prev, &prev_value);
        }
        tm_task_diff_get_value(last, &last_value);
        tm_task_diff_get_date(last, &date);

        value = prev_value;

        if (0 != last_value)
        {
            if (0 == prev_value)
            {
                type = "add";
                value = last_value;
            }
            else
            {
                type = "change";
            }
        }
        else if (0 != prev_value)
        {
            type = "remove";
        }

        add_task_history_diff_json(ret, type, name, date, value,
            0 != prev_value && 0 != last_value ? last_value : 0);
    }

    *plast = prev;
}

static void create_task_history_json(tm_task *task, json_t **pret)
{
    tm_task_diff **diffs = 0;
    unsigned int len = 0;
    json_t *ret = json_array();

    tm_task_diff *last_name = 0;
    tm_task_diff *last_description = 0;
    tm_task_diff *last_status = 0;

    tm_get_task_diffs(taskman, task, &diffs, &len);

    for (unsigned int i = 0; i < len; i++)
    {
        char *name = 0;
        tm_task_diff_get_name(diffs[i], &name);

        if (0 == strcmp("name", name))
        {
            add_task_history_diff(name, diffs[i], &last_name, ret);
        }
        if (0 == strcmp("description", name))
        {
            add_task_history_diff(name, diffs[i], &last_description, ret);
        }
        if (0 == strcmp("status", name))
        {
            add_task_history_diff(name, diffs[i], &last_status, ret);
        }
    }

    if (0 != last_name)
    {
        add_task_history_diff("name", 0, &last_name, ret);
    }
    if (0 != last_description)
    {
        add_task_history_diff("description", 0, &last_description, ret);
    }
    if (0 != last_status)
    {
        add_task_history_diff("status", 0, &last_status, ret);
    }

    tm_task_diff_arr_free(&diffs, len);
    *pret = ret;
}

static void create_task_reply(tm_task *task, char **pret)
{
    json_t *ret_json = json_object();
    char *ret_str = 0;
    char *tmp2 = 0;
    time_t tmp3 = 0;
    json_t *tmp = 0;

    tm_task_get_name(task, &tmp2);
    tmp = json_string(tmp2);
    json_object_set(ret_json, "name", tmp);
    json_decref(tmp);

    tm_task_get_description(task, &tmp2);
    if (0 != tmp2)
    {
        tmp = json_string(tmp2);
        json_object_set(ret_json, "description", tmp);
        json_decref(tmp);
    }

    tm_task_get_status(task, &tmp2);
    if (0 != tmp2)
    {
        tmp = json_string(tmp2);
        json_object_set(ret_json, "status", tmp);
        json_decref(tmp);
    }

    tm_task_get_date_created(task, &tmp3);
    tmp = json_integer(tmp3);
    json_object_set(ret_json, "date-created", tmp);
    json_decref(tmp);

    tm_task_get_date_updated(task, &tmp3);
    tmp = json_integer(tmp3);
    json_object_set(ret_json, "date-updated", tmp);
    json_decref(tmp);

    create_task_history_json(task, &tmp);
    json_object_set(ret_json, "history", tmp);
    json_decref(tmp);

    ret_str = json_dumps(ret_json, JSON_COMPACT);
    json_decref(ret_json);
    *pret = ret_str;
}

void do_api_task(struct mg_connection *conn, struct http_message *hm)
{
    tm_task *task = 0;
    char *reply = 0;
    int ret = 1;
    unsigned long id = -1;

    if (0 != mg_vcmp(&(hm->method), "GET"))
    {
        send_error(conn, hm, 405);
        return;
    }

    if (11 > hm->uri.len || 0 == extract_id((char *)(hm->uri.p + 10), hm->uri.len - 10, 0, &id))
    {
        send_error(conn, hm, 400);
        return;
    }

    ret = tm_get_task(taskman, id, &task);

    /* LCOV_EXCL_START */
    if (0 == ret)
    {
        send_error(conn, hm, 500);
        return;
    }
    /* LCOV_EXCL_STOP */
    else if (0 == task)
    {
        send_error(conn, hm, 404);
        return;
    }

    create_task_reply(task, &reply);
    tm_task_free(&task);

    send_ok(conn, hm, "Content-Type: application/json\r\n"
        "Cache-Control: no-store", strlen(reply), reply);
    free(reply);
}

static int extract_modifiers(json_t *root, int is_add, modifiers *pmodifiers)
{
    json_t *name = 0;
    json_t *description = 0;
    json_t *status = 0;

    if (0 == json_object_size(root))
    {
       return 0;
    }

    name = json_object_get(root, "name");
    if (1 == is_add && 0 == name)
    {
       return 0;
    }
    else if (0 != name && 0 == json_is_string(name))
    {
       return 0;
    }

    if (0 != (description = json_object_get(root, "description")) &&
        0 == json_is_string(description))
    {
       return 0;
    }

    if (0 != (status = json_object_get(root, "status")) &&
        0 == json_is_string(status))
    {
       return 0;
    }

    if (0 != name)
    {
        pmodifiers->name = strdup((char *)json_string_value(name));
    }
    if (0 != description)
    {
        pmodifiers->description = strdup((char *)json_string_value(description));
    }
    if (0 != status)
    {
        pmodifiers->status = strdup((char *)json_string_value(status));
    }
    return 1;
}

static int extract_modify(json_t *root, tm_query **pquery,
    modifiers *pmodifiers)
{
    json_t *filters = 0;
    json_t *modifiers = 0;

    if (2 != json_object_size(root))
    {
        return 0;
    }

    if (0 == (filters = json_object_get(root, "filters")) ||
        0 == json_is_object(filters))
    {
        return 0;
    }
    if (0 == (modifiers = json_object_get(root, "modifiers")) ||
        0 == json_is_object(modifiers))
    {
        return 0;
    }

    if (0 == extract_filters(filters, pquery))
    {
        return 0;
    }

    if (0 == extract_modifiers(modifiers, 0, pmodifiers))
    {
        tm_query_free(pquery);
        return 0;
    }
    return 1;
}

void do_api_modify(struct mg_connection *conn, struct http_message *hm)
{
    json_error_t err;
    json_t *root;
    tm_query *query = 0;
    modifiers the_modifiers;
    tm_task **arr = 0;
    unsigned int arr_len = 0;
    int ret;

    memset(&the_modifiers, 0, sizeof(the_modifiers));

    if (0 != mg_vcmp(&(hm->method), "POST"))
    {
        send_error(conn, hm, 405);
        return;
    }

    root = json_loadb(hm->body.p, hm->body.len, 0, &err);
    if (0 == root)
    {
        send_error(conn, hm, 400);
        return;
    }

    ret = extract_modify(root, &query, &the_modifiers);
    json_decref(root);
    if (0 == ret)
    {
        send_error(conn, hm, 400);
        return;
    }

    ret = tm_get_tasks(taskman, query, &arr, &arr_len);
    tm_query_free(&query);

    if (0 == ret)
    {
        send_error(conn, hm, 500);
    }
    else if (0 == arr_len)
    {
        send_error(conn, hm, 404);
    }
    else
    {
        if (0 != the_modifiers.name)
        {
            tm_task_arr_set_name(arr, arr_len, the_modifiers.name);
        }
        if (0 != the_modifiers.description)
        {
            tm_task_arr_set_description(arr, arr_len, the_modifiers.description);
        }
        if (0 != the_modifiers.status)
        {
            tm_task_arr_set_status(arr, arr_len, the_modifiers.status);
        }

        ret = tm_modify_tasks(taskman, arr, arr_len);

        if (0 == ret)
        {
            send_error(conn, hm, 500);
        }
        else
        {
            send_ok(conn, hm, "Cache-Control: no-store", 0, 0);
        }
    }

    modifiers_free(&the_modifiers);
    tm_task_arr_free(&arr, arr_len);
}

void do_api_remove(struct mg_connection *conn, struct http_message *hm)
{
    tm_query *query = 0;
    tm_task **arr = 0;
    unsigned int arr_len = 0;
    int ret;
    json_error_t err;
    json_t *root;

    if (0 != mg_vcmp(&(hm->method), "POST"))
    {
        send_error(conn, hm, 405);
        return;
    }

    root = json_loadb(hm->body.p, hm->body.len, 0, &err);
    if (0 == root)
    {
        send_error(conn, hm, 400);
        return;
    }

    ret = extract_filters(root, &query);
    json_decref(root);
    if (0 == ret)
    {
        send_error(conn, hm, 400);
        return;
    }

    ret = tm_get_tasks(taskman, query, &arr, &arr_len);
    tm_query_free(&query);

    /* LCOV_EXCL_START */
    if (0 == ret)
    {
        send_error(conn, hm, 500);
        return;
    }
    /* LCOV_EXCL_STOP */

    if (0 == arr_len)
    {
        send_error(conn, hm, 404);
    }
    else
    {
        ret = tm_remove_tasks(taskman, arr, arr_len);

        /* LCOV_EXCL_START */
        if (0 == ret)
        {
            send_error(conn, hm, 500);
        }
        /* LCOV_EXCL_STOP */
        else
        {
            send_ok(conn, hm, "Cache-Control: no-store", 0, 0);
        }
    }

    tm_task_arr_free(&arr, arr_len);
}

void do_api_add(struct mg_connection *conn, struct http_message *hm)
{
    json_error_t err;
    json_t *root;
    modifiers the_modifiers = {0};
    tm_task *task = 0;

    if (0 != mg_vcmp(&(hm->method), "POST"))
    {
        send_error(conn, hm, 405);
        return;
    }

    root = json_loadb(hm->body.p, hm->body.len, 0, &err);
    if (0 == root)
    {
        send_error(conn, hm, 400);
        return;
    }

    if (0 == extract_modifiers(root, 1, &the_modifiers))
    {
        send_error(conn, hm, 400);
    }
    else
    {
        tm_task_new(&task);
        tm_task_set_name(task, the_modifiers.name);

        if (0 != the_modifiers.description)
        {
            tm_task_set_description(task, the_modifiers.description);
        }
        if (0 != the_modifiers.status)
        {
            tm_task_set_status(task, the_modifiers.status);
        }

        modifiers_free(&the_modifiers);

        /* LCOV_EXCL_START */
        if (0 == tm_add_task(taskman, task))
        {
            send_error(conn, hm, 500);
        }
        /* LCOV_EXCL_STOP */
        else
        {
            unsigned long id = -1;
            json_t *tmp = 0;
            json_t *ret_json;
            char *ret_str = 0;

            ret_json = json_object();

            tm_task_get_id(task, &id);
            tmp = json_integer(id);
            json_object_set(ret_json, "id", tmp);
            json_decref(tmp);

            ret_str = json_dumps(ret_json, JSON_COMPACT);
            json_decref(ret_json);

            send_ok(conn, hm, "Cache-Control: no-store", strlen(ret_str), ret_str);
            free(ret_str);
        }
        tm_task_free(&task);
    }
    json_decref(root);
}

void do_api_ping(struct mg_connection *conn, struct http_message *hm)
{
    if (0 != mg_vcmp(&(hm->method), "GET"))
    {
        send_error(conn, hm, 405);
    }
    else
    {
        send_ok(conn, hm, 0, 0, 0);
    }
}
