/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <stdio.h>
#include <string.h>
#include <signal.h>

#include <static_index_html.h>
#include <tms.h>

typedef struct edit_html_userdata edit_html_userdata;
typedef struct diff_text diff_text;

enum edit_html_entered
{
    THE_NULL,
    THE_EDIT
};

struct edit_html_userdata
{
    enum edit_html_entered entered;

    tm_task *task;
    int edit_i;
};

static int tag_top(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    if (0 == strcmp("add-style", tag) ||
        0 == strcmp("error-style", tag) ||
        0 == strcmp("index-style", tag) ||
        0 == strcmp("task-style", tag) ||
        0 == strcmp("warning-style", tag))
    {
        char *tmp = "style='display: none;'";
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("edit-style", tag))
    {
        return 1;
    }
    return 0;
}

static int tag_edit(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    edit_html_userdata *puserdata = (edit_html_userdata *)userdata;
    if (0 == strcmp("id", tag))
    {
        char tmp[64];
        unsigned long id = -1;
        tm_task_get_id(puserdata->task, &id);
        sprintf(tmp, "%lu", id);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("name", tag))
    {
        char *tmp;
        tm_task_get_name(puserdata->task, &tmp);
        return tpl_appende(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("description", tag))
    {
        char *tmp;
        tm_task_get_description(puserdata->task, &tmp);
        return tpl_appende(pout, plen, tmp,
            0 != tmp ? strlen(tmp) : 0);
    }
    else if (0 == strcmp("status", tag))
    {
        char *tmp;
        tm_task_get_status(puserdata->task, &tmp);
        return tpl_appende(pout, plen, tmp, strlen(tmp));
    }
    return 0;
}

static int tag(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    edit_html_userdata *puserdata = (edit_html_userdata *)userdata;
    switch (puserdata->entered)
    {
        case THE_NULL:
            return tag_top(pout, plen, userdata, tag);
        case THE_EDIT:
            return tag_edit(pout, plen, userdata, tag);
        default:
            break;
    }
    return 0;
}

static int enter(char **pout, unsigned int *plen, void *userdata,
    char *name)
{
    edit_html_userdata *puserdata = (edit_html_userdata *)userdata;
    switch (puserdata->entered)
    {
        case THE_NULL:
            if (0 == strcmp("edit", name))
            {
                puserdata->entered = THE_EDIT;
                puserdata->edit_i = -1;
                return 1;
            }
        default:
            break;
    }
    return 0;
}

static int next(char **pout, unsigned int *plen, void *userdata)
{
    edit_html_userdata *puserdata = (edit_html_userdata *)userdata;
    switch (puserdata->entered)
    {
        case THE_EDIT:
            puserdata->edit_i++;
            return 0 == puserdata->edit_i;
        default:
            break;
    }
    return 0;
}

static void leave(char **pout, unsigned int *plen, void *userdata)
{
    edit_html_userdata *puserdata = (edit_html_userdata *)userdata;
    switch (puserdata->entered)
    {
        case THE_EDIT:
            puserdata->entered = THE_NULL;
            break;
        default:
            break;
    }
}

static void extract_name(struct http_message *hm, tm_task *task,
    unsigned int *plen)
{
    char *value = 0;
    char *existing = 0;

    extract_post_s(hm, "name", &value);
    tm_task_get_name(task, &existing);
    if (0 != value && 0 != *value && 0 != strcmp(existing, value))
    {
        tm_task_set_name(task, value);
        (*plen)++;
    }
}

static void extract_description(struct http_message *hm, tm_task *task,
    unsigned int *plen)
{
    char *value = 0;
    char *existing = 0;

    extract_post_s(hm, "description", &value);
    tm_task_get_description(task, &existing);
    if (0 != value &&
        0 != strcmp(0 != existing ? existing : "", value))
    {
        tm_task_set_description(task, value);
        (*plen)++;
    }
}

static void extract_status(struct http_message *hm, tm_task *task,
    unsigned int *plen)
{
    char *value = 0;
    char *existing = 0;

    extract_post_s(hm, "status", &value);
    tm_task_get_status(task, &existing);
    if (0 != value && 0 != *value && 0 != strcmp(existing, value))
    {
        tm_task_set_status(task, value);
        (*plen)++;
    }
}

static void extract_query_from_http_form(struct mg_connection *conn, struct http_message *hm,
    tm_task *task)
{
    unsigned int len = 0;

    extract_name(hm, task, &len);
    extract_description(hm, task, &len);
    extract_status(hm, task, &len);

    if (0 == len)
    {
        send_error(conn, hm, 422);
    }
    else if (0 == tm_modify_task(taskman, task))
    {
        send_error(conn, hm, 500);
    }
    else
    {
        unsigned long id = -1;
        tm_task_get_id(task, &id);

        char headers[64];
        sprintf(headers, "Location: /task/%lu.html", id);
        send_303(conn, hm, headers);
    }
}

void do_edit_html(struct mg_connection *conn, struct http_message *hm)
{
    unsigned int size = 0;
    char *page = 0;
    edit_html_userdata userdata;
    unsigned long id = 0;
    tm_task *task = 0;
    int ret = -1;

    if (12 > hm->uri.len || 0 == extract_id((char *)(hm->uri.p + 6), hm->uri.len - 6, 1, &id))
    {
        send_error(conn, hm, 400);
        return;
    }

    ret = tm_get_task(taskman, id, &task);

    if (0 == ret)
    {
        send_error(conn, hm, 500);
    }
    else if (0 == task)
    {
        send_error(conn, hm, 404);
    }
    else
    {
        userdata.task = task;
        userdata.entered = THE_NULL;

        if (0 == mg_vcmp(&(hm->method), "POST"))
        {
            extract_query_from_http_form(conn, hm, task);
        }
        else if (0 != mg_vcmp(&(hm->method), "GET"))
        {
            send_error(conn, hm, 405);
            tm_task_free(&task);
        }
        else if (0 != tpl_file_static_index_html(tag,
            enter,
            next,
            leave,
            &userdata, &size, &page))
        {
            send_ok(conn, hm, "Content-Type: text/html\r\n"
                "Cache-Control: no-store", size, page);
            free(page);
        }
        else
        {
            send_error(conn, hm, 500);
        }
        tm_task_free(&task);
    }
}
