/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <stdio.h>
#include <string.h>
#include <signal.h>

#include <static_index_html.h>
#include <tms.h>

typedef struct index_html_userdata index_html_userdata;
typedef struct row row;

enum index_html_entered
{
    IHE_NULL,
    IHE_INDEX,
    IHE_INDEX_ERROR,
    IHE_INDEX_ROWS,
    IHE_INDEX_ROWS_COLUMNS,
    IHE_INDEX_COLUMNS
};

struct row
{
    unsigned long id;
    char *name;

    char *status;
    unsigned int column_i;
};

struct index_html_userdata
{
    enum index_html_entered entered;

    unsigned int index_i;

    char *error;
    unsigned int error_i;

    unsigned int column_i;

    row **rows;
    unsigned int rows_len;
    unsigned int row_i;
};

static void extract_filter_id_equals(struct http_message *hm, tm_query **children,
    unsigned int *plen)
{
    unsigned long *pvalue = 0;
    extract_post_ul(hm, "filter-id-equals", &pvalue);
    if (0 != pvalue)
    {
        tm_query_id_equals(taskman, *pvalue, &(children[(*plen)++]));
        free(pvalue);
    }
}

static void extract_filter_name_contains(struct http_message *hm, tm_query **children,
    unsigned int *plen)
{
    char *pvalue = 0;
    extract_post_s(hm, "filter-name-contains", &pvalue);
    if (0 != pvalue)
    {
        tm_query_name_contains(taskman, pvalue, &(children[(*plen)++]));
        free(pvalue);
    }
}

static void extract_filter_description_contains(struct http_message *hm, tm_query **children,
    unsigned int *plen)
{
    char *pvalue = 0;
    extract_post_s(hm, "filter-description-contains", &pvalue);
    if (0 != pvalue)
    {
        tm_query_description_contains(taskman, pvalue, &(children[(*plen)++]));
        free(pvalue);
    }
}

static void extract_filter_status_contains(struct http_message *hm, tm_query **children,
    unsigned int *plen)
{
    char *pvalue = 0;
    extract_post_s(hm, "filter-status-contains", &pvalue);
    if (0 != pvalue)
    {
        tm_query_status_contains(taskman, pvalue, &(children[(*plen)++]));
        free(pvalue);
    }
}

static void extract_filter_date_created_after(struct http_message *hm, tm_query **children,
    unsigned int *plen)
{
    time_t *pvalue = 0;
    extract_post_t(hm, "filter-date-created-after", &pvalue);
    if (0 != pvalue)
    {
        tm_query_date_created_after(taskman, *pvalue, &(children[(*plen)++]));
        free(pvalue);
    }
}

static void extract_filter_date_created_before(struct http_message *hm, tm_query **children,
    unsigned int *plen)
{
    time_t *pvalue = 0;
    extract_post_t(hm, "filter-date-created-before", &pvalue);
    if (0 != pvalue)
    {
        tm_query_date_created_before(taskman, *pvalue, &(children[(*plen)++]));
        free(pvalue);
    }
}

static void extract_filter_date_updated_after(struct http_message *hm, tm_query **children,
    unsigned int *plen)
{
    time_t *pvalue = 0;
    extract_post_t(hm, "filter-date-updated-after", &pvalue);
    if (0 != pvalue)
    {
        tm_query_date_updated_after(taskman, *pvalue, &(children[(*plen)++]));
        free(pvalue);
    }
}

static void extract_filter_date_updated_before(struct http_message *hm, tm_query **children,
    unsigned int *plen)
{
    time_t *pvalue = 0;
    extract_post_t(hm, "filter-date-updated-before", &pvalue);
    if (0 != pvalue)
    {
        tm_query_date_updated_before(taskman, *pvalue, &(children[(*plen)++]));
        free(pvalue);
    }
}

static void extract_query_from_http_form(struct http_message *hm, tm_query **pquery)
{
    unsigned int len = 0;
    int free_children = 1;
    tm_query **children = 0;

    children = (tm_query **)malloc(sizeof(tm_query *) * 8);
    memset(children, 0, sizeof(tm_query *) * 8);

    extract_filter_id_equals(hm, children, &len);
    extract_filter_name_contains(hm, children, &len);
    extract_filter_description_contains(hm, children, &len);
    extract_filter_status_contains(hm, children, &len);
    extract_filter_date_created_after(hm, children, &len);
    extract_filter_date_created_before(hm, children, &len);
    extract_filter_date_updated_after(hm, children, &len);
    extract_filter_date_updated_before(hm, children, &len);

    if (1 == len)
    {
        *pquery = children[0];
    }
    else if (1 < len)
    {
        tm_query **tmp = (tm_query **)realloc(children, sizeof(tm_query *) * len);
        if (0 != tmp)
        {
            children = tmp;
            free_children = 0;
            tm_query_and(taskman, len, children, pquery);
        }
    }

    if (1 == free_children)
    {
        free(children);
    }
}

static int tag_rows_columns(char **pout, unsigned int *plen, void *userdata,
    char *tag)
{
    index_html_userdata *puserdata = (index_html_userdata *)userdata;
    row *a_row = puserdata->rows[puserdata->row_i];

    if (0 == strcmp("name", tag))
    {
        return tpl_appende(pout, plen, "status", 6);
    }
    else if (0 == strcmp("value", tag))
    {
        char *status = a_row->status;
        return tpl_appende(pout, plen, status, strlen(status));
    }
    return 0;
}

static int tag_rows(char **pout, unsigned int *plen, void *userdata,
    char *tag)
{
    index_html_userdata *puserdata = (index_html_userdata *)userdata;
    row *a_row = puserdata->rows[puserdata->row_i];

    if (0 == strcmp("id", tag))
    {
        char tmp[64];
        sprintf(tmp, "%lu", a_row->id);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("name", tag))
    {
        return tpl_appende(pout, plen, a_row->name, strlen(a_row->name));
    }
    return 0;
}

static int tag_columns(char **pout, unsigned int *plen, void *userdata,
    char *tag)
{
    if (0 == strcmp("name", tag))
    {
        return tpl_appende(pout, plen, "status", 6);
    }
    else if (0 == strcmp("name-human-readable", tag))
    {
        return tpl_appende(pout, plen, "Status", 6);
    }
    return 0;
}

static int tag_top(char **pout, unsigned int *plen, void *userdata,
    char *tag)
{
    if (0 == strcmp("index-style", tag))
    {
        return 1;
    }
    else if (0 == strcmp("add-style", tag) ||
        0 == strcmp("edit-style", tag) ||
        0 == strcmp("error-style", tag) ||
        0 == strcmp("task-style", tag) ||
        0 == strcmp("warning-style", tag))
    {
        char *tmp = "style='display: none;'";
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    return 0;
}

static int tag(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    index_html_userdata *puserdata = (index_html_userdata *)userdata;
    switch (puserdata->entered)
    {
        case IHE_INDEX_ERROR:
            if (0 == strcmp("error", tag))
            {
                return tpl_appende(pout, plen, puserdata->error, strlen(puserdata->error));
            }
            break;

        case IHE_INDEX_ROWS:
            return tag_rows(pout, plen, userdata, tag);

        case IHE_INDEX_ROWS_COLUMNS:
            return tag_rows_columns(pout, plen, userdata, tag);

        case IHE_INDEX_COLUMNS:
            return tag_columns(pout, plen, userdata, tag);

        case IHE_NULL:
            return tag_top(pout, plen, userdata, tag);

        default:
            break;
    }
    return 0;
}

static int enter_index(char **pout, unsigned int *plen, void *userdata,
    char *name)
{
    index_html_userdata *puserdata = (index_html_userdata *)userdata;

    if (0 == strcmp("error", name) && 0 != puserdata->error)
    {
        puserdata->error_i = -1;
        puserdata->entered = IHE_INDEX_ERROR;
        return 1;
    }
    else if (0 == strcmp("rows", name) && 0 != puserdata->rows_len)
    {
        puserdata->row_i = -1;
        puserdata->entered = IHE_INDEX_ROWS;
        return 1;
    }
    else if (0 == strcmp("columns", name))
    {
        puserdata->column_i = -1;
        puserdata->entered = IHE_INDEX_COLUMNS;
        return 1;
    }
    return 0;

}

static int enter(char **pout, unsigned int *plen, void *userdata,
    char *name)
{
    index_html_userdata *puserdata = (index_html_userdata *)userdata;
    switch (puserdata->entered)
    {
        case IHE_INDEX:
            return enter_index(pout, plen, userdata, name);

        case IHE_INDEX_ROWS:
            if (0 == strcmp("columns", name))
            {
                puserdata->rows[puserdata->row_i]->column_i = -1;
                puserdata->entered = IHE_INDEX_ROWS_COLUMNS;
                return 1;
            }
            break;

        case IHE_NULL:
            if (0 == strcmp("index", name))
            {
                puserdata->index_i = -1;
                puserdata->entered = IHE_INDEX;
                return 1;
            }
            break;

        default:
            break;
    }
    return 0;
}

static int next(char **pout, unsigned int *plen, void *userdata)
{
    index_html_userdata *puserdata = (index_html_userdata *)userdata;
    row *a_row = 0;

    switch (puserdata->entered)
    {
        case IHE_INDEX:
            puserdata->index_i++;
            if (0 == puserdata->index_i)
            {
                return 1;
            }
            break;

        case IHE_INDEX_ROWS:
            puserdata->row_i++;
            if (puserdata->rows_len > puserdata->row_i)
            {
                return 1;
            }
            break;

        case IHE_INDEX_ROWS_COLUMNS:
            a_row = puserdata->rows[puserdata->row_i];
            a_row->column_i++;
            if (0 == a_row->column_i)
            {
                return 1;
            }
            break;

        case IHE_INDEX_COLUMNS:
            puserdata->column_i++;
            if (0 == puserdata->column_i)
            {
                return 1;
            }
            break;

        default:
            break;
    }
    return 0;
}

static void leave(char **pout, unsigned int *plen, void *userdata)
{
    index_html_userdata *puserdata = (index_html_userdata *)userdata;
    switch (puserdata->entered)
    {
        case IHE_INDEX_ROWS_COLUMNS:
            puserdata->entered = IHE_INDEX_ROWS;
            break;
        case IHE_INDEX_ERROR:
        case IHE_INDEX_ROWS:
        case IHE_INDEX_COLUMNS:
            puserdata->entered = IHE_INDEX;
            break;
        case IHE_INDEX:
            puserdata->entered = IHE_NULL;
            break;
        default:
            break;
    }
}

static void build_rows(row ***prows, unsigned int *prows_len,
    tm_task **tasks, unsigned int tasks_len)
{
    row **rows = 0;
    unsigned int i;

    rows = (row **)malloc(sizeof(row *) * tasks_len);
    for (i = 0; i < tasks_len; i++)
    {
        char *tmp = 0;
        row *a_row = 0;

        a_row = (row *)malloc(sizeof(row));
        rows[i] = a_row;

        tm_task_get_id(tasks[i], &(a_row->id));

        tm_task_get_name(tasks[i], &tmp);
        a_row->name = strdup(tmp);

        tm_task_get_status(tasks[i], &tmp);
        a_row->status = strdup(tmp);
    }

    *prows = rows;
    *prows_len = tasks_len;
}

static void free_rows(row **rows, unsigned int rows_len)
{
    unsigned int i;

    for (i = 0; i < rows_len; i++)
    {
        row *a_row = rows[i];
        free(a_row->name);
        free(a_row->status);
        free(a_row);
    }
    free(rows);
}

void do_index_html(struct mg_connection *conn, struct http_message *hm)
{
    unsigned int size = 0;
    char *page = 0;
    index_html_userdata userdata;
    tm_query *query = 0;
    int ret = 0;
    tm_task **tasks = 0;
    unsigned int tasks_len = 0;

    userdata.entered = IHE_NULL;
    userdata.error = 0;
    userdata.error_i = 0;
    userdata.column_i = -1;
    userdata.rows = 0;
    userdata.rows_len = 0;
    userdata.row_i = -1;

    if (0 == mg_vcmp(&(hm->method), "POST"))
    {
        extract_query_from_http_form(hm, &query);
    }

    ret = tm_get_tasks(taskman, query, &tasks, &tasks_len);
    tm_query_free(&query);

    if (0 == ret)
    {
        userdata.error = "Running query failed";
    }
    else
    {
        build_rows(&(userdata.rows), &(userdata.rows_len),
            tasks, tasks_len);
        tm_task_arr_free(&tasks, tasks_len);
    }

    if (0 != tpl_file_static_index_html(tag,
        enter,
        next,
        leave,
        &userdata, &size, &page))
    {
        send_ok(conn, hm, "Content-Type: text/html\r\n"
            "Cache-Control: no-store", size, page);
        free(page);
    }
    else
    {
        send_error(conn, hm, 500);
    }

    if (0 != userdata.rows)
    {
        free_rows(userdata.rows, userdata.rows_len);
    }
}
