/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <stdio.h>
#include <string.h>
#include <signal.h>

#include <tms.h>

typedef struct argz argz;
struct argz
{
    int a;
    char *d;
    int h;
    char *i;
    unsigned int p;
};

static int should_exit = 0;
tm_lib *taskman;
static struct mg_mgr *mgr;
static struct mg_connection *mgr_conn;

static void argz_free(argz *argz)
{
    if (0 != argz->d)
    {
        free(argz->d);
    }
    if (0 != argz->i)
    {
        free(argz->i);
    }
}

static void usage(char *argv0, FILE *out)
{
    fprintf(out, "Usage: %s [-h]\n", argv0);
    fprintf(out, "       %s [-a/--access-log] [-d/--dbfile dbfile] "
                 "[-i/--ip ip] [-p/--port port]\n", argv0);
    fprintf(out, "\n");
}

static int parse_opt_key(char *k, char *pl, argz *arg)
{
    if (0 == strcmp("-a", k) ||
        0 == strcmp("--access-log", k))
    {
        arg->a = 1;
    }
    else if (0 == strcmp("-d", k) ||
        0 == strcmp("--dbfile", k))
    {
        *pl = 'd';
    }
    else if (0 == strcmp("-i", k) ||
        0 == strcmp("--ip", k))
    {
        *pl = 'i';
    }
    else if (0 == strcmp("-p", k) ||
        0 == strcmp("--port", k))
    {
        *pl = 'p';
    }
    else
    {
        return 0;
    }
    return 1;
}

static void parse_opt_value(char k, char *v, argz *arg)
{
    if ('d' == k)
    {
        free(arg->d);
        arg->d = strdup(v);
    }
    else if ('i' == k)
    {
        free(arg->i);
        arg->i = strdup(v);
    }
    else if ('p' == k)
    {
        arg->p = atoi(v);
    }
}

static int parse_opts(int argc, char **argv, argz *arg)
{
    char k = 0;
    unsigned int i;
    for (i = 1; i < argc; i++)
    {
        char *argvi = argv[i];
        int key = (k == 0);
        if (1 == key && 0 == parse_opt_key(argvi, &k, arg))
        {
            return 0;
        }
        else if (0 == key)
        {
            parse_opt_value(k, argvi, arg);
            k = 0;
        }
    }
    return 0 != k ? 0 : 1;
}

static int parse_args(int argc, char **argv, argz *arg)
{
    if (argc == 1)
    {
        return 1;
    }

    if (0 == strcmp("-h", argv[1]) || 0 == strcmp("--help", argv[1]))
    {
        arg->h = 1;
        return 1;
    }

    if (0 == parse_opts(argc, argv, arg))
    {
        return 0;
    }

    return 1;
}

static void print_accesslog(char *ip, unsigned int method_len, char *method,
    unsigned int path_len, char *path, unsigned int proto_len, char *proto,
    unsigned int status, unsigned int len)
{
    printf("%s unknown unknown [unknown] \"%.*s %.*s %.*s\" %u %u\n",
        ip, method_len, method, path_len, path, proto_len, proto, status, len);
}

void send_error(struct mg_connection *conn, struct http_message *hm, int code)
{
    argz *arg = (argz *)conn->mgr->user_data;

    mg_http_send_error(conn, code, "");

    if (1 == arg->a)
    {
        char addr[32];
        mg_conn_addr_to_str(conn, addr, sizeof(addr),
            MG_SOCK_STRINGIFY_IP | MG_SOCK_STRINGIFY_REMOTE);
        print_accesslog(addr, (unsigned int) hm->method.len, (char *)hm->method.p,
            (unsigned int) hm->uri.len, (char *) hm->uri.p, (unsigned int) hm->proto.len,
            (char *)hm->proto.p, code, 0);
    }
}

void send_ok(struct mg_connection *conn, struct http_message *hm,
    char *headers, unsigned int len, char *bytes)
{
    argz *arg = (argz *)conn->mgr->user_data;

    mg_send_head(conn, 200, len, headers);
    if (0 != bytes)
    {
        mg_send(conn, bytes, len);
    }

    if (1 == arg->a)
    {
        char addr[32];
        mg_conn_addr_to_str(conn, addr, sizeof(addr),
            MG_SOCK_STRINGIFY_IP | MG_SOCK_STRINGIFY_REMOTE);
        print_accesslog(addr, (unsigned int) hm->method.len, (char *)hm->method.p,
            (unsigned int) hm->uri.len, (char *) hm->uri.p, (unsigned int) hm->proto.len,
            (char *)hm->proto.p, 200, len);
    }
}

void send_304(struct mg_connection *conn, struct http_message *hm, char *headers)
{
    argz *arg = (argz *)conn->mgr->user_data;
    mg_send_head(conn, 304, 0, headers);

    if (1 == arg->a)
    {
        char addr[32];
        mg_conn_addr_to_str(conn, addr, sizeof(addr),
            MG_SOCK_STRINGIFY_IP | MG_SOCK_STRINGIFY_REMOTE);
        print_accesslog(addr, (unsigned int) hm->method.len, (char *)hm->method.p,
            (unsigned int) hm->uri.len, (char *) hm->uri.p, (unsigned int) hm->proto.len,
            (char *)hm->proto.p, 304, 0);
    }
}

void send_303(struct mg_connection *conn, struct http_message *hm, char *headers)
{
    argz *arg = (argz *)conn->mgr->user_data;
    mg_send_head(conn, 303, 0, headers);

    if (1 == arg->a)
    {
        char addr[32];
        mg_conn_addr_to_str(conn, addr, sizeof(addr),
            MG_SOCK_STRINGIFY_IP | MG_SOCK_STRINGIFY_REMOTE);
        print_accesslog(addr, (unsigned int) hm->method.len, (char *)hm->method.p,
            (unsigned int) hm->uri.len, (char *) hm->uri.p, (unsigned int) hm->proto.len,
            (char *)hm->proto.p, 303, 0);
    }
}

static int front_matches(const struct mg_str *a, char *b)
{
    int alen = a->len;
    int blen = strlen(b);
    return 0 == strncmp(a->p, b, blen) && blen <= alen;
}

static void event_handler(struct mg_connection *conn, int ev, void *ev_data)
{
    struct http_message *hm;

    if (ev != MG_EV_HTTP_REQUEST)
    {
        return;
    }

    hm = (struct http_message *)ev_data;
    if (0 == mg_vcmp(&(hm->uri), "/api/add"))
    {
        do_api_add(conn, hm);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/api/modify"))
    {
        do_api_modify(conn, hm);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/api/query"))
    {
        do_api_query(conn, hm);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/api/ping"))
    {
        do_api_ping(conn, hm);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/api/remove"))
    {
        do_api_remove(conn, hm);
    }
    else if (1 == front_matches(&(hm->uri), "/api/task/"))
    {
        do_api_task(conn, hm);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/add.html"))
    {
        do_add_html(conn, hm);
    }
    else if (1 == front_matches(&(hm->uri), "/edit"))
    {
        do_edit_html(conn, hm);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/") ||
        0 == mg_vcmp(&(hm->uri), "/index.html"))
    {
        do_index_html(conn, hm);
    }
    else if (1 == front_matches(&(hm->uri), "/task"))
    {
        do_task_html(conn, hm);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/lib.js"))
    {
        do_lib_js(conn, hm);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/other.html"))
    {
        do_other_html(conn, hm);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/other.js"))
    {
        do_other_js(conn, hm);
    }
    else
    {
        send_error(conn, hm, 404);
    }
    conn->flags |= MG_F_SEND_AND_CLOSE;
}

static void on_ctrl_c(int signum)
{
    should_exit = 1;
}

static void init_tm(argz *arg)
{
    tm_storage *storage = 0;

    taskman = 0;
    tm_storage_sqlite(arg->d, &storage);
    tm_new(storage, &taskman);
}

static void deinit_tm(void)
{
    tm_free(&taskman);
}

static int init_mgr(argz *arg)
{
    struct mg_bind_opts bind_opts;
    char addr[22];
    char *error_string = 0;

    mgr = (struct mg_mgr *)malloc(sizeof(struct mg_mgr));
    memset(mgr, 0, sizeof(struct mg_mgr));

    sprintf(addr, "%s:%u", arg->i, arg->p);
    memset(&bind_opts, 0, sizeof(bind_opts));
    bind_opts.error_string = (const char **)&error_string;
    mg_mgr_init(mgr, 0);
    mgr->user_data = arg;

    /* LCOV_EXCL_START */
    if (0 == (mgr_conn = mg_bind_opt(mgr, addr, event_handler, bind_opts)))
    {
        fprintf(stderr, "Cannot start server: %s\n",
            error_string);
        return 0;
    }
    /* LCOV_EXCL_STOP */

    mg_set_protocol_http_websocket(mgr_conn);
    return 1;
}

static void deinit_mgr(void)
{
    mg_mgr_free(mgr);
    free(mgr);
}

static void loop(void)
{
    char addr[32];
    mg_conn_addr_to_str(mgr_conn, addr, sizeof(addr),
        MG_SOCK_STRINGIFY_IP | MG_SOCK_STRINGIFY_PORT);

    printf("Started on %s\n", addr);
    while (0 == should_exit)
    {
         mg_mgr_poll(mgr, 200);
    }
    printf("Stopped\n");
}

int main(int argc, char **argv)
{
    argz arg;
    int ret = 1;

    arg.a = 0;
    arg.d = strdup("tm.db");
    arg.h = 0;
    arg.p = 3805;
    arg.i = strdup("0.0.0.0");

    if (0 == parse_args(argc, argv, &arg))
    {
        usage(argv[0], stderr);
    }
    else if (1 == arg.h)
    {
        usage(argv[0], stdout);
        ret = 0;
    }
    else
    {
        init_tm(&arg);

        if (1 == init_mgr(&arg))
        {
            setvbuf(stdout, 0, _IOLBF, 1024);
            signal(SIGINT, on_ctrl_c);

            loop();

            deinit_mgr();
        }

        deinit_tm();
        ret = 0;
    }
    argz_free(&arg);
    return ret;
}
