/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <stdio.h>
#include <string.h>
#include <signal.h>

#include <static_index_html.h>
#include <tms.h>

typedef struct task_html_userdata task_html_userdata;
typedef struct history_item history_item;

enum task_html_entered
{
    THE_NULL,
    THE_TASK,
    THE_TASK_HISTORY,
    THE_TASK_HISTORY_CHANGE
};

struct task_html_userdata
{
    enum task_html_entered entered;
    tm_task *task;
    tm_task_diff **diffs;
    unsigned int diffs_len;

    unsigned int task_i;

    unsigned int history_i;
    unsigned int history_change_i;
    history_item **history_items;
    unsigned int history_items_len;
};

enum history_item_type
{
    HIT_ADDED,
    HIT_REMOVED,
    HIT_CHANGED
};

struct history_item
{
    enum history_item_type type;
    char *name;
    char *name_human_readable;
    char *value;
    char *new_value;
    time_t date;
    char *date_formatted;
};

static int tag_task(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    task_html_userdata *puserdata = (task_html_userdata *)userdata;
    if (0 == strcmp("id", tag))
    {
        char tmp[64];
        unsigned long id = 0;
        tm_task_get_id(puserdata->task, &id);
        sprintf(tmp, "%lu", id);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("name", tag))
    {
        char *tmp = 0;
        tm_task_get_name(puserdata->task, &tmp);
        return tpl_appende(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("description", tag))
    {
        char *tmp = 0;
        tm_task_get_description(puserdata->task, &tmp);
        return tpl_appende(pout, plen, tmp,
            0 == tmp ? 0 : strlen(tmp));
    }
    else if (0 == strcmp("status", tag))
    {
        char *tmp = 0;
        tm_task_get_status(puserdata->task, &tmp);
        return tpl_appende(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("date-created", tag))
    {
        char tmp[64];
        time_t date = 0;
        tm_task_get_date_created(puserdata->task, &date);
        sprintf(tmp, "%ld", date);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("date-created-formatted", tag))
    {
        char tmp[21];
        time_t date = 0;
        struct tm tminfo;
        memset(&tminfo, 0, sizeof(tminfo));
        tm_task_get_date_created(puserdata->task, &date);
        localtime_r(&date, &tminfo);
        strftime(tmp, 20, "%Y-%m-%d %H:%M:%S", &tminfo);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("date-updated", tag))
    {
        char tmp[64];
        time_t date = 0;
        tm_task_get_date_updated(puserdata->task, &date);
        sprintf(tmp, "%ld", date);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("date-updated-formatted", tag))
    {
        char tmp[21];
        time_t date = 0;
        struct tm tminfo;
        memset(&tminfo, 0, sizeof(tminfo));
        tm_task_get_date_updated(puserdata->task, &date);
        localtime_r(&date, &tminfo);
        strftime(tmp, 20, "%Y-%m-%d %H:%M:%S", &tminfo);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    return 0;
}

static int tag_top(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    if (0 == strcmp("add-style", tag) ||
        0 == strcmp("edit-style", tag) ||
        0 == strcmp("error-style", tag) ||
        0 == strcmp("index-style", tag) ||
        0 == strcmp("warning-style", tag))
    {
        char *tmp = "style='display: none;'";
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("task-style", tag))
    {
        return 1;
    }
    return 0;
}

static int tag_task_history(char **pout, unsigned int *plen, void *userdata,
    char *tag)
{
    task_html_userdata *puserdata = (task_html_userdata *)userdata;
    history_item *hi = puserdata->history_items[puserdata->history_i];

    if (0 == strcmp("type", tag))
    {
        switch (hi->type)
        {
            case HIT_ADDED:
                return tpl_append(pout, plen, "added", 5);
            case HIT_REMOVED:
                return tpl_append(pout, plen, "removed", 7);
            case HIT_CHANGED:
                return tpl_append(pout, plen, "changed", 7);
        }
        return 0;
    }
    else if (0 == strcmp("name", tag))
    {
        return tpl_appende(pout, plen, hi->name, strlen(hi->name));
    }
    else if (0 == strcmp("name-human-readable", tag))
    {
        return tpl_appende(pout, plen, hi->name_human_readable, strlen(hi->name_human_readable));
    }
    else if (0 == strcmp("value", tag))
    {
        return tpl_appende(pout, plen, hi->value, strlen(hi->value));
    }
    else if (0 == strcmp("date", tag))
    {
        char tmp[64];
        sprintf(tmp, "%ld", hi->date);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("date-formatted", tag))
    {
        return tpl_append(pout, plen, hi->date_formatted, strlen(hi->date_formatted));
    }
    return 0;
}

static int tag_task_history_change(char **pout, unsigned int *plen, void *userdata,
    char *tag)
{
    task_html_userdata *puserdata = (task_html_userdata *)userdata;
    history_item *hi = puserdata->history_items[puserdata->history_i];

    if (0 == strcmp("new-value", tag))
    {
        return tpl_appende(pout, plen, hi->new_value, strlen(hi->new_value));
    }
    return 0;
}

static int tag(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    task_html_userdata *puserdata = (task_html_userdata *)userdata;
    switch (puserdata->entered)
    {
        case THE_TASK:
            return tag_task(pout, plen, userdata, tag);
        case THE_TASK_HISTORY:
            return tag_task_history(pout, plen, userdata, tag);
        case THE_TASK_HISTORY_CHANGE:
            return tag_task_history_change(pout, plen, userdata, tag);
        case THE_NULL:
            return tag_top(pout, plen, userdata, tag);
        default:
            break;
    }
    return 0;
}

static int append_history_item(history_item ***parr, unsigned int *plen, unsigned int *pallocated,
    enum history_item_type type, char *name, char *name_human_readable,
    char *value, char *new_value, time_t date, char *date_formatted)
{
    history_item *n = 0;

    if (*plen == *pallocated)
    {
        history_item **tmp = 0;
        *pallocated = *pallocated * 3 + 1;
        tmp = (history_item **)realloc(*parr, sizeof(history_item *) * (*pallocated));
        if (0 == tmp)
        {
            return 0;
        }
        *parr = tmp;
    }

    n = (history_item *)malloc(sizeof(history_item));
    n->type = type;
    n->name = name;
    n->name_human_readable = name_human_readable;
    n->value = 0 != value ? strdup(value) : 0;
    n->new_value = 0 != new_value ? strdup(new_value) : 0;
    n->date = date;
    n->date_formatted = date_formatted;

    (*parr)[*plen] = n;
    (*plen)++;

    return 1;
}

static void prepare_history_item(history_item ***parr, unsigned int *parr_len,
    unsigned int *parr_allocated, tm_task_diff *item, tm_task_diff **plast, char *name,
    char *name_human_readable)
{
    if (0 != *plast)
    {
        char *value = 0;
        char *new_value = 0;
        time_t date = 0;
        char *date_formatted;
        struct tm tminfo;

        if (0 != item)
        {
            tm_task_diff_get_value(item, &value);
        }
        tm_task_diff_get_date(*plast, &date);
        tm_task_diff_get_value(*plast, &new_value);

        memset(&tminfo, 0, sizeof(tminfo));
        localtime_r(&date, &tminfo);
        date_formatted = (char *)malloc(sizeof(char) * 21);
        strftime(date_formatted, 20, "%Y-%m-%d %H:%M:%S", &tminfo);

        if (0 == new_value)
        {
            append_history_item(parr, parr_len, parr_allocated, HIT_REMOVED, name,
                name_human_readable, value, 0, date, date_formatted);
        }
        else if (0 == value)
        {
            append_history_item(parr, parr_len, parr_allocated, HIT_ADDED, name,
                name_human_readable, new_value, 0, date, date_formatted);
        }
        else
        {
            append_history_item(parr, parr_len, parr_allocated, HIT_CHANGED, name,
                name_human_readable, value, new_value, date, date_formatted);
        }
    }
    *plast = item;
}

static void prepare_history_items(task_html_userdata *puserdata)
{
    tm_task_diff **diffs = puserdata->diffs;
    unsigned int len = puserdata->diffs_len;
    history_item **arr;
    unsigned int arr_allocated = 10;
    unsigned int arr_len = 0;

    tm_task_diff *last_name = 0;
    tm_task_diff *last_description = 0;
    tm_task_diff *last_status = 0;

    arr = (history_item **)malloc(sizeof(history_item *) * arr_allocated);

    puserdata->entered = THE_TASK_HISTORY;
    puserdata->history_i = -1;
    puserdata->history_items = 0;
    puserdata->history_items_len = 0;

    for (unsigned int i = 0; i < len; i++)
    {
        tm_task_diff *diff = diffs[i];
        char *name = 0;

        tm_task_diff_get_name(diff, &name);

        if (0 == strcmp("name", name))
        {
            prepare_history_item(&arr, &arr_len, &arr_allocated, diff, &last_name, "name", "Name");
        }
        else if (0 == strcmp("description", name))
        {
            prepare_history_item(&arr, &arr_len, &arr_allocated, diff, &last_description,
                "description", "Description");
        }
        else if (0 == strcmp("status", name))
        {
            prepare_history_item(&arr, &arr_len, &arr_allocated, diff, &last_status,
                "status", "Status");
        }
    }

    if (0 != last_name)
    {
        prepare_history_item(&arr, &arr_len, &arr_allocated, 0, &last_name, "name", "Name");
    }
    if (0 != last_description)
    {
        prepare_history_item(&arr, &arr_len, &arr_allocated, 0, &last_description, "description",
            "Description");
    }
    if (0 != last_status)
    {
        prepare_history_item(&arr, &arr_len, &arr_allocated, 0, &last_status, "status", "Status");
    }

    puserdata->history_items = arr;
    puserdata->history_items_len = arr_len;
}

static int enter_top(char **pout, unsigned int *plen, void *userdata,
    char *name)
{
    task_html_userdata *puserdata = (task_html_userdata *)userdata;
    if (0 == strcmp("task", name))
    {
        puserdata->entered = THE_TASK;
        puserdata->task_i = -1;
        return 1;
    }
    return 0;
}

static int enter_task(char **pout, unsigned int *plen, void *userdata,
    char *name)
{
    task_html_userdata *puserdata = (task_html_userdata *)userdata;
    if (0 == strcmp("history", name))
    {
        puserdata->history_i = -1;
        prepare_history_items(puserdata);
        return 1;
    }
    return 0;
}

static int enter_task_history(char **pout, unsigned int *plen, void *userdata,
    char *name)
{
    task_html_userdata *puserdata = (task_html_userdata *)userdata;
    history_item *hi = puserdata->history_items[puserdata->history_i];

    if (HIT_CHANGED == hi->type && 0 == strcmp("change", name))
    {
        puserdata->history_change_i = -1;
        puserdata->entered = THE_TASK_HISTORY_CHANGE;
        return 1;
    }
    return 0;
}

static int enter(char **pout, unsigned int *plen, void *userdata,
    char *name)
{
    task_html_userdata *puserdata = (task_html_userdata *)userdata;
    switch (puserdata->entered)
    {
        case THE_NULL:
            return enter_top(pout, plen, userdata, name);
        case THE_TASK:
            return enter_task(pout, plen, userdata, name);
        case THE_TASK_HISTORY:
            return enter_task_history(pout, plen, userdata, name);
        default:
            break;
    }
    return 0;
}

static int next(char **pout, unsigned int *plen, void *userdata)
{
    task_html_userdata *puserdata = (task_html_userdata *)userdata;
    switch (puserdata->entered)
    {
        case THE_TASK:
            puserdata->task_i++;
            return 0 == puserdata->task_i;
        case THE_TASK_HISTORY:
            puserdata->history_i++;
            return puserdata->history_items_len > puserdata->history_i;
        case THE_TASK_HISTORY_CHANGE:
            puserdata->history_change_i++;
            return 0 == puserdata->history_change_i;
        default:
            break;
    }
    return 0;
}

static void leave_task_history(char **pout, unsigned int *plen, void *userdata)
{
    task_html_userdata *puserdata = (task_html_userdata *)userdata;
    for (unsigned int i = 0; i < puserdata->history_items_len; i++)
    {
        history_item *it = puserdata->history_items[i];

        free(it->value);
        free(it->date_formatted);
        if (0 != it->new_value)
        {
            free(it->new_value);
        }
        free(it);
    }

    free(puserdata->history_items);
    puserdata->entered = THE_TASK;
}

static void leave(char **pout, unsigned int *plen, void *userdata)
{
    task_html_userdata *puserdata = (task_html_userdata *)userdata;
    switch (puserdata->entered)
    {
        case THE_TASK_HISTORY_CHANGE:
            puserdata->entered = THE_TASK_HISTORY;
            break;
        case THE_TASK_HISTORY:
            leave_task_history(pout, plen, userdata);
            break;
        case THE_TASK:
            puserdata->entered = THE_NULL;
            break;
        default:
            break;
    }
}

void do_task_html(struct mg_connection *conn, struct http_message *hm)
{
    tm_task *task = 0;
    unsigned int size = 0;
    char *page = 0;
    unsigned long id = 0;
    int ret = 0;

    if (0 != mg_vcmp(&(hm->method), "GET"))
    {
        send_error(conn, hm, 405);
        return;
    }

    if (12 > hm->uri.len || 0 == extract_id((char *)(hm->uri.p + 6), hm->uri.len - 6, 1, &id))
    {
        send_error(conn, hm, 400);
        return;
    }

    ret = tm_get_task(taskman, id, &task);

    if (0 == ret)
    {
        send_error(conn, hm, 500);
    }
    else if (0 == task)
    {
        send_error(conn, hm, 404);
    }
    else
    {
        task_html_userdata userdata;
        userdata.entered = THE_NULL;
        userdata.task = task;
        userdata.task_i = -1;
        userdata.history_i = -1;

        tm_get_task_diffs(taskman, task, &(userdata.diffs), &(userdata.diffs_len));

        if (0 != tpl_file_static_index_html(tag,
            enter,
            next,
            leave,
            &userdata, &size, &page))
        {
            send_ok(conn, hm, "Content-Type: text/html\r\n"
                "Cache-Control: no-store", size, page);
            free(page);
        }
        else
        {
            send_error(conn, hm, 500);
        }

        tm_task_diff_arr_free(&(userdata.diffs), userdata.diffs_len);
    }
    tm_task_free(&task);
}
