/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <criterion/criterion.h>
#include <jansson.h>

#include "common.h"

TestSuite(add);

Test(add, task)
{
    unsigned int port = -1;
    date_verifier last_1_minute = {DVT_AFTER, time(0) - 60};

    history_entry entry1 = {"add", "name", "Buy milk", 0, &last_1_minute};
    history_entry entry2 = {"add", "description", "Goat milk", 0, &last_1_minute};
    history_entry entry3 = {"add", "status", "open", 0, &last_1_minute};
    history_entry *history[] = {&entry1, &entry2, &entry3};

    expect_process_started(0, &port);
    do_add("127.0.0.1", port, 3,
        "name", "Buy milk",
        "description", "Goat milk",
        "status", "open");
    do_task("127.0.0.1", port, "1", "Buy milk", "Goat milk", "open",
        &last_1_minute, &last_1_minute, 3, history);
    expect_process_died(0);
}

Test(add, queries)
{
    unsigned int port = -1;

    expect_process_started(0, &port);
    do_add("127.0.0.1", port, 3,
        "name", "Buy milk",
        "description", "Goat milk",
        "status", "open");
    do_query("127.0.0.1", port, 0, 0, 0, 1,
        2,
          "id", 1,
          "name", "Buy milk");
    do_query("127.0.0.1", port, 0, 1, 1, 1,
        "show-column-description",
        "description",
        3,
          "id", 1,
          "name", "Buy milk",
          "description", "Goat milk");
    do_query("127.0.0.1", port, 0, 1, 1, 1,
        "show-column-status",
        "status",
        3,
          "id", 1,
          "name", "Buy milk",
          "status", "open");
    expect_process_died(0);
}

Test(add, queries_filters)
{
    unsigned int port = -1;

    expect_process_started(0, &port);

    do_add("127.0.0.1", port, 3,
        "name", "Buy milk",
        "description", "Goat milk",
        "status", "open");
    do_query("127.0.0.1", port, 1, 0, 0, 1,
        "id-equals", 1,
        2,
          "id", 1,
          "name", "Buy milk");
    do_query("127.0.0.1", port, 1, 0, 0, 1,
        "name-contains", "Buy milk",
        2,
          "id", 1,
          "name", "Buy milk");
    do_query("127.0.0.1", port, 1, 0, 0, 1,
        "description-contains", "Goat milk",
        2,
          "id", 1,
          "name", "Buy milk");
    do_query("127.0.0.1", port, 1, 0, 0, 1,
        "status-contains", "open",
        2,
          "id", 1,
          "name", "Buy milk");
    do_query("127.0.0.1", port, 1, 0, 0, 1,
        "date-created-after", "yesterday",
        2,
          "id", 1,
          "name", "Buy milk");
    do_query("127.0.0.1", port, 1, 0, 0, 1,
        "date-created-before", "tomorrow",
        2,
          "id", 1,
          "name", "Buy milk");
    do_query("127.0.0.1", port, 1, 0, 0, 1,
        "date-updated-after", "yesterday",
        2,
          "id", 1,
          "name", "Buy milk");
    do_query("127.0.0.1", port, 1, 0, 0, 1,
        "date-updated-before", "tomorrow",
        2,
          "id", 1,
          "name", "Buy milk");
    do_query("127.0.0.1", port, 8, 0, 0, 1,
        "id-equals", 1,
        "name-contains", "Buy milk",
        "description-contains", "Goat milk",
        "status-contains", "open",
        "date-created-after", "today",
        "date-created-before", "today",
        "date-updated-after", "today",
        "date-updated-before", "today",
        2,
          "id", 1,
          "name", "Buy milk");
    expect_process_died(0);
}

Test(add, same)
{
    unsigned int port = -1;
    date_verifier last_1_minute = {DVT_AFTER, time(0) - 60};
    history_entry entry1 = {"add", "name", "Buy milk", 0, &last_1_minute};
    history_entry entry2 = {"add", "status", "new", 0, &last_1_minute};
    history_entry *history[] = {&entry1, &entry2};

    expect_process_started(0, &port);
    do_add("127.0.0.1", port, 1,
        "name", "Buy milk");
    do_add("127.0.0.1", port, 1,
        "name", "Buy milk");
    do_task("127.0.0.1", port, "1", "Buy milk", 0, "new",
        &last_1_minute, &last_1_minute, 2, history);
    do_task("127.0.0.1", port, "2", "Buy milk", 0, "new",
        &last_1_minute, &last_1_minute, 2, history);
    do_query("127.0.0.1", port, 0, 0, 0, 2,
        2,
          "id", 1,
          "name", "Buy milk",
        2,
          "id", 2,
          "name", "Buy milk");
    expect_process_died(0);
}

Test(add, three_task)
{
    unsigned int port = -1;
    date_verifier last_1_minute = {DVT_AFTER, time(0) - 60};

    history_entry entry3_1 = {"add", "name", "Buy beer", 0, &last_1_minute};
    history_entry entry3_2 = {"add", "status", "new", 0, &last_1_minute};
    history_entry *history3[] = {&entry3_1, &entry3_2};

    history_entry entry2_1 = {"add", "name", "Buy gas", 0, &last_1_minute};
    history_entry *history2[] = {&entry2_1, &entry3_2};

    history_entry entry1_1 = {"add", "name", "Buy milk", 0, &last_1_minute};
    history_entry *history1[] = {&entry1_1, &entry3_2};

    expect_process_started(0, &port);

    do_add("127.0.0.1", port, 1,
        "name", "Buy milk");
    do_add("127.0.0.1", port, 1,
        "name", "Buy gas");
    do_add("127.0.0.1", port, 1,
        "name", "Buy beer");
    do_task("127.0.0.1", port, "1", "Buy milk", 0, "new",
        &last_1_minute, &last_1_minute, 2, history1);
    do_task("127.0.0.1", port, "2", "Buy gas", 0, "new",
        &last_1_minute, &last_1_minute, 2, history2);
    do_task("127.0.0.1", port, "3", "Buy beer", 0, "new",
        &last_1_minute, &last_1_minute, 2, history3);
    expect_process_died(0);
}

Test(add, three_queries)
{
    unsigned int port = -1;
    expect_process_started(0, &port);

    do_add("127.0.0.1", port, 1,
        "name", "Buy milk");
    do_add("127.0.0.1", port, 1,
        "name", "Buy gas");
    do_add("127.0.0.1", port, 1,
        "name", "Buy beer");
    do_query("127.0.0.1", port, 0, 0, 0, 3,
        2,
          "id", 1,
          "name", "Buy milk",
        2,
          "id", 2,
          "name", "Buy gas",
        2,
          "id", 3,
          "name", "Buy beer");
    do_query("127.0.0.1", port, 1, 0, 0, 3,
        "name-contains", "Buy",
        2,
          "id", 1,
          "name", "Buy milk",
        2,
          "id", 2,
          "name", "Buy gas",
        2,
          "id", 3,
          "name", "Buy beer");
    do_query("127.0.0.1", port, 1, 0, 0, 1,
        "name-contains", "milk",
        2,
          "id", 1,
          "name", "Buy milk");
    do_query("127.0.0.1", port, 1, 0, 0, 1,
        "name-contains", "gas",
        2,
          "id", 2,
          "name", "Buy gas");
    do_query("127.0.0.1", port, 1, 0, 0, 1,
        "name-contains", "beer",
        2,
          "id", 3,
          "name", "Buy beer");
    expect_process_died(0);
}

Test(add, invalid_name_not_string)
{
    unsigned int port = -1;
    char *request_str = 0;
    char url[256];

    json_t *request_json = json_object();
    json_t *tmp = json_integer(0);
    json_object_set(request_json, "name", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/add", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);

    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(add, invalid_empty)
{
    unsigned int port = -1;
    char url[256];

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/add", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", "{}", 400, 0, 0, 0);
    expect_process_no_output();
    expect_process_died(0);
}

Test(add, invalid_unknown_args)
{
    unsigned int port = -1;
    char *request_str = 0;
    char url[256];

    json_t *request_json = json_object();
    json_t *tmp = json_integer(0);
    json_object_set(request_json, "unknown-arg", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/add", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(add, invalid_not_json)
{
    unsigned int port = -1;
    char url[256];

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/add", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", "}", 400, 0, 0, 0);
    expect_process_no_output();
    expect_process_died(0);
}

Test(add, invalid_not_post)
{
    unsigned int port = -1;
    char url[256];

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/add", "127.0.0.1", port);
    expect_http_get(url, "Accept: application/json\r\n", 405, 0, 0, 0);
    expect_process_no_output();
    expect_process_died(0);
}
