/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <criterion/criterion.h>
#include <jansson.h>

#include "common.h"

TestSuite(modify);

Test(modify, empty)
{
    unsigned int port = -1;
    expect_process_started(0, &port);
    do_modify("127.0.0.1", port, 0, 0, 1,
        "name", "Buy beer");
    do_modify("127.0.0.1", port, 0, 1, 1,
        "id-equals", 1,
        "name", "Buy beer");
    do_modify("127.0.0.1", port, 0, 1, 1,
        "name-contains", "milk",
        "name", "Buy beer");
    expect_process_died(0);
}

Test(modify, id_equals_task)
{
    unsigned int port = -1;
    date_verifier after_add = {DVT_AFTER, 0};
    date_verifier after_modify = {DVT_AFTER, 0};
    history_entry entry1 = {"change", "name", "Buy milk", "Buy beer", &after_modify};
    history_entry entry2 = {"add", "name", "Buy milk", 0, &after_add};
    history_entry entry3 = {"add", "status", "new", 0, &after_add};
    history_entry *history[] = {&entry1, &entry2, &entry3};

    expect_process_started(0, &port);
    after_add.value = time(0);
    do_add("127.0.0.1", port, 1, "name", "Buy milk");
    after_modify.value = time(0);
    do_modify("127.0.0.1", port, 1, 1, 1,
        "id-equals", 1,
        "name", "Buy beer");
    do_task("127.0.0.1", port, "1", "Buy beer", 0, "new",
        &after_add, &after_modify, 3, history);
    expect_process_died(0);
}

Test(modify, id_equals_query)
{
    unsigned int port = -1;

    expect_process_started(0, &port);
    do_add("127.0.0.1", port, 1, "name", "Buy milk");
    do_modify("127.0.0.1", port, 1, 1, 1,
        "id-equals", 1,
        "name", "Buy beer");
    do_query("127.0.0.1", port, 0, 0, 0, 1,
        2,
          "id", 1,
          "name", "Buy beer");
    expect_process_died(0);
}

Test(modify, name_contains_task)
{
    unsigned int port = -1;
    date_verifier after_add = {DVT_AFTER, 0};
    date_verifier after_modify = {DVT_AFTER, 0};
    history_entry entry1 = {"change", "name", "Buy milk", "Buy beer", &after_modify};
    history_entry entry2 = {"add", "name", "Buy milk", 0, &after_add};
    history_entry entry3 = {"add", "status", "new", 0, &after_add};
    history_entry *history[] = {&entry1, &entry2, &entry3};

    expect_process_started(0, &port);
    after_add.value = time(0);
    do_add("127.0.0.1", port, 1, "name", "Buy milk");
    after_modify.value = time(0);
    do_modify("127.0.0.1", port, 1, 1, 1,
        "name-contains", "milk",
        "name", "Buy beer");
    do_task("127.0.0.1", port, "1", "Buy beer", 0, "new",
        &after_add, &after_modify, 3, history);
    expect_process_died(0);
}

Test(modify, name_contains_query)
{
    unsigned int port = -1;

    expect_process_started(0, &port);
    do_add("127.0.0.1", port, 1, "name", "Buy milk");
    do_modify("127.0.0.1", port, 1, 1, 1,
        "name-contains", "milk",
        "name", "Buy beer");
    do_query("127.0.0.1", port, 0, 0, 0, 1,
        2,
          "id", 1,
          "name", "Buy beer");
    expect_process_died(0);
}

Test(modify, description_contains_task)
{
    unsigned int port = -1;
    date_verifier after_add = {DVT_AFTER, 0};
    date_verifier after_modify = {DVT_AFTER, 0};
    history_entry entry1 = {"change", "description", "Goat milk", "Yak milk", &after_modify};
    history_entry entry2 = {"add", "name", "Buy milk", 0, &after_add};
    history_entry entry3 = {"add", "description", "Goat milk", 0, &after_add};
    history_entry entry4 = {"add", "status", "new", 0, &after_add};
    history_entry *history[] = {&entry1, &entry2, &entry3, &entry4};

    expect_process_started(0, &port);
    after_add.value = time(0);
    do_add("127.0.0.1", port, 2,
        "name", "Buy milk",
        "description", "Goat milk");
    after_modify.value = time(0);
    do_modify("127.0.0.1", port, 1, 1, 1,
        "description-contains", "Goat",
        "description", "Yak milk");
    do_task("127.0.0.1", port, "1", "Buy milk", "Yak milk", "new",
        &after_add, &after_modify, 4, history);
    expect_process_died(0);
}

Test(modify, description_contains_query)
{
    unsigned int port = -1;

    expect_process_started(0, &port);
    do_add("127.0.0.1", port, 2,
        "name", "Buy milk",
        "description", "Goat milk");
    do_modify("127.0.0.1", port, 1, 1, 1,
        "description-contains", "Goat",
        "description", "Yak milk");
    do_query("127.0.0.1", port, 0, 1, 1, 1,
        "show-column-description",
        "description",
        3,
            "id", 1,
            "name", "Buy milk",
            "description", "Yak milk");
    expect_process_died(0);
}

Test(modify, status_contains_task)
{
    unsigned int port = -1;
    date_verifier after_add = {DVT_AFTER, 0};
    date_verifier after_modify = {DVT_AFTER, 0};
    history_entry entry1 = {"change", "status", "open", "done", &after_modify};
    history_entry entry2 = {"add", "name", "Buy milk", 0, &after_add};
    history_entry entry3 = {"add", "status", "open", 0, &after_add};
    history_entry *history[] = {&entry1, &entry2, &entry3};

    expect_process_started(0, &port);
    after_add.value = time(0);
    do_add("127.0.0.1", port, 2,
        "name", "Buy milk",
        "status", "open");
    after_modify.value = time(0);
    do_modify("127.0.0.1", port, 1, 1, 1,
        "status-contains", "open",
        "status", "done");
    do_task("127.0.0.1", port, "1", "Buy milk", 0, "done",
        &after_add, &after_modify, 3, history);
    expect_process_died(0);
}

Test(modify, status_contains_query)
{
    unsigned int port = -1;

    expect_process_started(0, &port);
    do_add("127.0.0.1", port, 2,
        "name", "Buy milk",
        "status", "open");
    do_modify("127.0.0.1", port, 1, 1, 1,
        "status-contains", "open",
        "status", "done");
    do_query("127.0.0.1", port, 0, 1, 1, 1,
        "show-column-status",
        "status",
        3,
            "id", 1,
            "name", "Buy milk",
            "status", "done");
    expect_process_died(0);
}

static char *date_to_str(time_t date)
{
    static char buf[21];
    struct tm tinfo;

    memset(&tinfo, 0, sizeof(tinfo));
    localtime_r(&date, &tinfo);
    strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", &tinfo);
    return buf;
}

Test(modify, date_created_after_task)
{
    unsigned int port = -1;
    date_verifier after_add = {DVT_AFTER, 0};
    date_verifier after_modify = {DVT_AFTER, 0};
    history_entry entry1 = {"change", "name", "Buy milk", "Buy beer", &after_modify};
    history_entry entry2 = {"add", "name", "Buy milk", 0, &after_add};
    history_entry entry3 = {"add", "status", "new", 0, &after_add};
    history_entry *history[] = {&entry1, &entry2, &entry3};

    expect_process_started(0, &port);
    after_add.value = time(0);
    do_add("127.0.0.1", port, 1, "name", "Buy milk");
    after_modify.value = time(0);
    do_modify("127.0.0.1", port, 1, 1, 1,
        "date-created-after", date_to_str(after_add.value),
        "name", "Buy beer");
    do_task("127.0.0.1", port, "1", "Buy beer", 0, "new",
        &after_add, &after_modify, 3, history);
    expect_process_died(0);
}

Test(modify, date_created_before_task)
{
    unsigned int port = -1;
    date_verifier after_add = {DVT_AFTER, 0};
    date_verifier before_add = {DVT_BEFORE, 0};
    date_verifier after_modify = {DVT_AFTER, 0};
    history_entry entry1 = {"change", "name", "Buy milk", "Buy beer", &after_modify};
    history_entry entry2 = {"add", "name", "Buy milk", 0, &after_add};
    history_entry entry3 = {"add", "status", "new", 0, &after_add};
    history_entry *history[] = {&entry1, &entry2, &entry3};

    expect_process_started(0, &port);
    after_add.value = time(0);
    do_add("127.0.0.1", port, 1, "name", "Buy milk");
    before_add.value = time(0) + 1;
    after_modify.value = time(0);
    do_modify("127.0.0.1", port, 1, 1, 1,
        "date-created-before", date_to_str(before_add.value),
        "name", "Buy beer");
    do_task("127.0.0.1", port, "1", "Buy beer", 0, "new",
        &after_add, &after_modify, 3, history);
    expect_process_died(0);
}

Test(modify, date_updated_after_task)
{
    unsigned int port = -1;
    date_verifier after_add = {DVT_AFTER, 0};
    date_verifier after_modify = {DVT_AFTER, 0};
    history_entry entry1 = {"change", "name", "Buy milk", "Buy beer", &after_modify};
    history_entry entry2 = {"add", "name", "Buy milk", 0, &after_add};
    history_entry entry3 = {"add", "status", "new", 0, &after_add};
    history_entry *history[] = {&entry1, &entry2, &entry3};

    expect_process_started(0, &port);
    after_add.value = time(0);
    do_add("127.0.0.1", port, 1, "name", "Buy milk");
    after_modify.value = time(0);
    do_modify("127.0.0.1", port, 1, 1, 1,
        "date-updated-after", date_to_str(after_add.value),
        "name", "Buy beer");
    do_task("127.0.0.1", port, "1", "Buy beer", 0, "new",
        &after_add, &after_modify, 3, history);
    expect_process_died(0);
}

Test(modify, date_updated_before_task)
{
    unsigned int port = -1;
    date_verifier after_add = {DVT_AFTER, 0};
    date_verifier before_add = {DVT_BEFORE, 0};
    date_verifier after_modify = {DVT_AFTER, 0};
    history_entry entry1 = {"change", "name", "Buy milk", "Buy beer", &after_modify};
    history_entry entry2 = {"add", "name", "Buy milk", 0, &after_add};
    history_entry entry3 = {"add", "status", "new", 0, &after_add};
    history_entry *history[] = {&entry1, &entry2, &entry3};

    expect_process_started(0, &port);
    after_add.value = time(0);
    do_add("127.0.0.1", port, 1, "name", "Buy milk");
    before_add.value = time(0) + 1;
    after_modify.value = time(0);
    do_modify("127.0.0.1", port, 1, 1, 1,
        "date-updated-before", date_to_str(before_add.value),
        "name", "Buy beer");
    do_task("127.0.0.1", port, "1", "Buy beer", 0, "new",
        &after_add, &after_modify, 3, history);
    expect_process_died(0);
}

Test(modify, invalid_not_json)
{
    unsigned int port = -1;
    char url[256];

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", "}", 400, 0, 0, 0);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_not_2_items)
{
    unsigned int port = -1;
    char url[256];

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", "{}", 400, 0, 0, 0);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_no_filters)
{
    char *request_str = 0;
    unsigned int port = -1;
    char url[256];

    json_t *request_json = json_object();
    json_t *tmp = json_object();
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_filters_not_object)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp;
    unsigned int port = -1;
    char url[256];

    tmp = json_integer(0);
    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);

    tmp = json_object();
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_no_modifiers)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = json_object();
    unsigned int port = -1;
    char url[256];

    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_modifiers_not_object)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp;
    unsigned int port = -1;
    char url[256];

    tmp = json_integer(0);
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);

    tmp = json_object();
    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_modifiers_empty)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;
    unsigned int port = -1;
    char url[256];

    tmp = json_object();
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);

    tmp = json_object();
    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_filters_id_equals_not_int)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;
    json_t *tmp2 = 0;
    unsigned int port = -1;
    char url[256];

    tmp = json_object();
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);

    tmp = json_object();
    tmp2 = json_string("invalid");
    json_object_set(tmp, "id-equals", tmp2);
    json_decref(tmp2);

    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_filters_name_contains_not_string)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;
    json_t *tmp2 = 0;
    unsigned int port = -1;
    char url[256];

    tmp = json_object();
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);

    tmp = json_object();
    tmp2 = json_integer(0);
    json_object_set(tmp, "name-contains", tmp2);
    json_decref(tmp2);

    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_filters_description_contains_not_string)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;
    json_t *tmp2 = 0;
    unsigned int port = -1;
    char url[256];

    tmp = json_object();
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);

    tmp = json_object();
    tmp2 = json_integer(0);
    json_object_set(tmp, "description-contains", tmp2);
    json_decref(tmp2);

    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_filters_status_contains_not_string)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;
    json_t *tmp2 = 0;
    unsigned int port = -1;
    char url[256];

    tmp = json_object();
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);

    tmp = json_object();
    tmp2 = json_integer(0);
    json_object_set(tmp, "status-contains", tmp2);
    json_decref(tmp2);

    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_filters_date_created_after_not_integer)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;
    json_t *tmp2 = 0;
    unsigned int port = -1;
    char url[256];

    tmp = json_object();
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);

    tmp = json_object();
    tmp2 = json_string("abcde");
    json_object_set(tmp, "date-created-after", tmp2);
    json_decref(tmp2);

    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_filters_date_created_before_not_integer)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;
    json_t *tmp2 = 0;
    unsigned int port = -1;
    char url[256];

    tmp = json_object();
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);

    tmp = json_object();
    tmp2 = json_string("abcde");
    json_object_set(tmp, "date-created-before", tmp2);
    json_decref(tmp2);

    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_filters_date_updated_after_not_integer)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;
    json_t *tmp2 = 0;
    unsigned int port = -1;
    char url[256];

    tmp = json_object();
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);

    tmp = json_object();
    tmp2 = json_string("abcde");
    json_object_set(tmp, "date-updated-after", tmp2);
    json_decref(tmp2);

    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_filters_date_updated_before_not_integer)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;
    json_t *tmp2 = 0;
    unsigned int port = -1;
    char url[256];

    tmp = json_object();
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);

    tmp = json_object();
    tmp2 = json_string("abcde");
    json_object_set(tmp, "date-updated-before", tmp2);
    json_decref(tmp2);

    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}


Test(modify, invalid_modifiers_name_not_string)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;
    json_t *tmp2 = 0;
    unsigned int port = -1;
    char url[256];

    tmp = json_object();
    tmp2 = json_integer(0);
    json_object_set(tmp, "id-equals", tmp2);
    json_decref(tmp2);
    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);

    tmp = json_object();
    tmp2 = json_integer(0);
    json_object_set(tmp, "name", tmp2);
    json_decref(tmp2);
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_modifiers_description_not_string)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;
    json_t *tmp2 = 0;
    unsigned int port = -1;
    char url[256];

    tmp = json_object();
    tmp2 = json_integer(0);
    json_object_set(tmp, "id-equals", tmp2);
    json_decref(tmp2);
    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);

    tmp = json_object();
    tmp2 = json_integer(0);
    json_object_set(tmp, "description", tmp2);
    json_decref(tmp2);
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_modifiers_status_not_string)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;
    json_t *tmp2 = 0;
    unsigned int port = -1;
    char url[256];

    tmp = json_object();
    tmp2 = json_integer(0);
    json_object_set(tmp, "id-equals", tmp2);
    json_decref(tmp2);
    json_object_set(request_json, "filters", tmp);
    json_decref(tmp);

    tmp = json_object();
    tmp2 = json_integer(0);
    json_object_set(tmp, "status", tmp2);
    json_decref(tmp2);
    json_object_set(request_json, "modifiers", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(modify, invalid_not_post)
{
    unsigned int port = -1;
    char url[256];

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/modify", "127.0.0.1", port);
    expect_http_get(url, 0, 405, 0, 0, 0);
    expect_process_no_output();
    expect_process_died(0);
}
