/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <criterion/criterion.h>
#include <jansson.h>

#include "common.h"

TestSuite(query);

Test(query, empty)
{
    unsigned int port = -1;
    expect_process_started(0, &port);
    do_query("127.0.0.1", port, 0, 0, 0, 0);
    do_query("127.0.0.1", port, 1, 0, 0, 0,
        "id-equals", 1);
    do_query("127.0.0.1", port, 1, 0, 0, 0,
        "name-contains", "Buy beer");
    do_query("127.0.0.1", port, 1, 0, 0, 0,
        "description-contains", "Goat beer");
    expect_process_died(0);
}

Test(query, invalid_id_equals_not_int)
{
    char *request_str = 0;
    unsigned int port = -1;
    char url[256];

    json_t *request_json = json_object();
    json_t *filters = json_object();
    json_t *tmp = json_string("invalid");
    json_object_set(request_json, "filters", filters);
    json_object_set(filters, "id-equals", tmp);
    json_decref(tmp);
    json_decref(filters);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/query", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n"
        "Accept: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(query, invalid_name_contains_not_string)
{
    char *request_str = 0;
    unsigned int port = -1;
    char url[256];

    json_t *request_json = json_object();
    json_t *filters = json_object();
    json_t *tmp = json_integer(0);
    json_object_set(request_json, "filters", filters);
    json_object_set(filters, "name-contains", tmp);
    json_decref(tmp);
    json_decref(filters);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/query", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n"
        "Accept: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(query, invalid_description_contains_not_string)
{
    char *request_str = 0;
    unsigned int port = -1;
    char url[256];

    json_t *request_json = json_object();
    json_t *filters = json_object();
    json_t *tmp = json_integer(0);
    json_object_set(request_json, "filters", filters);
    json_object_set(filters, "description-contains", tmp);
    json_decref(tmp);
    json_decref(filters);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/query", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n"
        "Accept: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(query, invalid_unknown_arg)
{
    char *request_str = 0;
    unsigned int port = -1;
    char url[256];

    json_t *request_json = json_object();
    json_t *filters = json_object();
    json_t *tmp = json_integer(0);
    json_object_set(request_json, "filters", filters);
    json_object_set(filters, "unknown-arg", tmp);
    json_decref(tmp);
    json_decref(filters);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/query", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n"
        "Accept: application/json\r\n", request_str, 400, 0, 0, 0);

    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(query, invalid_not_json)
{
    unsigned int port = -1;
    char url[256];

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/query", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n"
        "Accept: application/json\r\n", "}", 400, 0, 0, 0);
    expect_process_no_output();
    expect_process_died(0);
}
