/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <criterion/criterion.h>
#include <jansson.h>

#include "common.h"

TestSuite(remove);

Test(remove, empty)
{
    unsigned int port = -1;
    expect_process_started(0, &port);
    do_remove("127.0.0.1", port, 0, 1,
        "id-equals", 1);
    do_remove("127.0.0.1", port, 0, 1,
        "name-contains", "Buy beer");
    do_remove("127.0.0.1", port, 0, 1,
        "description-contains", "Goat beer");
    do_remove("127.0.0.1", port, 0, 1,
        "status-contains", "open");
    do_remove("127.0.0.1", port, 0, 1,
        "date-created-after", 0);
    do_remove("127.0.0.1", port, 0, 1,
        "date-created-before", 0);
    do_remove("127.0.0.1", port, 0, 1,
        "date-updated-after", 0);
    do_remove("127.0.0.1", port, 0, 1,
        "date-updated-before", 0);
    expect_process_died(0);
}

Test(remove, id_equals)
{
    unsigned int port = -1;
    expect_process_started(0, &port);

    do_add("127.0.0.1", port, 1, "name", "Buy milk");
    do_remove("127.0.0.1", port, 1, 1,
        "id-equals", 1);
    do_query("127.0.0.1", port, 0, 0, 0, 0);
    do_remove("127.0.0.1", port, 0, 1,
        "id-equals", 1);
    do_remove("127.0.0.1", port, 0, 1,
        "name-contains", "Buy milk");

    expect_process_died(0);
}

Test(remove, name_contains)
{
    unsigned int port = -1;
    expect_process_started(0, &port);

    do_add("127.0.0.1", port, 1, "name", "Buy milk");
    do_remove("127.0.0.1", port, 1, 1,
        "name-contains", "Buy milk");
    do_query("127.0.0.1", port, 0, 0, 0, 0);
    do_remove("127.0.0.1", port, 0, 1,
        "id-equals", 1);
    do_remove("127.0.0.1", port, 0, 1,
        "name-contains", "Buy milk");

    expect_process_died(0);
}

Test(remove, description_contains)
{
    unsigned int port = -1;
    expect_process_started(0, &port);

    do_add("127.0.0.1", port, 2,
        "name", "Buy milk",
        "description", "Goat milk");
    do_remove("127.0.0.1", port, 1, 1,
        "description-contains", "Goat milk");
    do_query("127.0.0.1", port, 0, 0, 0, 0);
    do_remove("127.0.0.1", port, 0, 1,
        "id-equals", 1);
    do_remove("127.0.0.1", port, 0, 1,
        "description-contains", "Goat milk");

    expect_process_died(0);
}

Test(remove, status_contains)
{
    unsigned int port = -1;
    expect_process_started(0, &port);

    do_add("127.0.0.1", port, 2,
        "name", "Buy milk",
        "status", "open");
    do_remove("127.0.0.1", port, 1, 1,
        "status-contains", "open");
    do_query("127.0.0.1", port, 0, 0, 0, 0);
    do_remove("127.0.0.1", port, 0, 1,
        "id-equals", 1);
    do_remove("127.0.0.1", port, 0, 1,
        "status-contains", "open");

    expect_process_died(0);
}

static char *date_to_str(time_t date)
{
    static char buf[21];
    struct tm tinfo;

    memset(&tinfo, 0, sizeof(tinfo));
    localtime_r(&date, &tinfo);
    strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", &tinfo);
    return buf;
}

Test(remove, date_created_after)
{
    unsigned int port = -1;
    expect_process_started(0, &port);
    time_t after_add;

    do_add("127.0.0.1", port, 2,
        "name", "Buy milk",
        "status", "open");
    after_add = time(0) - 15;
    do_remove("127.0.0.1", port, 1, 1,
        "date-created-after", date_to_str(after_add));
    do_query("127.0.0.1", port, 0, 0, 0, 0);
    do_remove("127.0.0.1", port, 0, 1,
        "id-equals", 1);
    do_remove("127.0.0.1", port, 0, 1,
        "date-created-after", date_to_str(after_add));

    expect_process_died(0);
}

Test(remove, date_created_before)
{
    unsigned int port = -1;
    expect_process_started(0, &port);
    time_t after_add;

    do_add("127.0.0.1", port, 2,
        "name", "Buy milk",
        "status", "open");
    after_add = time(0) + 15;
    do_remove("127.0.0.1", port, 1, 1,
        "date-created-before", date_to_str(after_add));
    do_query("127.0.0.1", port, 0, 0, 0, 0);
    do_remove("127.0.0.1", port, 0, 1,
        "id-equals", 1);
    do_remove("127.0.0.1", port, 0, 1,
        "date-created-before", date_to_str(after_add));

    expect_process_died(0);
}

Test(remove, date_updated_after)
{
    unsigned int port = -1;
    expect_process_started(0, &port);
    time_t after_add;

    do_add("127.0.0.1", port, 2,
        "name", "Buy milk",
        "status", "open");
    after_add = time(0) - 15;
    do_remove("127.0.0.1", port, 1, 1,
        "date-updated-after", date_to_str(after_add));
    do_query("127.0.0.1", port, 0, 0, 0, 0);
    do_remove("127.0.0.1", port, 0, 1,
        "id-equals", 1);
    do_remove("127.0.0.1", port, 0, 1,
        "date-updated-after", date_to_str(after_add));

    expect_process_died(0);
}

Test(remove, date_updated_before)
{
    unsigned int port = -1;
    expect_process_started(0, &port);
    time_t after_add;

    do_add("127.0.0.1", port, 2,
        "name", "Buy milk",
        "status", "open");
    after_add = time(0) + 15;
    do_remove("127.0.0.1", port, 1, 1,
        "date-updated-before", date_to_str(after_add));
    do_query("127.0.0.1", port, 0, 0, 0, 0);
    do_remove("127.0.0.1", port, 0, 1,
        "id-equals", 1);
    do_remove("127.0.0.1", port, 0, 1,
        "date-updated-before", date_to_str(after_add));

    expect_process_died(0);
}

Test(remove, invalid_not_post)
{
    unsigned int port = -1;
    char url[256];

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/remove", "127.0.0.1", port);
    expect_http_get(url, 0, 405, 0, 0, 0);
    expect_process_no_output();
    expect_process_died(0);
}

Test(remove, invalid_not_json)
{
    unsigned int port = -1;
    char url[256];

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/remove", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", "}", 400, 0, 0, 0);
    expect_process_no_output();
    expect_process_died(0);
}

Test(remove, invalid_id_equals_not_int)
{
    unsigned int port = -1;
    char *request_str = 0;
    char url[256];

    json_t *request_json = json_object();
    json_t *tmp = json_string("invalid");
    json_object_set(request_json, "id-equals", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/remove", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(remove, invalid_name_contains_not_string)
{
    unsigned int port = -1;
    char *request_str = 0;
    char url[256];

    json_t *request_json = json_object();
    json_t *tmp = json_integer(0);
    json_object_set(request_json, "name-contains", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/remove", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(remove, invalid_description_contains_not_string)
{
    unsigned int port = -1;
    char *request_str = 0;
    char url[256];

    json_t *request_json = json_object();
    json_t *tmp = json_integer(0);
    json_object_set(request_json, "description-contains", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/remove", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(remove, invalid_status_contains_not_string)
{
    unsigned int port = -1;
    char *request_str = 0;
    char url[256];

    json_t *request_json = json_object();
    json_t *tmp = json_integer(0);
    json_object_set(request_json, "status-contains", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/remove", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(remove, invalid_date_created_after_not_integer)
{
    unsigned int port = -1;
    char *request_str = 0;
    char url[256];

    json_t *request_json = json_object();
    json_t *tmp = json_string("abcde");
    json_object_set(request_json, "date-created-after", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/remove", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(remove, invalid_date_created_before_not_integer)
{
    unsigned int port = -1;
    char *request_str = 0;
    char url[256];

    json_t *request_json = json_object();
    json_t *tmp = json_string("abcde");
    json_object_set(request_json, "date-created-before", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/remove", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(remove, invalid_date_updated_after_not_integer)
{
    unsigned int port = -1;
    char *request_str = 0;
    char url[256];

    json_t *request_json = json_object();
    json_t *tmp = json_string("abcde");
    json_object_set(request_json, "date-updated-after", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/remove", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(remove, invalid_date_updated_before_not_integer)
{
    unsigned int port = -1;
    char *request_str = 0;
    char url[256];

    json_t *request_json = json_object();
    json_t *tmp = json_string("abcde");
    json_object_set(request_json, "date-updated-before", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/remove", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}

Test(remove, invalid_unknown_arg)
{
    unsigned int port = -1;
    char *request_str = 0;
    char url[256];

    json_t *request_json = json_object();
    json_t *tmp = json_integer(0);
    json_object_set(request_json, "unknown-arg", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/api/remove", "127.0.0.1", port);
    expect_http_post(url, "Content-Type: application/json\r\n", request_str, 400, 0, 0, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died(0);
}
