/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <criterion/criterion.h>
#include <jansson.h>
#include <mongoose.h>
#include <subprocess.h>

#include "common.h"

void *DONT_CARE = &DONT_CARE;

/* LCOV_EXCL_START */

typedef struct http_body_verifier_userdata_static http_body_verifier_userdata_static;
typedef struct http_body_verifier_userdata_task http_body_verifier_userdata_task;
typedef void (*http_body_verifier)(void *, unsigned int, char *);

struct http_body_verifier_userdata_static
{
    unsigned int len;
    char *body;
};

static void http_body_verifier_static(void *, unsigned int, char *);

struct http_body_verifier_userdata_task
{
    char *id;
    char *name;
    char *description;
    char *status;
    date_verifier *date_created;
    date_verifier *date_updated;

    unsigned int history_len;
    history_entry **history_entries;
};

static void http_body_verifier_task(void *, unsigned int, char *);
static void http_body_verifier_task_get_date_updated(void *, unsigned int, char *);

static void http_body_verifier_add(void *, unsigned int, char *);

struct http_userdata
{
    int should_exit;
    int expected_code;

    unsigned int expected_headers_len;
    char **expected_headers;

    http_body_verifier body_verifier;
    void *body_verifier_userdata;
};

static char *usage = "Usage: ./tms [-h]\n"
                      "       ./tms [-a/--access-log] [-d/--dbfile dbfile] "
                      "[-i/--ip ip] [-p/--port port]\n"
                      "\n";

static subprocess *tms_sub;
static char tms_dir[16];

void build_body_from_template(tpl_format_f format_f, tpl_tag_f tag_f, tpl_enter_f enter_f,
    tpl_next_f next_f, tpl_leave_f leave_f, void *userdata, char **pout, unsigned int *plen)
{
    cr_expect_eq(1, (format_f)(tag_f, enter_f, next_f, leave_f, userdata, plen, pout));
}

static void on_connect(struct http_userdata *puserdata, int status)
{
    if (0 != status)
    {
         cr_expect_fail("Could not connect: %d", status);
         puserdata->should_exit = 1;
    }
}

static int append_header_pair(char **pmsg, unsigned int *plen, char *name,
    unsigned int name_len, char *value, unsigned int value_len)
{
    unsigned int nlen = name_len + value_len + 7;
    char *msg2;

    msg2 = (char *)realloc(*pmsg, sizeof(char) * ((*plen) + nlen + 1));
    if (0 == msg2)
    {
        cr_expect_fail("realloc() failed");
        return 0;
    }
    *pmsg = msg2;
    *plen += sprintf((*pmsg) + (*plen), "\n"
        "- %.*s: %.*s", name_len, name, value_len, value);
    return 1;
}

static void fail_header_not_found(struct http_message *hm, char *expected_name,
    char *expected_value)
{
    char *msg = 0;
    unsigned int len = 0;
    unsigned int expected_name_len = strlen(expected_name);
    unsigned int expected_value_len = strlen(expected_value);

    msg = (char *)malloc(sizeof(char) * (129 + expected_name_len + expected_value_len));
    len = sprintf(msg, "Not found header pair: %s: %s\n"
        "Header pairs with same name:", expected_name, expected_value);

    for (unsigned i = 0; 0 != hm->header_names[i].len; i++)
    {
        const char *name = hm->header_names[i].p;
        unsigned int name_len = hm->header_names[i].len;

        if (expected_name_len == name_len && 0 == strncmp(expected_name, name, name_len) &&
            0 == append_header_pair(&msg, &len, (char *)name, name_len,
                (char *)hm->header_values[i].p, hm->header_values[i].len))
        {
            break;
        }
    }

    cr_expect_fail("%s", msg);
    free(msg);
}

static void verify_a_header(struct http_message *hm, char *expected_name, char *expected_value)
{
    unsigned int expected_name_len = strlen(expected_name);
    unsigned int expected_value_len = strlen(expected_value);

    for (unsigned i = 0; 0 != hm->header_names[i].len; i++)
    {
        const char *name = hm->header_names[i].p;
        unsigned int name_len = hm->header_names[i].len;

        if (expected_name_len == name_len && 0 == strncmp(expected_name, name, name_len))
        {
            char *value = (char *)hm->header_values[i].p;
            unsigned int value_len = hm->header_values[i].len;

            if (expected_value_len == value_len &&
                0 == strncmp(expected_value, value, value_len))
            {
                return;
            }
        }
    }

    fail_header_not_found(hm, expected_name, expected_value);
}

static void verify_some_headers(struct http_userdata *puserdata, struct http_message *hm)
{
    unsigned int headers_len = 0;

    for (; 0 != hm->header_names[headers_len].len; headers_len++)
    {
    }

    cr_log_info("    expected at least %d headers", puserdata->expected_headers_len);
    cr_log_info("    actual headers len = %u", headers_len);

    for (unsigned int i = 0; i < puserdata->expected_headers_len; i++)
    {
        char *expected_name = puserdata->expected_headers[2 * i];
        char *expected_value = puserdata->expected_headers[2 * i + 1];

        cr_log_info("    expect header %u/%u: %s: %s", i, headers_len, expected_name,
            expected_value);
        verify_a_header(hm, expected_name, expected_value);
    }
}

static void http_body_verifier_static(void *userdata, unsigned int len, char *data)
{
    http_body_verifier_userdata_static *userdata2;
    userdata2 = (http_body_verifier_userdata_static *)userdata;

    cr_log_info("    expected:\n\n%.*s\n\n", userdata2->len, userdata2->body);
    if (0 != userdata2->body && (
        userdata2->len != len ||
        0 != strncmp(userdata2->body, data, userdata2->len)))
    {
        cr_expect_fail("Reply different, expected = %u, actual = %u\n"
            "Expected:\n\n%.*s\n\n"
            "Actual:\n\n%.*s\n\n",
            userdata2->len, len,
            userdata2->len, userdata2->body,
            len, data);
    }
    else if (0 == userdata2->body && 0 != len)
    {
        cr_expect_eq(0, len,
            "Reply not empty: actual = %u\n"
            "Actual:\n\n%.*s\n\n",
            len, len, data);
    }
}

static void on_reply(struct http_userdata *puserdata, struct http_message *hm)
{
    http_body_verifier verifier = puserdata->body_verifier;
    void *verifier_userdata = puserdata->body_verifier_userdata;

    cr_log_info("    expected code = %d", puserdata->expected_code);
    cr_log_info("    actual code = %u", hm->resp_code);
    cr_expect_eq(puserdata->expected_code, hm->resp_code,
        "Code different, expected = %u, actual = %u",
        puserdata->expected_code, hm->resp_code);

    verify_some_headers(puserdata, hm);

    cr_log_info("    actual:\n\n%.*s\n\n", (int)hm->body.len, hm->body.p);
    if (0 != hm->body.len)
    {
        if (0 != verifier)
        {
            (verifier)(verifier_userdata, hm->body.len, (char *)hm->body.p);
        }
        else
        {
            cr_expect_fail("No body verifier!");
        }
    }
    else if (0 != verifier)
    {
        cr_expect_fail("No body to verify!");
    }
}

static void http_event_handler(struct mg_connection *conn, int ev, void *ev_data)
{
    struct http_userdata *puserdata = (struct http_userdata *)conn->mgr->user_data;

    switch (ev)
    {
        case MG_EV_CONNECT:
            on_connect(puserdata, *(int *)ev_data);
            break;

        case MG_EV_HTTP_REPLY:
            on_reply(puserdata, (struct http_message *)ev_data);
            conn->flags |= MG_F_CLOSE_IMMEDIATELY;
            puserdata->should_exit = 1;
            break;

        case MG_EV_CLOSE:
            if (0 == puserdata->should_exit)
            {
                cr_expect_fail("Server closed connection");
                puserdata->should_exit = 1;
            }
            break;
        default:
            break;
    }
}

static void expect_http(char *url, char *headers, char *body, int expected_code,
    unsigned int expected_headers_len, char **expected_headers,
    http_body_verifier verifier, void *verifier_userdata)
{
    struct http_userdata userdata;
    struct mg_mgr mgr;

    cr_log_info("# Expect HTTP request");
    cr_log_info("    url = %s", url);
    cr_log_info("    headers = %s", headers);
    cr_log_info("    body = %s", body);
    cr_log_info("    expected_code = %d", expected_code);
    cr_log_info("    expected_headers_len = %d", expected_headers_len);
    for (unsigned int i = 0; i < expected_headers_len; i++)
    {
        cr_log_info("    header %u = %s: %s\n", i, expected_headers[2 * i],
            expected_headers[2 * i + 1]);
    }

    userdata.should_exit = 0;
    userdata.expected_headers_len = expected_headers_len;
    userdata.expected_headers = expected_headers;
    userdata.expected_code = expected_code;
    userdata.body_verifier = verifier;
    userdata.body_verifier_userdata = verifier_userdata;

    mg_mgr_init(&mgr, 0);
    mgr.user_data = &userdata;
    mg_connect_http(&mgr, http_event_handler, url, headers, body);
    while (0 == userdata.should_exit)
    {
        mg_mgr_poll(&mgr, 250);
    }
    mg_mgr_free(&mgr);
}

void expect_http_get(char *url, char *headers, int expected_code, unsigned int expected_headers_len,
    char **expected_headers, char *expected_body)
{
    http_body_verifier_userdata_static verifier_data =
    {
        0 != expected_body ? strlen(expected_body) : 0,
        expected_body
    };
    expect_http(url, headers, 0, expected_code, expected_headers_len,
        expected_headers,
        0 != expected_body ? http_body_verifier_static : 0,
        0 != expected_body ? (void *)&verifier_data : 0);
}

void expect_http_post(char *url, char *headers, char *body, int expected_code,
    unsigned int expected_headers_len, char **expected_headers, char *expected_body)
{
    http_body_verifier_userdata_static verifier_data =
    {
        0 != expected_body ? strlen(expected_body) : 0,
        expected_body
    };
    expect_http(url, headers, body, expected_code, expected_headers_len,
        expected_headers,
        0 != expected_body ? http_body_verifier_static : 0,
        0 != expected_body ? (void *)&verifier_data : 0);
}

void expect_process_usage(char **argv, int code)
{
    subprocess *sub = 0;
    subprocess_partial_stream *out = 0;
    subprocess_partial_stream *err = 0;

    char *expected_stdout = "";
    int expected_stdout_len = 0;
    char *expected_stderr = "";
    int expected_stderr_len = 0;

    if (0 == code)
    {
        expected_stdout = usage;
        expected_stdout_len = strlen(usage);
    }
    else
    {
        expected_stderr = usage;
        expected_stderr_len = strlen(usage);
    }

    if (0 == subprocess_new(argv, 0, 1, 1, &sub))
    {
        return;
    }

    subprocess_wait(sub, code);

    subprocess_read_output(sub, 20, &out);
    subprocess_partial_stream_cmp(out, expected_stdout_len, expected_stdout);
    subprocess_partial_stream_free(&out);

    subprocess_read_error(sub, 0, &err);
    subprocess_partial_stream_cmp(err, expected_stderr_len, expected_stderr);
    subprocess_partial_stream_free(&err);

    subprocess_free(&sub);
}

static void get_process_ip_port(subprocess_partial_stream *ps, char **pip,
    unsigned int *pport)
{
    subprocess_partial_stream_cmp(ps, 11, "Started on ");

    if (0 != pip)
    {
        char *ip = 0;
        unsigned int ip_len = 0;
        subprocess_partial_stream_get_until(ps, ':', &ip_len, &ip);
        (*pip) = (char *)malloc(sizeof(char) * (ip_len + 1));
        memcpy(*pip, ip, ip_len);
        (*pip)[ip_len] = 0;
    }
    else
    {
        subprocess_partial_stream_skip_until(ps, ':');
    }

    if (0 != pport)
    {
        char *portstr = 0;
        unsigned int portlen = 0;

        subprocess_partial_stream_get_until(ps, '\n', &portlen, &portstr);
        if (0 != portstr)
        {
            *pport = (unsigned int) atoi(portstr);
        }
    }
    else
    {
        subprocess_partial_stream_skip_until(ps, '\n');
    }
}

void expect_process_started_and_died(char **argv)
{
    subprocess *sub = 0;
    subprocess_partial_stream *out = 0;
    subprocess_partial_stream *err = 0;

    unsigned int port = -1;
    char url[256];

    if (0 == subprocess_new(argv, 0, 1, 1, &sub))
    {
        return;
    }

    subprocess_read_output(sub, 20, &out);
    get_process_ip_port(out, 0, &port);
    subprocess_partial_stream_free(&out);

    sprintf(url, "http://%s:%u/api/ping", "127.0.0.1", port);
    expect_http_get(url, 0, 200, 0, 0, 0);

    subprocess_kill(sub, SIGINT);
    subprocess_wait(sub, 0);

    subprocess_read_output(sub, 0, &out);
    subprocess_partial_stream_cmp(out, 8, "Stopped\n");
    subprocess_partial_stream_free(&out);

    subprocess_read_output(sub, 0, &err);
    subprocess_partial_stream_free(&err);

    subprocess_free(&sub);
}

static void expect_process_started2(int access_log, char **pip, unsigned int *pport)
{
    char *argv[6];
    unsigned int i = 0;
    subprocess_partial_stream *out = 0;

    sprintf(tms_dir, "tm-%d", getpid());
    if (0 != mkdir(tms_dir, 0700))
    {
        cr_expect_fail("Cannot mkdir() %s: %d %s", tms_dir, errno, strerror(errno));
        return;
    }
    if (0 != chdir(tms_dir))
    {
        rmdir(tms_dir);
        cr_expect_fail("Cannot chdir() %s: %d %s", tms_dir, errno, strerror(errno));
        return;
    }

    argv[i++] = "../tms";
    if (1 == access_log)
    {
        argv[i++] = "-a";
    }
    argv[i++] = "-p";
    argv[i++] = "0";
    argv[i++] = 0;

    tms_sub = 0;
    if (0 == subprocess_new(argv, 0, 1, 1, &tms_sub))
    {
        return;
    }

    subprocess_read_output(tms_sub, 20, &out);
    get_process_ip_port(out, pip, pport);
    subprocess_partial_stream_free(&out);
}

void expect_process_started(char **pip, unsigned int *pport)
{
    expect_process_started2(0, pip, pport);
}

void expect_process_started_accesslog(char **pip, unsigned int *pport)
{
    expect_process_started2(1, pip, pport);
}

void expect_process_no_output(void)
{
    subprocess_partial_stream *out = 0;
    subprocess_partial_stream *err = 0;

    subprocess_read_output(tms_sub, 0, &out);
    subprocess_partial_stream_free(&out);

    subprocess_read_output(tms_sub, 0, &err);
    subprocess_partial_stream_free(&err);
}

void expect_process_output_accesslog(char *ip, char *method, char *url, unsigned int code,
    unsigned int length)
{
    subprocess_partial_stream *out = 0;
    subprocess_partial_stream *err = 0;

    char tmp[32];

    subprocess_read_output(tms_sub, 0, &out);
    subprocess_partial_stream_cmp(out, strlen(ip), ip);
    subprocess_partial_stream_cmp(out, 1, " ");
    subprocess_partial_stream_cmp(out, 7, "unknown");
    subprocess_partial_stream_cmp(out, 1, " ");
    subprocess_partial_stream_cmp(out, 7, "unknown");
    subprocess_partial_stream_cmp(out, 2, " [");
    /* 10/Oct/2000:13:55:36 -0700 */
    subprocess_partial_stream_skip_until(out, ']');
    subprocess_partial_stream_cmp(out, 2, " \"");
    subprocess_partial_stream_cmp(out, strlen(method), method);
    subprocess_partial_stream_cmp(out, 1, " ");
    subprocess_partial_stream_cmp(out, strlen(url), url);
    subprocess_partial_stream_cmp(out, 1, " ");
    subprocess_partial_stream_cmp(out, 10, "HTTP/1.1\" ");

    sprintf(tmp, "%u", code);
    subprocess_partial_stream_cmp(out, strlen(tmp), tmp);
    subprocess_partial_stream_cmp(out, 1, " ");

    sprintf(tmp, "%u", length);
    subprocess_partial_stream_cmp(out, strlen(tmp), tmp);
    subprocess_partial_stream_cmp(out, 1, "\n");

    subprocess_partial_stream_free(&out);

    subprocess_read_output(tms_sub, 0, &err);
    subprocess_partial_stream_free(&err);
}

void expect_process_died(char *ip)
{
    subprocess_partial_stream *out = 0;
    subprocess_partial_stream *err = 0;

    if (0 != ip)
    {
        free(ip);
    }

    subprocess_kill(tms_sub, SIGINT);
    subprocess_wait(tms_sub, 0);

    subprocess_read_output(tms_sub, 0, &out);
    subprocess_partial_stream_cmp(out, 8, "Stopped\n");
    subprocess_partial_stream_free(&out);

    subprocess_read_output(tms_sub, 0, &err);
    subprocess_partial_stream_free(&err);

    subprocess_free(&tms_sub);
    remove("tm.db");
    chdir("..");
    rmdir(tms_dir);
}

static void modifiers_to_json(json_t *query, va_list list)
{
    char *key = va_arg(list, char *);

    if (0 == strcmp(key, "name") ||
        0 == strcmp(key, "description") ||
        0 == strcmp(key, "status"))
    {
        json_t *value;
        value = json_string(va_arg(list, char *));
        json_object_set(query, key, value);
        json_decref(value);
    }
}

static void expect_json_object_int(json_t *root, char *prefix, char *name)
{
    json_t *tmp = 0;

    tmp = json_object_get(root, name);
    if (0 == tmp)
    {
        cr_expect_fail("No JSON %s%s", prefix, name);
    }
    else if (0 == json_is_integer(tmp))
    {
        cr_expect_fail("JSON %s%s not integer", prefix, name);
    }
}

static void http_body_verifier_add(void *userdata, unsigned int len, char *data)
{
    json_error_t err;
    json_t *root = 0;

    root = json_loadb(data, len, 0, &err);
    if (0 == root)
    {
        cr_expect_fail("Cannot parse body as JSON");
        return;
    }

    if (0 == json_is_object(root))
    {
        cr_expect_fail("JSON body is not object");
        return;
    }

    cr_expect_eq(1, json_object_size(root), "JSON body size different");
    expect_json_object_int(root, "", "id");
    json_decref(root);
}

void do_add(char *ip, unsigned int port, unsigned int modifiers_len, ...)
{
    json_t *request_json = json_object();
    char *request_str = 0;
    va_list list;
    char url[256];
    char *expected_headers[] =
    {
        "Cache-Control", "no-store"
    };

    sprintf(url, "http://%s:%u/api/add", ip, port);

    va_start(list, modifiers_len);
    for (unsigned int i = 0; i < modifiers_len; i++)
    {
        modifiers_to_json(request_json, list);
    }
    va_end(list);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);
    expect_http(url, "Content-Type: application/json\r\n", request_str, 200, 1, expected_headers,
        http_body_verifier_add, 0);

    free(request_str);

    expect_process_no_output();
}

static void filters_to_json(json_t *filters, va_list list)
{
    char *key = va_arg(list, char *);
    json_t *value;

    if (0 == strcmp(key, "id-equals"))
    {
        value = json_integer(va_arg(list, unsigned int));
    }
    else if (0 == strcmp(key, "date-created-after") ||
        0 == strcmp(key, "date-created-before") ||
        0 == strcmp(key, "date-updated-after") ||
        0 == strcmp(key, "date-updated-before") ||
        0 == strcmp(key, "name-contains") ||
        0 == strcmp(key, "description-contains") ||
        0 == strcmp(key, "status-contains"))
    {
        value = json_string(va_arg(list, char *));
    }
    else
    {
        cr_expect_fail("Unexpected filter: %s", key);
        va_arg(list, void *);
        return;
    }
    json_object_set(filters, key, value);
    json_decref(value);
}

static void opts_to_json(json_t *opts, va_list list)
{
    char *key = va_arg(list, char *);
    json_t *value;

    if (0 == strcmp(key, "show-column-description") ||
        0 == strcmp(key, "show-column-status"))
    {
        value = json_true();
    }
    else
    {
        cr_expect_fail("Unexpected opt: %s", key);
        return;
    }
    json_object_set(opts, key, value);
    json_decref(value);
}

void do_remove(char *ip, unsigned int port, int found, unsigned int filters_len, ...)
{
    char *request_str = 0;
    va_list list;
    char url[256];
    sprintf(url, "http://%s:%u/api/remove", ip, port);

    va_start(list, filters_len);
    json_t *request_json = json_object();
    if (0 != filters_len)
    {
        unsigned int i;
        for (i = 0; i < filters_len; i++)
        {
            filters_to_json(request_json, list);
        }
        request_str = json_dumps(request_json, JSON_COMPACT);
        json_decref(request_json);
    }
    va_end(list);

    if (1 == found)
    {
        char *expected_headers[] =
        {
            "Cache-Control", "no-store"
        };
        expect_http(url, "Content-Type: application/json\r\n", request_str, 200, 1,
            expected_headers, 0, 0);
    }
    else
    {
        expect_http(url, "Content-Type: application/json\r\n", request_str, 404, 0, 0, 0, 0);
    }
    free(request_str);

    expect_process_no_output();
}

void do_modify(char *ip, unsigned int port, int found, unsigned int filters_len,
    unsigned int modifiers_len, ...)
{
    json_t *request_json = json_object();
    json_t *filters = json_object();
    json_t *modifiers = json_object();
    char *request_str = 0;
    va_list list;
    int i;
    char url[256];

    sprintf(url, "http://%s:%u/api/modify", ip, port);

    va_start(list, modifiers_len);
    json_object_set(request_json, "filters", filters);
    for (i = 0; i < filters_len; i++)
    {
        filters_to_json(filters, list);
    }
    json_decref(filters);

    json_object_set(request_json, "modifiers", modifiers);
    for (i = 0; i < modifiers_len; i++)
    {
        modifiers_to_json(modifiers, list);
    }
    json_decref(modifiers);
    va_end(list);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    if (1 == found)
    {
        char *expected_headers[] =
        {
            "Cache-Control", "no-store"
        };
        expect_http(url, "Content-Type: application/json\r\n", request_str, 200, 1,
            expected_headers, 0, 0);
    }
    else
    {
        expect_http(url, "Content-Type: application/json\r\n", request_str, 404, 0, 0, 0, 0);
    }
    free(request_str);

    expect_process_no_output();
}

static void query_args_to_json(va_list list, unsigned int filters_len, unsigned int opts_len,
    json_t *request_json)
{
    unsigned int i;
    if (0 != filters_len)
    {
        json_t *filters_json = json_object();
        for (i = 0; i < filters_len; i++)
        {
            filters_to_json(filters_json, list);
        }
        json_object_set(request_json, "filters", filters_json);
        json_decref(filters_json);
    }
    if (0 != opts_len)
    {
        json_t *opts_json = json_object();
        for (i = 0; i < opts_len; i++)
        {
            opts_to_json(opts_json, list);
        }
        json_object_set(request_json, "opts", opts_json);
        json_decref(opts_json);
    }
}

static void build_expected_columns(va_list list, unsigned int columns_len, json_t *ret_json)
{
    json_t *columns_json = json_array();
    unsigned int i;

    json_object_set(ret_json, "columns", columns_json);
    for (i = 0; i < columns_len; i++)
    {
        char *column = va_arg(list, char *);
        json_t *tmp = json_string(column);
        json_array_append(columns_json, tmp);
        json_decref(tmp);
    }
    json_decref(columns_json);
}

static void build_expected_item(va_list list, json_t **pret_json)
{
    unsigned int item_len = va_arg(list, unsigned int);
    unsigned int i;
    json_t *item_json = json_object();

    for (i = 0; i < item_len; i++)
    {
        char *key = va_arg(list, char *);
        json_t *tmp = 0;

        if (0 == strcmp("id", key))
        {
            tmp = json_integer(va_arg(list, unsigned int));
        }
        else if (0 == strcmp("name", key) ||
            0 == strcmp("description", key) ||
            0 == strcmp("status", key))
        {
            tmp = json_string(va_arg(list, char *));
        }
        else if (0 == strcmp("history", key))
        { /* @todo */
            va_arg(list, unsigned int);
        }
        else if (0 == strcmp("date-created", key) ||
            0 == strcmp("date-updated", key))
        {
           /* @todo */
            va_arg(list, void *);
        }
        else
        {
            continue;
        }

        json_object_set(item_json, key, tmp);
        json_decref(tmp);
    }
    *pret_json = item_json;
}

static void build_expected_items(va_list list, unsigned int items_len, json_t *ret_json)
{
    json_t *items_json = json_array();
    unsigned int i;

    json_object_set(ret_json, "items", items_json);
    for (i = 0; i < items_len; i++)
    {
        json_t *item_json = 0;
        build_expected_item(list, &item_json);
        json_array_append(items_json, item_json);
        json_decref(item_json);
    }
    json_decref(items_json);
}

static void build_query_response(va_list list, unsigned int columns_len,
    unsigned int items_len, char **pret)
{
    json_t *ret_json = json_object();
    char *ret_str = 0;

    build_expected_columns(list, columns_len, ret_json);
    build_expected_items(list, items_len, ret_json);

    ret_str = json_dumps(ret_json, JSON_COMPACT);
    json_decref(ret_json);
    *pret = ret_str;
}

void do_query(char *ip, unsigned int port, unsigned int filters_len, unsigned int opts_len,
    unsigned int columns_len, unsigned int items_len, ...)
{
    char *ret_str = 0;
    char ret_str_len_str[64];
    char *request_str = 0;
    va_list list;
    char url[256];
    char *expected_headers[] =
    {
        "Content-Length", ret_str_len_str,
        "Content-Type", "application/json",
        "Cache-Control", "no-store"
    };
    http_body_verifier_userdata_static verifier_data;

    sprintf(url, "http://%s:%u/api/query", ip, port);

    va_start(list, items_len);
    if (0 != filters_len || 0 != opts_len)
    {
        json_t *request_json = json_object();
        query_args_to_json(list, filters_len, opts_len, request_json);
        request_str = json_dumps(request_json, JSON_COMPACT);
        json_decref(request_json);
    }
    build_query_response(list, columns_len, items_len, &ret_str);
    va_end(list);

    verifier_data.len = strlen(ret_str);
    verifier_data.body = ret_str;

    sprintf(ret_str_len_str, "%lu", strlen(ret_str));

    if (0 != request_str)
    {
        expect_http(url, "Content-Type: application/json\r\n"
            "Accept: application/json\r\n", request_str,
            200, 3, expected_headers, http_body_verifier_static, (void *)&verifier_data);
        free(request_str);
    }
    else
    {
        expect_http(url, "Accept: application/json\r\n", 0,
            200, 3, expected_headers, http_body_verifier_static, (void *)&verifier_data);
    }
    free(ret_str);

    expect_process_no_output();
}

static void expect_json_object_str_eq(json_t *root, char *prefix, char *name, char *expected_value)
{
    json_t *tmp = 0;

    tmp = json_object_get(root, name);
    if (0 == expected_value)
    {
        if (0 != tmp)
        {
            cr_expect_fail("Unexpected JSON %s%s", prefix, name);
        }
    }
    else if (0 == tmp)
    {
        cr_expect_fail("No JSON %s%s", prefix, name);
    }
    else if (0 == json_is_string(tmp))
    {
        cr_expect_fail("JSON %s%s not string", prefix, name);
    }
    else
    {
        char *actual_value = (char *)json_string_value(tmp);
        cr_expect_str_eq(actual_value, expected_value,
            "JSON %s%s different\n"
            "Expected: \"%s\"\n"
            "Actual  : \"%s\"", prefix, name, actual_value, expected_value);
    }
}

static void expect_date_after(char *prefix, char *name, time_t a, time_t b)
{
    if (a < b)
    {
        char tmp[64];
        char tmp2[64];
        struct tm tma;
        struct tm tmb;

        localtime_r(&a, &tma);
        localtime_r(&b, &tmb);
        strftime(tmp, sizeof(tmp), "%c", &tma);
        strftime(tmp2, sizeof(tmp2), "%c", &tmb);

        cr_expect_fail("JSON %s%s date not after\n"
            "Expected to be after: %s (%lu)\n"
            "Actually is         : %s (%lu)",
            prefix, name,
            tmp2, b,
            tmp, a);
    }
}

static void expect_json_object_date(json_t *root, char *prefix, char *name,
    date_verifier *verifier)
{
    json_t *tmp = 0;

    if (DONT_CARE == verifier)
    {
        return;
    }

    tmp = json_object_get(root, name);
    if (0 == verifier)
    {
        if (0 != tmp)
        {
            cr_expect_fail("Unexpected JSON %s%s", prefix, name);
        }
    }
    else if (0 == tmp)
    {
        cr_expect_fail("No JSON %s%s", prefix, name);
    }
    else if (0 == json_is_integer(tmp))
    {
        cr_expect_fail("JSON %s%s not integer", prefix, name);
    }
    else
    {
        time_t actual_value = (time_t)json_integer_value(tmp);

        switch (verifier->type)
        {
            case DVT_AFTER:
                expect_date_after(prefix, name, actual_value, verifier->value);
                break;

            case DVT_BEFORE:
                expect_date_after(prefix, name, verifier->value, actual_value);
                break;

            default:
                cr_expect_fail("Unknown date verifier type: %d", verifier->type);
                break;
        }
    }
}

static void expect_json_object_history_item_eq(json_t *root, char *name, unsigned int i,
    history_entry **expected_entries)
{
    json_t *tmp = json_array_get(root, i);
    history_entry *entry = expected_entries[i];

    if (0 == json_is_object(tmp))
    {
        cr_expect_fail("JSON %s.%u not object", name, i);
    }
    else
    {
        char prefix[64];
        sprintf(prefix, "%s.%u ", name, i);
        cr_expect_eq(4 + (0 != entry->new_value ? 1 : 0), json_object_size(tmp));
        expect_json_object_str_eq(tmp, prefix, "type", entry->type);
        expect_json_object_str_eq(tmp, prefix, "name", entry->name);
        expect_json_object_str_eq(tmp, prefix, "value", entry->value);
        expect_json_object_str_eq(tmp, prefix, "new-value", entry->new_value);
        expect_json_object_str_eq(tmp, prefix, "new-value", entry->new_value);
        expect_json_object_date(tmp, prefix, "date", entry->date);
    }
}

static void expect_json_object_history_eq(json_t *root, char *name,
    unsigned int expected_len, history_entry **expected_entries)
{
    json_t *tmp = 0;

    if (DONT_CARE == expected_entries)
    {
        return;
    }

    tmp = json_object_get(root, name);
    if (0 == tmp)
    {
        cr_expect_fail("No JSON %s", name);
    }
    else if (0 == json_is_array(tmp))
    {
        cr_expect_fail("JSON %s not array", name);
    }
    else
    {
        unsigned int len = json_array_size(tmp);
        cr_expect_eq(expected_len, len, "JSON %s size different\n"
            "Expected: \"%u\"\n"
            "Actual  : \"%u\"", name, expected_len, len);

        for (unsigned int i = 0; i < len; i++)
        {
            expect_json_object_history_item_eq(tmp, name, i, expected_entries);
        }
    }
}

static void http_body_verifier_task(void *userdata, unsigned int len, char *data)
{
    http_body_verifier_userdata_task *puserdata = (http_body_verifier_userdata_task *)userdata;
    json_error_t err;
    json_t *root = 0;

    root = json_loadb(data, len, 0, &err);
    if (0 == root)
    {
        cr_expect_fail("Cannot parse body as JSON");
        return;
    }

    if (0 == json_is_object(root))
    {
        cr_expect_fail("JSON body is not object");
        return;
    }

    cr_expect_eq(5 + (0 != puserdata->description ? 1 : 0),
        json_object_size(root), "JSON body size different");

    expect_json_object_str_eq(root, "", "name", puserdata->name);
    expect_json_object_str_eq(root, "", "description", puserdata->description);
    expect_json_object_str_eq(root, "", "status", puserdata->status);
    expect_json_object_date(root, "", "date-created", puserdata->date_created);
    expect_json_object_date(root, "", "date-updated", puserdata->date_updated);
    expect_json_object_history_eq(root, "history", puserdata->history_len,
        puserdata->history_entries);

    json_decref(root);
}

static void http_body_verifier_task_get_date_updated(void *userdata, unsigned int len, char *data)
{
    time_t *pdate = (time_t *)userdata;
    json_error_t err;
    json_t *root = 0;
    json_t *date_updated = 0;

    root = json_loadb(data, len, 0, &err);
    if (0 == root)
    {
        cr_expect_fail("Cannot parse body as JSON");
        return;
    }

    if (0 == json_is_object(root))
    {
        cr_expect_fail("JSON body is not object");
        return;
    }

    date_updated = json_object_get(root, "date-updated");
    if (0 == date_updated)
    {
        cr_expect_fail("No date-updated");
    }
    else if (0 == json_is_integer(date_updated))
    {
        cr_expect_fail("date-updated is not integer");
    }
    else
    {
        time_t actual_value = (time_t)json_integer_value(date_updated);
        *pdate = actual_value;
    }

    json_decref(root);
}

void do_task(char *ip, unsigned int port, char *id, char *name, char *description,
    char *status, date_verifier *date_created, date_verifier *date_updated,
    unsigned int history_len, history_entry **history_entries)
{
    char url[256];
    http_body_verifier_userdata_task userdata;
    char *expected_headers[] =
    {
        "Content-Type", "application/json",
        "Cache-Control", "no-store"
    };

    memset(&userdata, 0, sizeof(userdata));
    userdata.id = id;
    userdata.name = name;
    userdata.description = description;
    userdata.status = status;
    userdata.date_created = date_created;
    userdata.date_updated = date_updated;
    userdata.history_len = history_len;
    userdata.history_entries = history_entries;

    sprintf(url, "http://%s:%u/api/task/%s", ip, port, id);

    expect_http(url,
        "Accept: application/json\r\n",
        0, 200,
        2, expected_headers,
        http_body_verifier_task, &userdata);
    expect_process_no_output();
}

void do_task_get_date_updated(char *ip, unsigned int port, char * id, time_t *pdate)
{
    char url[256];
    char *expected_headers[] =
    {
        "Content-Type", "application/json",
        "Cache-Control", "no-store"
    };

    sprintf(url, "http://%s:%u/api/task/%s", ip, port, id);

    expect_http(url,
        "Accept: application/json\r\n",
        0, 200,
        2, expected_headers,
        http_body_verifier_task_get_date_updated, (void *)pdate);
    expect_process_no_output();
}

/* LCOV_EXCL_STOP */
