/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <criterion/criterion.h>
#include <jansson.h>

#include "common.h"

#include <static_index_html.h>

TestSuite(index_html);

typedef struct reply_results reply_results;
typedef struct row row;

enum reply_entered
{
    RE_NULL,
    RE_INDEX,
    RE_INDEX_ROWS,
    RE_INDEX_ROWS_COLUMNS,
    RE_INDEX_COLUMNS
};

struct row
{
    unsigned long id;
    char *name;
    char *status;
    unsigned int column_i;
};

struct reply_results
{
    enum reply_entered entered;
    int index_i;

    int has_rows;
    row **rows;
    unsigned int rows_len;
    unsigned int row_i;

    unsigned int column_i;
};

static int tag_rows(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    reply_results *pres = (reply_results*)userdata;
    row *a_row = 0;
    a_row = pres->rows[pres->row_i];
    if (0 == strcmp("id", tag))
    {
        char tmp[64];
        sprintf(tmp, "%lu", a_row->id);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("name", tag))
    {
        return tpl_append(pout, plen, a_row->name, strlen(a_row->name));
    }
    return 0;
}

static int tag_rows_columns(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    reply_results *pres = (reply_results*)userdata;
    row *a_row = 0;

    a_row = pres->rows[pres->row_i];
    if (0 == strcmp("name", tag))
    {
        return tpl_append(pout, plen, "status", 6);
    }
    else if (0 == strcmp("value", tag))
    {
        char *tmp = a_row->status;
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    return 0;
}

static int tag_columns(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    if (0 == strcmp("name", tag))
    {
        return tpl_append(pout, plen, "status", 6);
    }
    else if (0 == strcmp("name-human-readable", tag))
    {
        return tpl_append(pout, plen, "Status", 6);
    }
    return 0;
}

static int tag_top(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    if (0 == strcmp("index-style", tag))
    {
        return 1;
    }
    else if (0 == strcmp("add-style", tag) ||
        0 == strcmp("edit-style", tag) ||
        0 == strcmp("error-style", tag) ||
        0 == strcmp("task-style", tag) ||
        0 == strcmp("warning-style", tag))
    {
        char *tmp = "style='display: none;'";
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }

    return 0;
}

static int tag(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    reply_results *pres = (reply_results*)userdata;
    switch (pres->entered)
    {
        case RE_INDEX_ROWS:
            return tag_rows(pout, plen, userdata, tag);
        case RE_INDEX_ROWS_COLUMNS:
            return tag_rows_columns(pout, plen, userdata, tag);
        case RE_INDEX_COLUMNS:
            return tag_columns(pout, plen, userdata, tag);
        case RE_NULL:
            return tag_top(pout, plen, userdata, tag);
        default:
            break;
    }
    return 0;
}

static int enter_index(char **pout, unsigned int *plen, void *userdata, char *name)
{
    reply_results *pres = (reply_results*)userdata;
    if (0 == strcmp("rows", name) && 1 == pres->has_rows)
    {
        pres->row_i = -1;
        pres->entered = RE_INDEX_ROWS;
        return 1;
    }
    else if (0 == strcmp("columns", name))
    {
        pres->column_i = -1;
        pres->entered = RE_INDEX_COLUMNS;
        return 1;
    }
    return 0;
}

static int enter_index_rows(char **pout, unsigned int *plen, void *userdata, char *name)
{
    reply_results *pres = (reply_results*)userdata;
    if (0 == strcmp("columns", name))
    {
        pres->rows[pres->row_i]->column_i = -1;
        pres->entered = RE_INDEX_ROWS_COLUMNS;
        return 1;
    }
    return 0;
}

static int enter_top(char **pout, unsigned int *plen, void *userdata, char *name)
{
    reply_results *pres = (reply_results*)userdata;
    if (0 == strcmp("index", name))
    {
        pres->index_i = -1;
        pres->entered = RE_INDEX;
        return 1;
    }
    return 0;
}

static int enter(char **pout, unsigned int *plen, void *userdata, char *name)
{
    reply_results *pres = (reply_results*)userdata;
    switch (pres->entered)
    {
        case RE_INDEX:
            return enter_index(pout, plen, userdata, name);

        case RE_INDEX_ROWS:
            return enter_index_rows(pout, plen, userdata, name);

        case RE_NULL:
            return enter_top(pout, plen, userdata, name);

        default:
            break;
    }
    return 0;
}

static int next(char **pout, unsigned int *plen, void *userdata)
{
    reply_results *pres = (reply_results*)userdata;
    row *a_row = 0;
    switch (pres->entered)
    {
        case RE_INDEX_ROWS:
            pres->row_i++;
            return pres->rows_len > pres->row_i;
        case RE_INDEX_ROWS_COLUMNS:
            a_row = pres->rows[pres->row_i];
            a_row->column_i++;
            return 0 == a_row->column_i;
        case RE_INDEX_COLUMNS:
            pres->column_i++;
            return 0 == pres->column_i;
        case RE_INDEX:
            pres->index_i++;
            return 0 == pres->index_i;
        default:
            break;
    }
    return 0;
}

static void leave(char **pout, unsigned int *plen, void *userdata)
{
    reply_results *pres = (reply_results*)userdata;
    switch (pres->entered)
    {
        case RE_INDEX_ROWS:
        case RE_INDEX_COLUMNS:
            pres->entered = RE_INDEX;
            break;
        case RE_INDEX_ROWS_COLUMNS:
            pres->entered = RE_INDEX_ROWS;
            break;
        case RE_INDEX:
            pres->entered = RE_NULL;
            break;
        default:
            break;
    }
}

Test(index_html, get_empty)
{
    char *body = 0;
    unsigned int len = 0;
    reply_results res;
    unsigned int port = -1;
    char url[256];
    char len_str[64];
    char *expected_headers[6] =
    {
        "Content-Type", "text/html",
        "Content-Length", len_str,
        "Cache-Control", "no-store"
    };

    res.entered = RE_NULL;
    res.has_rows = 0;
    res.column_i = -1;

    expect_process_started(0, &port);
    build_body_from_template(tpl_file_static_index_html,
        tag, enter, next, leave, &res, &body, &len);
    sprintf(len_str, "%u", len);

    sprintf(url, "http://%s:%u/", "127.0.0.1", port);
    expect_http_get(url, 0, 200, 3, expected_headers, body);

    sprintf(url, "http://%s:%u/index.html", "127.0.0.1", port);
    expect_http_get(url, 0, 200, 3, expected_headers, body);

    free(body);
    expect_process_died(0);
}

Test(index_html, post_empty)
{
    char *body = 0;
    unsigned int len = 0;
    char *post = "filter-id-equals=&"
        "filter-name-contains=&"
        "filter-description-contains=&"
        "filter-status-contains=&"
        "filter-date-created-after=&"
        "filter-date-created-before=&"
        "filter-date-updated-after=&"
        "filter-date-updated-before=";
    reply_results res;
    unsigned int port = -1;
    char url[256];
    char len_str[64];
    char *expected_headers[6] =
    {
        "Content-Type", "text/html",
        "Content-Length", len_str,
        "Cache-Control", "no-store"
    };

    res.entered = RE_NULL;
    res.has_rows = 0;
    res.column_i = -1;

    expect_process_started(0, &port);
    build_body_from_template(tpl_file_static_index_html,
        tag, enter, next, leave, &res, &body, &len);
    sprintf(len_str, "%u", len);

    sprintf(url, "http://%s:%u/", "127.0.0.1", port);
    expect_http_post(url, 0, post, 200, 3, expected_headers, body);

    sprintf(url, "http://%s:%u/index.html", "127.0.0.1", port);
    expect_http_post(url, 0, post, 200, 3, expected_headers, body);

    free(body);
    expect_process_died(0);
}

Test(index_html, post_nonempty)
{
    char *body = 0;
    unsigned int len = 0;
    char post[1024];
    row row1 = {1, "Buy milk", "new", -1};
    row *rows[] = {&row1};
    reply_results res;
    unsigned int port = -1;
    char url[256];
    char len_str[64];
    char *expected_headers[6] =
    {
        "Content-Type", "text/html",
        "Content-Length", len_str,
        "Cache-Control", "no-store"
    };
    time_t date_updated = 0;

    res.entered = RE_NULL;
    res.has_rows = 1;
    res.rows = rows;
    res.rows_len = 1;
    res.row_i = -1;
    res.column_i = -1;

    expect_process_started(0, &port);
    do_add("127.0.0.1", port, 2,
        "name", "Buy milk",
        "description", "Goat milk");
    do_task_get_date_updated("127.0.0.1", port, "1", &date_updated);
    sprintf(post, "filter-id-equals=1&"
        "filter-name-contains=Buy&"
        "filter-description-contains=Goat&"
        "filter-status-contains=new&"
        "filter-date-created-after=%ld&"
        "filter-date-created-before=%ld&"
        "filter-date-updated-after=%ld&"
        "filter-date-updated-before=%ld",
        date_updated - 15,
        date_updated + 15,
        date_updated - 15,
        date_updated + 15);

    build_body_from_template(tpl_file_static_index_html,
        tag, enter, next, leave, &res, &body, &len);
    sprintf(len_str, "%u", len);

    sprintf(url, "http://%s:%u/", "127.0.0.1", port);
    expect_http_post(url, 0, post, 200, 3, expected_headers, body);

    sprintf(url, "http://%s:%u/index.html", "127.0.0.1", port);
    expect_http_post(url, 0, post, 200, 3, expected_headers, body);

    free(body);
    expect_process_died(0);
}

Test(index_html, get_nonempty)
{
    char *body = 0;
    unsigned int len = 0;
    row row1 = {1, "Buy milk", "new", -1};
    row *rows[] = {&row1};
    reply_results res;
    unsigned int port = -1;
    char url[256];
    char len_str[64];
    char *expected_headers[6] =
    {
        "Content-Type", "text/html",
        "Content-Length", len_str,
        "Cache-Control", "no-store"
    };

    res.entered = RE_NULL;
    res.has_rows = 1;
    res.rows = rows;
    res.rows_len = 1;
    res.row_i = -1;
    res.column_i = -1;

    expect_process_started(0, &port);
    do_add("127.0.0.1", port, 1,
        "name", "Buy milk");
    build_body_from_template(tpl_file_static_index_html,
        tag, enter, next, leave, &res, &body, &len);
    sprintf(len_str, "%u", len);

    sprintf(url, "http://%s:%u/", "127.0.0.1", port);
    expect_http_get(url, 0, 200, 3, expected_headers, body);

    sprintf(url, "http://%s:%u/", "127.0.0.1", port);
    expect_http_get(url, 0, 200, 3, expected_headers, body);

    free(body);
    expect_process_died(0);
}
