/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <criterion/criterion.h>
#include <jansson.h>

#include "common.h"

#include <static_index_html.h>

TestSuite(task_html);

typedef struct reply_results reply_results;
typedef struct history_item history_item;

enum reply_entered
{
    RE_NULL,
    RE_TASK,
    RE_TASK_HISTORY,
    RE_TASK_HISTORY_CHANGE,
};

struct reply_results
{
    enum reply_entered entered;

    unsigned int i;
    char *id;
    char *name;
    char *description;
    char *status;
    time_t date_created;
    time_t date_updated;

    history_item **history_items;
    unsigned int history_items_len;
    unsigned int history_i;
    unsigned int history_inside_i;
};

struct history_item
{
    char *type;
    char *name;
    char *name_human_readable;
    char *value;
    char *new_value;
    time_t date;
};

static int tag_task(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    reply_results *pres = (reply_results*)userdata;
    if (0 == strcmp("id", tag))
    {
        return tpl_append(pout, plen, pres->id, strlen(pres->id));
    }
    else if (0 == strcmp("name", tag))
    {
        return tpl_append(pout, plen, pres->name, strlen(pres->name));
    }
    else if (0 == strcmp("status", tag))
    {
        return tpl_append(pout, plen, pres->status, strlen(pres->status));
    }
    else if (0 == strcmp("description", tag))
    {
        char *tmp = pres->description;
        return tpl_append(pout, plen, tmp,
            0 == tmp ? 0 : strlen(tmp));
    }
    else if (0 == strcmp("date-created", tag))
    {
        char tmp[64];
        sprintf(tmp, "%ld", pres->date_created);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("date-created-formatted", tag))
    {
        char tmp[21];
        struct tm tinfo;
        memset(&tinfo, 0, sizeof(tinfo));

        localtime_r(&(pres->date_created), &tinfo);
        strftime(tmp, 20, "%Y-%m-%d %H:%M:%S", &tinfo);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("date-updated", tag))
    {
        char tmp[64];
        sprintf(tmp, "%ld", pres->date_updated);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("date-updated-formatted", tag))
    {
        char tmp[21];
        struct tm tinfo;
        memset(&tinfo, 0, sizeof(tinfo));

        localtime_r(&(pres->date_updated), &tinfo);
        strftime(tmp, 20, "%Y-%m-%d %H:%M:%S", &tinfo);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    return 0;
}

static int tag_task_history(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    reply_results *pres = (reply_results*)userdata;
    if (0 == strcmp("type", tag))
    {
        history_item *item = pres->history_items[pres->history_i];
        char *tmp = item->type;
        return tpl_append(pout, plen, tmp,
            0 == tmp ? 0 : strlen(tmp));
    }
    else if (0 == strcmp("name", tag))
    {
        history_item *item = pres->history_items[pres->history_i];
        char *tmp = item->name;
        return tpl_append(pout, plen, tmp,
            0 == tmp ? 0 : strlen(tmp));
    }
    else if (0 == strcmp("name-human-readable", tag))
    {
        history_item *item = pres->history_items[pres->history_i];
        char *tmp = item->name_human_readable;
        return tpl_append(pout, plen, tmp,
            0 == tmp ? 0 : strlen(tmp));
    }
    else if (0 == strcmp("value", tag))
    {
        history_item *item = pres->history_items[pres->history_i];
        char *tmp = item->value;
        return tpl_append(pout, plen, tmp,
            0 == tmp ? 0 : strlen(tmp));
    }
    else if (0 == strcmp("date", tag))
    {
        history_item *item = pres->history_items[pres->history_i];
        char tmp[64];
        sprintf(tmp, "%ld", item->date);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else if (0 == strcmp("date-formatted", tag))
    {
        history_item *item = pres->history_items[pres->history_i];
        char tmp[21];
        struct tm tinfo;
        memset(&tinfo, 0, sizeof(tinfo));

        localtime_r(&(item->date), &tinfo);
        strftime(tmp, 20, "%Y-%m-%d %H:%M:%S", &tinfo);
        return tpl_append(pout, plen, tmp, strlen(tmp));
    }
    else
    {
        cr_expect_fail("Unexpected history type: %s\n", tag);
    }
    return 0;
}

static int tag_task_history_change(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    reply_results *pres = (reply_results*)userdata;
    if (0 == strcmp("new-value", tag))
    {
        history_item *item = pres->history_items[pres->history_i];
        char *tmp = item->new_value;
        return tpl_append(pout, plen, tmp,
            0 == tmp ? 0 : strlen(tmp));
    }
    else
    {
        cr_expect_fail("Unexpected history type: %s\n", tag);
    }
    return 0;
}

static int tag(char **pout, unsigned int *plen, void *userdata, char *tag)
{
    reply_results *pres = (reply_results*)userdata;
    switch (pres->entered)
    {
        case RE_TASK:
            return tag_task(pout, plen, userdata, tag);
        case RE_TASK_HISTORY:
            return tag_task_history(pout, plen, userdata, tag);
        case RE_TASK_HISTORY_CHANGE:
            return tag_task_history_change(pout, plen, userdata, tag);
        case RE_NULL:
            if (0 == strcmp("add-style", tag) ||
                0 == strcmp("edit-style", tag) ||
                0 == strcmp("error-style", tag) ||
                0 == strcmp("index-style", tag) ||
                0 == strcmp("warning-style", tag))
            {
                char *tmp = "style='display: none;'";
                return tpl_append(pout, plen, tmp, strlen(tmp));
            }
            else if (0 == strcmp("task-style", tag))
            {
                return 1;
            }
            break;
        default:
            break;
    }
    return 0;
}

static int enter_top(char **pout, unsigned int *plen, void *userdata, char *name)
{
    reply_results *pres = (reply_results*)userdata;
    if (0 == strcmp("task", name))
    {
        pres->i = -1;
        pres->entered = RE_TASK;
        return 1;
    }
    return 0;
}

static int enter_task(char **pout, unsigned int *plen, void *userdata, char *name)
{
    reply_results *pres = (reply_results*)userdata;
    if (0 == strcmp("history", name))
    {
        pres->history_i = -1;
        pres->entered = RE_TASK_HISTORY;
        return 1;
    }
    return 0;
}

static int enter_task_history(char **pout, unsigned int *plen, void *userdata, char *name)
{
    reply_results *pres = (reply_results*)userdata;
    int is_change = 0 != pres->history_items[pres->history_i]->new_value;
    if (1 == is_change && 0 == strcmp("change", name))
    {
        pres->history_inside_i = -1;
        pres->entered = RE_TASK_HISTORY_CHANGE;
        return 1;
    }
    return 0;
}

static int enter(char **pout, unsigned int *plen, void *userdata, char *name)
{
    reply_results *pres = (reply_results*)userdata;
    switch (pres->entered)
    {
        case RE_NULL:
            return enter_top(pout, plen, userdata, name);
        case RE_TASK:
            return enter_task(pout, plen, userdata, name);
        case RE_TASK_HISTORY:
            return enter_task_history(pout, plen, userdata, name);
        default:
            break;
    }
    return 0;
}

static int next(char **pout, unsigned int *plen, void *userdata)
{
    reply_results *pres = (reply_results*)userdata;
    switch (pres->entered)
    {
        case RE_TASK:
            pres->i++;
            return 0 == pres->i;
        case RE_TASK_HISTORY:
            pres->history_i++;
            return pres->history_items_len > pres->history_i;
        case RE_TASK_HISTORY_CHANGE:
            pres->history_inside_i++;
            return 0 == pres->history_inside_i;
        default:
            break;
    }
    return 0;
}

static void leave(char **pout, unsigned int *plen, void *userdata)
{
    reply_results *pres = (reply_results*)userdata;
    switch (pres->entered)
    {
        case RE_TASK_HISTORY_CHANGE:
            pres->entered = RE_TASK_HISTORY;
            break;
        case RE_TASK_HISTORY:
            pres->entered = RE_TASK;
            break;
        case RE_TASK:
            pres->entered = RE_NULL;
            break;
        default:
            break;
    }
}

static void get_date_updated(char *ip, unsigned int port, char *task, time_t *pdate)
{
    char tmp[21];
    struct tm tinfo;
    memset(&tinfo, 0, sizeof(tinfo));

    do_task_get_date_updated(ip, port, task, pdate);

    localtime_r(pdate, &tinfo);
    strftime(tmp, 20, "%Y-%m-%d %H:%M:%S", &tinfo);
    cr_log_info("Date updated after add: %s", tmp);
}

static void build_history_items(reply_results *res)
{
    history_item **history_items;
    history_item *item;

    history_items = (history_item **)malloc(sizeof(history_item *) * 5);

    item = (history_item *)malloc(sizeof(history_item));
    item->type = "removed";
    item->name = "description";
    item->name_human_readable = "Description";
    item->value = "Goat milk";
    item->date = res->date_updated;
    item->new_value = 0;
    history_items[0] = item;

    item = (history_item *)malloc(sizeof(history_item));
    item->type = "changed";
    item->name = "name";
    item->name_human_readable = "Name";
    item->value = "Buy milk";
    item->date = res->date_updated;
    item->new_value = "Buy beer";
    history_items[1] = item;

    item = (history_item *)malloc(sizeof(history_item));
    item->type = "added";
    item->name = "name";
    item->name_human_readable = "Name";
    item->value = "Buy milk";
    item->date = res->date_created;
    item->new_value = 0;
    history_items[2] = item;

    item = (history_item *)malloc(sizeof(history_item));
    item->type = "added";
    item->name = "description";
    item->name_human_readable = "Description";
    item->value = "Goat milk";
    item->date = res->date_created;
    item->new_value = 0;
    history_items[3] = item;

    item = (history_item *)malloc(sizeof(history_item));
    item->type = "added";
    item->name = "status";
    item->name_human_readable = "Status";
    item->value = "open";
    item->date = res->date_created;
    item->new_value = 0;
    history_items[4] = item;

    res->history_items = history_items;
    res->history_items_len = 5;
}

static void free_history_items(reply_results *res)
{
    for (unsigned int i = 0; i < res->history_items_len; i++)
    {
        free(res->history_items[i]);
    }
    free(res->history_items);
}

Test(task_html, nonempty)
{
    char *body = 0;
    unsigned int len = 0;
    reply_results res;
    unsigned int port = -1;
    char url[256];
    char len_str[64];
    char *expected_headers[6] =
    {
        "Content-Length", len_str,
        "Content-Type", "text/html",
        "Cache-Control", "no-store"
    };

    res.entered = RE_NULL;
    res.id = "1";
    res.name = "Buy beer";
    res.description = 0;
    res.status = "open";
    res.date_created = 0;
    res.date_updated = 0;

    expect_process_started(0, &port);

    do_add("127.0.0.1", port, 3,
        "name", "Buy milk",
        "description", "Goat milk",
        "status", "open");
    get_date_updated("127.0.0.1", port, "1", &(res.date_created));

    do_modify("127.0.0.1", port, 1, 1, 2,
        "id-equals", 1,
        "name", "Buy beer",
        "description", "");
    get_date_updated("127.0.0.1", port, "1", &(res.date_updated));

    build_history_items(&res);
    build_body_from_template(tpl_file_static_index_html,
        tag, enter, next, leave, &res, &body, &len);
    free_history_items(&res);
    sprintf(len_str, "%u", len);
    sprintf(url, "http://%s:%u/task/1.html", "127.0.0.1", port);
    expect_http_get(url, 0, 200, 3, expected_headers, body);

    free(body);
    expect_process_died(0);
}

Test(task_html, invalid_method)
{
    unsigned int port = -1;
    char url[256];

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/task", "127.0.0.1", port);
    expect_http_post(url, 0, "{}", 405, 0, 0, 0);
    expect_process_died(0);
}

Test(task_html, invalid_id)
{
    unsigned int port = -1;
    char url[256];

    expect_process_started(0, &port);

    sprintf(url, "http://%s:%u/task", "127.0.0.1", port);
    expect_http_get(url, 0, 400, 0, 0, 0);

    sprintf(url, "http://%s:%u/task/1.htm", "127.0.0.1", port);
    expect_http_get(url, 0, 400, 0, 0, 0);

    sprintf(url, "http://%s:%u/task/.html", "127.0.0.1", port);
    expect_http_get(url, 0, 400, 0, 0, 0);

    sprintf(url, "http://%s:%u/task/a.htm", "127.0.0.1", port);
    expect_http_get(url, 0, 400, 0, 0, 0);

    sprintf(url, "http://%s:%u/task/a.html", "127.0.0.1", port);
    expect_http_get(url, 0, 400, 0, 0, 0);

    expect_process_died(0);
}

Test(task_html, no_task)
{
    unsigned int port = -1;
    char url[256];

    expect_process_started(0, &port);
    sprintf(url, "http://%s:%u/task/1.html", "127.0.0.1", port);
    expect_http_get(url, 0, 404, 0, 0, 0);
    expect_process_died(0);
}
